/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * wdev.h - common wireless stack implementation for linux
 *
 * $Id: wdev.h,v 1.6 2003/08/22 17:35:38 reyk Exp $
 *
 * Copyright (c) 2003, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef _LINUX_WIRELESS_DEVICE_H
#define _LINUX_WIRELESS_DEVICE_H

#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/wireless.h>

#ifndef IW_MODE_MONITOR
#define IW_MODE_MONITOR 6
#endif

#define WIRELESS_ENABLE_MONITOR 1
#define WIRELESS_DISABLE_MONITOR 0

#define WIRELESS_DISABLE_NETWORK -1
#define WIRELESS_FIND_NETWORK 0
#define WIRELESS_FOUND_NETWORK 1

/* encapsulate or transmit raw frame? */
#define WIRELESS_ENCAP 0 
#define WIRELESS_RAW 1

#define	WIRELESS_ADDR_SIZE 6
#define	WIRELESS_SSID_SIZE IW_ESSID_MAX_SIZE
typedef u32 nwid_t;

extern u8 wireless_no_addr[WIRELESS_ADDR_SIZE];

/* enum fields */
enum wireless_protocol {
	WIRELESS_PROTO_IEEE80211 = 0,
	WIRELESS_PROTO_WLSEC = 1,
	WIRELESS_PROTO_CUSTOM = 2
};

enum wireless_encoding_protocol {
	WIRELESS_ENCODING_NONE = 0,
	WIRELESS_ENCODING_WEP = 1,
	WIRELESS_ENCODING_WLSEC_ESP = 2,
	WIRELESS_ENCODING_CUSTOM = 3
};

enum wireless_phy_type {
	WIRELESS_PHY_MODE_FHSS = 0,
	WIRELESS_PHY_MODE_DSSS = 1,
	WIRELESS_PHY_MODE_OFDM = 2,
	WIRELESS_PHY_MODE_MISC = 3
};

typedef struct wireless_mac_cap {
	u8 nwid:1, essid:1, bssid:1, master:1, monitor:1, reserved:3;
} wireless_mac_cap_t;

typedef struct wireless_phy_cap {
	u8 ofdm:1, dsss:1, fhss:1, misc:1, reserved:4;
} wireless_phy_cap_t;

enum wireless_mac_header {
	WIRELESS_STD_HDR = 0,
	WIRELESS_PRISM_HDR = 1
};

typedef struct wireless_mac {
	/* used by some protocols */
	u8 address[WIRELESS_ADDR_SIZE];
	/* agrego esto para tener las tres direcciones: dest add, source add y bssid */
	u8 dest_address[WIRELESS_ADDR_SIZE];
	/* network id (by name and/or address and/or numeric id) */
	u8 essid[WIRELESS_SSID_SIZE];
	u8 bssid[WIRELESS_ADDR_SIZE];
	u8 desired_bssid[WIRELESS_ADDR_SIZE]; 
	nwid_t nwid, desired_nwid;

	/* network status */
	u8 searching, associated;
	u32 network_timeout;

	/* wireless link layer protocol (ie. IEEE 802.11) */
	enum wireless_protocol proto;
	
	/* host based wireless encoding protocol (ignore if NONE) */
	enum wireless_encoding_protocol encoding;

	/* protocol capabilities */
	wireless_mac_cap_t cap;

	/* add special header in front of received monitor packet? */
	enum wireless_mac_header header;

	int mode; /* IW_MODE_* in linux/wireless.h */
	void *encoding_priv; /* protocol specific data */
	void *priv; /* protocol specific data */
} wireless_mac_t;

typedef struct wireless_phy {
	/* radio type */
	enum wireless_phy_type type;
	wireless_phy_cap_t cap;

	/* current radio frequency */ 
	struct iw_freq freq; 
	struct iw_freq min_freq; 
	struct iw_freq max_freq; 

	/* current radio quality */
	struct iw_quality qual;
} wireless_phy_t;

struct wireless_device {
	/* top level network device */
	struct net_device *dev;

	/* wireless statistics */
	struct iw_statistics *iw_stats; /* device statistics */

	/* common proc directory */
	struct proc_dir_entry *proc;

	wireless_mac_t mac; /* mac layer */
	wireless_phy_t phy; /* physical layer */

	int (*set_mode)(struct wireless_device *wdev, int mode);
	int (*set_channel)(struct wireless_device *wdev, struct iw_freq *freq);
	void (*set_monitor)(struct wireless_device *wdev, int val);
	int (*set_network)(struct wireless_device *wdev, int val);
	int (*transmit)(struct net_device *dev, struct sk_buff *skb, int encap);

	/* load/unload protocol specific data...
	 * either called at initialization time or at netdevice up/down event */
	int (*init)(struct wireless_device *wdev); 
	void (*exit)(struct wireless_device *wdev);
	int (*encoding_init)(struct wireless_device *wdev); 
	void (*encoding_exit)(struct wireless_device *wdev);

	/* encapsulate outgoing frame */
	int (*encap)(struct wireless_device *wdev, 
		     struct sk_buff *skb, /* tx socket buffer */
		     void *buf, int len); /* input frame buffer */

	/* after frame has been transmitted */
	int (*transmit_ok)(struct wireless_device *wdev);

	/* decapsulate received frame */
	struct sk_buff* (*decap)(struct wireless_device *wdev, 
				 void *buf, int *len); /* output frame buffer */

	/* host based link layer encoding...
	 * used instead of integrated (wep) engine */
	int (*encode)(struct wireless_device *wdev, 
		      void *buf, int len); /* frame buffer */
	int (*decode)(struct wireless_device *wdev, 
		      void *buf, int len); /* frame buffer */
};

#define WIRELESS_IOCTLS \
        case SIOCGIWNAME: \
	case SIOCSIWFREQ: \
	case SIOCGIWFREQ: \
	case SIOCSIWMODE: \
	case SIOCGIWMODE: \
	case SIOCSIWAP: \
	case SIOCGIWAP: \
        case SIOCGIWNWID: \
        case SIOCSIWNWID: \
        case SIOCGIWESSID: \
        case SIOCSIWESSID

/* ARPHRD_IEEE80211_PRISM (from linux-wlan-ng) */
typedef struct linux_wlan_ng_val {
	u32 did;
	u16 status, len;
	u32 data;
} __attribute__ ((__packed__)) linux_wlan_ng_val_t;

typedef struct linux_wlan_ng_prism_hdr {
	u32 msgcode, msglen;
	char devname[16];
	linux_wlan_ng_val_t hosttime, mactime, channel, rssi, sq,
		signal, noise, rate, istx, frmlen;
} __attribute__ ((__packed__)) linux_wlan_ng_prism_hdr_t;

#ifdef __KERNEL__
/*
 * global wireless data
 */
extern struct proc_dir_entry *proc_wireless;
extern int init_wireless(void);
extern void exit_wireless(void);

/* common ioctls */
extern int wireless_ioctl(struct wireless_device *wdev, 
			  struct iwreq *iwr, int cmd);

/* enable/disable monitor mode (if supported) */
extern void wireless_monitor(struct wireless_device *wdev, int val);

/* 
 * wireless setup 
 */
extern int wireless_register(struct wireless_device *wdev);
extern void wireless_unregister(struct wireless_device *wdev);
#endif /* __KERNEL__ */

/* 
 * IEEE 802.11 
 */
/* IEEE 802.11 private date */
enum wireless_ieee80211_status {
	WIRELESS_IEEE80211_OFFLINE = 0,
	WIRELESS_IEEE80211_AUTHENTICATING,
	WIRELESS_IEEE80211_AUTHENTICATED,
	WIRELESS_IEEE80211_ASSOCIATING,
	WIRELESS_IEEE80211_ASSOCIATED
};

#define WIRELESS_IEEE80211_FCS 4

typedef struct wireless_ieee80211 {
	enum wireless_ieee80211_status status;
	u16 seq:12;
} __attribute__ ((__packed__)) wireless_ieee80211_t;

#ifdef __KERNEL__
extern int wireless_ieee80211_init(struct wireless_device *wdev);
extern void wireless_ieee80211_exit(struct wireless_device *wdev);
extern int wireless_ieee80211_encap(struct wireless_device *wdev, 
				    struct sk_buff *skb,
				    void *buf, int buflen);
extern int wireless_ieee80211_transmit_ok(struct wireless_device *wdev);
extern struct sk_buff *wireless_ieee80211_decap(struct wireless_device *wdev, 
						void *buf, int *len);
extern void wireless_ieee80211_dump_hdr(char *info, u8 *buf, int len);
#endif /* __KERNEL__ */

/* 
 * (host based) encoding 
 */
typedef struct wireless_encoding_wep {
	int key_id;
	/* TODO... */
} __attribute__ ((__packed__)) wireless_encoding_wep_t;

#ifdef __KERNEL__
extern int wireless_wep_init(struct wireless_device *wdev);
extern void wireless_wep_exit(struct wireless_device *wdev);
extern int wireless_wep_encode(struct wireless_device *wdev, 
			       void *buf, int len);
extern int wireless_wep_decode(struct wireless_device *wdev, 
			       void *buf, int len);
#endif /* __KERNEL__ */

#ifdef USE_WLSEC
/* WLSec improved IEEE 802.11 */
#ifdef __KERNEL__
extern int wireless_wlsec_encap(struct wireless_device *wdev, 
				struct sk_buff *skb,
				void *buf, int buflen);
extern int wireless_wlsec_decap(struct wireless_device *wdev, 
				void *buf, int len);

extern int wireless_wlsec_esp_encode(struct wireless_device *wdev, 
				     void *buf, int len);
extern int wireless_wlsec_esp_decode(struct wireless_device *wdev, 
				     void *buf, int len);
#endif /* __KERNEL__ */
#endif


static inline char *wireless_type(int type) 
{
	switch(type) {
	case WIRELESS_PHY_MODE_FHSS:
		return("FHSS");
	case WIRELESS_PHY_MODE_DSSS:
		return("DSSS");
	case WIRELESS_PHY_MODE_OFDM:
		return("OFDM");
	case WIRELESS_PHY_MODE_MISC:
		return("misc");
	default:
		return("unknown");
	}
}

static inline char *wireless_proto(int proto) 
{
	switch(proto) {
	case WIRELESS_PROTO_IEEE80211:
		return("802.11");
	case WIRELESS_PROTO_WLSEC:
		return("WLSec");
	case WIRELESS_PROTO_CUSTOM:
		return("custom");
	default:
		return("unknown");
	}
}

#endif /* _LINUX_WIRELESS_DEVICE_H */
