/*                                                                                  DMA handling
    Copyright (C) 2003-2004  Chris Kennedy <ckennedy at kmos org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ivtv-driver.h"
#include "ivtv-queue.h"
#include "ivtv-reset.h"
#include "ivtv-mailbox.h"
#include "ivtv-i2c.h"
#include "ivtv-kthreads.h"
#include "ivtv-irq.h"
#include "ivtv-video.h"

/* type is 0=enc, 1=dec ; reg can be 0x00 or 0x04 or 0x10 */
/* if 0x00 then 0x00 is clear, while with 0x04 the opposite is true, 0x03 */
int ivtv_dma_status(struct ivtv *itv, int type, int reg, int sleep) {
        u32 data[IVTV_MBOX_MAX_DATA], result;
        u32 reg_result = 0;
        int ret = 0;

	/* Either check register or use API command */
	if (!sleep) {
        	/* Get DMA status from register */
        	reg_result = readl(itv->reg_mem + reg);
	} else {
        	/* API Check DMA Status */
        	ret = ivtv_api(itv, itv->dec_mbox,
                	&itv->dec_msem, IVTV_API_DEC_DMA_STATUS,
                                &result, 0, &data[0]);
        	/* Error with API cmd */
        	if (ret)
                	return -EIO;

        	reg_result = data[0];
	}

        /* Check if complete */
        if (type) { /* 1=Decoder */
                if (!(reg_result & IVTV_DMA_SUCCESS_READ)) {
                        IVTV_DEBUG(IVTV_DEBUG_DMA,
                                "DEC DMA reg 0x%08x Result=0x%08x %lu\n",
                                reg, reg_result, jiffies);
                        ret |= IVTV_DMA_SUCCESS_READ;
                }
        } else { /* 0=Encoder */
                if (!(reg_result & IVTV_DMA_SUCCESS_WRITE)) {
                        IVTV_DEBUG(IVTV_DEBUG_DMA,
                                "ENC DMA reg 0x%08x Result=0x%08x %lu\n",
                                reg, reg_result, jiffies);
                        ret |= IVTV_DMA_SUCCESS_WRITE;
                }
        }

        if (!reg) /* This register only is for dma write/read complete */
                return ret;

        /* Check for errors, gather and return value */
        if (reg_result & IVTV_DMA_ERR_READ) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC DMA Error. Result=0x%08x %lu\n",
                        reg_result, jiffies);
                ret |= IVTV_DMA_ERR_READ;
        }
        if (reg_result & IVTV_DMA_ERR_WRITE) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "ENC DMA Error. Result=0x%08x %lu\n",
                        reg_result, jiffies);
                ret |= IVTV_DMA_ERR_WRITE;
        }
        if (reg_result & IVTV_DMA_ERR_LIST) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "Link List DMA Error. Result=0x%08x %lu\n",
                        reg_result, jiffies);
                ret |= IVTV_DMA_ERR_LIST;
        }

        return ret;
}

int ivtv_dec_buffer(struct ivtv *itv, 
	int type, u32 mem_offset, u32 mem_size, u32 buffer_bytes)
{
        /* Invalid DMA Xfer Type */
        if (type < 0 || type > 2) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: Decoder Invalid type %d?\n", type);
                return -EIO;
        }

        /* Check if size request is too big for decoder buffer */
        if ( ((mem_size)+buffer_bytes) > IVTV_DEC_MAX_BUF)
        {
                IVTV_DEBUG(IVTV_DEBUG_INFO,
                        "DEC: wants 0x%08x has 0x%08x = 0x%08x bytes!\n",
                        mem_size, buffer_bytes, ((mem_size)+buffer_bytes));
                mem_size = (IVTV_DEC_MAX_BUF - buffer_bytes);
                IVTV_DEBUG(IVTV_DEBUG_INFO,
                        "DEC: Changed mem_size to 0x%08x.\n", mem_size);
        }

	/* More than Maximum Xfer, reduce it */
	if (mem_size > IVTV_DMA_MAX_XFER)
		mem_size = IVTV_DMA_MAX_XFER;

        /* Manually calculate if decoder isn't updating */
        if ((mem_offset + mem_size) > 0x0050fb00)
                mem_size = (mem_offset + mem_size) - 0x0050fb00;

        if ((mem_offset + mem_size) < 0x0040fb00 ||
                                (mem_offset + mem_size) > 0x0050fb00)
        {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: Decoder wants 0x%08x bytes, "
                        "bad memory location 0x%08x.\n",
                        mem_size, (mem_offset + mem_size));
                return -EIO;
        }

        /* mem_size less than minimum size, can't xfer */
        if (mem_size < IVTV_DMA_ALIGN_XFER) {
                IVTV_DEBUG(IVTV_DEBUG_INFO,
                        "DEC: Decoder Invalid mem_size 0x%08x?\n",
                        mem_size);
                return -EIO;
        }

        /* Check if size is one SG buffer or more, else won't work */
        if (mem_size < IVTV_DMA_DEC_BUF_SIZE) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                               "DEC: Decoder wants 0x%08x bytes, too small.\n",
                               mem_size);
                return -EIO;
        }

	/* Align xfer to minimum size */
	mem_size = (mem_size - (mem_size%IVTV_DMA_ALIGN_XFER));

        /* is this request more than the maximum buffer 
		or maximum buffer filled? */
        if (buffer_bytes >= IVTV_DEC_MAX_BUF ||
		mem_size > IVTV_DEC_MAX_BUF) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: Decoder Invalid request for 0x%08x bytes, "
			"has 0x%08x bytes in buffer!\n",
                        mem_size, buffer_bytes);
                return -EIO;
        }

        /* Return needed Buffers */
        return (mem_size / IVTV_DMA_DEC_BUF_SIZE);
}

int ivtv_reset_stream(struct ivtv *itv, int type)
{
        DECLARE_WAITQUEUE(wait, current);
        int rc = 0;
        u32 then;

        then = jiffies;
        add_wait_queue(&itv->streams[type].waitq, &wait);
        do {
                set_current_state(TASK_INTERRUPTIBLE);

                /* check DMA status register */
                if (!test_and_set_bit(IVTV_F_S_DMAP,
                                        &itv->streams[type].s_flags))
                {
                        break;
                }

		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ/100);

                if ((jiffies - then) > (HZ)) {
                        IVTV_DEBUG(IVTV_DEBUG_ERR, "Decoder wait failed.\n");
                        break;
                }

                if (signal_pending(current))
                        rc = -EINTR;
        } while (!rc);
        set_current_state(TASK_RUNNING);
        remove_wait_queue(&itv->streams[type].waitq, &wait);

        if (rc)
                return -EBUSY;

        IVTV_DEBUG(IVTV_DEBUG_INFO, "Decoder locked.\n");

        return 0;
}

