/*
    file operation functions
    Copyright (C) 2003-2004  Kevin Thayer <nufan_wfk at yahoo.com>
    Copyright (C) 2004  Chris Kennedy ckennedy@kmos.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Main Driver file for the ivtv project:
 * Driver for the iTVC15 chip.
 * Author: Kevin Thayer (nufan_wfk at yahoo.com)
 * License: GPL
 * http://www.sourceforge.net/projects/ivtv/
 * 
 * -----
 * MPG600/MPG160 support by  T.Adachi <tadachi@tadachi-net.com>
 *                      and Takeru KOMORIYA<komoriya@paken.org>
 *
 * AVerMedia M179 GPIO info by Chris Pinkham <cpinkham@bc2va.org>
 *                using information provided by Jiun-Kuei Jung @ AVerMedia.
 */

#include "ivtv-driver.h"
#include "ivtv-dma.h"
#include "ivtv-fileops.h"
#include "ivtv-i2c.h"
#include "ivtv-queue.h"
#include "ivtv-irq.h"
#include "ivtv-reset.h"
#include "ivtv-vbi.h"
#include "ivtv-mailbox.h"
#include "ivtv-kthreads.h"
#include "ivtv-audio.h"
#include "ivtv-video.h"
#include "ivtv-streams.h"
#include "ivtv-cards.h"
#include <media/audiochip.h>

typedef unsigned long uintptr_t;

int ivtv_claim_stream(struct ivtv_open_id *id, int type)
{
	struct ivtv_stream *stream = &id->itv->streams[type];

	if (!test_and_set_bit(IVTV_F_S_IN_USE, &stream->s_flags)) {
		/* claim this stream if it wasn't already claimed */
		stream->id = id->open_id;
	}
	if (stream->id != id->open_id) {
		/* someone else is using this stream already */
		return -EBUSY;
	}
	return 0;
}

void ivtv_release_stream(struct ivtv *itv, int type)
{
	struct ivtv_stream *stream = &itv->streams[type];

	stream->id = -1;
	clear_bit(IVTV_F_S_IN_USE, &stream->s_flags);
}

static int ivtv_check_digitizer_sync(struct ivtv *itv)
{
	int sig=0, ret=0;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "Checking digitizer\n");

	ret = ivtv_saa7115(itv, DECODER_GET_STATUS, &sig);

	if (ret) {
		IVTV_DEBUG(IVTV_DEBUG_ERR,
			   "I2C call failed. is saa7115 loaded?\n");
	}
	if (sig & DECODER_STATUS_GOOD) {
		IVTV_DEBUG(IVTV_DEBUG_INFO, "Digitizer sync GOOD\n");
	} else {
		IVTV_DEBUG(IVTV_DEBUG_INFO, "Digitizer sync BAD\n");
		return -EIO;
	}

        return 0; 
}

static int ivtv_reset_digitizer(struct ivtv *itv)
{
        int x=0, ret=0;

        /* x is just a placeholder. It's unused */
        /* this just resets the scaler.. shouldn't need to re-do any settings */
        ret = ivtv_saa7115(itv, DECODER_RESET, &x);

        return ret;
}


#ifdef YUV_FIXUP
static int ivtv_YUV_fixup(struct ivtv_stream *st, int count,
			  char *ubuf, struct ivtv_buffer *buf)
{
/*  
 * int count = # of bytes to transfer to client 
 * st->ubytes = # of bytes written this frame
 * ubuf = buffer to write to (user's buffer)
 * buf = read buffer
 *
 */
	int src_width=720; /* all known formats have src width of 720 */
	int Hoff, Voff; /* collectors for offsets to read position */
	int width, height; /* resolution of the capture stream */
	int curline; /* vertical line currently being processed */
	int maxline; /* height of combined frame */
	int cur_m_block; /* current horizontal offset of working mblock in this row */
	int maxblock; /* # of macroblocks in a row */
	int Hbytes; /* # of bytes to write to user this time around */
	int retval=0; /* accumulator for total bytes written */
	int start; /* position in buf to read from */
	int buf_start; /* byte offset of first byte in this *buf */

	height = st->format.fmt.pix.height;
	width  = st->format.fmt.pix.width;
	maxblock = (width + 0xf) >> 4;
	maxline = (int)(1.5 * height); /* 1 for Y, .5 for UV */
	/* Offset is always bufsize * buffer index 
	buf_start = (st->ubytes - buf->readpos);  tested/works */

	buf_start = IVTV_DMA_ENC_BUF_SIZE * buf->buffer.index;

	/* FIXME it may not be possible to get YUV width > 720 */
	// if (width > src_width) src_width=width;

	curline = (int) (st->ubytes / width);

	while (curline < maxline) {
// 	printk(" cl: %d, ml: %d\n", curline, maxline);
	Voff =	16 * (curline & 0xf) +		   /* Voffset within MBlock */
		  ((curline & 0xfff0) * src_width); /* Voffset of Mblock */

	cur_m_block = (st->ubytes - (curline * width)) >> 4;

/*	printk("voff %d, macroVoff %d, Voff %d, cmb %d\n", (16 * (curline & 0xf)), 
			((curline & 0xff0) * src_width), Voff, cur_m_block);
*/

	while ((cur_m_block < maxblock) && (count > 0)) {
		Hoff   = (cur_m_block * 256) +	/* mblock offset within line */
			 /* Hoffset within mblock, usually 0 */
			 ((st->ubytes - (curline * width)) & 0xf);
		Hbytes = 16 - ((st->ubytes - (curline * width)) & 0xf);

		if (Hbytes > count) Hbytes = count;

		start = Hoff + Voff;

		if (copy_to_user((char *)((uintptr_t)ubuf + retval),
				 (u32 *)((uintptr_t)buf->buffer.m.userptr +
				 (start - buf_start)),
				 Hbytes)) {
			IVTV_DEBUG(IVTV_DEBUG_ERR, "copy to user failed\n");
			return -EFAULT;
		}
		
		count -= Hbytes;
		retval += Hbytes;
		st->ubytes += Hbytes;

		cur_m_block++;
	}

	/* if user can't handle anymore data or buffer empty */
	curline++;
	if ((count == 0) ) /*|| ((curline * src_width) % IVTV_DMA_ENC_BUF_SIZE) == 0)*/
		return retval;
	}

	/* FIXME i don't think you should ever get here */
	IVTV_DEBUG(IVTV_DEBUG_ERR, "You've just sailed off the edge of this function\n");
	return retval;
}
#endif

static int ivtv_read(struct file *filp, char *ubuf, size_t count, loff_t *pos) 
{
        struct ivtv_open_id *id = filp->private_data;
        struct ivtv *itv = id->itv;
	int x, sleepctr, datalen, retval = 0, freed = 0;
	int reset = 0;
	size_t newcount;
	int type = id->type;
	struct ivtv_buffer *buf;
	struct ivtv_stream *st = &itv->streams[type];
	DECLARE_WAITQUEUE(wait, current);
	int idx = itv->vbi_inserted_frame % 20;
	int blocking = !(filp->f_flags & O_NONBLOCK);

	IVTV_DEBUG(IVTV_DEBUG_INFO, " Read stream %d.. \n", type);


        IVTV_DEBUG(IVTV_DEBUG_DMA,
                   "ENC: free_q: %d elements\n",
                   atomic_read(&st->free_q.elements));
        IVTV_DEBUG(IVTV_DEBUG_DMA,
                   "ENC: dma_q: %d elements\n",
                   atomic_read(&st->dma_q.elements));
        IVTV_DEBUG(IVTV_DEBUG_DMA,
                   "ENC: full_q: %d elements\n",
                   atomic_read(&st->full_q.elements));
        IVTV_DEBUG(IVTV_DEBUG_DMA,
                   "ENC: io_q: %d elements\n",
                   atomic_read(&st->io_q.elements));

	st->seq = 0;
	st->ubytes = 0;

	if (atomic_read(&itv->capturing) == 0 && (st->id == -1)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR,
			   "Stream not initialized before read (shouldn't happen)\n");
		return -EIO;
	}

	sleepctr = retval = 0;
	buf = NULL;

	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&st->waitq, &wait);
	do {
 		if (type != IVTV_DEC_STREAM_TYPE_VBI &&
 		    type != IVTV_DEC_STREAM_TYPE_VOUT &&
	  	    (itv->trans_id == 0) && 
		    (sleepctr >= IVTV_MAX_DATA_SLEEP)) 
		{
			IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"Timeout waiting for data!\n");

			/* Possible digitizer sync issue? */
			x = ivtv_check_digitizer_sync(itv);
			if (x) {
				IVTV_DEBUG(IVTV_DEBUG_ERR, 
					"Resetting digitizer\n");
				x = ivtv_reset_digitizer(itv);
				reset++;
			}
			if (x && reset > 3) {
				IVTV_DEBUG(IVTV_DEBUG_ERR, 
					"Failed reseting digitizer!\n");

				/* guess we couldn't reset it */
				IVTV_DEBUG(IVTV_DEBUG_ERR, 
					"Capture has failed!!! Try again.\n");
                		retval = -EIO;
			}
		}
		if (retval) /* failed all attempts at fixing stream */
			break;

		buf = ivtv_deq_peek_head_lock_bh(&st->full_q, st);
		if (buf) 
			break;

		if (test_bit(IVTV_F_S_STREAMOFF, &st->s_flags)) {
			IVTV_DEBUG(IVTV_DEBUG_ERR,
					"EOS received in waitq\n");
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&st->waitq, &wait);

			retval = 0;
			goto read_streamoff;
		}
		
		/* Done capturing? */
		if (type != IVTV_DEC_STREAM_TYPE_VBI && 
				!test_bit(IVTV_F_S_CAPTURING, &st->s_flags)) {
			IVTV_DEBUG(IVTV_DEBUG_INFO,
					"No more data to send, returning 0\n");
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&st->waitq, &wait);

			retval = 0;
			goto read_ok;
		}

		if (!blocking) {
			IVTV_DEBUG(IVTV_DEBUG_ERR,
					"ENC: sleep interrupted in waitq\n");
			retval = -EAGAIN;
			break;
		}

		if (ivtv_sleep_timeout(IVTV_SLEEP_WAIT, 1)) {
			retval = -EINTR;
		}

		sleepctr++;
	} while (!retval);

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&st->waitq, &wait);

	/* an error (or signal) occured */
	if (retval) {
		goto read_fail;
	}

        /* byteswap ABCD -> DCBA for MPG data, if bit set*/
        if (test_and_clear_bit (IVTV_F_B_NEED_BUF_SWAP, &buf->b_flags))  {
		int y;
                for (y = 0; y < buf->buffer.bytesused; y += 4)
                       swab32s( (u32 *)((uintptr_t)buf->buffer.m.userptr + y));

                IVTV_DEBUG(IVTV_DEBUG_DMA, 
			"BSWAPPING count=%d\n", buf->buffer.bytesused);
        }

	/* First Read */
        if (itv->first_read == 1)
		itv->first_read = 0;	

	/* copy buffer to the client */
	while ((count > 0) && (buf->buffer.bytesused > 0 || itv->vbi_sliced_mpeg_offset[idx])) {
		u8 *p;
		static u8 ba[] = { 0, 0, 1, 0xba };
		static u8 be[] = { 0, 0, 1, 0xbe };
		int read_from_buf = 1;

		newcount = count;
		if (itv->vbi_sliced_mpeg_offset[idx]) {
			datalen = itv->vbi_sliced_mpeg_size[idx] - itv->vbi_sliced_mpeg_offset[idx];
		}
		else {
			datalen = buf->buffer.bytesused;
		}

		IVTV_DEBUG(IVTV_DEBUG_DMA, "datalen 0x%08x\n", datalen);

		if (newcount > datalen)
		       	newcount = datalen;

		if (itv->vbi_sliced_mpeg_offset[idx]) {
			p = itv->vbi_sliced_mpeg_data[idx] + itv->vbi_sliced_mpeg_offset[idx];

			if (copy_to_user((char *)((uintptr_t)ubuf+retval), p, newcount)) {
				IVTV_DEBUG(IVTV_DEBUG_ERR, "VBI copy to user failed\n");
				retval = -EFAULT;
				goto read_fail;
			}
			read_from_buf = 0;
			itv->vbi_sliced_mpeg_offset[idx] += newcount;
			if (itv->vbi_sliced_mpeg_offset[idx] == itv->vbi_sliced_mpeg_size[idx]) {
				itv->vbi_sliced_mpeg_offset[idx] = itv->vbi_sliced_mpeg_size[idx] = 0;
				itv->vbi_inserted_frame++;
			}
		}
#ifdef YUV_FIXUP
		else if ((type == 1) && (itv->options.yuv_fixup)) {
			newcount = ivtv_YUV_fixup(st,newcount,ubuf+retval,buf);
			if (newcount < 0) {
				IVTV_DEBUG(IVTV_DEBUG_ERR, "Error fixing up YUV!\n");
				retval = newcount;
				goto read_fail;
			}
		} 
#endif
		else if (type == 0 && itv->vbi_sliced_mpeg_size[idx]) {
			int n = newcount;
			u8 *f;

			f = p = (u8 *)((uintptr_t)buf->buffer.m.userptr + buf->readpos);

			while (n >= 4) {
				if (!itv->vbi_search_ba) {
					if (!memcmp(f, be, 4)) itv->vbi_search_ba = 1;
					f++; n--;
					continue;
				}
				if (memcmp(f, ba, 4)) {
					f++; n--;
					continue;
				}
				newcount = f - p;
				itv->vbi_search_ba = 0;
				itv->vbi_sliced_mpeg_offset[idx] = 1;
				break;
			}
			if (copy_to_user((char *)((uintptr_t)ubuf+retval), p, newcount)) {
				IVTV_DEBUG(IVTV_DEBUG_ERR, "VBI2 copy to user failed\n");
				retval = -EFAULT;
				goto read_fail;
			}
		}
		else {
			int field_lines = (itv->std & V4L2_STD_625_50) ? 312 : 263;

			p = (u8 *)((uintptr_t)buf->buffer.m.userptr + buf->readpos);

                    	if (type == IVTV_ENC_STREAM_TYPE_VBI) {
				if (itv->vbi_service_set_in > 0) {
					int cnt = ivtv_SLICED_VBI_fixup(itv, p, newcount, field_lines);
					buf->readpos += newcount - cnt;
					buf->buffer.bytesused -= newcount - cnt;
					newcount = cnt;
					p = (u8 *)itv->vbi_sliced_data;
				}
			}
			else if (type == IVTV_DEC_STREAM_TYPE_VBI) {
				int cnt = ivtv_DEC_VBI_fixup(itv, p, newcount, field_lines);
				buf->readpos += newcount - cnt;
				buf->buffer.bytesused -= newcount - cnt;
				newcount = cnt;
				p = (u8 *)itv->vbi_sliced_dec_data;
			}
			if (copy_to_user((char *)((uintptr_t)ubuf+retval), p, newcount)) {
				IVTV_DEBUG(IVTV_DEBUG_ERR, "MPG copy to user failed\n");
				retval = -EFAULT;
				goto read_fail;
			}
		}

		retval += newcount;
		count -= newcount;
		if (read_from_buf) {
			buf->readpos += newcount;
			buf->buffer.bytesused -= newcount;
		}

		IVTV_DEBUG(IVTV_DEBUG_DMA,
			   "new datalen 0x%08x\n", buf->buffer.bytesused);

		/* if buffer is empty or we've read the whole frame */
		if ((buf->buffer.bytesused == 0)) {
			ivtv_move_buf_lock_bh(&st->full_q, &st->free_q, buf, st);
			freed++;

			/* Read all the buffers we are willing to read */
			if (count == 0)
				goto read_ok;

			/* Grab the next buffer */
			buf = ivtv_deq_peek_head_lock_bh(&st->full_q, st);

			/* Buffers are empty, no more to read */
			if (!buf)
				goto read_ok;

			/* Still have buffers */
		        /* byteswap ABCD->DCBA for MPG data,if bit set*/
                        if (test_and_clear_bit(IVTV_F_B_NEED_BUF_SWAP, 
				&buf->b_flags))  
			{
                                int y;
                                for (y = 0; y < buf->buffer.bytesused; y+= 4) {
                                       swab32s((u32 *)((uintptr_t)
						buf->buffer.m.userptr + y));
                                }
                                IVTV_DEBUG(IVTV_DEBUG_DMA,
                                        "BSWAPPING count=%d\n",
                                        buf->buffer.bytesused);
                        }
		}
	} /* end of while */

	if (buf->buffer.bytesused > 0) {
		IVTV_DEBUG(IVTV_DEBUG_DMA, 
			"ivtv_read: buffer still has "
			"valid data %d bytes, read %d returned %d\n", 
			buf->buffer.bytesused, retval ,count);
	}

	/* Error Condition */
	if (retval != 0) { 
		IVTV_DEBUG(IVTV_DEBUG_INFO, "Returning %d\n", retval);
		goto read_fail;
	}

	/* if we put some buffers back in the free queue, kick off dma
	 * scheduling if card was stopped due to overflow before */
	if (freed && test_and_clear_bit(IVTV_F_S_OVERFLOW, &st->s_flags)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"ENC: Buffer Overflow, freeing %d\n", freed);

        	set_bit(IVTV_F_T_ENC_DMA_NEEDED, &itv->t_flags);
        	tasklet_schedule(&itv->dma_tq);
	} else if (test_and_clear_bit(IVTV_F_S_OVERFLOW, &st->s_flags)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"ENC: Buffer Overflow, no free buffers!!!\n");
	}

	/* Shouldn't ever get here */
	return -EIO;

read_fail:
	return retval;
read_streamoff:
	IVTV_DEBUG(IVTV_DEBUG_INFO, "v4l2 read attempting EOS w/STREAMOFF\n");
read_ok:
	if (retval > 0)
		*pos += retval;

	if (retval == 0) {
		IVTV_DEBUG(IVTV_DEBUG_INFO, "v4l2 read returning 0\n");
                if (test_and_clear_bit(IVTV_F_S_STREAMOFF, &st->s_flags)) {
                        IVTV_DEBUG(IVTV_DEBUG_INFO,
                                        "Returning from STREAMOFF w/%d ret\n", 
					retval);
                }
	}

	return retval;
}

ssize_t ivtv_v4l2_read(struct file *filp, char *buf, size_t count, loff_t *pos) 
{
        struct ivtv_open_id *id = filp->private_data;
        struct ivtv *itv = id->itv;
	struct ivtv_stream *stream, *vbi_stream;
	int type = id->type;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "v4l2 read\n");

	if (type == IVTV_ENC_STREAM_TYPE_RAD ||
	    type == IVTV_DEC_STREAM_TYPE_MPG ||
	    type == IVTV_DEC_STREAM_TYPE_YUV ||
	    type == IVTV_DEC_STREAM_TYPE_VOUT) {
		/* you cannot read from these stream types. */
		return -EPERM;
	}
	
	stream = &itv->streams[type];

	if (test_bit(IVTV_F_I_PASSTHROUGH, &itv->i_flags))
		return -EBUSY;

	/* Try to claim this stream. */
	if (ivtv_claim_stream(id, type))
		return -EBUSY;

	/* This stream does not need to start capturing */
	if (type == IVTV_DEC_STREAM_TYPE_VBI)
		return ivtv_read(filp, buf, count, pos);

	/* If capture is already in progress, then we also have to
	   do nothing extra. */
	if (test_and_set_bit(IVTV_F_S_CAPTURING, &stream->s_flags))
		return ivtv_read(filp, buf, count, pos);

	down(&itv->streams_lock);
	/* Tell the card to start capturing */
	if (ivtv_start_v4l2_encode_stream(itv, type)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "Failed to start capturing for stream %d\n", type);
		/* failure, clean up */
		clear_bit(IVTV_F_S_CAPTURING, &stream->s_flags);
		up(&itv->streams_lock);
		return -EIO;
	}
	up(&itv->streams_lock);

	vbi_stream = &itv->streams[IVTV_ENC_STREAM_TYPE_VBI];

	/* Note: yes, we could combine the ifs below into one
	   big if, but this is more readable. */

	/* We're done if this is any stream but the mpg stream */
	if (type != IVTV_ENC_STREAM_TYPE_MPG)
		return ivtv_read(filp, buf, count, pos);

	/* We're also done if we do not want to insert vbi data in
	   the mpeg stream. */
	if (!itv->vbi_insert_mpeg)
		return ivtv_read(filp, buf, count, pos);

	/* In order to insert vbi data in the mpeg stream, the vbi
	   stream must also be started. So try to claim that stream
	   first. On failure, we just continue without vbi insertion.
	   This is debatable. It might be better to just give an
	   error instead. */
	if (ivtv_claim_stream(id, IVTV_ENC_STREAM_TYPE_VBI)) {
		IVTV_DEBUG(IVTV_DEBUG_INFO, "No VBI insertion, VBI stream already in use.\n");
		return ivtv_read(filp, buf, count, pos);
	}

	/* Mark as having no reader process, the data is copied internally */
	down(&itv->streams_lock);
	set_bit(IVTV_F_S_NO_READ, &vbi_stream->s_flags);
	set_bit(IVTV_F_S_CAPTURING, &vbi_stream->s_flags);
	if (ivtv_start_v4l2_encode_stream(itv, IVTV_ENC_STREAM_TYPE_VBI)) {
		/* Failure, clean up and return an error */
		clear_bit(IVTV_F_S_NO_READ, &vbi_stream->s_flags);
		clear_bit(IVTV_F_S_CAPTURING, &vbi_stream->s_flags);
		ivtv_stop_capture(itv, type);
		clear_bit(IVTV_F_S_CAPTURING, &stream->s_flags);
		up(&itv->streams_lock);
		return -EIO;
	}
	up(&itv->streams_lock);
	return ivtv_read(filp, buf, count, pos);
}

static int ivtv_get_dec_buffer(struct ivtv *itv, struct ivtv_stream *stream,
	       struct ivtv_buffer **buf, int max_bufs, int blocking)
{
	DECLARE_WAITQUEUE(wait, current);
	int retval = 0;

	add_wait_queue(&stream->waitq, &wait);
	for (;;) {
		set_current_state(TASK_INTERRUPTIBLE);

		*buf = ivtv_deq_buf_lock(&stream->free_q, stream);
		if (*buf) {
			break;
		}

		if (!blocking) {
			retval = -EAGAIN;
			break;
		}

		/* Run Decoder Work Queue to get free buffers and wait */
               	dec_schedule_work(itv);
		schedule();

		if (signal_pending(current)) {
			retval = -EINTR;
			break;
		}
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&stream->waitq, &wait);

	return retval;
}

static int ivtv_fill_dec_buffers(struct ivtv *itv, int type, const char *ubuf,
				 size_t count, int block) 
{
	struct ivtv_stream *stream = &itv->streams[type];
	struct ivtv_buffer *buf = NULL;
	int copybytes = 0, bytesread = 0, retval = 0;
	int allocated_bufs = 0;
	int offset = 0;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "ivtv_fill_dec_buffers, %zd bytes\n", count);

        /* Read what the user gives us. queue it for DMA after each buffer
	 * also enqueue partly-full buffers. */

	IVTV_DEBUG(IVTV_DEBUG_DMA,
		   "DEC: free_q: %d elements\n", 
		   atomic_read(&stream->free_q.elements));
	IVTV_DEBUG(IVTV_DEBUG_DMA,
		   "DEC: dma_q: %d elements\n", 
		   atomic_read(&stream->dma_q.elements));
	IVTV_DEBUG(IVTV_DEBUG_DMA,
		   "DEC: full_q: %d elements\n", 
		   atomic_read(&stream->full_q.elements));
	IVTV_DEBUG(IVTV_DEBUG_DMA,
		   "DEC: io_q: %d elements\n", 
		   atomic_read(&stream->io_q.elements));

	/* Does many writes, up to chosen DMA xfer size */
	while (count - bytesread > 0) {
		buf = ivtv_deq_buf_lock(&stream->io_q, stream);
                if (buf) {
			IVTV_DEBUG(IVTV_DEBUG_DMA, 
				"Got IO Buff with %d bytes.\n",
			         buf->buffer.bytesused);

                        offset = buf->buffer.bytesused;
                } else {
			IVTV_DEBUG(IVTV_DEBUG_DMA, "Grabing a FREE Buf...\n");
			retval = ivtv_get_dec_buffer(itv, 
					stream, &buf, 0, block);

			if (retval)
				return retval;

			IVTV_DEBUG(IVTV_DEBUG_DMA, 
				"Got FREE Buff with %u bytes, clearing...\n",
			         buf->buffer.bytesused);
			buf->buffer.bytesused = 0;
			offset = 0;
		}

		/* bytes left to send > bytes in a free buffer */
		if (count - bytesread > (IVTV_DMA_DEC_BUF_SIZE - offset)) {
			copybytes = (IVTV_DMA_DEC_BUF_SIZE - offset);
		} else {
			copybytes = count - bytesread;
		}

		/* copy data */
		IVTV_DEBUG(IVTV_DEBUG_DMA, "copying %d bytes to 0x%08x\n",
			   	    copybytes, (int)buf->buffer.m.userptr);

		/* Make sure we really got all the user data */	
		if (copy_from_user(((char *)buf->buffer.m.userptr+offset),
			(char *)((uintptr_t)ubuf + bytesread), copybytes)) {
			ivtv_enq_buf_lock(&stream->free_q, buf, stream);
			return -EFAULT;
		}

		bytesread += copybytes;
		buf->buffer.bytesused += copybytes;

		/* Send DMA or IO Queue Buffer */
		if (buf->buffer.bytesused == IVTV_DMA_DEC_BUF_SIZE) {
			int y;

        		/* Byteswap IO buffer */ 
                	for (y = 0; y < buf->buffer.bytesused; y += 4) {
                       		swab32s((u32 *)	
					((uintptr_t)buf->buffer.m.userptr + y));
			}

			/* Send to the FULL Queue */
			ivtv_enq_buf_lock(&stream->full_q, buf, stream);
			allocated_bufs++;

        		/* Start DMA or Push it ahead when stalled */
			if (atomic_read(&itv->streams[type].full_q.elements) >= 
				(IVTV_DMA_MAX_XFER/IVTV_DMA_DEC_BUF_SIZE) &&
				test_bit(IVTV_F_S_NEEDS_DATA, 
					&itv->streams[type].s_flags)) 
			{
        			dec_schedule_work(itv);
			}
		} else {
			/* Send to the IO Queue */
			ivtv_enq_buf_lock(&stream->io_q, buf, stream);
		}
	}

	return bytesread;
}

ssize_t ivtv_v4l2_write(struct file *filp, const char *buf, size_t count, loff_t *pos)
{
	struct ivtv_open_id *id = filp->private_data;
        struct ivtv *itv = id->itv;
	struct ivtv_stream *stream;
	int bytes_written = 0, ret = 0;
	int type = id->type;
	int blocking = !(filp->f_flags & O_NONBLOCK);
	
	if (type == IVTV_ENC_STREAM_TYPE_RAD ||
	    type == IVTV_ENC_STREAM_TYPE_YUV ||
	    type == IVTV_ENC_STREAM_TYPE_VBI ||
	    type == IVTV_ENC_STREAM_TYPE_PCM ||
	    type == IVTV_DEC_STREAM_TYPE_VBI ||
	    type == IVTV_ENC_STREAM_TYPE_MPG) {
		/* you cannot read from these stream types. */
		return -EPERM;
	}
	
	stream = &itv->streams[type];

	if (test_bit(IVTV_F_I_PASSTHROUGH, &itv->i_flags))
		return -EBUSY;

	/* Try to claim this stream */
	if (ivtv_claim_stream(id, type))
		return -EBUSY;

	/* This stream does not need to start any decoding */
	if (type == IVTV_DEC_STREAM_TYPE_VOUT) {
		return ivtv_write_vbi(itv, buf, count);
	}

	/* Test if we have to start decoding */
	if (!test_and_set_bit(IVTV_F_S_DECODING, &stream->s_flags)) {
		down(&itv->streams_lock);
	    	if (ivtv_start_v4l2_decode_stream(itv, type) != 0) {
			IVTV_DEBUG(IVTV_DEBUG_ERR, 
					"Failed to start decoding for stream %d\n", type);
			/* failure, clean up */
			clear_bit(IVTV_F_S_DECODING, &stream->s_flags);
			up(&itv->streams_lock);
			return -EIO;
		}
		up(&itv->streams_lock);
	}

	/* do all the work */

	IVTV_DEBUG(IVTV_DEBUG_INFO, "ivtv_write\n");

	/* All data must be aligned to 256 bytes, so anything less
	   is just returned. */
	while (count - bytes_written > 0) { /* completely use up user data 
					 * before returning */

		/* Are we stoping decoding, then lets bail */
		if (test_bit(IVTV_F_S_STREAMOFF, &stream->s_flags))
			break;

		/* buffer the data - this may block waiting on free buffers */
		ret = ivtv_fill_dec_buffers(itv, type, buf + bytes_written, 
					    (count - bytes_written), blocking);

		if (ret < 0)
			break;

		bytes_written += ret;

	}

	if (ret > 0)
		*pos += ret;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "DEC: returning %d\n", bytes_written);
	return bytes_written ? bytes_written : ret;
}

unsigned int ivtv_v4l2_dec_poll(struct file *filp, poll_table *wait)
{
	struct ivtv_open_id *id = filp->private_data;
	struct ivtv *itv = id->itv;
	unsigned int mask = 0;

	/* Are we capturing */
	if (atomic_read(&itv->decoding) == 0) {
		mask |= POLLERR;
       		goto poll_exit;
	}

	/* add stream's waitq to the poll list */
	poll_wait(filp, &itv->streams[id->type].waitq, wait);

	/* Allow write if buffers are not full */
	if (atomic_read(&itv->streams[id->type].full_q.elements) < 
		itv->options.dec_mpg_buffers)
	{
		mask |= POLLOUT | POLLWRNORM; /* Writable */
	}

	IVTV_DEBUG(IVTV_DEBUG_INFO, "DEC: dec_poll returning 0x%x\n", mask);
	if (!mask) {
        	IVTV_DEBUG(IVTV_DEBUG_INFO, "Decoder fullq size %d\n",
		        atomic_read(&itv->streams[id->type].full_q.elements));
	        IVTV_DEBUG(IVTV_DEBUG_INFO, "Decoder freeq size %d\n",
			atomic_read(&itv->streams[id->type].free_q.elements));
	        IVTV_DEBUG(IVTV_DEBUG_INFO, "Decoder dmaq size %d\n",
			atomic_read(&itv->streams[id->type].dma_q.elements));
	}
poll_exit:	
	return mask;
}

unsigned int ivtv_v4l2_enc_poll(struct file *filp, poll_table *wait)
{
	struct ivtv_open_id *id = filp->private_data;
	unsigned int mask = 0;
	struct ivtv *itv = id->itv;
	int min_buf_fill;

        min_buf_fill = (IVTV_DMA_ENC_MAX_XFER/IVTV_DMA_ENC_BUF_SIZE);

	/* Are we capturing */
	if (atomic_read(&itv->capturing) == 0) {
		mask |= POLLERR;
       		goto poll_exit;
	}

	/* add stream's waitq to the poll list */
	poll_wait(filp, &id->itv->streams[id->type].waitq, wait);

	if (atomic_read(&itv->streams[id->type].full_q.elements) > 0)
	{
		mask |= POLLIN | POLLRDNORM;
	}

	IVTV_DEBUG(IVTV_DEBUG_INFO, "ENC: enc_poll returning 0x%x\n", mask);
poll_exit:
	if (!mask) {
        	IVTV_DEBUG(IVTV_DEBUG_INFO, "Encoder fullq size %d\n",
		        atomic_read(&itv->streams[id->type].full_q.elements));
	        IVTV_DEBUG(IVTV_DEBUG_INFO, "Encoder freeq size %d\n",
		        atomic_read(&itv->streams[id->type].free_q.elements));
	        IVTV_DEBUG(IVTV_DEBUG_INFO, "Encoder dmaq size %d\n",
	    	        atomic_read(&itv->streams[id->type].dma_q.elements));
	}
	return mask;
}

int ivtv_v4l2_close(struct inode *inode, struct file *filp)
{
	struct ivtv_open_id *id = filp->private_data;
	struct ivtv_stream *stream, *vbi_stream;
	struct ivtv *itv;
	
	IVTV_DEBUG(IVTV_DEBUG_INFO, "v4l2 close\n");
	
	if (NULL == id) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "invalid id on v4l2 close\n");
		return -ENODEV;
	}

	itv = id->itv;

	stream = &itv->streams[id->type];

	/* Easy case first: this stream was never claimed by us */
	if (stream->id != id->open_id) {
		kfree(id);
		return 0;
	}
	    
	/* 'Unclaim' this stream */
	
	/* Stop radio */
	if (id->type == IVTV_ENC_STREAM_TYPE_RAD) {
		struct video_channel vc = { 0 };

		/* We're going to pause, change inputs and frequencies, and
		   resume, so make sure no one else is doing similar things. */
		down(&itv->streams_lock);
		// Closing radio device, return to TV mode
		vc.norm = std2norm(itv->std);
		mute_and_pause(itv);
		/* Switch tuner to TV */
		ivtv_tuner(itv, VIDIOCSCHAN, &vc);
		/* Mark that the radio is no longer in use */
		clear_bit(IVTV_F_I_RADIO_USER, &itv->i_flags);
		/* Select TV frequency */
		ivtv_tuner(itv, VIDIOCSFREQ, &itv->freq_tv);
		/* Select correct audio input (i.e. TV tuner or Line in) */
		ivtv_audio_set_io(itv);
		/* Start automatic sound detection (if supported) */
		ivtv_audio_freq_changed(itv);
		/* Done! Unmute and continue. */
		unmute_and_resume(itv, 1);
		up(&itv->streams_lock);
		ivtv_release_stream(itv, id->type);
		kfree(id);
		return 0;
	}

	down(&itv->streams_lock);
	/* Stop decoding */
	if (test_bit(IVTV_F_S_DECODING, &stream->s_flags)) {
		IVTV_DEBUG(IVTV_DEBUG_INFO, "close stopping decode\n");
		ivtv_stop_v4l2_decode_stream(itv, id->type);
		clear_bit(IVTV_F_S_DECODING, &stream->s_flags);
		ivtv_flush_queues_lock(itv, id->type);
	}

	/* Stop capturing */
	if (test_bit(IVTV_F_S_CAPTURING, &stream->s_flags)) {
		vbi_stream = &itv->streams[IVTV_ENC_STREAM_TYPE_VBI];

		IVTV_DEBUG(IVTV_DEBUG_INFO, "close stopping capture\n");
		ivtv_stop_capture(itv, id->type);
		clear_bit(IVTV_F_S_CAPTURING, &stream->s_flags);

		/* Special case: a running VBI capture for VBI insertion
		   in the mpeg stream. Need to stop that too. */
		if (id->type == IVTV_ENC_STREAM_TYPE_MPG &&
		    vbi_stream->id == id->open_id) {
			ivtv_stop_capture(itv, IVTV_ENC_STREAM_TYPE_VBI);
			clear_bit(IVTV_F_S_CAPTURING, &vbi_stream->s_flags);
			clear_bit(IVTV_F_S_NO_READ, &vbi_stream->s_flags);
			ivtv_flush_queues_lock(itv, IVTV_ENC_STREAM_TYPE_VBI);
			ivtv_release_stream(itv, IVTV_ENC_STREAM_TYPE_VBI);
		}
		ivtv_flush_queues_lock_bh(itv, id->type);
	}
	up(&itv->streams_lock);
	ivtv_release_stream(itv, id->type);

	kfree(id);
	return 0;
}


int ivtv_v4l2_open(struct inode *inode, struct file *filp)
{
	int x, y = 0;
	struct ivtv_open_id *item;
	struct ivtv *itv = NULL;
	struct ivtv_stream *stream = NULL;
	struct ivtv_stream *vbi_stream = NULL;
	int minor = MINOR(inode->i_rdev);
	int err = 0;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "v4l2 open on minor %d\n", minor);

	/* Find which card this open was on */
	spin_lock(&ivtv_cards_lock);
	for (x = 0; itv == NULL && x < ivtv_cards_active; x++) {
		/* find out which stream this open was on */
		for (y = 0; y < ivtv_cards[x].streamcount; y++) {
			stream = &ivtv_cards[x].streams[y];
			IVTV_DEBUG(IVTV_DEBUG_INFO, "current minor %d\n",
				       stream->v4l2dev->minor);
			if (stream->v4l2dev->minor == minor) {
				itv = &ivtv_cards[x];
				break;
			}
		}
	}
	spin_unlock(&ivtv_cards_lock);

	if (itv == NULL) {
		/* Couldn't find a device registered on that minor, shouldn't happen! */
		IVTV_KERN_ERR("No ivtv device found on minor %d!\n", minor);
		return -ENXIO;
	}

	IVTV_DEBUG(IVTV_DEBUG_INFO, "opened card # %d, stream %d\n", x, y);

	// Allocate memory
	item = kmalloc(sizeof(struct ivtv_open_id), GFP_KERNEL);
	if (NULL == item) {
		IVTV_DEBUG(IVTV_DEBUG_ERR,"nomem on v4l2 open\n");
		return -ENOMEM;
	}
	item->itv = itv;
	item->type = y;
	
	down(&itv->streams_lock);
	if (!atomic_read(&stream->allocated_buffers)) {
		/* first time open, try to allocate buffer memory */
		err = ivtv_stream_alloc(itv, y);
	}
	if (!err && y == IVTV_DEC_STREAM_TYPE_MPG) {
		/* This is required when decoding an MPG stream. It's used
		   to handle embedded VBI data. */
		vbi_stream = &itv->streams[IVTV_DEC_STREAM_TYPE_VBI];
		if (!atomic_read(&vbi_stream->allocated_buffers)) {
			/* first time open, try to allocate buffer memory */
			err = ivtv_stream_alloc(itv, IVTV_DEC_STREAM_TYPE_VBI);
		}
        	clear_bit(IVTV_F_I_DEC_YUV, &itv->i_flags);
	} else if (!err && y == IVTV_DEC_STREAM_TYPE_YUV)
        	set_bit(IVTV_F_I_DEC_YUV, &itv->i_flags);

	if (!err && y == IVTV_ENC_STREAM_TYPE_MPG) {
		/* This is required when encoding an MPG stream. It's used
		   to handle embedded VBI data. */
		vbi_stream = &itv->streams[IVTV_ENC_STREAM_TYPE_VBI];
		if (!atomic_read(&vbi_stream->allocated_buffers)) {
			/* first time open, try to allocate buffer memory */
			err = ivtv_stream_alloc(itv, IVTV_ENC_STREAM_TYPE_VBI);
		}
	} 

	item->open_id = itv->open_id++;
	up(&itv->streams_lock);
	if (err) {
		kfree(item);
		return err;
	}
	
	filp->private_data = item;
	
	if (item->type == IVTV_ENC_STREAM_TYPE_RAD) {
		/* Try to claim this stream */
		if (ivtv_claim_stream(item, item->type)) {
			/* No, it's already in use */
			kfree(item);
			return -EBUSY;
		}

		/* We're going to pause, change inputs and frequencies, and
		   resume, so make sure no one else is doing similar things. */
		down(&itv->streams_lock);
		/* We have the radio */
		mute_and_pause(itv);
		/* Switch tuner to radio */
		ivtv_tuner(itv, AUDC_SET_RADIO, 0);
		/* Mark that the radio is being used. */
		set_bit(IVTV_F_I_RADIO_USER, &itv->i_flags);
		/* Select radio frequency */
		ivtv_tuner(itv, VIDIOCSFREQ, &itv->freq_radio);
		/* Select the correct audio input (i.e. radio tuner) */
		ivtv_audio_set_io(itv);
		/* Start automatic sound detection (if supported) */
		ivtv_audio_freq_changed(itv);
		/* Done! Unmute and continue. */
		unmute_and_resume(itv, 1);
		up(&itv->streams_lock);
	}

	return 0;
}

static u32 ivtv_pause_encoder(struct ivtv *itv, int cmd)
{
	u32 data[IVTV_MBOX_MAX_DATA], result=0;
	int x;

        /* Check if needed to resume */
	switch (cmd) {
		case 0: /* pause */
			break;
		case 1: /* unpause */
			break;
		default: /* wrong input */
			return -EIO;
	}	

        data[0] = 0; /* 0 = pause, 1 = unpause */
	if (cmd) data[0] = 1;

        x = ivtv_api(itv, itv->enc_mbox, &itv->enc_msem, IVTV_API_PAUSE_ENCODER,
		     &result,1, &data[0]);

	return result;
}

void mute_and_pause(struct ivtv *itv)
{
	int dig;

	/* streams_lock must be held */
	IVTV_ASSERT(ivtv_sem_count(&itv->streams_lock) <= 0);

	/* Mute sound to avoid pop */
	ivtv_audio_set_mute(itv, 1);

	if (test_bit(IVTV_F_I_PASSTHROUGH, &itv->i_flags)) return;

	if (atomic_read(&itv->capturing) > 0) {
		if (0 != ivtv_pause_encoder(itv, 0)) 
			IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"Mute: Error pausing stream\n");

       		IVTV_DEBUG(IVTV_DEBUG_INFO, "Disabling digitizer\n");
       		dig = 0;
      		ivtv_saa7115(itv, DECODER_ENABLE_OUTPUT, &dig);
	}
}

void unmute_and_resume(struct ivtv *itv, int sleep)
{
	int x, dig;
	int reset_trys = 100; /* 1 second */

	/* streams_lock must be held */
	IVTV_ASSERT(ivtv_sem_count(&itv->streams_lock) <= 0);

	if (!test_bit(IVTV_F_I_PASSTHROUGH, &itv->i_flags) &&
			(atomic_read(&itv->capturing) > 0)) {
		/* this seems to be needed when in mid encode, instead of init
        	   refresh input (no args) */
        	x = ivtv_vapi(itv, IVTV_API_REFRESH_INPUT, 0);
        	if (x) IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"Error refreshing input. Code %d\n\n",x);

       		IVTV_DEBUG(IVTV_DEBUG_INFO, "Enabling digitizer\n");
       		dig = 1;
       		ivtv_saa7115(itv, DECODER_ENABLE_OUTPUT, &dig);

		if (0 != ivtv_pause_encoder(itv, 1)) 
			IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"Unmute: Error unpausing stream\n");

       		IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"Finished with reset in %d ms\n", reset_trys);

		ivtv_sleep_timeout(HZ/10, 0);
	}

	/* Unmute */
	ivtv_audio_set_mute(itv, 0);
       	IVTV_DEBUG(IVTV_DEBUG_ERR, "Finished with Mute\n");
}

int ivtv_get_timing_info(struct ivtv *itv, struct ivtv_ioctl_framesync *info)
{
	u32 ret;

	int suicidecounter = 0;
	
	memset(info, 0x00, sizeof(struct ivtv_ioctl_framesync));

	/* Occasionally, we'll get a weird, invalid number for 
	 * frames played. fortunately, it sets the SCR timestamp to 0
	 * in that case, which it never is otherwise. cool, huh */
	while (info->scr == 0) { /* eliminate bogus values, FIXME ugly */
		u32 data[IVTV_MBOX_MAX_DATA];

		ret = ivtv_api_result(itv, IVTV_API_DEC_TIMING_INFO, 0, data);
		if (ret) {
			IVTV_DEBUG(IVTV_DEBUG_ERR, "DEC: err sending timing info\n");
			return ret;
		}

		info->frame	= data[0];
		info->pts	= ((u64)data[2] << 32) | (u64)data[1];
		info->scr	= ((u64)data[4] << 32) | (u64)data[3];

		if (suicidecounter++ > 10) { /* endless loops are bad! */
			IVTV_DEBUG(IVTV_DEBUG_ERR,"Timeout getting frames played\n");
			return -1;
		}
		if (info->scr == 0)
			ivtv_sleep_timeout(HZ/50, 1);
	}

	return 0;
}
