/*
    interrupt handling
    Copyright (C) 2003-2004  Kevin Thayer <nufan_wfk at yahoo.com>

    Copyright (C) 2004  Chris Kennedy ckennedy@kmos.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ivtv-driver.h"
#include "ivtv-dma.h"
#include "ivtv-firmware.h"
#include "ivtv-fileops.h"
#include "ivtv-queue.h"
#include "ivtv-irq.h"
#include "ivtv-mailbox.h"
#include "ivtv-vbi.h"
#include "ivtv-reset.h"

static void ivtv_DMA_done(struct ivtv *itv);

/* ceiling function for ints.. */
static int ivtv_ceil(int x, int y)
{
	int floor = (int)(x / y);

	if ((floor * y) < x)
		return floor + 1;
        return floor;
}

static void ivtv_show_debug_flags(struct ivtv *itv)
{
	int y;

	printk(KERN_DEBUG "ivtv: i_flags=%lx", itv->i_flags);
	for (y = 0; y < itv->streamcount; y++)
		printk(", %d s_sflags=%lx", y, itv->streams[y].s_flags);
	printk("\n");
}

static void ivtv_swap_copy(const char *buf, const char *ubuf, size_t count)
{
	u32 *src, *dst;

	src = (u32 *)ubuf;
	dst = (u32 *)buf;

#ifdef CONFIG_X86
	while ((unsigned long)src <= (unsigned long)ubuf + count) { /* byteswap while copying */
		__asm__ __volatile__ ("bswap %0" :
				      "=r" (*dst) :
				      "0" (*src) );
		src++; dst++;
	}
#else
	{
		int y;

		/* Old slow memcpy then swab */
		memcpy((void *)buf, (void *)ubuf, count);
		for (y = 0; y < count; y += 4) {
			swab32s((u32 *)(buf + y));
		}
	}
#endif
}

IRQRETURN_T ivtv_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	struct ivtv *itv = (struct ivtv *)dev_id;
	u32 combo;
	u32 stat;

	/* get contents of irq status register*/
	stat = readl(itv->reg_mem + IVTV_REG_IRQSTATUS);

	combo = ~itv->irqmask & stat;

	if (0 == combo) {
		/* wasn't for us*/
		return IRQ_NONE;
	}

	/* Clear out IRQ */
	writel(combo, (itv->reg_mem + IVTV_REG_IRQSTATUS));
        
	IVTV_DEBUG(IVTV_DEBUG_IRQ, 
		"======= valid IRQ bits: 0x%08x ======\n", combo);

	if (combo & IVTV_IRQ_DEC_DMA_ERR) {
		int type;

		/* Return if in middle of reload */
		if (test_bit(DEC_DMA_ERR, &itv->r_flags))
			return IRQ_HANDLED;

		/* Get DMA Status from card */
		stat = readl(itv->reg_mem + IVTV_REG_DMASTATUS);
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"Decoder DMA Error 0x%08x\n", stat);

		/* Mask Decoder DMA Error */
		ivtv_set_irq_mask(itv, IVTV_IRQ_DEC_DMA_ERR);

        	/* YUV or MPG decoding stream */
        	if (test_bit(IVTV_F_I_DEC_YUV, &itv->i_flags))
                	type = IVTV_DEC_STREAM_TYPE_YUV;
        	else
                	type = IVTV_DEC_STREAM_TYPE_MPG;

		/* Unmap OSD DMA */
		ivtv_unmap_user_dma(itv, &itv->streams[type].udma, 0);

		/* Unmap Enc DMA */
                set_bit (IVTV_F_T_ENC_DMA_DONE, &itv->t_flags);
		tasklet_schedule(&itv->dma_tq);
#if 1
		/* Flag a DMA Error and schedule firmware/streams reload */
		set_bit(FW_RESET_NEEDED, &itv->r_flags);
		set_bit(DEC_DMA_ERR, &itv->r_flags);

		/* Reload Firmware Work Queue */
		fw_schedule_work(itv);
#endif
	}

	/* Define each tasklets irq bits */
	#define IVTV_IRQ_ENC_DMA (IVTV_IRQ_ENC_DMA_COMPLETE| \
			  IVTV_IRQ_ENC_START_CAP| \
			  IVTV_IRQ_ENC_VBI_CAP| \
			  IVTV_IRQ_DEC_VBI_RE_INSERT|\
			  IVTV_IRQ_ENC_EOS)

	#define IVTV_IRQ_DEC_DMA (IVTV_IRQ_DEC_DMA_COMPLETE| \
			  IVTV_IRQ_DEC_DATA_REQ)

        if (combo & IVTV_IRQ_ENC_DMA) {
		int run_tasklet = 0;

		/* Encoder */
                if (combo & IVTV_IRQ_ENC_DMA_COMPLETE) {
                        /* set the dma done flag */
                        IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ ENC DMA Complete\n");
                        set_bit (IVTV_F_T_ENC_DMA_DONE, &itv->t_flags);
			run_tasklet = 1;
                }

                if (combo & IVTV_IRQ_ENC_START_CAP) {
                        /* set dma needed flag */
                        IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ ENC DMA\n");
                        set_bit (IVTV_F_T_ENC_DMA_NEEDED, &itv->t_flags);
			run_tasklet = 1;
                }

                if (combo & IVTV_IRQ_ENC_VBI_CAP) {
                        /* set vbi dma needed flag */
                        IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ ENC VBI DMA\n");
                        set_bit(IVTV_F_T_ENC_VBI_DMA, &itv->t_flags);
			run_tasklet = 1;
                }
 
                if (combo & IVTV_IRQ_DEC_VBI_RE_INSERT) {
                        IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ DEC VBI Reinsertion\n");
                        set_bit(IVTV_F_T_DEC_VBI_INSERT, &itv->t_flags);
			run_tasklet = 1;
                }

		if (combo & IVTV_IRQ_ENC_EOS) {
			IVTV_DEBUG(IVTV_DEBUG_IRQ, "Encoder End Of Stream\n");
			set_bit(IVTV_F_I_EOS, &itv->i_flags);

			wake_up(&itv->cap_w);
		}

		if (run_tasklet) {
                	tasklet_schedule (&itv->dma_tq);
		}
	}

        if (combo & IVTV_IRQ_DEC_DMA) {
		int type = IVTV_DEC_STREAM_TYPE_MPG;

        	/* YUV or MPG */
        	if (test_bit(IVTV_F_I_DEC_YUV, &itv->i_flags))
                	type = IVTV_DEC_STREAM_TYPE_YUV;
        	else
                	type = IVTV_DEC_STREAM_TYPE_MPG;

                if (combo & IVTV_IRQ_DEC_DATA_REQ) {
                        IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ DEC Data Request\n");
			set_bit(IVTV_F_S_NEEDS_DATA, 
				&itv->streams[type].s_flags);

			/* Decoder Work Queue */
			if (atomic_read(&itv->streams[type].full_q.elements) >=
				(IVTV_DMA_MAX_XFER/IVTV_DMA_DEC_BUF_SIZE))
			{
				dec_schedule_work(itv);
			}

                }

		/* Decoder */
                if (combo & IVTV_IRQ_DEC_DMA_COMPLETE) {
                        IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ DEC DMA Done\n");
			set_bit(IVTV_F_S_DMA_DONE, 
				&itv->streams[type].s_flags);

			/* Decoder DMA Done, clears flag */
			ivtv_dec_DMA_done(itv, &itv->streams[type]);
                }
        }

	/* Decoder Vertical Sync */
	if (combo & IVTV_IRQ_DEC_VSYNC) {
		IVTV_DEBUG(IVTV_DEBUG_IRQ, "IRQ Decoder Vsync\n");
		itv->dec_dma_stat.vsync_count++;
		set_bit(IVTV_F_T_DEC_VSYNC, &itv->t_flags);
		tasklet_schedule (&itv->dec_vsync_tq);
	}

	if (combo & IVTV_IRQ_ENC_VIM_RST) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "VIM Restart\n");

		/* Signal action was done */
                set_bit(IVTV_F_I_DIG_RST, &itv->i_flags);
	}

	if (combo & IVTV_IRQ_DEC_COPY_PROTECT) {
		IVTV_DEBUG(IVTV_DEBUG_IRQ, "Copy Protection\n");
	}

	if (combo & IVTV_IRQ_DEC_AUD_MODE_CHG) {
		IVTV_DEBUG(IVTV_DEBUG_IRQ, "Audio mode change\n");
	}

	if (combo & IVTV_IRQ_DEC_IFRAME_DONE) {
		IVTV_DEBUG(IVTV_DEBUG_IRQ, "I-Frame done\n");
	}

	return IRQ_HANDLED;
}

void ivtv_dec_VSYNC_tasklet (unsigned long arg) 
{
        struct ivtv *itv = (struct ivtv *) arg;
 
	if (test_and_clear_bit(IVTV_F_T_DEC_VSYNC, &itv->t_flags)) {
		u32 *data = itv->dec_mbox[IVTV_MBOX_FIELD_DISPLAYED].data;
		u32 newframe = readl(&data[0]);
		u64 newpts = ((u64)readl(&data[2]) << 32) | (u64)(readl(&data[1]));
		u64 newscr = ((u64)readl(&data[4]) << 32) | (u64)(readl(&data[3]));

		itv->dec_timestamp.pts = newpts;
		itv->dec_timestamp.scr = newscr;
		if (newframe != itv->dec_timestamp.frame) {
			itv->dec_timestamp.frame = newframe;
			wake_up_interruptible(&itv->vsync_w);
		}

		IVTV_DEBUG(IVTV_DEBUG_IRQ, 
			"ivtv_dec_VSYNC_tasklet: frames %d, pts %ld, scr %ld\n",
		       	itv->dec_timestamp.frame, 
			(long int)itv->dec_timestamp.pts,
		       	(long int)itv->dec_timestamp.scr);
	}
}

static void ivtv_DMA_done(struct ivtv *itv) 
{
	u32 result;
	int y,stmtype=-1;
	struct ivtv_stream *stream=NULL;
	struct ivtv_buffer *buf;
	int freed = 0;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "DMA Done tasklet\n");

	for (y=0; y < itv->streamcount; y++) {
		if (itv->streams[y].dma != PCI_DMA_FROMDEVICE) continue;
		if (test_bit(IVTV_F_S_DMAP, &itv->streams[y].s_flags)) {
			if (stmtype != -1) {
				IVTV_DEBUG(IVTV_DEBUG_INFO, "Got duplicate DMA-done %d %08lx %08lx\n",
					       	y,itv->streams[stmtype].trans_id, itv->streams[y].trans_id);
				if (itv->streams[y].trans_id < itv->streams[stmtype].trans_id) {
					stmtype = y;
				}
			}
			else {
				IVTV_DEBUG(IVTV_DEBUG_INFO, "Got DMA-done %d\n", y);
				stmtype = y;
			}
		}
	}

	if (stmtype < 0) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "Got DMA-done, but not expecting one\n");
		ivtv_show_debug_flags(itv);
		return;
	}

	/* check DMA status register */
	result = readl(itv->reg_mem + IVTV_REG_DMASTATUS);

	if (!(result & IVTV_DMA_SUCCESS_WRITE)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR,
			"ENC DMA not finished. Stream %d. Result=0x%08x\n", 
			stmtype, result);
		/* Return and wait till really done */	
		return;
	}

	if (!test_and_clear_bit(IVTV_F_S_DMAP, &itv->streams[stmtype].s_flags)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"Got DMA-done for stream %d, but DMA bit was cleared\n",
			stmtype);
		ivtv_show_debug_flags(itv);
		return;
	}

	stream = &itv->streams[stmtype];

	/* DMA was fine if we made it this far */
	mod_timer(&itv->enc_timeout, jiffies + ENC_DMA_TIMEOUT);

        /* Unmap Buffers SG List */
        pci_unmap_sg(itv->dev, stream->SGlist, stream->SGlen, stream->dma);
	stream->SGlen = 0;
	sg_dma_address(stream->SGlist) = IVTV_DMA_UNMAPPED;

	/* remove from dma_pending queue */
	while ((buf = ivtv_deq_buf_lock_bh(&stream->dma_q, stream))) {
		int buffer_size;
		if (stmtype == IVTV_ENC_STREAM_TYPE_VBI ||
			stmtype == IVTV_DEC_STREAM_TYPE_VBI ||
			stmtype == IVTV_DEC_STREAM_TYPE_VOUT)
		{
			buffer_size = IVTV_DMA_VBI_BUF_SIZE;
		} else
			buffer_size = IVTV_DMA_ENC_BUF_SIZE;

		/* flag byteswap ABCD -> DCBA for MPG & VBI data outside irq */
		if (stmtype == IVTV_ENC_STREAM_TYPE_MPG)
		{
			set_bit (IVTV_F_B_NEED_BUF_SWAP, &buf->b_flags);
		}

		/* put in the 'done' queue */
		ivtv_enq_buf_lock_bh(&stream->full_q, buf, stream);
		freed++;
	}

        /* Unmap SG Array */
	if (freed) {
		IVTV_ASSERT(stream->SG_handle != IVTV_DMA_UNMAPPED);
        	pci_unmap_single(itv->dev, stream->SG_handle,
                	(sizeof(struct ivtv_SG_element) * stream->SG_length), stream->dma);
		stream->SG_handle = IVTV_DMA_UNMAPPED;
		stream->SG_length = 0;
	}

	/*wake up client*/
	wake_up(&stream->waitq);
}


static int ivtv_ignore_DMA_req(struct ivtv *itv, u32 type)
{
	int ret = 0;
	int cmddirection = 0;

	if (type < IVTV_DEC_STREAM_TYPE_MPG)
		cmddirection = IVTV_API_SCHED_DMA_TO_HOST;
	else
		cmddirection = IVTV_API_DEC_DMA_FROM_HOST;

       	if (ivtv_vapi(itv, cmddirection, 3, 0, 0, type)) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "error sending DMA info\n");
		ret = -EIO;
	}

	if (!ret && type < itv->streamcount)
		set_bit(IVTV_F_S_DMAP, &itv->streams[type].s_flags);

	return ret;
}

/* Must be spinlocked */
static int gather_free_buffers(struct ivtv *itv, int streamtype,
		struct list_head *free_list, int bufs_needed)
{
	struct ivtv_stream *st = &itv->streams[streamtype];
	struct ivtv_buffer *buf;
	int bufs_received;
	int x;

	/* gather the needed buffers first, so we don't have to bail
	 * in mid-air. put them on a list on the stack */
	for (x = 0; x < bufs_needed; x++) {
		buf = ivtv_deq_buf_lock_bh(&st->free_q, st);
		if (!buf)
			break;

		list_add_tail(&buf->list, free_list);
	}

	/* How many did we actually get */
	bufs_received = x;
	
	/* damn, failed */
	if (bufs_received != bufs_needed) {
		IVTV_DEBUG(IVTV_DEBUG_ERR,
			"needed %d bufs for stream %d, received %d differ by (%d) bufs\n",
			bufs_needed, streamtype, bufs_received, 
			(bufs_needed - bufs_received));

		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"SCHED: free_q: %d elements\n", 
			atomic_read(&st->free_q.elements));
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"SCHED: dma_q: %d elements\n", 
			atomic_read(&st->dma_q.elements));
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"SCHED: full_q: %d elements\n", 
			atomic_read(&st->full_q.elements));
		
		while (!list_empty(free_list)) {
			buf = list_entry(free_list->next, struct ivtv_buffer, list);
			list_del_init(&buf->list);
			ivtv_enq_buf_lock_bh(&st->free_q, buf, st);
		}

		return 0;
	}
	return bufs_received;
}

/* FIXME this function is getting too long. split it up? */
static void ivtv_sched_DMA(struct ivtv *itv) 
{
	u32 data[IVTV_MBOX_MAX_DATA], result;
	u32 type, size, offset, streamtype, sg_size = 0;
	u32 UVsize=0, UVoffset=0;
	u64 pts_stamp = 0;
	struct ivtv_stream *st;
	int x, i, bufs_needed, bufs_received;
	int uvflag=0;
	struct ivtv_buffer *buf;
	LIST_HEAD(free_list);
	long sequence;
	struct scatterlist *sg;

        /* check DMA status register */
        result = readl(itv->reg_mem + IVTV_REG_DMASTATUS);

        if (!(result & IVTV_DMA_SUCCESS_WRITE)) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "ENC DMA not finished. Sched: Result=0x%08x\n", result);
                /* Return and wait till really done */
                return;
        }

	IVTV_DEBUG(IVTV_DEBUG_INFO, "Sched DMA tasklet\n");

	/* Get DMA destination and size arguments from card */
	x = ivtv_api_getresult_nosleep(&itv->enc_mbox[9], &result, &data[0]);
	if (x) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "error:%d getting DMA info\n", x);
		return;
	}
	
	type = data[0];

	switch (type) {
		case 0:  /* MPEG */
			offset = data[1];
			size   = data[2];
			IVTV_DEBUG(IVTV_DEBUG_INFO,
				   "DMA/MPG type 0x%08x,size 0x%08x,offset 0x%08x\n",
	 			   type, size, offset);
			bufs_needed = ivtv_ceil(size, IVTV_DMA_ENC_BUF_SIZE);
			streamtype = IVTV_ENC_STREAM_TYPE_MPG;
			break;

		case 1: /* YUV */
			offset		= data[1];
			size		= data[2];
			UVoffset	= data[3];
			UVsize		= data[4];
			pts_stamp	= ((u64)data[5] << 32) | data[6];
			IVTV_DEBUG(IVTV_DEBUG_INFO,
				   "DMA/YUV type 0x%08x,Ysize 0x%08x,Yoffset 0x%08x,"
				   "UVsize 0x%08x,UVoffset 0x%08x,PTS 0x%09llx\n",
	 			   type, size, offset, UVsize, UVoffset, pts_stamp);
			bufs_needed = ivtv_ceil(size, IVTV_DMA_ENC_BUF_SIZE);
			bufs_needed += ivtv_ceil(UVsize, IVTV_DMA_ENC_BUF_SIZE);
			streamtype = IVTV_ENC_STREAM_TYPE_YUV;
			break;

		case 2:  /* PCM (audio) */
			offset = data[1] + 12;
			size   = data[2] - 12;
			pts_stamp = *(u32 *)(itv->dec_mem + offset - 8);
			pts_stamp |= (u64)(*(u32 *)(itv->dec_mem + offset - 12)) << 32;
			IVTV_DEBUG(IVTV_DEBUG_INFO,
				"DMA/PCM type 0x%08x,size 0x%08x,offset 0x%08x "
				"PTS 0x%09llx\n", 
				type, size, offset, pts_stamp);
			bufs_needed = ivtv_ceil(size, IVTV_DMA_ENC_BUF_SIZE);
                        streamtype = IVTV_ENC_STREAM_TYPE_PCM;
			break;

		default:
			IVTV_DEBUG(IVTV_DEBUG_ERR,
				"DMA/UNKNOWN type 0x%08x, NOT SUPPORTED\n", 
				type);
			ivtv_ignore_DMA_req(itv, type);
			return;
	}
	
	/* FIXME should check for null on the stream element */
	if (streamtype >= itv->streamcount) {
		IVTV_DEBUG(IVTV_DEBUG_ERR,
			"No stream handler for type %d\n", streamtype);
		ivtv_ignore_DMA_req(itv, streamtype);
		return;
	}

	st = &itv->streams[streamtype];
	st->pts = pts_stamp;

	if (test_bit(IVTV_F_S_NO_READ, &st->s_flags)) {
		return;
	}

        /* Make sure request will work, otherwise wasting our time */
        if (atomic_read(&st->free_q.elements) < bufs_needed) {
                set_bit(IVTV_F_S_OVERFLOW, &st->s_flags);
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "ivtv_sched_dma: needed %d bufs for stream %d, "
                        "received %d differ by (%d) bufs\n",
                        bufs_needed, streamtype,
                       atomic_read(&st->free_q.elements),
                        (bufs_needed - atomic_read(&st->free_q.elements)));

                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "SCHED: free_q: %d elements\n",
                        atomic_read(&st->free_q.elements));
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "SCHED: dma_q: %d elements\n",
                        atomic_read(&st->dma_q.elements));
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "SCHED: full_q: %d elements\n",
                        atomic_read(&st->full_q.elements));

                return;
        }

	bufs_received = 
		gather_free_buffers(itv, streamtype, &free_list, bufs_needed);

	if (bufs_received == 0) {
		set_bit(IVTV_F_S_OVERFLOW, &st->s_flags);
		return;
	}

        /* Set SGarray to all 0's */
        memset(st->SGarray, 0,
                        sizeof(struct ivtv_SG_element)*IVTV_DMA_SG_ENT);

        /* Set SGarray to all 0's */
        memset(st->SGlist, 0,
                        sizeof(*st->SGlist)*IVTV_DMA_SG_ENT);

	/* increment the sequence # */
	sequence = ++st->seq;

	for (x = 0; x < bufs_received; x++) {
		IVTV_DEBUG(IVTV_DEBUG_DMA, "size: %d 0x%08x\n", size, size);

		if ((size == 0) && (type == 1) && (uvflag == 0)) { /* YUV */
			/* process the UV section */
			offset	= UVoffset;
			size	= UVsize;
			uvflag	= 1;
		}

		/* extract the buffers we procured earlier */
		buf = list_entry(free_list.next, struct ivtv_buffer, list);
		list_del_init(&buf->list);

		buf->readpos = 0;
		buf->buffer.index = x;
		buf->buffer.sequence = sequence;
		buf->ts = jiffies;
		buf->b_flags = 0;

		if (size < IVTV_DMA_ENC_BUF_SIZE) {
			buf->buffer.bytesused = size;
			/* bytecount must be multiple of 0x100 (256)? */
			/* Seems to be already fine as is */
			st->SGlist[x].length = size;
			st->SGlist[x].offset = 0;
			size = 0;
		}
		else {

			buf->buffer.bytesused	= IVTV_DMA_ENC_BUF_SIZE;
			st->SGlist[x].length	= IVTV_DMA_ENC_BUF_SIZE;
			size		       -= IVTV_DMA_ENC_BUF_SIZE;

			st->SGlist[x].offset	= 0;
		}

		/* Attach Buffer to SG Array */
		st->SGlist[x].page =  
			virt_to_page((void *)buf->buffer.m.userptr);

		/* FIXME need to add pts stuff, index, etc. */
		if (test_bit(IVTV_F_S_NO_READ, &st->s_flags)) {
			ivtv_enq_buf_lock_bh(&st->free_q, buf, st);
		}
		else {
			ivtv_enq_buf_lock_bh(&st->dma_q, buf, st);
		}
	}

	/* Allocate Buffers SG List */
	st->SGlen = bufs_received = x;
	st->SG_length = pci_map_sg(itv->dev, st->SGlist, st->SGlen, st->dma);

	/* Setup SG Array with Buffer SG List */
	for (i = 0, sg = st->SGlist; i < st->SG_length; i++, sg++) {
		st->SGarray[i].src = offset;		/* Encoder Addr */
		st->SGarray[i].dst = sg_dma_address(sg);/* Buffer Addr */
		st->SGarray[i].size = sg_dma_len(sg);	/* Buffer size */
		sg_size += st->SGarray[i].size;		/* Size of SG List */
		offset += st->SGarray[i].size;		/* Increment Enc Addr */
	}

	IVTV_DEBUG(IVTV_DEBUG_DMA,
		"ENC: SGlen = 0x%08x, SG_length = 0x%08x, sg_size = 0x%08x\n",
			st->SGlen, st->SG_length, sg_size);

	/* This should wrap gracefully */
	st->trans_id = itv->trans_id++;

	/* Mark last buffer size for Interrupt flag */
	st->SGarray[st->SG_length-1].size |= 0x80000000;

	/* Map SG Array */
	if (st->SG_handle != IVTV_DMA_UNMAPPED) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, 
			"Encoder DMA not finished, still mapped!!!\n");
		IVTV_ASSERT(st->SG_handle == IVTV_DMA_UNMAPPED);
                pci_dma_sync_single(itv->dev, st->SG_handle,
                        sizeof(struct ivtv_SG_element) * st->SG_length,
                        st->dma);
	} else {
        	st->SG_handle = pci_map_single(itv->dev, st->SGarray,
                        (sizeof(struct ivtv_SG_element) * st->SG_length), 
			st->dma);
	}

	IVTV_DEBUG(IVTV_DEBUG_INFO, 
		"Sched dma: addr: 0x%08llx, SG_length: %d, array_size 0x%08x,"
		" type 0x%08x\n", (u64)st->SG_handle, 
		st->SG_length, sg_size, type);

	mod_timer(&itv->enc_timeout, jiffies + ENC_DMA_TIMEOUT);

	set_bit(IVTV_F_S_DMAP, &st->s_flags);
	ivtv_vapi(itv, IVTV_API_SCHED_DMA_TO_HOST, 3,
		st->SG_handle, sg_size, type);
}

/* FIXME this function is getting too long. split it up? */
static void ivtv_sched_VBI(struct ivtv *itv, int streamtype) 
{
	u32 size, offset;
	u64 pts_stamp;
	struct ivtv_stream *st;
	int x, bufs_needed, bufs_received;
	struct ivtv_buffer *buf;
	LIST_HEAD(free_list);
	long sequence;

	IVTV_DEBUG(IVTV_DEBUG_INFO, "Sched DMA tasklet\n");

	if (streamtype == IVTV_ENC_STREAM_TYPE_VBI) {
		size   = itv->vbi_enc_size;
		offset = itv->vbi_enc_start + itv->vbi_index * (size + 12) + 12;
		itv->vbi_index = (itv->vbi_index + 1) % itv->vbi_total_frames;
		pts_stamp = *(u32 *)(itv->enc_mem + offset - 4);
		pts_stamp |= (u64)(*(u32 *)(itv->enc_mem + offset - 8)) << 32;
		IVTV_DEBUG(IVTV_DEBUG_DMA,
			   "DMA/VBI type 0x%08x, size 0x%08x, offset 0x%08x,"
			   "PTS 0x%09llx\n", streamtype, size, offset, pts_stamp);
		itv->vbi_frame++;
	}
	else {
		offset = readl(itv->dec_mem + itv->vbi_dec_start);
		offset += itv->vbi_dec_start;
		size = readl(itv->dec_mem + itv->vbi_dec_start + 4);
		pts_stamp = 0;
		IVTV_DEBUG(IVTV_DEBUG_DMA,
			   "DMA/VBI-Reinsertion type 0x%08x, size 0x%08x, offset 0x%08x\n",
			   streamtype, size, offset);
	}
	
	st = &itv->streams[streamtype];
	st->pts = pts_stamp;

	bufs_needed = ivtv_ceil(size, IVTV_DMA_VBI_BUF_SIZE);

        /* Make sure request will work, otherwise wasting our time */
        if (atomic_read(&st->free_q.elements) < bufs_needed) {
                set_bit(IVTV_F_S_OVERFLOW, &st->s_flags);
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "ivtv_sched_vbi: needed %d bufs for stream %d, "
                        "received %d differ by (%d) bufs\n",
                        bufs_needed, streamtype,
                        atomic_read(&st->free_q.elements),
                        (bufs_needed - atomic_read(&st->free_q.elements)));

                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "SCHED: free_q: %d elements\n",
                        atomic_read(&st->free_q.elements));
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "SCHED: dma_q: %d elements\n",
                        atomic_read(&st->dma_q.elements));
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "SCHED: full_q: %d elements\n",
                        atomic_read(&st->full_q.elements));

                return;
        }

        bufs_received =
                gather_free_buffers(itv, streamtype, &free_list, bufs_needed);

	if (bufs_received == 0) {
		set_bit(IVTV_F_S_OVERFLOW, &st->s_flags);
		return;
	}

	/* increment the sequence # */
	sequence = ++st->seq;

	for (x = 0; x < bufs_received; x++) {
		IVTV_DEBUG(IVTV_DEBUG_INFO, "size: %d 0x%08x\n", size, size);

		/* extract the buffers we procured earlier */
		buf = list_entry(free_list.next, struct ivtv_buffer, list);
		list_del_init(&buf->list);

		buf->readpos = 0;
		buf->buffer.index = x;
		buf->buffer.sequence = sequence;
		buf->ts = jiffies;
		buf->b_flags = 0;

		if (size < IVTV_DMA_VBI_BUF_SIZE) {
			buf->buffer.bytesused = size;
			size = 0;
		}
		else {
			buf->buffer.bytesused = IVTV_DMA_VBI_BUF_SIZE;
			size -= IVTV_DMA_VBI_BUF_SIZE;
		}
		if (streamtype == IVTV_ENC_STREAM_TYPE_VBI) {
			ivtv_swap_copy((void *)buf->buffer.m.userptr, 
				(void *)(itv->enc_mem + offset), 
				buf->buffer.bytesused);
		} else {
			ivtv_swap_copy((void *)buf->buffer.m.userptr, 
				(void *)(itv->dec_mem + offset), 
				buf->buffer.bytesused);
		}
		ivtv_process_vbi_data(itv, buf, pts_stamp, streamtype);

		spin_lock_bh(&st->slock);
		if (test_bit(IVTV_F_S_NO_READ, &st->s_flags) || st->id == -1) {
			ivtv_enq_buf_nolock(&st->free_q, buf);
		}
		else {
			ivtv_enq_buf_nolock(&st->full_q, buf);
		}
		spin_unlock_bh(&st->slock);
	}
	if (test_bit(IVTV_F_S_NO_READ, &st->s_flags))
	       	return;
	wake_up(&st->waitq);
}
 
void ivtv_DMA_tasklet (unsigned long arg) 
{
        struct ivtv *itv = (struct ivtv *) arg;

	/* Encoder */
        do {
                if (test_and_clear_bit(IVTV_F_T_ENC_DMA_DONE, &itv->t_flags)) {
                        ivtv_DMA_done(itv);
                }
                if (test_and_clear_bit(IVTV_F_T_ENC_DMA_NEEDED, &itv->t_flags)) {
                        ivtv_sched_DMA(itv);
                }
                if (test_and_clear_bit(IVTV_F_T_ENC_VBI_DMA, &itv->t_flags)) {
                        ivtv_sched_VBI(itv, IVTV_ENC_STREAM_TYPE_VBI);
                }
                if (test_and_clear_bit(IVTV_F_T_DEC_VBI_INSERT, &itv->t_flags)) {
                        ivtv_sched_VBI(itv, IVTV_DEC_STREAM_TYPE_VBI);
                }
        } while (test_bit(IVTV_F_T_ENC_DMA_DONE, &itv->t_flags) ||
                 test_bit(IVTV_F_T_ENC_DMA_NEEDED, &itv->t_flags) ||
                 test_bit(IVTV_F_T_DEC_VBI_INSERT, &itv->t_flags) || 
                 test_bit(IVTV_F_T_ENC_VBI_DMA, &itv->t_flags));
}

/* only called from process context in decoder work queue*/
int  ivtv_dec_DMA_done_check(struct ivtv *itv, struct ivtv_stream *stream) 
{
	int freed = 0;
	struct ivtv_buffer *buf;

        /* Unmap Buffers SG List */
        pci_unmap_sg(itv->dev, stream->SGlist, stream->SGlen, stream->dma);
        stream->SGlen = 0;
        sg_dma_address(stream->SGlist) = IVTV_DMA_UNMAPPED;

        /* Free last DMA call */
        while ((buf = 
		ivtv_deq_buf_lock(&stream->dma_q, stream)) != NULL)
	{	
              buf->dma_handle = IVTV_DMA_UNMAPPED;
              buf->buffer.bytesused   = 0;
              buf->readpos = 0;
	      buf->ts = 0;	

              /* put in the 'done' queue */
              ivtv_enq_buf_lock(&stream->free_q, buf, stream);
              freed++;
        }

	/* Oddly less buffers than we sent? */
	if (freed != stream->SG_length)
                IVTV_DEBUG(IVTV_DEBUG_ERR, "DEC: Freed %d but have %d in DMA\n",
			 freed, atomic_read(&itv->dec_dma_stat.dma_buffers));

	/* Unmap SG Array */
	if (freed && stream->SG_length) {
		IVTV_ASSERT(stream->SG_handle != IVTV_DMA_UNMAPPED);
        	pci_unmap_single(itv->dev, stream->SG_handle,
                        (sizeof(struct ivtv_SG_element) * 
			stream->SG_length), stream->dma);
		stream->SG_handle = IVTV_DMA_UNMAPPED;
		stream->SG_length = 0;
		atomic_set(&itv->dec_dma_stat.dma_buffers, 0);
	}

        /* Check for Overflow, bogus */
        if (freed && test_and_clear_bit(IVTV_F_S_OVERFLOW, &stream->s_flags)) {
                IVTV_DEBUG(IVTV_DEBUG_ERR, "DEC: Buffer Overflow %d\n", freed);
        }

	return freed;
}

/* called from interrupt or process context */
int ivtv_dec_DMA_done(struct ivtv *itv, struct ivtv_stream *stream) {
	int retrys = 10;
        while (ivtv_dma_status(itv, 1, 0x04, 0) ||
                !ivtv_dma_status(itv, 1, 0x00, 0))
                                {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: (%d) waiting for DMA Done"
                                 ", DMA busy.\n",
                                 retrys);

                if (retrys-- == 0)
                        break;
        }

        /* DMA Done */
        if (test_and_clear_bit(IVTV_F_S_DMA_DONE, &stream->s_flags)) {
                IVTV_DEBUG(IVTV_DEBUG_INFO,
                        "ivtv: DMA Done %lu type %d 0x%08x bytes\n",
                        jiffies, atomic_read(&itv->dec_dma_stat.type),
                        itv->dec_dma_stat.last_xfer);

                clear_bit(IVTV_F_S_DMAP, &stream->s_flags);

                /* Signal DMA done, exit waitq */
		if (!test_and_clear_bit(DMA_IN_USE, &stream->udma.u_flags))
                	itv->dec_dma_stat.last_xfer = 0;
		else
			ivtv_unmap_user_dma(itv, &stream->udma, 0);

                wake_up(&stream->waitq);

		return 0;
        }

        IVTV_DEBUG(IVTV_DEBUG_ERR,
                "ivtv: DMA Done failed %lu type %d 0x%08x bytes\n",
                jiffies, atomic_read(&itv->dec_dma_stat.type),
                itv->dec_dma_stat.last_xfer);

	/* No DMA to clear */
	return 1;
}

void ivtv_enc_timeout(unsigned long arg)
{
	struct ivtv *itv = (struct ivtv *)arg;
	struct ivtv_stream *stream = 
		&itv->streams[IVTV_ENC_STREAM_TYPE_MPG];

	if (!test_bit(IVTV_F_S_DMAP, &stream->s_flags)) {
		u32 dmastatus = readl(itv->reg_mem + IVTV_REG_DMASTATUS);
	
		/* This checks for encoder DMA errors, and doesn't return 
		   if they have occured */
		if (dmastatus & IVTV_DMA_WRITE_ERR) {
			IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"ENC DMA Write error: %u\n", dmastatus);

        		set_bit(FW_SOFT_RESET_NEEDED, &itv->r_flags);
        		/* Reload Firmware */
        		fw_schedule_work(itv);
		} else
			return;
	}

	IVTV_DEBUG(IVTV_DEBUG_ERR, 
		"ivtv_enc_timeout: Running encoder tasklet...\n");

	set_bit(IVTV_F_T_ENC_DMA_NEEDED, &itv->t_flags);
	set_bit(IVTV_F_T_ENC_DMA_DONE, &itv->t_flags);
	tasklet_schedule(&itv->dma_tq);
}

