/*
    audio dual language watch thread and decoder thread.

    Dual language watch:
    Copyright (C) 2004  Stephane Zermatten

    Decoder thread:
    Copyright (C) 2004  Chris Kennedy ckennedy@kmos.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ivtv-driver.h"
#include "ivtv-dma.h"
#include "ivtv-queue.h"
#include "ivtv-reset.h"
#include "ivtv-mailbox.h"
#include "ivtv-i2c.h"
#include "ivtv-kthreads.h"
#include "ivtv-irq.h"
#include "ivtv-video.h"

#include <linux/smp_lock.h>

static void ivtv_dualwatch_start_kthread(struct ivtv* ivtv);
static void ivtv_dualwatch_stop_kthread(struct ivtv* ivtv);
static void ivtv_dualwatch_init_kthread(struct ivtv *ivtv);
static void ivtv_dualwatch_exit_kthread(struct ivtv *ivtv);
static void ivtv_dualwatch_set_dual(struct ivtv *itv);


/**
 * Change the DUAL flag 
 */
static void ivtv_dualwatch_set_dual(struct ivtv *itv) 
{
        u32 audio_bitmask;
        int goal_stereo_flag;
        int stereo_flag;

        if(!itv->dualwatch.encoding)
                return;

        audio_bitmask = itv->dualwatch.audio_bitmask;
        stereo_flag = audio_bitmask&IVTV_CODEC_AUDIO_MPEG_STEREO_MASK;
        if(itv->dualwatch.enabled) {
                struct v4l2_tuner vt;
                memset(&vt, 0, sizeof(vt));
                ivtv_msp34xx(itv, VIDIOC_G_TUNER, &vt);
                if(vt.rxsubchans&V4L2_TUNER_SUB_LANG2) {
                        goal_stereo_flag = IVTV_CODEC_AUDIO_MPEG_STEREO_DUAL;
                } else {
                        goal_stereo_flag = IVTV_CODEC_AUDIO_MPEG_STEREO_STEREO;
                }
        } else {
                if(stereo_flag==IVTV_CODEC_AUDIO_MPEG_STEREO_DUAL) {
                        goal_stereo_flag = IVTV_CODEC_AUDIO_MPEG_STEREO_STEREO;
                } else {
                        goal_stereo_flag = stereo_flag;
                }
        }

        if(stereo_flag==IVTV_CODEC_AUDIO_MPEG_STEREO_STEREO
           && ((itv->codec.audio_bitmask&IVTV_CODEC_AUDIO_MPEG_STEREO_MASK ) == IVTV_CODEC_AUDIO_MPEG_STEREO_JOINT)) {
                goal_stereo_flag = IVTV_CODEC_AUDIO_MPEG_STEREO_JOINT;
        }
        
        if(goal_stereo_flag != stereo_flag) {
                u32 new_bitmap = 
                        goal_stereo_flag 
                        | (itv->dualwatch.audio_bitmask & ~IVTV_CODEC_AUDIO_MPEG_STEREO_MASK);

                IVTV_DEBUG( IVTV_DEBUG_API, 
                            "dualwatch: change stereo flag from 0x%x to 0x%x. new audio_bitmask=0x%ux\n",
                            stereo_flag, 
                            goal_stereo_flag,
                            new_bitmap);
                
                if (ivtv_vapi(itv, IVTV_API_ASSIGN_AUDIO_PROPERTIES, 1, new_bitmap) == 0) {
                        itv->dualwatch.audio_bitmask = new_bitmap;
                } else {
                        IVTV_DEBUG( IVTV_DEBUG_API, 
                                    "dualwatch: changing stereo flag failed\n");
                }
        }
}

/* start the thread if necessary */
static void ivtv_dualwatch_update(struct ivtv* itv)
{
        ivtv_dualwatch_set_dual(itv);

        /* start timer, but only if necessary (timer stops automatically) */
        if(itv->dualwatch.encoding 
           && itv->dualwatch.enabled) {
                if(!itv->dualwatch.thread)
                        ivtv_dualwatch_start_kthread(itv);
        } else {
                if(itv->dualwatch.thread)
                        ivtv_dualwatch_stop_kthread(itv);
        }
}
/* enable/disable dualwatch (= enable/disable full stereo (not lang1/lang2 or mono) mode) */
void ivtv_dualwatch_set_enabled(struct ivtv* itv, int enabled)
{
        itv->dualwatch.enabled=enabled;
        ivtv_dualwatch_update(itv);
}

/* start encoding => start the thread if DUALWATCH is enabled */
void ivtv_dualwatch_start_encoding(struct ivtv* itv)
{
        if(itv->dualwatch.encoding)
                return;
        itv->dualwatch.encoding = 1;
        itv->dualwatch.audio_bitmask = itv->codec.audio_bitmask;
        ivtv_dualwatch_update(itv);
}

/* stop encoding => stop the thread if DUALWATCH is enabled */
void ivtv_dualwatch_stop_encoding(struct ivtv* itv)
{
        if(!itv->dualwatch.encoding)
                return;
        if(itv->dualwatch.thread)
                ivtv_dualwatch_stop_kthread(itv);
        itv->dualwatch.encoding = 0; /* after stopping the thread */
}

/* function executed on the thread */
static int ivtv_dualwatch_kthread(void *ptr) {
   struct ivtv* ivtv = (struct ivtv*)ptr;
   struct ivtv_dualwatch *kthread = &ivtv->dualwatch;

   ivtv_dualwatch_init_kthread(ivtv);

   for(;;) {
           interruptible_sleep_on_timeout(&kthread->queue, HZ);
           
           mb();
           
           if (kthread->terminate)
                   break;    
           
           ivtv_dualwatch_set_dual(ivtv);
   }
   ivtv_dualwatch_exit_kthread(ivtv);
   return 0;
}

static void ivtv_dualwatch_start_kthread(struct ivtv* ivtv) {
        /* adapted from example by Martin Frey 
         * http://www.scs.ch/~frey/linux/kernelthreads.html
         */
        struct ivtv_dualwatch *kthread = &ivtv->dualwatch;
        
        if(!kthread->thread) {
                init_MUTEX_LOCKED(&kthread->startstop_sem);
                kernel_thread(ivtv_dualwatch_kthread, ivtv, 0);
                down(&kthread->startstop_sem);
        }
}

static void ivtv_dualwatch_stop_kthread(struct ivtv* ivtv) {
        struct ivtv_dualwatch *kthread = &ivtv->dualwatch;
        
        if (!kthread->thread || kthread->terminate)
                return;
        /* this function needs to be protected with the big
           kernel lock (lock_kernel()). The lock must be
           grabbed before changing the terminate
           flag and released after the down() call. */
        lock_kernel();
        if(kthread->thread) { /* paranoid mode */
                init_MUTEX_LOCKED(&kthread->startstop_sem);
                /* We need to do a memory barrier here to be sure that
                   the flags are visible on all CPUs. 
                */
                mb();
                /* set flag to request thread termination */
                kthread->terminate = 1;
                
                /* We need to do a memory barrier here to be sure that
                   the flags are visible on all CPUs. 
                */
                mb();
                kill_proc(kthread->thread->pid, SIGKILL, 1);
                down(&kthread->startstop_sem);
        }
        unlock_kernel();
 
        /* now we are sure the thread is in zombie state. We
           notify init to clean the process up. */
        kill_proc(0, SIGCHLD, 1);

}

/* initialize new created thread. Called by the new thread. */
static void ivtv_dualwatch_init_kthread(struct ivtv *ivtv)
{
        struct ivtv_dualwatch *kthread = &ivtv->dualwatch;

#ifndef LINUX26
        daemonize();
        exit_files(current);
        reparent_to_init();
#else
        daemonize("ivtv_dualwatch");
#endif

        /* lock the kernel. A new kernel thread starts without
           the big kernel lock, regardless of the lock state
           of the creator (the lock level is *not* inheritated)
        */
        lock_kernel();

        /* fill in thread structure */
        kthread->thread = current;
        
        siginitsetinv(&current->blocked, sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGTERM));
        init_waitqueue_head(&kthread->queue);
        
        kthread->terminate = 0;
        sprintf(current->comm, "ivtv_dualwatch");

        /* let others run */
        unlock_kernel(); 

        /* tell the creator that we are ready and let him continue */
        up(&kthread->startstop_sem);
}

/* cleanup of thread. Called by the exiting thread. */
static void ivtv_dualwatch_exit_kthread(struct ivtv *ivtv)
{
        struct ivtv_dualwatch *kthread = &ivtv->dualwatch;

	/* lock the kernel, the exit will unlock it */
        lock_kernel();
        
        kthread->thread = NULL;
        mb();
        
        /* notify the stop_kthread() routine that we are terminating. */
        up(&kthread->startstop_sem);
        
        /* there is no race here between execution of the "killer" and real termination
           of the thread (race window between up and do_exit), since both the
           thread and the "killer" function are running with the kernel lock held.
           The kernel lock will be freed after the thread exited, so the code
           is really not executed anymore as soon as the unload functions gets
           the kernel lock back.
           The init process may not have made the cleanup of the process here,
           but the cleanup can be done safely with the module unloaded.
        */
}

void fw_work_handler(void *arg) {
        struct ivtv *itv = (struct ivtv *)arg;

        /* Check Firmware */
        ivtv_firmware_flags(itv, 100);

	return;
}

/* Must be spinlocked */
static int gather_free_buffers(struct ivtv *itv, int streamtype,
                struct list_head *free_list, int bufs_needed)
{
        struct ivtv_stream *st = &itv->streams[streamtype];
        struct ivtv_buffer *buf;
        int bufs_received;
        int x;

        /* gather the needed buffers first, so we don't have to bail
         * in mid-air. put them on a list on the stack */
        for (x = 0; x < bufs_needed; x++) {
                buf = ivtv_deq_buf_lock(&st->full_q, st);
                if (!buf)
                        break;

                list_add_tail(&buf->list, free_list);
        }

        /* How many did we actually get */
        bufs_received = x;

        /* damn, failed */
        if (bufs_received != bufs_needed) {
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: needed %d bufs for stream %d, "
			"received %d differ by (%d) bufs\n",
                        bufs_needed, streamtype, bufs_received,
                        (bufs_needed - bufs_received));

                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: free_q: %d elements\n",
                        atomic_read(&st->free_q.elements));
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: dma_q: %d elements\n",
                        atomic_read(&st->dma_q.elements));
                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "DEC: full_q: %d elements\n",
                        atomic_read(&st->full_q.elements));

                while (!list_empty(free_list)) {
                        buf = list_entry(free_list->next, 
				struct ivtv_buffer, list);
                        list_del_init(&buf->list);
                        ivtv_enq_buf_lock(&st->full_q, buf, st);
                }

                return 0;
        }
        return bufs_received;
}

void dec_work_handler(void *arg) {
        struct ivtv *itv = (struct ivtv *)arg;
        int type = IVTV_DEC_STREAM_TYPE_MPG;
        struct ivtv_stream *stream= &itv->streams[type];
	int ret = -1;
	unsigned long then = 0;	
	int max = 0;
	int rc = 0;
        int x=0, bytes_written=0;
        struct ivtv_buffer *buf;
        u32 data[IVTV_MBOX_MAX_DATA], result;
        u32 mem_offset, mem_size, orig_mem_size, hw_stream_type, buffer_bytes;
        int stmtype = type;
	int bufs_needed = 0;
        int bufs_received;
        LIST_HEAD(free_list);
	long sequence;
	u32 offset, sg_size = 0;
#ifndef DEC_DMA_PIO
	int i = 0;
        struct scatterlist *sg;
#endif

	/* YUV or MPG Stream? */
	if (test_bit(IVTV_F_I_DEC_YUV, &itv->i_flags)) {
		type = IVTV_DEC_STREAM_TYPE_YUV;
	} else {
		type = IVTV_DEC_STREAM_TYPE_MPG;
	}
       	stream= &itv->streams[type];

	/* Wait for it to be ready */
        {
               int retrys = 100;

               /* Check DMA Status, both registers */
               while (ivtv_dma_status(itv, 1, 0x04, 0) ||
                     !ivtv_dma_status(itv, 1, 0x00, 0))
               {
                     IVTV_DEBUG(IVTV_DEBUG_DEC,
                             "DEC: (%d) Sleeping for API cmd"
                             ", DMA busy.\n",
                             retrys);

                     ivtv_sleep_timeout(HZ/100, 1);
                            if (retrys-- == 0)
                                   break;
                }

	        if (!retrys) {
			IVTV_DEBUG(IVTV_DEBUG_DEC,
				"dec_work_queue: DMA not finished %lu\n", 
				jiffies);
			goto exit_dec_wait;
		}
        }

        /* Wait for FW Reload */
        if (test_bit(FW_RESET_LOCK, &itv->r_flags)) {
                int retrys = 500;

                IVTV_DEBUG(IVTV_DEBUG_ERR,
                        "ivtv_write: Sleeping for FW Reset");

                /* Check FW Reset Status  */
                while (test_bit(FW_RESET_LOCK, &itv->r_flags)) {
                        ivtv_sleep_timeout(HZ/100, 1);
                        if (retrys-- == 0)
                                break;
                }
        }

        /* Schedule DMA */
	then = jiffies;

	/* Schedule DMA */
        DECLARE_WAITQUEUE(wait, current);

      	IVTV_DEBUG(IVTV_DEBUG_INFO, 
		"dec_work_queue: Schedule DMA %lu\n", jiffies);

       	then = jiffies;
       	add_wait_queue(&stream->waitq, &wait);
       	do {
               	int sleep = (HZ/100);
               	set_current_state(TASK_INTERRUPTIBLE);

               	/* Lock Decoder */
               	if ((jiffies-itv->dec_dma_stat.ts) > 9 &&
			!test_and_set_bit(IVTV_F_S_DMAP, 
					&stream->s_flags)) 
		{
                       	break;
               	}

               	set_current_state(TASK_UNINTERRUPTIBLE);
               	schedule_timeout(sleep);

               	/* Lock Decoder */
               	if ((jiffies-itv->dec_dma_stat.ts) > 9 &&
			!test_and_set_bit(IVTV_F_S_DMAP, 
					&stream->s_flags)) 
		{
                       	break;
               	}

               	/* This exits the wait if the last DMA xfer is done */
               	if ((jiffies-itv->dec_dma_stat.ts) > 20 && (jiffies - then) > 
			(itv->dec_dma_stat.last_xfer/DEC_API_TIMEOUT)) 
		{
                       	IVTV_DEBUG(IVTV_DEBUG_INFO,
                               	"DEC: wait done at %u jiffies.\n",
                               	(itv->dec_dma_stat.last_xfer/DEC_API_TIMEOUT));
                       	if (!test_bit(FW_RESET_NEEDED, &itv->r_flags)) {
                               	/* check DMA status register */
                               	result = 
					readl(itv->reg_mem + 
						IVTV_REG_DMASTATUS);
                               	IVTV_DEBUG(IVTV_DEBUG_DMA, 	
					"DEC: DMA Status: 0x%08x\n",
                                        	result);
                       	}
                       	rc = -EIO;
                       	break;
               	}

       	} while (!rc);
       	set_current_state(TASK_RUNNING);
       	remove_wait_queue(&stream->waitq, &wait);

	/* Failed, bail out */ 
	if (rc)
		goto exit_dec_wait;

	if (test_and_clear_bit(IVTV_F_S_NEEDS_DATA, &stream->s_flags)) {
		/* Get DMA Xfer Info to know what to send */
                ret = ivtv_api(itv, itv->dec_mbox,
                        &itv->dec_msem, IVTV_API_DEC_XFER_INFO,
                                        &result, 0, &data[0]);

		/* Did xfer status fail? */
		if (ret) {
                	IVTV_DEBUG(IVTV_DEBUG_ERR,
                        	"DEC DMA Xfer Info failed=0x%08x %lu\n", 
			ret, jiffies);

			goto clear_dmap;
		}

        	IVTV_DEBUG(IVTV_DEBUG_INFO,
                   "DEC: Mailbox 10: 0x%08x 0x%08x 0x%08x 0x%08x\n",
                   	data[0],data[1],data[2],data[3]);

        	/* Get card mem addr and size from data array */
		hw_stream_type = data[0];
        	mem_offset = data[1];
       	 	mem_size = orig_mem_size = data[2];
        	buffer_bytes = data[3]; /* # bytes in card's buffer */

		/* Destination Offset */
                offset = (mem_offset + IVTV_DECODER_OFFSET);

#ifndef DEC_DMA_PIO
		/* Unmap Last DMA */
                if (stream->SG_length &&
			!ivtv_dec_DMA_done_check(itv, stream)) 
		{
                	IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"DEC: Couldn't free %d buffers in DMA\n",
                         	stream->SG_length);
		}
		IVTV_ASSERT(stream->SG_handle == IVTV_DMA_UNMAPPED);
#endif
		/* Calculate memory size, or if safe to send data at all */
		if ((bufs_needed = ivtv_dec_buffer(itv, hw_stream_type, 
			mem_offset, mem_size, buffer_bytes)) <= 0)
		{
                        IVTV_DEBUG(IVTV_DEBUG_ERR,
                                "DEC: Decoder error 0x%08x, mem_offset 0x%08x "
				"mem_size 0x%08x buffer_bytes 0x%08x\n",
                                        bufs_needed, mem_offset, mem_size, 
					buffer_bytes);

                        goto clear_dmap;
		}

		/* Allow smaller xfers than max xfer size */
		if ((atomic_read(&stream->full_q.elements) > 0) && 
			((atomic_read(&stream->full_q.elements) *
				IVTV_DMA_DEC_BUF_SIZE) >= IVTV_DMA_MIN_XFER) &&
			(atomic_read(&stream->full_q.elements) < bufs_needed))
		{
			int have_space = 0;	
			int xfer_align_spc = 0;	
			int xfer_align_num = 0;	

			/* If too many buffers available, only send needed */
			if (bufs_needed > atomic_read(&stream->full_q.elements))
				bufs_needed = 
					atomic_read(&stream->full_q.elements);

			/* Calculate actual bytes needed */
			have_space = (bufs_needed * IVTV_DMA_DEC_BUF_SIZE);

			/* Calculate number of aligned data chunks */
			xfer_align_spc = (have_space/IVTV_DMA_ALIGN_XFER);

			/* Calculate each data chunks number of buffers */
			xfer_align_num = 
				(IVTV_DMA_ALIGN_XFER/IVTV_DMA_DEC_BUF_SIZE);

			/* Calculate the actual aligned buffers needed */
			bufs_needed = xfer_align_num * xfer_align_spc;
		}

		if (atomic_read(&stream->dma_q.elements) > 0) {
                        IVTV_DEBUG(IVTV_DEBUG_ERR,
                                "DEC: DMA_Q %d Dec buffers still in progress.\n",
				atomic_read(&stream->dma_q.elements));

                        goto clear_dmap;
		}

                /* Wait till data input */
                if (atomic_read(&stream->full_q.elements) < bufs_needed) {
                        IVTV_DEBUG(IVTV_DEBUG_ERR,
                                "DEC: Only %d Dec buffers available need %d.\n",
				atomic_read(&stream->full_q.elements), 
				bufs_needed);

                        goto clear_dmap;
                }
	
		/* Gather enough buffers */
        	bufs_received =
                	gather_free_buffers(itv, 
				type, &free_list, bufs_needed);

		/* Did it fail, shouldn't ever */	
		if (bufs_received != bufs_needed) {
                        IVTV_DEBUG(IVTV_DEBUG_ERR,
                                "DEC: Only %d Dec buffers need %d.\n",
				bufs_received, bufs_needed);

                        goto clear_dmap;
		} 

		/* Max buffers to send */
		max = bufs_needed;

		/* translate buffers back into mem_size */
		mem_size = (bufs_needed * IVTV_DMA_DEC_BUF_SIZE);

#ifndef DEC_DMA_PIO
        	/* Set SGarray to all 0's */
        	memset(stream->SGarray, 0,
                        sizeof(struct ivtv_SG_element)*IVTV_DMA_SG_ENT);

        	/* Set SGarray to all 0's */
        	memset(stream->SGlist, 0,
                        sizeof(*stream->SGlist)*IVTV_DMA_SG_ENT);
#endif
        	IVTV_DEBUG(IVTV_DEBUG_DEC,
                	"DEC: want: "
                	"0x%08x  length: 0x%08x  full: 0x%08x  space 0x%08x\n",
                	mem_offset, orig_mem_size, buffer_bytes,
                	(0x00100000-buffer_bytes));

		/* increment the sequence # */
        	sequence = ++stream->seq;

                /* Send next DMA up to max buffers */
                for (x = 0; x < bufs_received; x++) 
		{
                	/* extract the buffers we procured earlier */
                	buf = list_entry(free_list.next, 
				struct ivtv_buffer, list);
                	list_del_init(&buf->list);

			/* Check if buffer has overflown, do what though? */
                        if (buf->buffer.bytesused != IVTV_DMA_DEC_BUF_SIZE) {
                                IVTV_DEBUG(IVTV_DEBUG_ERR,
                                        "DEC: Error, 0x%08x in Buffer.\n",
                                        buf->buffer.bytesused);
                                buf->buffer.bytesused = IVTV_DMA_DEC_BUF_SIZE;
                        }
        
                        bytes_written += buf->buffer.bytesused;
#ifndef DEC_DMA_PIO
        		/* Write using DMA */
                        stream->SGlist[x].length = buf->buffer.bytesused;
                        stream->SGlist[x].offset = 0;
                        stream->SGlist[x].page = 
				virt_to_page((void *)buf->buffer.m.userptr);;
#else
        		/* Write using Processor IO */
	      		memcpy_toio((void *)
				((itv->dec_mem + offset)-IVTV_DECODER_OFFSET), 
				(void *)buf->buffer.m.userptr, 
				buf->buffer.bytesused);

			/* Incr decoder offset */
                        offset += buf->buffer.bytesused;

                        buf->buffer.bytesused   = 0;
                        buf->readpos = 0;
                        buf->ts = 0;

                        /* Don't indicate we are doing DMA */
                        stream->SG_handle = 0;
			sg_size += buf->buffer.bytesused;
#endif

			/* Decoding State */
			buf->buffer.index 	= x;
			buf->buffer.sequence    = sequence;
#ifndef DEC_DMA_PIO
			buf->ts 		= jiffies;
                        buf->readpos            += stream->SGlist[x].length;
                        buf->buffer.bytesused   -= stream->SGlist[x].length;
        
			/* Enqueue Buffer to DMA Queue */	
                        ivtv_enq_buf_lock(&stream->dma_q, buf, stream);
#else
                        /* put in the 'done' queue */
                        ivtv_enq_buf_lock(&stream->free_q, buf, stream);
#endif
                }

		bufs_received = x;
#ifndef DEC_DMA_PIO
                /* Allocate Buffers SG List */
                stream->SGlen = bufs_received;
                stream->SG_length =
                        pci_map_sg(itv->dev, 
				stream->SGlist, stream->SGlen, stream->dma);

                /* Setup SG Array with Buffer SG List */
                for (i = 0, sg = stream->SGlist; i < stream->SG_length; i++, sg++) {
                        stream->SGarray[i].src = sg_dma_address(sg);/*Buffer Addr */
                        stream->SGarray[i].dst = offset;          /* Decoder Addr */
                        stream->SGarray[i].size = sg_dma_len(sg);  /* Buffer size */
                        sg_size += stream->SGarray[i].size; /* Size of SG List */
                        offset += stream->SGarray[i].size;  /* Increment Dec Addr */
                }

                IVTV_DEBUG(IVTV_DEBUG_DMA,
                        "DEC: SGlen = 0x%08x, SG_length = 0x%08x, "
                        "sg_size = 0x%08x\n",
                        stream->SGlen, stream->SG_length, sg_size);

                /*Set Last Element Bit*/
                stream->SGarray[stream->SG_length-1].size |= 0x80000000;

                /* Map, if already mapped then at least sync */
                if (stream->SG_handle == IVTV_DMA_UNMAPPED) {
                        stream->SG_handle =
                                pci_map_single(itv->dev, stream->SGarray,
                                (sizeof(struct ivtv_SG_element) * stream->SG_length),
                                stream->dma);
                } else {
                        IVTV_ASSERT(stream->SG_handle == IVTV_DMA_UNMAPPED);
                        pci_dma_sync_single(itv->dev, stream->SG_handle,
                                sizeof(struct ivtv_SG_element) * stream->SG_length,
                                stream->dma);
                }
#endif

		IVTV_DEBUG(IVTV_DEBUG_DEC,
			"DEC: sent: "
			"0x%08x <length: 0x%08x  fill: 0x%08x <space 0x%08x <\n",
			(mem_offset+bytes_written),
			bytes_written, 
			(buffer_bytes+bytes_written),
			(0x00100000-(buffer_bytes+bytes_written)));

        	/* Decoder DMA State */
        	itv->dec_dma_stat.last_xfer = bytes_written;
		atomic_set(&itv->dec_dma_stat.dma_buffers, bufs_received);
        	atomic_set(&itv->dec_dma_stat.type, 0);

        	/* Decoder DMA State */
        	itv->dec_dma_stat.last_addr     = mem_offset;
        	itv->dec_dma_stat.bytes_filled  = (buffer_bytes+bytes_written);
        	itv->dec_dma_stat.bytes_needed  = (mem_size-bytes_written);
        	itv->dec_dma_stat.total_xfer    += bytes_written;
        	itv->dec_dma_stat.sg_bytes      = 
			(atomic_read(&stream->full_q.elements) * 
				IVTV_DMA_DEC_BUF_SIZE);
		itv->dec_dma_stat.data[0]	= mem_offset;
		itv->dec_dma_stat.data[1]	= mem_size;
		itv->dec_dma_stat.data[2]	= buffer_bytes;

		IVTV_DEBUG(IVTV_DEBUG_DEC,
			"DEC: DMA State:\n"
			" type: 0x%08x "
			"last_xfer: 0x%08x "
			"last_addr: 0x%08x\n"
			" bytes_filled: 0x%08x "
			"bytes_needed: 0x%08x "
			"total_xfer: 0x%08x\n"
			" sg_bytes: 0x%08x\n"
			" vsync_count: %lu\n"
			" ts: %u\n",
			atomic_read(&itv->dec_dma_stat.type),
			itv->dec_dma_stat.last_xfer,
			itv->dec_dma_stat.last_addr,
			itv->dec_dma_stat.bytes_filled,
			itv->dec_dma_stat.bytes_needed,
			itv->dec_dma_stat.total_xfer,
			itv->dec_dma_stat.sg_bytes,
			itv->dec_dma_stat.vsync_count,
			itv->dec_dma_stat.ts);  

		/* Check for YUV */
		if (test_bit(IVTV_F_I_DEC_YUV, &itv->i_flags))
			hw_stream_type = 2;

        	/* Decoder/OSD DMA */
        	if (ivtv_vapi(itv, IVTV_API_DEC_DMA_FROM_HOST, 3,
                	stream->SG_handle, sg_size, hw_stream_type))
        	{
                	IVTV_DEBUG(IVTV_DEBUG_ERR, 
				"DEC: Error sending DMA!!!\n");
#ifndef DEC_DMA_PIO
                	/* Unmap Last DMA */
                	if (stream->SG_length &&
                        	!ivtv_dec_DMA_done_check(itv, stream))
                	{
                        	IVTV_DEBUG(IVTV_DEBUG_ERR,
                                	"DEC: Couldn't free %d buf in DMA\n",
                                	stream->SG_length);
                	}
                	IVTV_ASSERT(stream->SG_handle == IVTV_DMA_UNMAPPED);
#endif
        		/* Decoder DMA State */
        		itv->dec_dma_stat.last_xfer = 0;
        		atomic_set(&itv->dec_dma_stat.type, -1);

			goto clear_dmap;
        	}
#ifdef DEC_DMA_PIO
		clear_bit(IVTV_F_S_DMAP, &itv->streams[stmtype].s_flags);
#endif
        	IVTV_DEBUG(IVTV_DEBUG_INFO,
                   	"Sched DEC dma: addr: 0x%08llx, "
			"array_size 0x%08x, type 0x%08x\n",
			(u64)stream->SG_handle, sg_size, hw_stream_type);

      		IVTV_DEBUG(IVTV_DEBUG_INFO,
        		"dec_work_queue: DMA Sched at %lu (%lu) jiffies\n", 
			jiffies, (jiffies - then));
	} else if (!rc) {
		/* Failed to be needed */
		IVTV_DEBUG(IVTV_DEBUG_INFO, "Sched DEC dma: no data needed\n");
		clear_bit(IVTV_F_S_DMAP, &itv->streams[stmtype].s_flags);
	}

	return;
clear_dmap:	
		clear_bit(IVTV_F_S_DMAP, &itv->streams[stmtype].s_flags);
               	set_bit(IVTV_F_S_NEEDS_DATA, &stream->s_flags);
exit_dec_wait:	
      	IVTV_DEBUG(IVTV_DEBUG_INFO,
        	"dec_work_queue: Failed at %lu (%lu) jiffies\n", 
		jiffies, (jiffies - then));

	return;
}
