/*
    buffer queues.
    Copyright (C) 2003-2004  Kevin Thayer <nufan_wfk at yahoo.com>

    Copyright (C) 2004  Chris Kennedy ckennedy@kmos.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ivtv-driver.h"
#include "ivtv-queue.h"

void ivtv_enq_buf_nolock(struct ivtv_buffer_list *queue, struct ivtv_buffer *buf)
{
	WARN_ON(!list_empty(&buf->list));
	list_add_tail(&buf->list, &queue->list);
	atomic_inc(&queue->elements);
}


/* Adds buffers to the tail, effectively making a queue */
int ivtv_enq_buf_lock_bh(struct ivtv_buffer_list *queue,
		 struct ivtv_buffer *buf, struct ivtv_stream *stream)
{
	spin_lock_bh(&stream->slock);
	ivtv_enq_buf_nolock(queue, buf);
	spin_unlock_bh(&stream->slock);

	return 0;
}

/* Adds buffers to the tail, effectively making a queue */
int ivtv_enq_buf_lock(struct ivtv_buffer_list *queue,
		 struct ivtv_buffer *buf, struct ivtv_stream *stream)
{
	spin_lock(&stream->slock);
	ivtv_enq_buf_nolock(queue, buf);
	spin_unlock(&stream->slock);

	return 0;
}

void ivtv_del_buf_nolock(struct ivtv_buffer_list *queue,
			   struct ivtv_buffer *buffer)
{
	WARN_ON(list_empty(&buffer->list));
	list_del_init(&buffer->list);
	atomic_dec(&queue->elements);
}

void ivtv_del_buf_lock_bh(struct ivtv_buffer_list *queue,
                           struct ivtv_buffer *buffer, struct ivtv_stream *stream)
{
	spin_lock_bh(&stream->slock);
	ivtv_del_buf_nolock(queue, buffer);
	spin_unlock_bh(&stream->slock);
}

void ivtv_del_buf_lock(struct ivtv_buffer_list *queue,
                           struct ivtv_buffer *buffer, struct ivtv_stream *stream)
{
	spin_lock(&stream->slock);
	ivtv_del_buf_nolock(queue, buffer);
	spin_unlock(&stream->slock);
}

/* returns first item in queue, doesn't dequeue */
void ivtv_move_buf_lock_bh(struct ivtv_buffer_list *from,
		   struct ivtv_buffer_list *to, struct ivtv_buffer *buffer,
		   struct ivtv_stream *stream)
{
	WARN_ON(list_empty(&buffer->list));

	spin_lock_bh(&stream->slock);
	list_move_tail(&buffer->list, &to->list);
	atomic_dec(&from->elements);
	atomic_inc(&to->elements);
	spin_unlock_bh(&stream->slock);
}

/* returns first item in queue, doesn't dequeue */
void ivtv_move_buf_lock(struct ivtv_buffer_list *from,
		   struct ivtv_buffer_list *to, struct ivtv_buffer *buffer,
		   struct ivtv_stream *stream)
{
	WARN_ON(list_empty(&buffer->list));

	spin_lock(&stream->slock);
	list_move_tail(&buffer->list, &to->list);
	atomic_dec(&from->elements);
	atomic_inc(&to->elements);
	spin_unlock(&stream->slock);
}

/* returns first item in queue, doesn't dequeue */
struct ivtv_buffer *ivtv_deq_peek_head_nolock(struct ivtv_buffer_list *queue, int type)
{
	/* make sure list has something to DeQ */
	if (!list_empty(&queue->list))
		return list_entry(queue->list.next, struct ivtv_buffer, list);
		
	IVTV_DEBUG(IVTV_DEBUG_INFO,
		"deqpeek: %d DeQ from empty list\n", type);
	return NULL;
}

struct ivtv_buffer *ivtv_deq_peek_head_lock(struct ivtv_buffer_list *queue,
	       	struct ivtv_stream *stream)
{
	struct ivtv_buffer *buffer;

	spin_lock(&stream->slock);
	buffer = ivtv_deq_peek_head_nolock(queue, stream->subtype);
	spin_unlock(&stream->slock);

	return buffer;
}

struct ivtv_buffer *ivtv_deq_peek_head_lock_bh(struct ivtv_buffer_list *queue,
	       	struct ivtv_stream *stream)
{
	struct ivtv_buffer *buffer;

	spin_lock_bh(&stream->slock);
	buffer = ivtv_deq_peek_head_nolock(queue, stream->subtype);
	spin_unlock_bh(&stream->slock);

	return buffer;
}

/* removes buffer from the head */
struct ivtv_buffer *ivtv_deq_buf_nolock(struct ivtv_buffer_list *queue, int type)
{
	struct ivtv_buffer *buf;

	/* make sure list has something to DeQ */
	if (!list_empty(&queue->list)) {
		buf = list_entry(queue->list.next, struct ivtv_buffer, list);
		list_del_init(queue->list.next);
		atomic_dec(&queue->elements);
		return buf;
	}

	IVTV_DEBUG(IVTV_DEBUG_DMA, 
		"deqbuf: %d DeQ from empty list!\n", type);
	return NULL;
}

struct ivtv_buffer *ivtv_deq_buf_lock(struct ivtv_buffer_list *queue,
	       	struct ivtv_stream *stream)
{
	struct ivtv_buffer *buf;

	spin_lock(&stream->slock);
	buf = ivtv_deq_buf_nolock(queue, stream->subtype);
	spin_unlock(&stream->slock);

	return buf;
}

struct ivtv_buffer *ivtv_deq_buf_lock_bh(struct ivtv_buffer_list *queue,
	       	struct ivtv_stream *stream)
{
	struct ivtv_buffer *buf;

	spin_lock_bh(&stream->slock);
	buf = ivtv_deq_buf_nolock(queue, stream->subtype);
	spin_unlock_bh(&stream->slock);

	return buf;
}

static void ivtv_free_buffer(struct ivtv *itv, struct ivtv_buffer *item, int dma)
{
	/* streams of dma type PCI_DMA_NONE will always have the dma_handle
	   set to IVTV_DMA_UNMAPPED */
	if (item->dma_handle != IVTV_DMA_UNMAPPED)
		pci_unmap_page(itv->dev, item->dma_handle, item->buffer.length, dma);
	if ((item->buffer.length/PAGE_SIZE) == 1)
		free_page(item->buffer.m.userptr);
	else
		free_pages(item->buffer.m.userptr, 
			(((item->buffer.length/PAGE_SIZE)/2)-1));
	kfree(item);
}

void ivtv_free_queues_nolock(struct ivtv *itv, int type)
{
	struct ivtv_stream *st = &itv->streams[type];
        struct ivtv_buffer *item;

	while ((item = ivtv_deq_buf_nolock(&st->free_q, type)))
		ivtv_free_buffer(itv, item, st->dma);
	while ((item = ivtv_deq_buf_nolock(&st->full_q, type)))
		ivtv_free_buffer(itv, item, st->dma);
	while ((item = ivtv_deq_buf_nolock(&st->dma_q, type)))
		ivtv_free_buffer(itv, item, st->dma);
}

static struct ivtv_buffer *ivtv_init_buffer(int bufsize)
{
	struct ivtv_buffer *buf;

	buf = kmalloc(sizeof(struct ivtv_buffer), GFP_KERNEL);
	if (buf == NULL) {
		IVTV_DEBUG(IVTV_DEBUG_ERR, "No memory on ivtv_buffer alloc!\n");
		return NULL;
	}

	if ((bufsize/PAGE_SIZE) == 1)
		buf->buffer.m.userptr = __get_free_page(GFP_KERNEL);
	else 
		buf->buffer.m.userptr = 
			__get_free_pages(GFP_KERNEL, 
			(((bufsize/PAGE_SIZE)/2)-1));

	if (buf->buffer.m.userptr == 0) {
		kfree(buf);
		IVTV_DEBUG(IVTV_DEBUG_ERR, "No memory on buffer alloc!\n");
		return NULL;
	}

    	/* Clear Buffer */
   	/*memset((void *)buf->buffer.m.userptr, 0, bufsize);*/

	buf->dma_handle = IVTV_DMA_UNMAPPED;
	buf->b_flags = 0;
	buf->ts = 0;
	buf->buffer.length = bufsize;
	buf->buffer.bytesused = 0;
	buf->readpos = 0;
	INIT_LIST_HEAD(&buf->list);

	return buf;
}

static int ivtv_add_queue_nolock(struct ivtv_buffer_list *queue, 
	int x, int bufsize, struct ivtv_stream *st)
{
        struct ivtv_buffer *item;

        /* allocate buffer */
        item = ivtv_init_buffer(bufsize);
        if (item == NULL) {
                IVTV_DEBUG(IVTV_DEBUG_ERR, "Buffer alloc failed!\n");
                return -ENOMEM;
        }

        /* setup buffer */
        item->buffer.index = x;
        /* FIXME: leave it at PRIVATE for now */
        item->buffer.type = V4L2_BUF_TYPE_PRIVATE;
        item->buffer.field = V4L2_FIELD_INTERLACED;
        item->buffer.memory = V4L2_MEMORY_MMAP;

        /* enqueue buffer */
        ivtv_enq_buf_nolock(queue, item);

        return 0;
}

static int ivtv_init_queue_nolock(struct ivtv_buffer_list *queue, int buffers, int bufsize)
{
	struct ivtv_buffer *item;
	int x;

	for (x = 0; x < buffers; x++) {
		/* allocate buffer */
		item = ivtv_init_buffer(bufsize);
		if (item == NULL) {
			IVTV_DEBUG(IVTV_DEBUG_ERR, "Buffer alloc failed!\n");
			return -ENOMEM;
		}
		
		/* setup buffer */
		item->buffer.index = x;
		/* FIXME: leave it at PRIVATE for now */
		item->buffer.type = V4L2_BUF_TYPE_PRIVATE;
		item->buffer.field = V4L2_FIELD_INTERLACED;
		item->buffer.memory = V4L2_MEMORY_MMAP;
		
		/* enqueue buffer */
		ivtv_enq_buf_nolock(queue, item);
	}

	return 0;
}

int ivtv_init_queues_nolock(struct ivtv *itv, int streamtype, int buffers, int bufsize)
{
	struct ivtv_stream *st = &itv->streams[streamtype];

	/* The full and dma queues are initially empty, so we only have to
	   allocate the free queue. */
	if (ivtv_init_queue_nolock(&st->free_q, buffers, bufsize) < 0) {
		ivtv_free_queues_nolock(itv, streamtype);
		return -ENOMEM;
	}
	return 0;
}

int ivtv_move_queue(struct ivtv_buffer_list *src,
		    struct ivtv_buffer_list *dst, struct ivtv_stream *stream)
{
	struct ivtv_buffer *buf;

	spin_lock_bh(&stream->slock);

  	while ((buf = ivtv_deq_buf_nolock(src, stream->subtype)))
		ivtv_enq_buf_nolock(dst, buf);

	spin_unlock_bh(&stream->slock);
	return 0;
}

void ivtv_flush_queues_nolock(struct ivtv *itv, int type)
{
	struct ivtv_stream *st = &itv->streams[type];
	struct ivtv_buffer *buf;

	/* move free_q to full_q to clean up partially-filled buffers */
	while ((buf = ivtv_deq_buf_nolock(&st->free_q, st->subtype)))
		ivtv_enq_buf_nolock(&st->full_q, buf);

	/* move io_q to full_q to clean up partially-filled buffers */
	while ((buf = ivtv_deq_buf_nolock(&st->io_q, st->subtype)))
		ivtv_enq_buf_nolock(&st->full_q, buf);

	while ((buf = ivtv_deq_buf_nolock(&st->full_q, st->subtype))) {
		buf->buffer.bytesused = 0;
		buf->readpos = 0;
        	buf->dma_handle = IVTV_DMA_UNMAPPED;
        	buf->b_flags = 0;
        	buf->ts = 0;

		ivtv_enq_buf_nolock(&st->free_q, buf);
	}

	/* Unmap DMA SG List if still mapped */
	if (st->SGlen > 0 && (sg_dma_address(st->SGlist) != IVTV_DMA_UNMAPPED)) {
		pci_unmap_sg(itv->dev, st->SGlist, st->SGlen, st->dma);	
		st->SGlen = 0;
		sg_dma_address(st->SGlist) = IVTV_DMA_UNMAPPED;
	}

	while ((buf = ivtv_deq_buf_nolock(&st->dma_q, st->subtype))) {
		IVTV_ASSERT(buf->dma_handle == IVTV_DMA_UNMAPPED);
		if (buf->dma_handle != IVTV_DMA_UNMAPPED)
			pci_unmap_page(itv->dev, buf->dma_handle, PAGE_SIZE, st->dma);
		buf->buffer.bytesused = 0;
		buf->readpos = 0;
        	buf->dma_handle = IVTV_DMA_UNMAPPED;
        	buf->b_flags = 0;
        	buf->ts = 0;

		ivtv_enq_buf_nolock(&st->free_q, buf);
	}
}

void ivtv_flush_queues_lock_bh(struct ivtv *itv, int type)
{
	struct ivtv_stream *st = &itv->streams[type];

	/* move free_q to full_q to clean up partially-filled buffers */
	spin_lock_bh(&st->slock);
	ivtv_flush_queues_nolock(itv, type);
	spin_unlock_bh(&st->slock);
}

void ivtv_flush_queues_lock(struct ivtv *itv, int type)
{
	struct ivtv_stream *st = &itv->streams[type];

	/* move free_q to full_q to clean up partially-filled buffers */
	spin_lock(&st->slock);
	ivtv_flush_queues_nolock(itv, type);
	spin_unlock(&st->slock);
}

/* User DMA Buffers */
int ivtv_alloc_user_dma(struct ivtv *itv, struct ivtv_user_dma *dma, int size)
{
    IVTV_DEBUG(IVTV_DEBUG_INFO, "ivtvfb_alloc_user_dma_to_device\n");

    /* Allocate SG Array */
    dma->SGarray = kmalloc(sizeof(struct ivtv_SG_element)*size,
                          GFP_KERNEL);

    if (!dma->SGarray) {
        printk(KERN_ERR
               "ivtvfb: cannot allocate scatter/gather list for %d pages\n",
               size);
        return -ENOMEM;
    }
    /* Clear SG Array */
    memset(dma->SGarray, 0, sizeof(struct ivtv_SG_element)*size);

    /* Allocate DMA Page Array Buffer */
    dma->map = kmalloc(sizeof(struct page *) * size, GFP_KERNEL);
    if (!dma->map) {
        IVTV_DEBUG(IVTV_DEBUG_ERR, "can't alloc dma page array\n");
        kfree(dma->SGarray);
        return -ENOMEM;
    }
    /* Clear Page Array */
    memset(dma->map, 0, sizeof(struct page *)*size);

    /* Clear DMA Page Array Size */
    dma->page_count = -1;

    return 0;
}

int ivtv_prep_user_dma(struct ivtv *itv, struct ivtv_user_dma *dma,
                                   unsigned long ivtv_dest_addr,
                                   char *userbuf, int size_in_bytes)
{
    unsigned long first, last; 	
    int i, offset;
    unsigned long uaddr;
    int err;	

    IVTV_DEBUG(IVTV_DEBUG_DMA, "ivtvfb_prep_user_dma, dst: 0x%08x\n",
                    (unsigned int) ivtv_dest_addr);

    if (dma->map == NULL)
	return -ENOMEM;	

    uaddr = ((unsigned long) userbuf & PAGE_MASK);
    first = (uaddr & PAGE_MASK) >> PAGE_SHIFT;
    last = ((uaddr+size_in_bytes-1) & PAGE_MASK) >> PAGE_SHIFT;

    /* Align to PAGE_SIZE */
    offset = dma->offset = uaddr & ~PAGE_MASK;
    dma->page_count = last-first+1;

    /* Get user pages for DMA Xfer */
    down_read(&current->mm->mmap_sem);
    err = get_user_pages(current, 
		current->mm, uaddr & PAGE_MASK, dma->page_count, 0, 1, dma->map, NULL);
    up_read(&current->mm->mmap_sem);

    if (dma->page_count != err) {
        IVTV_DEBUG(IVTV_DEBUG_ERR, "failed to map user pages, returned %d instead of %d\n",
		dma->page_count, err);
        return -EINVAL;
    }

    /* Flag SG Array is mapped */
    set_bit(DMA_IN_USE, &dma->u_flags);

    /* Fill SG Array with new values */
    for (i = 0; i < dma->page_count; i++) {
        dma->SGarray[i].size = PAGE_SIZE - offset;
        dma->SGarray[i].src = pci_map_page(itv->dev, dma->map[i], 
		offset, PAGE_SIZE, PCI_DMA_TODEVICE);
        dma->SGarray[i].dst = ivtv_dest_addr + (i * PAGE_SIZE) + offset;
        offset = 0;
    }

    /* Tag SG Array with Interrupt Bit */
    dma->SGarray[dma->page_count - 1].size |= 0x80000000;

    /* Map DMA Page Array Buffer */
    dma->SG_handle =
        pci_map_single(itv->dev, (void *) dma->SGarray,
                       (sizeof(struct ivtv_SG_element) *
                         dma->page_count), PCI_DMA_TODEVICE);

    return dma->page_count;
}

int ivtv_unmap_user_dma(struct ivtv *itv, struct ivtv_user_dma *dma, int count)
{
    int i;
    IVTV_DEBUG(IVTV_DEBUG_INFO, "ivtvfb_unmap_user_dma\n");

    /* DMA in Use */
    if (test_bit(DMA_IN_USE, &dma->u_flags))
	return 1;

    /* Unmap DMA Page Array Buffer */
    if (dma->page_count && (dma->SG_handle != IVTV_DMA_UNMAPPED)) {
        pci_unmap_single(itv->dev,  dma->SG_handle,
               	dma->page_count, PCI_DMA_TODEVICE);

    	dma->SG_handle = IVTV_DMA_UNMAPPED;
    }

    /* Release User Pages */
    for (i = (dma->page_count-1); i >= 0; i--) {
 	pci_unmap_page(itv->dev, dma->SGarray[i].src,
		PAGE_SIZE, PCI_DMA_TODEVICE);
	dma->SGarray[i].size = 0;
       	dma->SGarray[i].src =  0;
       	dma->SGarray[i].dst =  0;
	flush_dcache_page(dma->map[i]);
	page_cache_release(dma->map[i]);
    }	
    dma->page_count = -1;

    /* Clear SG Array */	
    if (count > 0)	
    	memset(dma->SGarray, 0, (sizeof(struct ivtv_SG_element) * count));

    /* Clear Page Array */
    if (count > 0)	
    	memset(dma->map, 0, (sizeof(struct page *) * count));

    return 0;
}

int ivtv_check_user_dma(struct ivtv *itv, struct ivtv_user_dma *dma, int count)
{
    int i;
    IVTV_DEBUG(IVTV_DEBUG_INFO, "ivtvfb_check_user_dma\n");

    /* DMA in Use */
    if (test_bit(DMA_IN_USE, &dma->u_flags))
        printk("ivtvfb_check: DMA_IN_USE set\n");

    /* Unmap DMA Page Array Buffer */
    if (dma->page_count && (dma->SG_handle != IVTV_DMA_UNMAPPED)) {
        printk("ivtvfb_check: dma->page_count = %d, dma->SG_handle = 0x%08x\n",
		dma->page_count, dma->SG_handle);
    } else if (dma->SG_handle != IVTV_DMA_UNMAPPED) {
        printk("ivtvfb_check: DMA not mapped\n");
    } else
        printk("ivtvfb_check: DMA is mapped, page_count = %d, dma_handle = 0x%08x\n",
		dma->page_count, dma->SG_handle);

    /* Release User Pages */
    for (i = (dma->page_count-1); i >= 0; i--) {
	printk("ivtvfb_check: DMA is: src = 0x%08x dst = 0x%08x size = 0x%08x\n",
		dma->SGarray[i].src, dma->SGarray[i].dst, dma->SGarray[i].size);
    }

    return 0;
}

void ivtv_free_user_dma(struct ivtv *itv, struct ivtv_user_dma *dma)
{
    IVTV_DEBUG(IVTV_DEBUG_INFO, "ivtvfb_free_user_dma\n");

    /* Unmap any remaining DMA */	
    if (ivtv_unmap_user_dma(itv, dma, dma->page_count))
    	IVTV_DEBUG(IVTV_DEBUG_ERR, "OSD: Error during DMA unmap, IN_USE!\n");

    /* Free SG Array */
    kfree(dma->SGarray);

    /* Free DMA Buffer of Pages */
    kfree(dma->map);

    return;
}

