/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_mem.c 1.61 2004/07/17 17:32:57EDT sf_support Exp  $
 */

#include <linux/version.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <scsi/scsi_device.h>
#include "lpfc_hw.h"
#include "lpfc_mem.h"
#include "lpfc_sli.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
#include "lpfc_cfgparm.h"
#include "lpfc.h"
#include "lpfc_logmsg.h"
#include "lpfc_crtn.h"

static void lpfc_mem_safety_pool_create(struct lpfc_hba *, uint32_t );
static void lpfc_mem_safety_pool_destroy(struct lpfc_hba *);
static void lpfc_mbuf_safety_pool_create(struct lpfc_hba *, uint32_t );
static void lpfc_mbuf_safety_pool_destroy(struct lpfc_hba *);
void *lpfc_mbuf_alloc(struct lpfc_hba *, int, dma_addr_t *);
void lpfc_mbuf_free(struct lpfc_hba *, void *, dma_addr_t);
static void *lpfc_scsibuf_pool_alloc(int , void *);
static void  lpfc_scsibuf_pool_free(void *, void *);

static void *lpfc_bind_pool_alloc(int , void *);
static void *lpfc_iocb_pool_alloc(int , void *);
/* static void *lpfc_scsibuf_pool_alloc(int , void *); */
static void *lpfc_mbox_pool_alloc(int , void *);
static void *lpfc_nlp_pool_alloc(int , void *);
static void  lpfc_bind_pool_free(void *, void *);
static void  lpfc_iocb_pool_free(void *, void *);
/* static void  lpfc_scsibuf_pool_free(void *, void *); */
static void  lpfc_mbox_pool_free(void *, void *);
static void  lpfc_nlp_pool_free(void *, void *);

int
lpfc_mem_alloc(struct lpfc_hba * phba)
{

	phba->lpfc_scsi_dma_ext_pool = 0;
	phba->lpfc_mbuf_pool = 0;

	phba->lpfc_scsi_dma_ext_pool =
		pci_pool_create("lpfc_scsi_dma_ext_pool",
				phba->pcidev,
				LPFC_SCSI_DMA_EXT_SIZE,
				8,
				0);

	phba->lpfc_mbuf_pool =
		pci_pool_create("lpfc_mbuf_pool",
				phba->pcidev,
				LPFC_BPL_SIZE,
				8,
				0);

	if (!phba->lpfc_scsi_dma_ext_pool || !phba->lpfc_mbuf_pool) {
		lpfc_mem_free(phba);
		return 0;
	}

	lpfc_mbuf_safety_pool_create(phba, LPFC_MBUF_POOL_SIZE);
	lpfc_mem_safety_pool_create(phba, LPFC_MEM_POOL_SIZE);

	return (1);
}

int
lpfc_mem_free(struct lpfc_hba * phba)
{
	LPFC_MBOXQ_t *mbox, *next_mbox;
	struct lpfc_sli *psli;

	/* free the mapped address match area for each ring */
	psli = &phba->sli;


	/* Free everything on mbox queue */
	list_for_each_entry_safe(mbox, next_mbox, &psli->mboxq, list) {
		list_del_init(&mbox->list);
		mempool_free(mbox, phba->mbox_mem_pool);
	}

	psli->sliinit.sli_flag &= ~LPFC_SLI_MBOX_ACTIVE;
	if (psli->mbox_active) {
		mempool_free(psli->mbox_active, phba->mbox_mem_pool);
		psli->mbox_active = 0;
	}

	lpfc_mbuf_safety_pool_destroy(phba);
	lpfc_mem_safety_pool_destroy(phba);

	if (phba->lpfc_scsi_dma_ext_pool)
		pci_pool_destroy(phba->lpfc_scsi_dma_ext_pool);

	if (phba->lpfc_mbuf_pool)
		pci_pool_destroy(phba->lpfc_mbuf_pool);

	return 1;
}

void
lpfc_mem_safety_pool_create(struct lpfc_hba *phba, uint32_t count)
{

	phba->iocb_mem_pool = mempool_create( count,
					      lpfc_iocb_pool_alloc,
					      lpfc_iocb_pool_free,
					      NULL);

	phba->scsibuf_mem_pool = mempool_create( count,
						lpfc_scsibuf_pool_alloc,
						lpfc_scsibuf_pool_free,
						NULL);

	phba->mbox_mem_pool = mempool_create( count,
					      lpfc_mbox_pool_alloc,
					      lpfc_mbox_pool_free,
					      NULL);

	phba->nlp_mem_pool = mempool_create( count,
					     lpfc_nlp_pool_alloc,
					     lpfc_nlp_pool_free,
					     NULL);

	phba->bind_mem_pool = mempool_create( count,
					      lpfc_bind_pool_alloc,
					      lpfc_bind_pool_free,
					      NULL);
	return;
}


void
lpfc_mem_safety_pool_destroy(struct lpfc_hba *phba) {

	if (phba->iocb_mem_pool)
		mempool_destroy(phba->iocb_mem_pool);
	if (phba->mbox_mem_pool)
		mempool_destroy(phba->mbox_mem_pool);
	if (phba->nlp_mem_pool)
		mempool_destroy(phba->nlp_mem_pool);
	if (phba->bind_mem_pool)
		mempool_destroy(phba->bind_mem_pool);
	return;
}


static void
lpfc_mbuf_safety_pool_create(struct lpfc_hba * phba, uint32_t count)
{
	int i;

	struct lpfc_dma_pool *pool =
		&(phba->lpfc_mbuf_safety_pool);

	pool->elements = kmalloc(sizeof(struct lpfc_dmabuf) * count,
				 GFP_KERNEL);
	pool->max_count = 0;
	pool->current_count = 0;
	for ( i=0; i<count; i++) {
		pool->elements[i].virt =
			pci_pool_alloc(phba->lpfc_mbuf_pool,
				       GFP_KERNEL,
				       &(pool->elements[i].phys));

		if (!pool->elements[i].virt)
			break;
		pool->max_count++;
		pool->current_count++;
	}
}

static void
lpfc_mbuf_safety_pool_destroy(struct lpfc_hba * phba)
{
	struct lpfc_dma_pool *pool =
		&(phba->lpfc_mbuf_safety_pool);
	int i;

	if ( pool->max_count != pool->current_count)
		printk("Memory leaked in mbuf safety pool \n");

	for (i=0; i< pool->current_count; i++) {
		pci_pool_free(phba->lpfc_mbuf_pool,
			      pool->elements[i].virt,
			      pool->elements[i].phys);
	}

	kfree(pool->elements);
}

void *
lpfc_mbuf_alloc(struct lpfc_hba * phba, int mem_flags, dma_addr_t * handle)
{
	void *ret;
	struct lpfc_dma_pool *pool =
		&(phba->lpfc_mbuf_safety_pool);

	ret = pci_pool_alloc(phba->lpfc_mbuf_pool, GFP_ATOMIC, handle);
	/*
	   If we are low in memory and is priority memory allocation
	   use safety pool
	*/
	if ((!ret) && ( mem_flags & MEM_PRI)
	    && (pool->current_count)) {
		pool->current_count--;
		ret = pool->elements[pool->current_count].virt;
		*handle = pool->elements[pool->current_count].phys;
	}
	return ret;
}

void
lpfc_mbuf_free(struct lpfc_hba * phba, void *virt, dma_addr_t dma)
{
	struct lpfc_dma_pool *pool =
		&(phba->lpfc_mbuf_safety_pool);

	if ( pool->current_count < pool->max_count) {
		pool->elements[pool->current_count].virt = virt;
		pool->elements[pool->current_count].phys = dma;
		pool->current_count++;
		return;
	}

	pci_pool_free(phba->lpfc_mbuf_pool, virt, dma);
	return;
}

static void *
lpfc_scsibuf_pool_alloc(int gfp_flags, void *data) {
	void * ret;

	ret = kmalloc(sizeof (struct lpfc_scsi_buf), gfp_flags);

	return ret;
}

static void
lpfc_scsibuf_pool_free(void *obj, void *data) {

	kfree(obj);
}




static void *
lpfc_bind_pool_alloc(int gfp_flags, void *data) {
	void * ret;

	ret = kmalloc(sizeof (struct lpfc_bindlist), gfp_flags);

	return ret;
}

static void
lpfc_bind_pool_free(void *obj, void *data) {

	kfree(obj);
}

static void *
lpfc_mbox_pool_alloc(int gfp_flags, void *data) {
	void * ret;

	ret = kmalloc(sizeof (LPFC_MBOXQ_t), gfp_flags);

	return ret;
}

static void
lpfc_mbox_pool_free(void *obj, void *data) {

	kfree(obj);
}

static void *
lpfc_nlp_pool_alloc(int gfp_flags, void *data) {
	void * ret;

	ret = kmalloc(sizeof (struct lpfc_nodelist), gfp_flags);

	return ret;
}

static void
lpfc_nlp_pool_free(void *obj, void *data) {

	kfree(obj);
}

static void *
lpfc_iocb_pool_alloc(int gfp_flags, void *data) {
	void * ret;

	ret = kmalloc(sizeof (struct lpfc_iocbq), gfp_flags);

	return ret;
}

static void
lpfc_iocb_pool_free(void *obj, void *data) {

	kfree(obj);
}

/* static void * */
/* lpfc_scsibuf_pool_alloc(int gfp_flags, void *data) { */
/* 	void * ret; */

/* 	ret = kmalloc(sizeof (struct lpfc_scsi_buf), gfp_flags); */

/* 	return ret; */
/* } */

/* static void */
/* lpfc_scsibuf_pool_free(void *obj, void *data) { */

/* 	kfree(obj); */
/* } */

