/*
 *   ALSA tvmixer for video grabber cards with audio chips such as the msp3400.

 *   Copyright (C) 2004 Perry Gilfillan
 *	(c) 2000 Gerd Knorr
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <sound/driver.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/moduleparam.h>
#include <sound/core.h>
#include <linux/i2c.h>
#include <linux/videodev.h>
#include <sound/initval.h>
#include <sound/control.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include "i2c-compat.h"
#endif

#define DEB(x) if (debug>=1) x
#define DEB2(x) if (debug>=2) x

// module parameters (see "Module Parameters")
static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;
static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;
static int enable[SNDRV_CARDS] = SNDRV_DEFAULT_ENABLE_PNP;
static int debug = 0;
static int boot_devs;

module_param_array(index, int, boot_devs, 0444);
MODULE_PARM_DESC(index, "Index value for tvmixer.");
MODULE_PARM_SYNTAX(index, SNDRV_INDEX_DESC);
module_param_array(id, charp, boot_devs, 0444);
MODULE_PARM_DESC(id, "ID string for tvmixer.");
MODULE_PARM_SYNTAX(id, SNDRV_ID_DESC);
module_param_array(enable, bool, boot_devs, 0444);
MODULE_PARM_DESC(enable, "Enable tvmixer.");
MODULE_PARM_SYNTAX(enable, SNDRV_ENABLE_DESC);
module_param(debug, int, 0444);
MODULE_PARM_DESC(debug, "Debug level (0-3");

MODULE_AUTHOR("Perry Gilfillan <perrye@gilfillan.org>");
MODULE_LICENSE("GPL");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif

// definition of the chip-specific record
typedef struct snd_tvmixer tvmixer_t;
#define chip_t tvmixer_t

struct snd_tvmixer {
	snd_card_t *snd_card;
	struct i2c_client *tvac_client;
#define TVMIXER_CONTROLS 4
	snd_kcontrol_t *kctls[TVMIXER_CONTROLS];
};

static tvmixer_t *mixers[SNDRV_CARDS];

/* ----------------------------------------------------------------------- */

static int snd_tvmixer_adapters(struct i2c_adapter *adap);
static int snd_tvmixer_clients(struct i2c_client *tvac_client);

/* ----------------------------------------------------------------------- */

static int snd_tvmixer_volume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 65535;
	return 0;
}
 
static int snd_tvmixer_volume_get(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct video_audio va;
	tvmixer_t *tvmixer = snd_kcontrol_chip(kcontrol);

	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCGAUDIO, &va);

	ucontrol->value.integer.value[0] =
		(min(65536 - va.balance,32768) * va.volume) / 32768;
	ucontrol->value.integer.value[1] =
		(min(va.balance,(u16)32768) * va.volume) / 32768;
	return 0;
}

static int snd_tvmixer_volume_put(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	int vol[2];
	struct video_audio va;
	tvmixer_t *tvmixer = snd_kcontrol_chip(kcontrol);

	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client,VIDIOCGAUDIO,&va);

	vol[0] = ucontrol->value.integer.value[0];
	vol[1] = ucontrol->value.integer.value[1];
	va.volume = max(vol[0], vol[1]);
	va.balance = (32768*min(vol[0],vol[1])) / 
			(va.volume ? va.volume : 1);
	va.balance = (vol[0]<vol[1]) ? 
		(65535-va.balance) : va.balance;
	if (va.volume) /* FIXME: Workaround for stuck muting */
		va.flags &= ~VIDEO_AUDIO_MUTE;
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCSAUDIO, &va);
		
	return 0;
}

static int snd_tvmixer_mute_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int snd_tvmixer_mute_get(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct video_audio va;
	tvmixer_t *tvmixer = snd_kcontrol_chip(kcontrol);

	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCGAUDIO, &va);

	ucontrol->value.integer.value[0] = (va.flags & VIDEO_AUDIO_MUTE);
	return 0;
}

static int snd_tvmixer_mute_put(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct video_audio va;
	tvmixer_t *tvmixer = snd_kcontrol_chip(kcontrol);

	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client,VIDIOCGAUDIO,&va);

	if (ucontrol->value.integer.value[0])
		va.flags &= ~VIDEO_AUDIO_MUTE;
	else
		va.flags |= VIDEO_AUDIO_MUTE;

	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCSAUDIO, &va);
		
	return 0;
}

static int snd_tvmixer_bass_treble_info(snd_kcontrol_t *kcontrol, 
		snd_ctl_elem_info_t * uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 65535;
	return 0;
}

static int snd_tvmixer_bass_treble_get(snd_kcontrol_t * kcontrol, 
		snd_ctl_elem_value_t * ucontrol)
{
	struct video_audio va;
	tvmixer_t *tvmixer = snd_kcontrol_chip(kcontrol);

	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCGAUDIO, &va);
	ucontrol->value.integer.value[0] = 
		kcontrol->private_value ? va.bass : va.treble;
	return 0;
}

static int snd_tvmixer_bass_treble_put(snd_kcontrol_t * kcontrol, 
		snd_ctl_elem_value_t * ucontrol)
{
	struct video_audio va;
	tvmixer_t *tvmixer = snd_kcontrol_chip(kcontrol);
	
	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCGAUDIO, &va);

	if (kcontrol->private_value)
		va.bass = 
			ucontrol->value.integer.value[0];
	else
		va.treble = 
			ucontrol->value.integer.value[0];

	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCSAUDIO, &va);
	return 0;
}

static snd_kcontrol_new_t snd_tvmixer_controls[TVMIXER_CONTROLS] = {
{
	.name = "Master Volume",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	.info = snd_tvmixer_volume_info,
	.get = snd_tvmixer_volume_get,
	.put = snd_tvmixer_volume_put,
},
{
	.name = "Master Switch",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	.info = snd_tvmixer_mute_info,
	.get = snd_tvmixer_mute_get,
	.put = snd_tvmixer_mute_put,
},
{
	.name = "Tone Control - Treble",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	.private_value = 0,
	.info = snd_tvmixer_bass_treble_info,
	.get = snd_tvmixer_bass_treble_get,
	.put = snd_tvmixer_bass_treble_put,
},
{
	.name = "Tone Control - Bass",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	.private_value = 1,
	.info = snd_tvmixer_bass_treble_info,
	.get = snd_tvmixer_bass_treble_get,
	.put = snd_tvmixer_bass_treble_put,
},
};

int snd_tvmixer_build(tvmixer_t *tvmixer)
{
	snd_kcontrol_t *kctl;
	unsigned int idx;
	int err;

	for (idx = 0; idx < TVMIXER_CONTROLS; idx++) {
		kctl = snd_ctl_new1(&snd_tvmixer_controls[idx], tvmixer);
		if (kctl == NULL)
			return -ENOMEM;
		err = snd_ctl_add(tvmixer->snd_card, kctl);
		if (err < 0)
			return err;
		tvmixer->kctls[idx] = kctl;
	}
	return 0;
}

/* ----------------------------------------------------------------------- */

static int snd_tvmixer_free(tvmixer_t *tvmixer)
{
	struct video_audio va;

	memset(&va, 0, sizeof(va));
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client,VIDIOCGAUDIO,&va);

	va.flags = VIDEO_AUDIO_MUTE;
	tvmixer->tvac_client->driver->command(
		tvmixer->tvac_client, VIDIOCSAUDIO, &va);

	printk(KERN_INFO "snd-tvmixer: %s unregistered\n",
		i2c_clientname(tvmixer->tvac_client));
	snd_magic_kfree(tvmixer);
	return 0;
}

static int snd_tvmixer_dev_free(snd_device_t *device)
{
	tvmixer_t *tvmixer = snd_magic_cast(tvmixer_t,
		device->device_data, return -ENXIO);
	return snd_tvmixer_free(tvmixer);
}

static int snd_tvmixer_create(snd_card_t *snd_card,
				struct i2c_client *tvac_client,
				tvmixer_t **rtvmixer)
{
	int err = 0;
	tvmixer_t *tvmixer;

	static snd_device_ops_t snd_tvmixer_dev_ops = {
		.dev_free = snd_tvmixer_dev_free,
	};

	tvmixer = *rtvmixer = NULL;
	  
	tvmixer = snd_magic_kcalloc(tvmixer_t, 0, GFP_KERNEL);
	if (NULL == tvmixer) {
		printk(KERN_INFO 
			"snd-tvmixer: snd_magic_kcalloc failed");
		return -ENOMEM;
	}

	strcpy(snd_card->mixername, "TV Mixer");

	tvmixer->snd_card = snd_card;
	tvmixer->tvac_client = tvac_client;

	err = snd_device_new(snd_card, SNDRV_DEV_LOWLEVEL,
				tvmixer, &snd_tvmixer_dev_ops);
	if (err < 0) {
		printk(KERN_INFO
			"snd-tvmixer: snd_device_new failed: err = %d\n", err);
		snd_tvmixer_free(tvmixer);
		return err;
	}

	*rtvmixer = tvmixer;
	return 0;
}

static int snd_tvmixer_adapters(struct i2c_adapter *adap)
{
#ifdef I2C_ADAP_CLASS_TV_ANALOG
	if (adap->class != I2C_ADAP_CLASS_TV_ANALOG)
		return 0;
#else
	switch (adap->id) {
		case I2C_ALGO_BIT | I2C_HW_B_VOO: 
//		{
//			if (!strnicmp(adap->name, "DDC\0", 3))
//				return 0;
//		}
		case I2C_ALGO_BIT | I2C_HW_B_BT848:
		case I2C_ALGO_BIT | I2C_HW_B_RIVA:
			break;
		default:
			return 0;
	}
#endif
	DEB(printk(KERN_INFO "snd-tvmixer: snd_tvmixer_adapters found %s\n",
		adap->name));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	{
		struct list_head *p;
		list_for_each(p, &adap->clients) {
			struct i2c_client *client;
			client = list_entry(p, struct i2c_client, list);
			snd_tvmixer_clients(client);
		}
	}
#else
	{
		int i;
		for (i=0; i<I2C_CLIENT_MAX; i++) {
			if (NULL == adap->clients[i])
				continue;
			snd_tvmixer_clients(adap->clients[i]);
		}
	}
#endif
	return 0;
}

static int snd_tvmixer_clients(struct i2c_client *tvac_client)
{

	snd_card_t *snd_card;
	tvmixer_t *tvmixer;
	struct video_audio va;
	static int cards = 0;
	int slot;
	int err = 0;

	/* unregister ?? */
	for (slot = 0; slot < SNDRV_CARDS; slot++) {
		tvmixer = mixers[slot];
		if (NULL == tvmixer)
			continue;
		if (tvmixer->tvac_client == tvac_client) {
			printk(KERN_INFO "snd-tvmixer: "
			       "snd_tvmixer_clients releasing %s\n",
			       tvmixer->snd_card->longname);
			snd_card_free(tvmixer->snd_card);
			mixers[slot] = NULL;
			cards--;
		}
	}

	/* look for a free slot */
	for (slot = 0; slot < SNDRV_CARDS; slot++)
		if (NULL == mixers[slot])
			break;
	if (slot == SNDRV_CARDS) {
		printk(KERN_WARNING "snd-tvmixer: mixers[] too small\n");
		return -1;
	}

	/* audio chip with mixer ??? */
	if (NULL == tvac_client->driver->command)
		return -1;

	DEB(printk("snd-tvmixer: Found i2c_client %s\n", tvac_client->name));

	memset(&va, 0, sizeof(va));
	if (0 != tvac_client->driver->command(tvac_client, VIDIOCGAUDIO, &va)) {
		printk(KERN_WARNING "snd-tvmixer: i2c_client not a Mixer?\n");
		return -1;
	}
	if (0 == (va.flags & VIDEO_AUDIO_VOLUME)) {
		printk(KERN_WARNING "snd-tvmixer: i2c_client does not have a volume control\n");
		return -1;
	}

	/* ALSA setup */
	/* did the last client max out the cards */
	if (cards >= SNDRV_CARDS) {
		printk(KERN_WARNING "snd-tvmixer: SNDRV_CARDS too small\n");
		return -ENODEV;
	}
	if (!enable[cards]) {
		cards++;
		printk(KERN_WARNING "snd-tvmixer: enable[%d] = 0\n", cards);
		return -ENOENT;
	}

	snd_card = snd_card_new(index[cards], id[cards], THIS_MODULE, 0);
	if (NULL == snd_card) {
		printk(KERN_WARNING "snd-tvmixer: snd_card_new failed\n");
		return -ENOMEM;
	}

	err = snd_tvmixer_create(snd_card, tvac_client, &tvmixer);
	if (err) {
		printk(KERN_WARNING "snd-tvmixer: snd_tvmixer_create failed: %d\n", err);
		snd_card_free(snd_card);
		return err;
	}
	
	strcpy(snd_card->driver, "ALSA TV Mixer");
	strcpy(snd_card->shortname, "ALSA TV Mixer");
	sprintf(snd_card->longname, "%s(%d) on %s",
		snd_card->shortname, cards, tvmixer->tvac_client->name);

	snd_tvmixer_build(tvmixer);

	err = snd_card_register(snd_card);
	if (err) {
		printk(KERN_WARNING "snd-tvmixer: snd_card_register failed\n");
		snd_card_free(snd_card);
		return err;
	}

	mixers[slot] = tvmixer;
	cards++;
	
	printk(KERN_INFO "snd-tvmixer: %s (%s) registered\n",
		tvmixer->tvac_client->name,
		tvmixer->tvac_client->adapter->name);
	
	return 0;
}

/* ----------------------------------------------------------------------- */

static struct i2c_driver snd_tvmixer_i2c_driver = {
	.owner           = THIS_MODULE,
	.name            = "tv card mixer driver",
        .id              = I2C_DRIVERID_TVMIXER,
#ifdef I2C_DF_DUMMY
	.flags           = I2C_DF_DUMMY,
#else
	.flags           = I2C_DF_NOTIFY,
        .detach_adapter  = snd_tvmixer_adapters,
#endif
        .attach_adapter  = snd_tvmixer_adapters,
        .detach_client   = snd_tvmixer_clients,
};

static int snd_tvmixer_init_module(void)
{
	i2c_add_driver(&snd_tvmixer_i2c_driver);
	return 0;
}

static void snd_tvmixer_cleanup_module(void)
{
	tvmixer_t *tvmixer;
	int slot;
	
	i2c_del_driver(&snd_tvmixer_i2c_driver);
	for (slot = 0; slot < SNDRV_CARDS; slot++) {
		tvmixer = mixers[slot];
		if (tvmixer) {
			snd_card_free(tvmixer->snd_card);
			mixers[slot] = NULL;
		}
	}
}

module_init(snd_tvmixer_init_module);
module_exit(snd_tvmixer_cleanup_module);

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
