// ****************************************************************************
//
//  	CDarlaDspCommObject.cpp
//
// Implementation file for Darla20 DSP interface class.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"

#include "DSP/Darla20DSP.c"


static int InitHw(struct EchoGals *eg, u16 Device, u16 SubDevice)
{
	int err;

	ECHO_DEBUGPRINTF(("InitHw() - Darla20\n"));

	if ((SubDevice & 0xfff0) != DARLA20) {
		ECHO_DEBUGPRINTF(("Wrong card type\n"));
		return -ENODEV;
	}

	// This part is common to all the cards
	if ((err = InitDspCommPage(eg))) {
		ECHO_DEBUGPRINTF(("InitHw - could not initialize DSP comm page\n"));
		return err;
	}

	eg->DeviceId = Device;
	eg->SubDeviceId = SubDevice;
	eg->CardType = DARLA20;
	eg->BadBoard = 1;
	eg->NumPipesOut = 8;
	eg->NumPipesIn = 2;
	eg->NumBussesOut = 8;
	eg->NumBussesIn = 2;
	eg->FirstDigitalBusOut = 8;
	eg->FirstDigitalBusIn = 2;
	eg->HasVmixer = FALSE;
	eg->NumMidiOut = 0;
	eg->NumMidiIn = 0;
	eg->DspCodeToLoad = Darla20DSP;
	eg->SpdifStatus = GD_SPDIF_STATUS_UNDEF;
	eg->ClockState = GD_CLOCK_UNDEF;
	// Since this card has no ASIC, mark it as loaded so everything works OK
	eg->ASICLoaded = TRUE;
	eg->Flags =	ECHOGALS_FLAG_BADBOARD |
			ECHOCAPS_OUTPUT_GAIN |
			ECHOCAPS_VU_METER |
			ECHOGALS_FLAG_SPDIF_NODITHER;
	eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL;

	// Load the DSP and the ASIC on the PCI card
	err = LoadFirmware(eg);
	if (err || eg->BadBoard)
		return -EIO;

	eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;

	// Must call this here after DSP is init to init gains and mutes
	if ((err = InitLineLevels(eg)))
		return err;

	// Get default sample rate from DSP
	eg->SampleRate = GetSampleRate(eg);

	ECHO_DEBUGPRINTF(("InitHw done\n"));
	return err;
}



/* QueryAudioSampleRate is used to find out if this card can handle a given sample rate. */
static int QueryAudioSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	if (SampleRate != 44100 && SampleRate != 48000) {
		ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate must be 44,100 Hz or 48,000 Hz\n"));
		return -EINVAL;
	}

	ECHO_DEBUGPRINTF(("QueryAudioSampleRate()\n"));
	return ECHOSTATUS_OK;
}



/* The Darla20 has no external clock sources */
static u32 GetInputClockDetect(const struct EchoGals * eg)
{
	return ECHO_CLOCK_BIT_INTERNAL;
}



// The Darla20 has no ASIC. Just do nothing
static int LoadASIC(struct EchoGals *eg)
{
	return TRUE;
}



//===========================================================================
//
// DspSetSampleRate
//
// Set the audio sample rate for Gina20/Darla20
//
// For Gina and Darla, three parameters need to be set: the resampler state,
// the clock state, and the S/PDIF output status state.  The idea is that
// these parameters are changed as infrequently as possible.
//
// Resampling is no longer supported.
//
//===========================================================================

static int DspSetSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	u8 ClockState, SpdifStatus;

	if (WaitForHandshake(eg))
		return -EIO;

	// Pick the clock
	if (eg->InputClock == ECHO_CLOCK_SPDIF) {	// Gina20 only
		ClockState = GD_CLOCK_SPDIFIN;
	} else {
		switch (SampleRate) {
		case 44100:
			ClockState = GD_CLOCK_44;
			break;
		case 48000:
			ClockState = GD_CLOCK_48;
			break;
		default:
			ClockState = GD_CLOCK_NOCHANGE;
			break;
		}
	}
	if (eg->ClockState == ClockState)
		ClockState = GD_CLOCK_NOCHANGE;
//@@ottim
	// Select S/PDIF output status
	switch (SampleRate) {
	case 44100:
		SpdifStatus = GD_SPDIF_STATUS_44;
		break;
	case 48000:
		SpdifStatus = GD_SPDIF_STATUS_48;
		break;
	default:
		SpdifStatus = GD_SPDIF_STATUS_NOCHANGE;
		break;
	}
	if (SpdifStatus == eg->SpdifStatus)
		SpdifStatus = GD_SPDIF_STATUS_NOCHANGE;

	// Write the audio state to the comm page
	eg->DspCommPage->SampleRate = SWAP32(SampleRate);
	eg->DspCommPage->GDClockState = ClockState;
	eg->DspCommPage->GDSpdifStatus = SpdifStatus;
	eg->DspCommPage->GDResamplerState = 3;	// magic number - should always be 3

	// Send command to DSP
	ClearHandshake(eg);
	SendVector(eg, DSP_VC_SET_GD_AUDIO_STATE);

	// Save the new audio state
	if (ClockState != GD_CLOCK_NOCHANGE)
		eg->ClockState = ClockState;

	if (SpdifStatus != GD_SPDIF_STATUS_NOCHANGE)
		eg->SpdifStatus = SpdifStatus;

	return SampleRate;
}



// The Darla20 has no external clock sources
static int SetInputClock(struct EchoGals *eg, u16 Clock)
{
	if (Clock != ECHO_CLOCK_INTERNAL)
		return -EINVAL;
	else
		return ECHOSTATUS_OK;
}


/*
// Layla20 only
static int SetOutputClock(struct EchoGals *eg, u16 Clock)
{
	return -EINVAL;
}
*/

/*
// This card has no vmixer.
static int SetVmixerGain(struct EchoGals *eg, u16 Output, u16 Pipe, int Gain)
{
	return -EINVAL;
}
*/

/*
// This card has no vmixer.
static int UpdateVmixerLevel(struct EchoGals *eg)
{
	return -EINVAL;
}
*/

/*
// Set input bus gain (one unit is 0.5dB !)
static int SetInputGain(struct EchoGals *eg, u16 Channel, int Gain)
{
	return -EINVAL;
}
*/

/*
static // Gina24, Mona, Layla24 - only
int DspSetDigitalMode(struct EchoGals *eg, u8 Mode)
{
	return -EINVAL;
}
*/


/* This card has no digital channels
static int SetProfessionalSpdif(struct EchoGals *eg, char Prof)
{
	return -EINVAL;
}*/
