// ****************************************************************************
//
//  	CDarlaDspCommObject.cpp
//
// Implementation file for Darla20 DSP interface class.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"

#include "DSP/Darla24DSP.c"

static int SetNominalLevel(struct EchoGals *eg, u16 Index, char Consumer);


static int InitHw(struct EchoGals *eg, u16 Device, u16 SubDevice)
{
	int err, i;

	ECHO_DEBUGPRINTF(("InitHw() - Darla24\n"));

	if ((SubDevice & 0xfff0) != DARLA24) {
		ECHO_DEBUGPRINTF(("Wrong card type\n"));
		return -ENODEV;
	}
	
	// This part is common to all the cards
	if ((err = InitDspCommPage(eg))) {
		ECHO_DEBUGPRINTF(("InitHw - could not initialize DSP comm page\n"));
		return err;
	}

	eg->DeviceId = Device;
	eg->SubDeviceId = SubDevice;
	eg->CardType = DARLA24;
	eg->BadBoard = 1;
	eg->NumPipesOut = 8;
	eg->NumPipesIn = 2;
	eg->NumBussesOut = 8;
	eg->NumBussesIn = 2;
	eg->FirstDigitalBusOut = 8;
	eg->FirstDigitalBusIn = 2;
	eg->HasVmixer = FALSE;
	eg->NumMidiOut = 0;
	eg->NumMidiIn = 0;
	eg->DspCodeToLoad = Darla24DSP;
	// Since this card has no ASIC, mark it as loaded so everything works OK
	eg->ASICLoaded = TRUE;
	eg->Flags =	ECHOGALS_FLAG_BADBOARD |
			ECHOCAPS_OUTPUT_GAIN |
			ECHOCAPS_VU_METER |
			ECHOGALS_FLAG_SPDIF_NODITHER |
			ECHOCAPS_INPUT_NOMINAL_LEVEL |
			ECHOCAPS_OUTPUT_NOMINAL_LEVEL |
			ECHOCAPS_SUPER_INTERLEAVE;
	eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_ESYNC;

	// Load the DSP and the ASIC on the PCI card
	err = LoadFirmware(eg);
	if (err || eg->BadBoard)
		return -EIO;

	// Clear the "bad board" flag; set the flags to indicate that Darla24 can handle super-interleave
	eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;

	// Must call this here after DSP is init to init gains and mutes
	if ((err = InitLineLevels(eg)))
		return err;

	// Set professional nominal levels. This card has no digital i/o, so analog i/o are contiguous
	for (i = 0; i < eg->NumBussesOut + eg->NumBussesIn; i++)
		err = SetNominalLevel(eg, i, FALSE);	// FALSE is +4dBu here

	// Get default sample rate from DSP
	eg->SampleRate = GetSampleRate(eg);

	ECHO_DEBUGPRINTF(("InitHw done\n"));
	return err;
}



//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a given sample rate.
//
//===========================================================================

static int QueryAudioSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	if (SampleRate != 8000 && SampleRate != 11025 && SampleRate != 16000 &&
	    SampleRate != 22050 && SampleRate != 32000 && SampleRate != 44100 && 
	    SampleRate != 48000 && SampleRate != 88200 && SampleRate != 96000) {
		ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Invalid sample rate\n"));
		return -EINVAL;
	}

	ECHO_DEBUGPRINTF(("QueryAudioSampleRate()\n"));
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Darla24 only supports Esync input clock.
//
//===========================================================================

static u32 GetInputClockDetect(const struct EchoGals *eg)
{
	u32 ClocksFromDsp, ClockDetectBits;

	// Map the DSP clock detect bits to the generic driver clock detect bits
	ClocksFromDsp = SWAP32(eg->DspCommPage->StatusClocks);

	ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

	if (ClocksFromDsp & GLDM_CLOCK_DETECT_BIT_ESYNC)
		ClockDetectBits |= ECHO_CLOCK_BIT_ESYNC;

	return ClockDetectBits;
}



// The Darla24 has no ASIC. Just do nothing
static int LoadASIC(struct EchoGals *eg)
{
	return TRUE;
}



//===========================================================================
//
// DspSetSampleRate
// 
// Set the audio sample rate for Darla24; this is fairly simple.  You
// just pick the right magic number.
//
//===========================================================================

static int DspSetSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	u8 Clock;

	// Pick the magic number
	switch (SampleRate) {
	case 96000:
		Clock = GD24_96000;
		break;
	case 88200:
		Clock = GD24_88200;
		break;
	case 48000:
		Clock = GD24_48000;
		break;
	case 44100:
		Clock = GD24_44100;
		break;
	case 32000:
		Clock = GD24_32000;
		break;
	case 22050:
		Clock = GD24_22050;
		break;
	case 16000:
		Clock = GD24_16000;
		break;
	case 11025:
		Clock = GD24_11025;
		break;
	case 8000:
		Clock = GD24_8000;
		break;
	default:
		ECHO_DEBUGPRINTF(("DspSetSampleRate: Error, invalid sample rate %d\n", SampleRate));
		return -EINVAL;
	}

	if (WaitForHandshake(eg))
		return -EIO;

	// Override the sample rate if this card is set to Echo sync.
	// DspCommPage->wInputClock is just being used as a parameter here;
	// the DSP ignores it.
	if (eg->InputClock == ECHO_CLOCK_ESYNC)
		Clock = GD24_EXT_SYNC;

	eg->DspCommPage->SampleRate = SWAP32(SampleRate);

	// Write the audio state to the comm page
	eg->DspCommPage->GDClockState = Clock;

	// Send command to DSP
	ClearHandshake(eg);
	SendVector(eg, DSP_VC_SET_GD_AUDIO_STATE);

	ECHO_DEBUGPRINTF(("DspSetSampleRate: %d clock %d\n", SampleRate, Clock));
	return GetSampleRate(eg);
}



//===========================================================================
//
// Set input clock
//
// Darla24 supports internal and Esync clock.
//
//===========================================================================

static int SetInputClock(struct EchoGals *eg, u16 Clock)
{
	if (Clock != ECHO_CLOCK_INTERNAL && Clock != ECHO_CLOCK_ESYNC)
		return -EINVAL;

	eg->InputClock = Clock;

	return DspSetSampleRate(eg, GetSampleRate(eg));
}


/*
// Layla20 only
static int SetOutputClock(struct EchoGals *eg, u16 Clock)
{
	return -EINVAL;
}
*/

/*
// This card has no vmixer.
static int SetVmixerGain(struct EchoGals *eg, u16 Output, u16 Pipe, int Gain)
{
	return -EINVAL;
}
*/

/*
// Gina24, Mona, Layla24 - only
static int DspSetDigitalMode(struct EchoGals *eg, u8 Mode)
{
	return -EINVAL;
}
*/

/*
// This card has no vmixer.
static int UpdateVmixerLevel(struct EchoGals *eg)
{
	return -EINVAL;
}
*/

/*
// Set input bus gain (one unit is 0.5dB !)
static int SetInputGain(struct EchoGals *eg, u16 Channel, int Gain)
{
	return -EINVAL;
}
*/


/* This card has no digital i/o
static int SetProfessionalSpdif(struct EchoGals *eg, char Prof)
{
	return -EINVAL;
}*/
