// ****************************************************************************
//
// CDspCommObject.cpp
//
// Implementation file for EchoGals generic driver DSP interface class.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#ifndef ECHO24_FAMILY
#define ECHOGALS_FAMILY
#endif

#include "Echoaudio.h"
#include <linux/delay.h>

#ifdef DSP_56361
#include "DSP/LoaderDSP.c"
#endif

static int RestoreDspSettings(struct EchoGals *eg);


static u64 GetSystemTime(void)
{
	struct timeval tv;

	do_gettimeofday(&tv);
	return (1000000ULL * tv.tv_sec + tv.tv_usec);
}



// Tell the DSP to go into low-power mode
static int GoComatose(struct EchoGals *eg)
{
	ECHO_DEBUGPRINTF(("GoComatose\n"));
	if (eg->DspCode) {
		// Make LoadFirmware do a complete reload
		eg->DspCode = NULL;
		// Put the DSP to sleep
		return SendVector(eg, DSP_VC_GO_COMATOSE);
	}
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// Some vector commands involve the DSP reading or writing data to and
// from the comm page; if you send one of these commands to the DSP,
// it will complete the command and then write a non-zero value to
// the Handshake field in the comm page.  This function waits for the
// handshake to show up.
//
//===========================================================================

static int WaitForHandshake(struct EchoGals *eg)
{
	int i;

	// Wait up to 10ms for the handshake from the DSP
	for (i = 0; i < HANDSHAKE_TIMEOUT; i++) {
		// Look for the handshake value
		if (eg->DspCommPage->Handshake) {
			//if (i)  ECHO_DEBUGPRINTF(("Handshake time: %d\n", i));
			return 0;
		}
		udelay(1);
	}

	ECHO_DEBUGPRINTF(("WaitForHandshake: Timeout waiting for DSP\n"));
	return -EBUSY;
}



//===========================================================================
//
// Much of the interaction between the DSP and the driver is done via vector
// commands; SendVector writes a vector command to the DSP.  Typically,
// this causes the DSP to read or write fields in the comm page.
//
// Returns ECHOSTATUS_OK if sent OK.
//
//===========================================================================

static int SendVector(struct EchoGals *eg, u32 Command)
{
	int i;

	WMB;
	// Wait up to 100ms for the "vector busy" bit to be off
	for (i = 0; i < VECTOR_BUSY_TIMEOUT; i++) {
		if (!(GetDspRegister(eg, CHI32_VECTOR_REG) & CHI32_VECTOR_BUSY)) {
			SetDspRegister(eg, CHI32_VECTOR_REG, Command);
			WMB;
			//if (i)  ECHO_DEBUGPRINTF(("SendVector time: %d\n", i));
			return 0;
		}
		udelay(1);
	}

	ECHO_DEBUGPRINTF(("timeout on SendVector\n"));
	return -EBUSY;
}



//===========================================================================
//
// Write_DSP writes a 32-bit value to the DSP; this is used almost
// exclusively for loading the DSP.
//
//===========================================================================

static int Write_DSP(struct EchoGals *eg, u32 Data)
{
	u32 Status;
	u64 CurTime, Timeout;

//      ECHO_DEBUGPRINTF(("Write_DSP\n"));
	CurTime = GetSystemTime();
	Timeout = CurTime + 10000000L;	// 10 sec.
	while (Timeout >= CurTime) {
		MB;
		Status = GetDspRegister(eg, CHI32_STATUS_REG);
		if ((Status & CHI32_STATUS_HOST_WRITE_EMPTY) != 0) {
			SetDspRegister(eg, CHI32_DATA_REG, Data);
			MB;
//                      ECHO_DEBUGPRINTF(("Write DSP: 0x%x", Data));
			return ECHOSTATUS_OK;
		}
		CurTime = GetSystemTime();
	}

	eg->BadBoard = TRUE;	// Set TRUE until DSP re-loaded
	ECHO_DEBUGPRINTF(("Write_DSP: Set BadBoard to TRUE\n"));
	return -EIO;
}



//===========================================================================
//
// Read_DSP reads a 32-bit value from the DSP; this is used almost
// exclusively for loading the DSP and checking the status of the ASIC.
//
//===========================================================================

static int Read_DSP(struct EchoGals *eg, u32 *Data)
{
	u32 Status;
	u64 CurTime, Timeout;

//  ECHO_DEBUGPRINTF(("Read_DSP\n"));
	CurTime = GetSystemTime();

	Timeout = CurTime + READ_DSP_TIMEOUT;
	while (Timeout >= CurTime) {
		Status = GetDspRegister(eg, CHI32_STATUS_REG);
		if ((Status & CHI32_STATUS_HOST_READ_FULL) != 0) {
			*Data = GetDspRegister(eg, CHI32_DATA_REG);
//       ECHO_DEBUGPRINTF(("Read DSP: 0x%x\n", *pdwData));
			return ECHOSTATUS_OK;
		}
		CurTime = GetSystemTime();
	}

	eg->BadBoard = TRUE;	// Set TRUE until DSP re-loaded
	ECHO_DEBUGPRINTF(("Read_DSP: Set BadBoard to TRUE\n"));
	return -EIO;
}



/****************************************************************************

	Firmware loading functions

 ****************************************************************************/

//===========================================================================
//
// This function is used to read back the serial number from the DSP;
// this is triggered by the SET_COMMPAGE_ADDR command.
//
// Only some early Echogals products have serial numbers in the ROM;
// the serial number is not used, but you still need to do this as
// part of the DSP load process.
//
//===========================================================================

static int ReadSn(struct EchoGals *eg)
{
	int j;
	u32 Sn[6];

	for (j = 0; j < 5; j++) {
		if (Read_DSP(eg, &Sn[j])) {
			ECHO_DEBUGPRINTF(("Failed to read serial number word %d\n", j));
			return -EIO;
		}
	}
	ECHO_DEBUGPRINTF(("Read serial number %08x %08x %08x %08x %08x\n", Sn[0], Sn[1], Sn[2], Sn[3], Sn[4]));
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// ASIC status check - some cards have one or two ASICs that need to be
// loaded.  Once that load is complete, this function is called to see if
// the load was successful.
//
// If this load fails, it does not necessarily mean that the hardware is
// defective - the external box may be disconnected or turned off.
//
// This routine sometimes fails for Layla20; for Layla20, the loop runs 5 times
// and succeeds if it wins on three of the loops.
//
//===========================================================================

#ifdef ECHOCARD_HAS_ASIC

static int CheckAsicStatus(struct EchoGals *eg)
{
#if NUM_ASIC_TESTS!=1
	int GoodCt = 0;
#endif
	u32 AsicStatus;
	int st, NumTests, i;

	if (!eg->HasASIC) {
		eg->ASICLoaded = TRUE;
		return TRUE;
	}

	NumTests = NUM_ASIC_TESTS;

	for (i = 0; i < NumTests; i++) {
		// Send the vector command
		SendVector(eg, DSP_VC_TEST_ASIC);

		// The DSP will return a value to indicate whether or not the ASIC is currently loaded
		st = Read_DSP(eg, &AsicStatus);
		if (st != ECHOSTATUS_OK) {
			ECHO_DEBUGPRINTF(("CheckAsicStatus: failed on Read_DSP\n"));
			eg->ASICLoaded = FALSE;
			return FALSE;
		}

		if (AsicStatus == ASIC_ALREADY_LOADED) {
#if NUM_ASIC_TESTS==1
			eg->ASICLoaded = TRUE;
			break;
#else
			// For Layla20
			if (++GoodCt == 3) {
				eg->ASICLoaded = TRUE;
				break;
			}
#endif
		} else {
			eg->ASICLoaded = FALSE;
		}
	}

	return eg->ASICLoaded;
}

#else /* ECHOCARD_HAS_ASIC */

static inline int CheckAsicStatus(struct EchoGals *eg)
{
	eg->ASICLoaded = TRUE;
	return TRUE;
}

#endif /* ECHOCARD_HAS_ASIC */



//===========================================================================
//
//      Load ASIC code - done after the DSP is loaded
//
//===========================================================================

#ifdef ECHOCARD_HAS_ASIC

static int LoadASICgeneric(struct EchoGals *eg, u32 Cmd, u8 *Code, u32 Size)
{
	u32 i;

	if (!eg->HasASIC)
		return TRUE;

#ifdef _DEBUG
	u32 Checksum = 0;
	u64 StartTime, CurTime;
	StartTime = GetSystemTime();
#endif

	// Send the "Here comes the ASIC" command
	if (Write_DSP(eg, Cmd) != ECHOSTATUS_OK)
		return FALSE;

	// Write length of ASIC file in bytes
	if (Write_DSP(eg, Size) != ECHOSTATUS_OK)
		return FALSE;

	for (i = 0; i < Size; i++) {
#ifdef _DEBUG
		Checksum += Code[i];
#endif
		if (Write_DSP(eg, Code[i]) != ECHOSTATUS_OK) {
			ECHO_DEBUGPRINTF(("failed on Write_DSP\n"));
			return FALSE;
		}
	}

#ifdef _DEBUG
	CurTime = GetSystemTime();
	ECHO_DEBUGPRINTF(("LoadASIC took %ld usec.\n", (ULONG) (CurTime - StartTime)));
	ECHO_DEBUGPRINTF(("Checksum is 0x%x\n", Checksum));
	ECHO_DEBUGPRINTF(("ASIC load OK\n"));
#endif
	ECHO_DEBUGPRINTF(("ASIC loaded\n"));
	return TRUE;
}

#endif /* ECHOCARD_HAS_ASIC */



//===========================================================================
//
// InstallResidentLoader
//
// Install the resident loader for 56361 DSPs;  The resident loader
// is on the EPROM on the board for 56301 DSP.
//
// The resident loader is a tiny little program that is used to load
// the real DSP code.
//
//===========================================================================

#ifdef DSP_56361

static int InstallResidentLoader(struct EchoGals *eg)
{
	u32 Address;
	int Index,  Num, i;
	u16 *Code;
	u32 Status;
	int Success;
	u64 CurTime, Timeout;

	// 56361 cards only!  This check is required by the old 56301-based Gina24
	if (eg->DeviceId != DEVICE_ID_56361)
		return ECHOSTATUS_OK;

	// Look to see if the resident loader is present.  If the resident loader
	// is already installed, host flag 5 will be on.
	Status = GetDspRegister(eg, CHI32_STATUS_REG);
	if (Status & CHI32_STATUS_REG_HF5) {
		ECHO_DEBUGPRINTF(("Resident loader already installed; status is 0x%x\n", Status));
		return ECHOSTATUS_OK;
	}
	// Set DSP format bits for 24 bit mode
	SetDspRegister(eg, CHI32_CONTROL_REG, GetDspRegister(eg, CHI32_CONTROL_REG) | 0x900);

	//---------------------------------------------------------------------------
	//
	// Loader
	//
	// The DSP code is an array of 16 bit words.  The array is divided up into
	// sections.  The first word of each section is the size in words, followed
	// by the section type.
	//
	// Since DSP addresses and data are 24 bits wide, they each take up two
	// 16 bit words in the array.
	//
	// This is a lot like the other loader loop, but it's not a loop,
	// you don't write the memory type, and you don't write a zero at the end.
	//
	//---------------------------------------------------------------------------

	Code = LoaderDSP;

	// Skip the header section; the first word in the array is the size of
	// the first section, so the first real section of code is pointed to
	//  by Code[0].
	Index = Code[0];

	// Skip the section size, LRS block type, and DSP memory type
	Index += 3;

	// Get the number of DSP words to write
	Num = Code[Index++];

	// Get the DSP address for this block; 24 bits, so build from two words
	Address = (Code[Index] << 16) + Code[Index + 1];
	Index += 2;

	// Write the count to the DSP
	if (Write_DSP(eg, (u32)Num)) {
		ECHO_DEBUGPRINTF(("InstallResidentLoader: Failed to write word count!\n"));
		return -EIO;
	}
	// Write the DSP address
	if (Write_DSP(eg, Address)) {
		ECHO_DEBUGPRINTF(("InstallResidentLoader: Failed to write DSP address!\n"));
		return -EIO;
	}
	// Write out this block of code to the DSP
	for (i = 0; i < Num; i++) {
		u32 Data;

		Data = (Code[Index] << 16) + Code[Index + 1];
		if (Write_DSP(eg, Data)) {
			ECHO_DEBUGPRINTF(("InstallResidentLoader: Failed to write DSP code\n"));
			return -EIO;
		}
		Index += 2;
	}

	// Wait for flag 5 to come up
	CurTime = GetSystemTime();
	Timeout = CurTime + 10000L;	// 10ms
	Success = FALSE;
	do {
		udelay(50);	// Give the DSP some time; no need to hog the CPU
		Status = GetDspRegister(eg, CHI32_STATUS_REG);
		if (Status & CHI32_STATUS_REG_HF5) {
			Success = TRUE;
			break;
		}
		CurTime = GetSystemTime();
	} while (CurTime < Timeout);

	if (!Success) {
		ECHO_DEBUGPRINTF(("Resident loader failed to set HF5\n"));
		return -EIO;
	}

	ECHO_DEBUGPRINTF(("Resident loader successfully installed\n"));
	return ECHOSTATUS_OK;
}

#endif /* DSP_56361 */


//===========================================================================
//
// LoadDSP
//
// This loads the DSP code.
//
//===========================================================================

static int LoadDSP(struct EchoGals *eg, u16 *Code)
{
	u32 Address;
	int Index, Num, i;
	u64 Timeout, CurTime;

	if (eg->DspCode == Code) {
		ECHO_DEBUGPRINTF(("DSP is already loaded!\n"));
		return ECHOSTATUS_FIRMWARE_LOADED;
	}
	eg->BadBoard = TRUE;	// Set TRUE until DSP loaded
	eg->DspCode = NULL;	// Current DSP code not loaded
	eg->ASICLoaded = FALSE;	// Loading the DSP code will reset the ASIC

	ECHO_DEBUGPRINTF(("LoadDSP  Set BadBoard to TRUE\n"));

	// If this board requires a resident loader, install it.
#ifdef DSP_56361
	InstallResidentLoader(eg);
#endif

	// Send software reset command
	if (SendVector(eg, DSP_VC_RESET) != ECHOSTATUS_OK) {
		ECHO_DEBUGPRINTF(("LoadDsp SendVector DSP_VC_RESET failed, Critical Failure\n"));
		return -EIO;
	}
	// Delay 10us
	udelay(10);

	// Wait 10ms for HF3 to indicate that software reset is complete
	CurTime = GetSystemTime();
	Timeout = CurTime + 10000L;	// 10ms

	while (1) {
		if (GetDspRegister(eg, CHI32_STATUS_REG) & CHI32_STATUS_REG_HF3)
			break;
		CurTime = GetSystemTime();
		if (CurTime > Timeout) {
			ECHO_DEBUGPRINTF(("LoadDSP Timeout waiting for CHI32_STATUS_REG_HF3\n"));
			return -EIO;
		}
	}

	// Set DSP format bits for 24 bit mode now that soft reset is done
	SetDspRegister(eg, CHI32_CONTROL_REG, GetDspRegister(eg, CHI32_CONTROL_REG) | 0x900);

	//---------------------------------------------------------------------------
	// Main loader loop
	//---------------------------------------------------------------------------
	Index = Code[0];

	for (;;) {
		int BlockType, MemType;

		// Total Block Size
		Index++;

		// Block Type
		BlockType = Code[Index];
		if (BlockType == 4)	// We're finished
			break;

		Index++;

		// Memory Type  P=0,X=1,Y=2
		MemType = Code[Index];
		Index++;

		// Block Code Size
		Num = Code[Index];
		Index++;
		if (Num == 0)	// We're finished
			break;

		// Start Address
		Address = ((u32)Code[Index] << 16) + Code[Index + 1];
		Index += 2;

		if (Write_DSP(eg, (u32)Num)) {
			ECHO_DEBUGPRINTF(("LoadDSP: failed to write number of DSP words\n"));
			return -EIO;
		}

		if (Write_DSP(eg, Address)) {
			ECHO_DEBUGPRINTF(("LoadDSP: failed to write DSP address\n"));
			return -EIO;
		}

		if (Write_DSP(eg, (u32)MemType)) {
			ECHO_DEBUGPRINTF(("LoadDSP: failed to write DSP memory type\n"));
			return -EIO;
		}
		// Code
		for (i = 0; i < Num; i++) {
			u32 Data;

			Data = ((u32)Code[Index] << 16) + Code[Index + 1];
			if (Write_DSP(eg, Data)) {
				ECHO_DEBUGPRINTF(("LoadDSP: failed to write DSP data\n"));
				return -EIO;
			}

			Index += 2;
		}
//              ECHO_DEBUGPRINTF( ("\tEnd Code Block\n") );
	}

	if (Write_DSP(eg, 0)) {	// We're done!!!
		ECHO_DEBUGPRINTF(("LoadDSP: Failed to write final zero\n"));
		return -EIO;
	}
	// Delay 10us
	udelay(10);

	CurTime = GetSystemTime();
	Timeout = CurTime + 500000L;	// 1/2 sec. timeout

	while (CurTime <= Timeout) {
		// Wait for flag 4 - indicates that the DSP loaded OK
		if (GetDspRegister(eg, CHI32_STATUS_REG) & CHI32_STATUS_REG_HF4) {
			SetDspRegister(eg, CHI32_CONTROL_REG, GetDspRegister(eg, CHI32_CONTROL_REG) & ~0x1b00);

			if (Write_DSP(eg, DSP_FNC_SET_COMMPAGE_ADDR)) {
				ECHO_DEBUGPRINTF(("LoadDSP: Failed to write DSP_FNC_SET_COMMPAGE_ADDR\n"));
				return -EIO;
			}

			if (Write_DSP(eg, eg->CommPagePhys)) {
				ECHO_DEBUGPRINTF(("LoadDSP: Failed to write comm page address\n"));
				return -EIO;
			}

			/* Get the serial number via slave mode.
			 * This is triggered by the SET_COMMPAGE_ADDR command.
			 * We don't actually use the serial number but we have to get
			 * it as part of the DSP init voodoo. */
			if (ReadSn(eg)) {
				ECHO_DEBUGPRINTF(("LoadDSP: Failed to read serial number\n"));
				return -EIO;
			}

			eg->DspCode = Code;	// Show which DSP code loaded
			eg->BadBoard = FALSE;	// DSP OK
			ECHO_DEBUGPRINTF(("LoadDSP:  Set BadBoard to FALSE\n"));
			return ECHOSTATUS_OK;
		}
		CurTime = GetSystemTime();
	}

	ECHO_DEBUGPRINTF(("LoadDSP: DSP load timed out waiting for HF4\n"));
	return -EIO;
}



//===========================================================================
//
// LoadFirmware takes care of loading the DSP and any ASIC code.
//
//===========================================================================

static int LoadFirmware(struct EchoGals *eg)
{
	int st;

	// Try to load the DSP
	if (!eg->DspCodeToLoad || !eg->DspCommPage)
		return -EPERM;

	// See if the ASIC is present and working - only if the DSP is already loaded
	if (eg->DspCode) {
		st = CheckAsicStatus(eg);
		if (st)
			return ECHOSTATUS_OK;
		// ASIC check failed; force the DSP to reload
		eg->DspCode = NULL;
	}
	// Try and load the DSP
	st = LoadDSP(eg, eg->DspCodeToLoad);
	if (st != ECHOSTATUS_OK && st != ECHOSTATUS_FIRMWARE_LOADED)
		return st;

	// Load the ASIC if the DSP load succeeded; LoadASIC will
	// always return TRUE for cards that don't have an ASIC.
	st = LoadASIC(eg);
	if (!st)
		st = -EIO;
	else
		// ASIC load was successful
		st = RestoreDspSettings(eg);

	return st;
}



/****************************************************************************

  Mixer functions

 ****************************************************************************/

#if defined(ECHOCARD_HAS_INPUT_NOMINAL_LEVEL) || defined(ECHOCARD_HAS_OUTPUT_NOMINAL_LEVEL)

// Set the nominal level for an input or output bus (true = -10dBV, false = +4dBu)
static int SetNominalLevel(struct EchoGals *eg, u16 Index, char Consumer)
{
	ECHO_DEBUGPRINTF(("SetNominalLevel[%d]=%d\n", (int)Index, (int)Consumer));

	if (Index >= eg->NumBussesOut + eg->NumBussesIn) {
		ECHO_DEBUGPRINTF(("SetNominalOutLineLevel: Invalid index %d\n", (int)Index));
		return -EINVAL;
	}

	// Wait for the handshake (OK even if ASIC is not loaded)
	if (WaitForHandshake(eg))
		return -EIO;

	eg->NominalLevel[Index] = Consumer;

	if (Consumer)
		eg->DspCommPage->cmdNominalLevel |= SWAP32(1 << Index);
	else
		eg->DspCommPage->cmdNominalLevel &= ~SWAP32(1 << Index);

	return ECHOSTATUS_OK;
}

#endif /* ECHOCARD_HAS_*_NOMINAL_LEVEL */

//===========================================================================
//
// Set the gain for a single physical output channel (dB).
//
// Cond: @@ ECHOGAIN_MINOUT <= Gain <= ECHOGAIN_MAXOUT, Channel < eg->NumBussesOut
//
//===========================================================================

static int SetOutputGain(struct EchoGals *eg, u16 Channel, int Gain)
{
	if (Gain < ECHOGAIN_MINOUT)
		Gain = ECHOGAIN_MINOUT;
	else if (Gain > ECHOGAIN_MAXOUT)
		Gain = ECHOGAIN_MAXOUT;

	if (Channel < eg->NumBussesOut) {
		if (WaitForHandshake(eg))
			return -EIO;

		// Save the new value
		eg->OutLevel[Channel] = Gain;
		eg->DspCommPage->OutLineLevel[Channel] = GENERIC_TO_DSP(Gain);

//    ECHO_DEBUGPRINTF(("SetBusOutGain: [%d] = 0x%x\n", Channel, Gain));
		return ECHOSTATUS_OK;
	}

	ECHO_DEBUGPRINTF(("SetPipeOutGain: Invalid out pipe %d\n", Channel));
	return -EINVAL;
}



// Set the monitor level from an input bus to an output bus.
static int SetMonitorGain(struct EchoGals *eg, u16 OutChannel, u16 InChannel, int Gain)
{
	if (Gain < ECHOGAIN_MINOUT)
		Gain = ECHOGAIN_MINOUT;
	else if (Gain > ECHOGAIN_MAXOUT)
		Gain = ECHOGAIN_MAXOUT;
	//ECHO_DEBUGPRINTF(("SetMonitorGain[%d][%d] = %d\n", OutChannel, InChannel, Gain));

	if (WaitForHandshake(eg))
		return -EIO;

	eg->MonitorLevel[OutChannel][InChannel] = Gain;
	// Write the gain value to the comm page
	eg->DspCommPage->Monitors[MONITOR_INDEX(OutChannel, InChannel)] = GENERIC_TO_DSP(Gain);

	return ECHOSTATUS_OK;
}



#ifdef ECHOCARD_HAS_DIGITAL_IN_AUTOMUTE

// Digital input auto-mute - Gina24, Layla24, and Mona only
static int SetDigitalInputAutoMute(struct EchoGals *eg, int AutoMute)
{
	ECHO_DEBUGPRINTF(("CDspCommObject::SetDigitalInputAutoMute %d\n", AutoMute));

	if (AutoMute)
		eg->Flags |= ECHOGALS_FLAG_DIGITAL_IN_AUTOMUTE;
	else
		eg->Flags &= ~ECHOGALS_FLAG_DIGITAL_IN_AUTOMUTE;

	// Re-set the input clock to the current value - indirectly causes the
	// auto-mute flag to be sent to the DSP
	return SetInputClock(eg, eg->InputClock);
}

#endif /* DIGITAL_INPUT_AUTO_MUTE_SUPPORT */



// Tell the DSP to read and update output, nominal & monitor levels in comm page.
static int UpdateOutputLineLevel(struct EchoGals *eg)
{
	//ECHO_DEBUGPRINTF(("UpdateAudioOutLineLevel\n"));
	if (WaitForHandshake(eg))
		return -EIO;
	ClearHandshake(eg);
	return SendVector(eg, DSP_VC_UPDATE_OUTVOL);
}



// Tell the DSP to read and update input levels in comm page
static int UpdateInputLineLevel(struct EchoGals *eg)
{
	//  ECHO_DEBUGPRINTF(("UpdateAudioInLineLevel\n"));
	if (WaitForHandshake(eg))
		return -EIO;
	ClearHandshake(eg);
	return SendVector(eg, DSP_VC_UPDATE_INGAIN);
}



// SetMetersOn turns the meters on or off.  If meters are turned on, the DSP
// will write the meter and clock detect values to the comm page at about 30Hz
static void SetMetersOn(struct EchoGals *eg, char On)
{
	int Index;

	if (On && !eg->MetersOn) {
		SendVector(eg, DSP_VC_METERS_ON);
		eg->MetersOn = 1;
	} else if (!On && eg->MetersOn) {
		SendVector(eg, DSP_VC_METERS_OFF);
		eg->MetersOn = 0;
		for (Index = 0; Index < DSP_MAXPIPES; Index++) {
			eg->DspCommPage->VULevel[Index] = GENERIC_TO_DSP(ECHOGAIN_MUTED);
			eg->DspCommPage->PeakMeter[Index] = GENERIC_TO_DSP(ECHOGAIN_MUTED);
		}
	}
}



//===========================================================================
//
// Fill out an the given array using the current values in the comm page.
//
// Meters are written to the comm page by the DSP as follows:
//
// Output busses
// Input busses
// Output pipes (vmixer cards only)
//
// This function is designed to match the requirements of ALSA and it assumes
// there are no more than 16 in/out busses or pipes (maximum is ECHO_MAXAUDIO*,
// but currently there are no Echoaudio cards have so many channels. Meters is
// an array of signed chars[3][16][2])
//
//===========================================================================
static void GetAudioMeters(struct EchoGals *eg, long *Meters)
{
	int i, m, n;

	m = 0;
	n = 0;
	for (i = eg->NumBussesOut; i > 0; i--, m++) {
		Meters[n++] = DSP_TO_GENERIC((int)eg->DspCommPage->VULevel[m]);
		Meters[n++] = DSP_TO_GENERIC((int)eg->DspCommPage->PeakMeter[m]);
	}
	for (; n < 32; n++)
		Meters[n] = 0;

	for (i = eg->NumBussesIn; i > 0; i--, m++) {
		Meters[n++] = DSP_TO_GENERIC((int)eg->DspCommPage->VULevel[m]);
		Meters[n++] = DSP_TO_GENERIC((int)eg->DspCommPage->PeakMeter[m]);
	}
	for (; n < 64; n++)
		Meters[n] = 0;

	if (eg->HasVmixer) {
		for (i = eg->NumPipesOut; i > 0; i--, m++) {
			Meters[n++] = DSP_TO_GENERIC((int)eg->DspCommPage->VULevel[m]);
			Meters[n++] = DSP_TO_GENERIC((int)eg->DspCommPage->PeakMeter[m]);
		}
	}
	for (; n < 96; n++)
		Meters[n] = 0;
}



//===========================================================================
//
// This is called after LoadFirmware to restore old gains, meters on, monitors, etc.
//
//===========================================================================

static int RestoreDspSettings(struct EchoGals *eg)
{
	ECHO_DEBUGPRINTF(("RestoreDspSettings\n"));

	if (!CheckAsicStatus(eg))
		return -EIO;

	//@ Gina20/Darla20 only. Should be harmless for other cards.
	eg->DspCommPage->GDClockState = GD_CLOCK_UNDEF;
	eg->DspCommPage->GDSpdifStatus = GD_SPDIF_STATUS_UNDEF;

	eg->DspCommPage->Handshake = 0xffffffff;

#ifdef MIDI_SUPPORT
	eg->NextMidiWriteTime = 0;
#endif

	DspSetSampleRate(eg, GetSampleRate(eg));
	if (eg->MetersOn)
		SendVector(eg, DSP_VC_METERS_ON);

	if (WaitForHandshake(eg))
		return -EIO;

	SetInputClock(eg, eg->InputClock);

#ifdef ECHOCARD_HAS_OUTPUT_CLOCK_SWITCH
	SetOutputClock(eg, eg->OutputClock);
#endif

	if (UpdateOutputLineLevel(eg))
		return -EIO;

	if (UpdateInputLineLevel(eg))
		return -EIO;
#ifndef EMUL
#ifdef ECHOCARD_HAS_VMIXER
	if (UpdateVmixerLevel(eg))
		return -EIO;
#endif
#endif

	if (WaitForHandshake(eg))
		return -EIO;
	ClearHandshake(eg);

	ECHO_DEBUGPRINTF(("RestoreDspSettings done\n"));
	return SendVector(eg, DSP_VC_UPDATE_FLAGS);
}



/****************************************************************************

	Transport functions

 ****************************************************************************/

static void DspSetFormat(struct EchoGals *eg, u16 PipeIndex, const Audioformat_t *Format, char NoDither)
{
	u16 DspFormat;

	DspFormat = DSP_AUDIOFORM_SS_16LE;

	// Look for super-interleave
	if (Format->DataInterleave > 2) {
		DspFormat = DSP_AUDIOFORM_SUPER_INTERLEAVE_32LE | Format->DataInterleave;
	} else if (Format->DataAreBigEndian) {	// For big-endian data, only 32 bit mono->mono samples and 32 bit stereo->stereo are supported
		switch (Format->DataInterleave) {
		case 1:
			DspFormat = DSP_AUDIOFORM_MM_32BE;
			break;
#ifdef STEREO_BIG_ENDIAN32_SUPPORT
		case 2:
			DspFormat = DSP_AUDIOFORM_SS_32BE;
			break;
#endif
		}
	} else {		// Check for 32 bit little-endian mono->mono case
		if (Format->DataInterleave == 1 && Format->BitsPerSample == 32 && !Format->MonoToStereo) {
			DspFormat = DSP_AUDIOFORM_MM_32LE;
		} else {	// Handle the other little-endian formats
			switch (Format->BitsPerSample) {
			case 8:
				if (Format->DataInterleave == 2)
					DspFormat = DSP_AUDIOFORM_SS_8;
				else
					DspFormat = DSP_AUDIOFORM_MS_8;	//@@ mono->stereo  You may not want this
				break;
			default:
			case 16:
				// If this is a digital input and the no dither flag is set, use the no-dither format
				if (NoDither && (PipeIndex >= (eg->NumPipesOut + eg->FirstDigitalBusIn))) {
					// 16 bit, no dither
					if (Format->DataInterleave == 2)
						DspFormat = DSP_AUDIOFORM_SS_16LE_ND;
					else
						DspFormat = DSP_AUDIOFORM_MS_16LE_ND;
				} else {
					// 16 bit with dither
					if (Format->DataInterleave == 2)
						DspFormat = DSP_AUDIOFORM_SS_16LE;
					else
						DspFormat = DSP_AUDIOFORM_MS_16LE;
				}
				break;
			case 32:
				if (Format->DataInterleave == 2)
					DspFormat = DSP_AUDIOFORM_SS_32LE;
				else
					DspFormat = DSP_AUDIOFORM_MS_32LE;
				break;
			}
		}
	}
	ECHO_DEBUGPRINTF(("SetAudioFormat[%d] = %x\n", PipeIndex, DspFormat));
	eg->DspCommPage->AudioFormat[PipeIndex] = SWAP16(DspFormat);
}



// StartTransport starts transport for a set of pipes
static int StartTransport(struct EchoGals *eg, u32 ChannelMask, u32 CyclicMask)
{
	ECHO_DEBUGPRINTF(("StartTransport %x\n", ChannelMask));

	// Wait for the previous command to complete
	if (WaitForHandshake(eg))
		return -EIO;

	// Write the appropriate fields in the comm page
	eg->DspCommPage->cmdStart |= SWAP32(ChannelMask);
	eg->DspCommPage->cmdCyclicBuffer |= SWAP32(CyclicMask);

	if (eg->DspCommPage->cmdStart) {
#if 0
		{	// Dump s-g list
			int x, y;
			for (y = 0; y < 4096; y += 32) {
				print("%03x ", y);
				for (x = 0; x < 32; x++)
					print("%02x ", ((char *)eg->DspCommPage)[y * 32 + x]);
				print("\n");
			}
		}
#endif
		ClearHandshake(eg);
		SendVector(eg, DSP_VC_START_TRANSFER);
		// Wait for transport to start
		if (WaitForHandshake(eg))
			return -EIO;
		// Keep track of which pipes are transporting
		eg->ActiveMask |= ChannelMask;
		eg->DspCommPage->cmdStart = 0;
		return ECHOSTATUS_OK;
	}			// if this monkey is being started

	ECHO_DEBUGPRINTF(("StartTransport: No pipes to start!\n"));
	return -EINVAL;
}



// PauseTransport pauses transport for a set of pipes
static int PauseTransport(struct EchoGals *eg, u32 ChannelMask)
{
	ECHO_DEBUGPRINTF(("PauseTransport %x\n", ChannelMask));

	// Wait for the last command to finish
	if (WaitForHandshake(eg))
		return -EIO;

	// Write to the comm page
	eg->DspCommPage->cmdStop |= SWAP32(ChannelMask);
	eg->DspCommPage->cmdReset = 0;
	if (eg->DspCommPage->cmdStop) {
		// Clear the handshake and send the vector command
		ClearHandshake(eg);
		SendVector(eg, DSP_VC_STOP_TRANSFER);
		// Wait for transport to stop
		if (WaitForHandshake(eg))
			return -EIO;
		// Keep track of which pipes are transporting
		eg->ActiveMask &= ~ChannelMask;
		eg->DspCommPage->cmdStop = 0;
		eg->DspCommPage->cmdReset = 0;
		return ECHOSTATUS_OK;
	}			// if this monkey is being started

	ECHO_DEBUGPRINTF(("PauseTransport: No pipes to stop!\n"));
	return ECHOSTATUS_OK;
}



// ResetTransport resets transport for a set of pipes
static int StopTransport(struct EchoGals *eg, u32 ChannelMask)
{
	ECHO_DEBUGPRINTF(("StopTransport %x\n", ChannelMask));

	// Wait for the last command to finish
	if (WaitForHandshake(eg))
		return -EIO;

	// Write to the comm page
	eg->DspCommPage->cmdStop |= SWAP32(ChannelMask);
	eg->DspCommPage->cmdReset |= SWAP32(ChannelMask);
	if (eg->DspCommPage->cmdReset) {
		// Clear the handshake and send the vector command
		ClearHandshake(eg);
		SendVector(eg, DSP_VC_STOP_TRANSFER);
		// Wait for transport to stop
		if (WaitForHandshake(eg))
			return -EIO;
		// Keep track of which pipes are transporting
		eg->ActiveMask &= ~ChannelMask;
		eg->DspCommPage->cmdStop = 0;
		eg->DspCommPage->cmdReset = 0;
		return ECHOSTATUS_OK;
	}			// if this monkey is being started

	ECHO_DEBUGPRINTF(("StopTransport: No pipes to stop!\n"));
	return ECHOSTATUS_OK;
}



static int DspPowerDown(struct EchoGals *eg)
{
	// Stops all active pipes (@@ they should already be turned off by DelEchoGals())
	StopTransport(eg, eg->ActiveMask);

	// Meters off
	SetMetersOn(eg, FALSE);

#ifdef MIDI_SUPPORT
	// MIDI input off
	eg->MidiOnCount = 1;
	SetMidiOn(eg, FALSE);
#endif
	// Go to sleep
	return GoComatose(eg);
}
