// ****************************************************************************
//
//  	CLaylaDspCommObject.cpp
//
// Implementation file for EchoGals generic driver Layla DSP
// interface class.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"

#include "DSP/Layla20DSP.c"
#include "ASIC/LaylaASIC.c"

static int SetProfessionalSpdif(struct EchoGals *eg, char Prof);
static int SetNominalLevel(struct EchoGals *eg, u16 Index, char Consumer);
static int LoadASICgeneric(struct EchoGals *eg, u32 Cmd, u8 *Code, u32 Size);
static int CheckAsicStatus(struct EchoGals *eg);
static int UpdateFlags(struct EchoGals *eg);


static int InitHw(struct EchoGals *eg, u16 Device, u16 SubDevice)
{
	int err, i;

	ECHO_DEBUGPRINTF(("InitHw() - Layla20\n"));

	if ((SubDevice & 0xfff0) != LAYLA20) {
		ECHO_DEBUGPRINTF(("Wrong card type\n"));
		return -ENODEV;
	}
	
	// This part is common to all the cards
	if ((err = InitDspCommPage(eg))) {
		ECHO_DEBUGPRINTF(("InitHw - could not initialize DSP comm page\n"));
		return err;
	}

	eg->DeviceId = Device;
	eg->SubDeviceId = SubDevice;
	eg->CardType = LAYLA20;
	eg->BadBoard = 1;
	eg->NumPipesOut = 12;
	eg->NumPipesIn = 10;
	eg->NumBussesOut = 12;
	eg->NumBussesIn = 10;
	eg->FirstDigitalBusOut = 10;
	eg->FirstDigitalBusIn = 8;
	eg->HasVmixer = FALSE;
	eg->NumMidiOut = 1;
	eg->NumMidiIn = 1;
	eg->HasASIC = TRUE;
	eg->DspCodeToLoad = Layla20DSP;
	eg->Flags =	ECHOGALS_FLAG_BADBOARD |
			ECHOCAPS_INPUT_GAIN |
			ECHOCAPS_OUTPUT_GAIN |
			ECHOCAPS_VU_METER |
			ECHOGALS_FLAG_SPDIF_NODITHER |
			ECHOCAPS_OUTPUT_NOMINAL_LEVEL |
			ECHOCAPS_SUPER_INTERLEAVE;
	eg->InClockTypes =	ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF |
				ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_SUPER;
	eg->OutClockTypes =	ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_SUPER;

	// Load the DSP and the ASIC on the PCI card
	err = LoadFirmware(eg);
	if (err || eg->BadBoard)
		return -EIO;

	eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;

	// Must call this here after DSP is init to init gains and mutes
	if ((err = InitLineLevels(eg)))
		return err;
/*
	//
	// Initialize the MIDI input
	//
	err = m_MidiIn.Init( this );
	if ( ECHOSTATUS_OK != err )
		return err;
*/

	// Set professional nominal levels
	for (i = 0; i < eg->FirstDigitalBusOut; i++)
		err = SetNominalLevel(eg, i, FALSE);	// FALSE is +4dBu here

	for (i = 0; i < eg->FirstDigitalBusIn; i++)
		err = SetNominalLevel(eg, eg->NumBussesOut + i, FALSE);

	// Set the S/PDIF output format to "professional"
	SetProfessionalSpdif(eg, TRUE);

	// Get default sample rate from DSP
	eg->SampleRate = GetSampleRate(eg);

	ECHO_DEBUGPRINTF(("InitHw done\n"));
	return err;
}



//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a
// given sample rate.
//
//===========================================================================

static int QueryAudioSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	if (SampleRate < 8000 || SampleRate > 50000) {
		ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Sample rate must be >= 8,000 Hz and <= 50,000 Hz\n"));
		return -EINVAL;
	}

	ECHO_DEBUGPRINTF(("QueryAudioSampleRate() %d Hz OK\n", SampleRate));
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Layla20 supports S/PDIF clock, word clock, and super clock.
//
//===========================================================================

static u32 GetInputClockDetect(const struct EchoGals *eg)
{
	u32 ClocksFromDsp, ClockDetectBits;

	// Map the DSP clock detect bits to the generic driver clock detect bits
	ClocksFromDsp = SWAP32(eg->DspCommPage->StatusClocks);

	ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

	if (ClocksFromDsp & GLDM_CLOCK_DETECT_BIT_SPDIF)
		ClockDetectBits |= ECHO_CLOCK_BIT_SPDIF;

	if (ClocksFromDsp & GLDM_CLOCK_DETECT_BIT_WORD) {
		if (ClocksFromDsp & GLDM_CLOCK_DETECT_BIT_SUPER)
			ClockDetectBits |= ECHO_CLOCK_BIT_SUPER;
		else
			ClockDetectBits |= ECHO_CLOCK_BIT_WORD;
	}

	return ClockDetectBits;
}



/****************************************************************************

  Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
// Layla20 has an ASIC in the external box
//
//===========================================================================

static int LoadASIC(struct EchoGals *eg)
{
	if (eg->ASICLoaded)
		return TRUE;

	if (!LoadASICgeneric(eg, DSP_FNC_LOAD_LAYLA_ASIC, LaylaASIC, LAYLA_ASIC_SIZE))
		return FALSE;

	// Check if ASIC is alive and well.
	return CheckAsicStatus(eg);
}



//===========================================================================
//
// DspSetSampleRate
//
// Set the audio sample rate for Layla20
//
// Layla is simple; just send it the sampling rate (assuming that the clock mode is correct).
//
//===========================================================================

static int DspSetSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	// Only set the clock for internal mode. Do not return failure, simply treat it as a non-event.
	if (eg->InputClock != ECHO_CLOCK_INTERNAL) {
		ECHO_DEBUGPRINTF(("DspSetSampleRate: Cannot set sample rate - clock not set to CLK_CLOCKININTERNAL\n"));
		eg->DspCommPage->SampleRate = SWAP32(SampleRate);
		return GetSampleRate(eg);
	}

	if (SampleRate < 8000 || SampleRate > 50000) {
		ECHO_DEBUGPRINTF(("DspSetSampleRate: %d invalid!\n", SampleRate));
		return GetSampleRate(eg);
	}

	if (WaitForHandshake(eg))
		return -EIO;

	eg->DspCommPage->SampleRate = SWAP32(SampleRate);

	ClearHandshake(eg);
	SendVector(eg, DSP_VC_SET_LAYLA_SAMPLE_RATE);

	ECHO_DEBUGPRINTF(("DspSetSampleRate: Layla sample rate changed to %d\n", SampleRate));
	return SampleRate;
}



//===========================================================================
//
// Set the input clock to internal, S/PDIF, ADAT
//
//===========================================================================

static int SetInputClock(struct EchoGals *eg, u16 Clock)
{
	u16 NewClock;
	u32 SampleRate;

	ECHO_DEBUGPRINTF(("SetInputClock:\n"));
	SampleRate = 0;
	switch (Clock) {
	case ECHO_CLOCK_INTERNAL:
		ECHO_DEBUGPRINTF(("Set Layla20 clock to INTERNAL\n"));
		SampleRate = GetSampleRate(eg);
		// If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
		if (SampleRate < 8000 || SampleRate > 50000) {
			SampleRate = 48000;
			eg->DspCommPage->SampleRate = SWAP32((u32) 48000);
		}
		NewClock = LAYLA20_CLOCK_INTERNAL;
		break;
	case ECHO_CLOCK_SPDIF:
		ECHO_DEBUGPRINTF(("Set Layla20 clock to SPDIF\n"));
		NewClock = LAYLA20_CLOCK_SPDIF;
		break;
	case ECHO_CLOCK_WORD:
		ECHO_DEBUGPRINTF(("Set Layla20 clock to WORD\n"));
		NewClock = LAYLA20_CLOCK_WORD;
		break;
	case ECHO_CLOCK_SUPER:
		ECHO_DEBUGPRINTF(("Set Layla20 clock to SUPER\n"));
		NewClock = LAYLA20_CLOCK_SUPER;
		break;
	default:
		ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Layla24\n", Clock));
		return -EINVAL;
	}
	eg->InputClock = Clock;

	// Send the new clock to the DSP
	eg->DspCommPage->InputClock = SWAP16(NewClock);
	ClearHandshake(eg);
	SendVector(eg, DSP_VC_UPDATE_CLOCKS);

	if (SampleRate)
		DspSetSampleRate(eg, SampleRate);

	return ECHOSTATUS_OK;
}



//===========================================================================
//
// Set new output clock
//
//===========================================================================

static int SetOutputClock(struct EchoGals *eg, u16 Clock)
{
	if (WaitForHandshake(eg))
		return -EIO;

	ECHO_DEBUGPRINTF(("SetOutputClock:\n"));
	eg->DspCommPage->OutputClock = SWAP16(Clock);
	eg->OutputClock = Clock;

	ClearHandshake(eg);
	return SendVector(eg, DSP_VC_UPDATE_CLOCKS);
}


/*
// This card has no vmixer.
static int SetVmixerGain(struct EchoGals *eg, u16 Output, u16 Pipe, int Gain)
{
	return -EINVAL;
}
*/


// Set input bus gain (one unit is 0.5dB !)
static int SetInputGain(struct EchoGals *eg, u16 Channel, int Gain)
{
	if (Gain < ECHOGAIN_MININP)
		Gain = ECHOGAIN_MININP;
	else if (Gain > ECHOGAIN_MAXINP)
		Gain = ECHOGAIN_MAXINP;

	if (Channel > eg->NumBussesIn)
		return -EINVAL;

	// Wait for the handshake (OK even if ASIC is not loaded)
	if (WaitForHandshake(eg))
		return -EIO;

	// Save the new value
	eg->InLevel[Channel] = Gain;
	Gain = GENERIC_TO_DSP(Gain * 2);
	Gain += GL20_INPUT_GAIN_MAGIC_NUMBER;

	// Put it in the comm page
	eg->DspCommPage->InLineLevel[Channel] = Gain;
	return ECHOSTATUS_OK;
}


/*
// This card has no vmixer.
static int UpdateVmixerLevel(struct EchoGals *eg)
{
	return -EINVAL;
}
*/

/*
// Gina24, Mona, Layla24 - only
static int DspSetDigitalMode(struct EchoGals *eg, u8 Mode)
{
	return -EINVAL;
}
*/



// Tell the DSP to reread the flags from the comm page
static int UpdateFlags(struct EchoGals *eg)
{
	//  ECHO_DEBUGPRINTF(("UpdateFlags\n"));
	if (WaitForHandshake(eg))
		return -EIO;
	ClearHandshake(eg);
	return SendVector(eg, DSP_VC_UPDATE_FLAGS);
}



static int SetProfessionalSpdif(struct EchoGals *eg, char Prof)
{
	ECHO_DEBUGPRINTF(("SetProfessionalSpdif %d\n", Prof));
	if (Prof)
		eg->DspCommPage->Flags |= SWAP32(DSP_FLAG_PROFESSIONAL_SPDIF);
	else
		eg->DspCommPage->Flags &= ~SWAP32(DSP_FLAG_PROFESSIONAL_SPDIF);
	return UpdateFlags(eg);
}



/*
BOOL CLaylaDspCommObject::IsMidiOutActive()
{
        ULONGLONG       ullCurTime;

        m_pOsSupport->OsGetSystemTime( &ullCurTime );
        return( ( ( ullCurTime - MidiOutTime ) > MIDI_ACTIVITY_TIMEOUT_USEC ) ? FALSE : TRUE );
}       // BOOL CLaylaDspCommObject::IsMidiOutActive()
*/
