// ****************************************************************************
//
//  	CLayla24DspCommObject.cpp
//
// Implementation file for EchoGals generic driver Layla24 DSP
// interface class.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//#include <sound/pcm.h>

// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"

#include <linux/delay.h>

#include "DSP/Layla24DSP.c"
#include "ASIC/Layla24_1ASIC.c"
#include "ASIC/Layla24_2A_ASIC.c"
#include "ASIC/Layla24_2S_ASIC.c"

static int SetInputClock(struct EchoGals *eg, u16 Clock);
static int SetProfessionalSpdif(struct EchoGals *eg, char Prof);
static int SetNominalLevel(struct EchoGals *eg, u16 Index, char Consumer);
static int SetDigitalMode(struct EchoGals *eg, u8 Mode);
static int LoadASICgeneric(struct EchoGals *eg, u32 Cmd, u8 *Code, u32 Size);
static int CheckAsicStatus(struct EchoGals *eg);


static int InitHw(struct EchoGals *eg, u16 Device, u16 SubDevice)
{
	int err, i;

	ECHO_DEBUGPRINTF(("InitHw() - Layla24\n"));

	if ((SubDevice & 0xfff0) != LAYLA24) {
		ECHO_DEBUGPRINTF(("Wrong card type\n"));
		return -ENODEV;
	}

	// This part is common to all the cards
	if ((err = InitDspCommPage(eg))) {
		ECHO_DEBUGPRINTF(("InitHw - could not initialize DSP comm page\n"));
		return err;
	}

	eg->DeviceId = Device;
	eg->SubDeviceId = SubDevice;
	eg->CardType = LAYLA24;
	eg->BadBoard = 1;
	eg->NumPipesOut = 16;
	eg->NumPipesIn = 16;
	eg->NumBussesOut = 16;
	eg->NumBussesIn = 16;
	eg->FirstDigitalBusOut = 8;
	eg->FirstDigitalBusIn = 8;
	eg->HasVmixer = FALSE;
	eg->NumMidiOut = 1;
	eg->NumMidiIn = 1;
	eg->HasASIC = TRUE;
	eg->DspCodeToLoad = Layla24DSP;
	eg->Flags =	ECHOGALS_FLAG_BADBOARD |
			ECHOCAPS_OUTPUT_GAIN |
			ECHOCAPS_VU_METER |
			ECHOGALS_FLAG_SPDIF_NODITHER |
			ECHOCAPS_INPUT_NOMINAL_LEVEL |
			ECHOCAPS_OUTPUT_NOMINAL_LEVEL |
			ECHOCAPS_SUPER_INTERLEAVE |
			ECHOCAPS_DIGITAL_IN_AUTOMUTE |
			ECHOGALS_FLAG_DIGITAL_IN_AUTOMUTE |
			ECHOCAPS_STEREO_BE32;
	eg->InClockTypes =	ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF |
				ECHO_CLOCK_BIT_WORD | ECHO_CLOCK_BIT_ADAT;
	eg->DigitalModes =	ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA |
				ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_OPTICAL |
				ECHOCAPS_HAS_DIGITAL_MODE_ADAT;
	eg->DigitalMode =	DIGITAL_MODE_SPDIF_RCA;
	eg->ProfessionalSpdif = FALSE;
//@@questo ifdef sar da togliere
#ifdef MIDI_SUPPORT
	eg->MtcState = MTC_STATE_NORMAL;
#endif

	// Load the DSP and the ASIC on the PCI card
	err = LoadFirmware(eg);
	if (err || eg->BadBoard)
		return -EIO;

	eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;

	// Must call this here after DSP is init to init gains and mutes
	if ((err = InitLineLevels(eg)))
		return err;
/*
	//
	// Initialize the MIDI input
	//
	err = m_MidiIn.Init( this );
	if ( ECHOSTATUS_OK != err )
		return err;
*/

	// Set professional nominal levels
	for (i = 0; i < eg->FirstDigitalBusOut; i++)
		err = SetNominalLevel(eg, i, FALSE);	// FALSE is +4dBu here

	for (i = 0; i < eg->FirstDigitalBusIn; i++)
		err = SetNominalLevel(eg, eg->NumBussesOut + i, FALSE);

	// Set the digital mode to S/PDIF RCA
	SetDigitalMode(eg, DIGITAL_MODE_SPDIF_RCA);

	// Set the S/PDIF output format to "professional"
	SetProfessionalSpdif(eg, TRUE);

	// Get default sample rate from DSP
	eg->SampleRate = GetSampleRate(eg);

	ECHO_DEBUGPRINTF(("InitHw done\n"));
	return err;
}



//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a
// given sample rate.
//
//===========================================================================

static int QueryAudioSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	if (SampleRate ==  8000 || SampleRate == 11025 ||
	    SampleRate == 16000 || SampleRate == 22050 ||
	    (SampleRate >= 25000 && SampleRate <= 50000) ||
	    (eg->DigitalMode != DIGITAL_MODE_ADAT &&
	    SampleRate >= 50000 && SampleRate <= 100000))
		return ECHOSTATUS_OK;

	ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Invalid sample rate\n"));
	return -EINVAL;
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Layla24 supports S/PDIF, word, and ADAT input clocks.
//
//===========================================================================

static u32 GetInputClockDetect(const struct EchoGals *eg)
{
	u32 ClocksFromDsp, ClockDetectBits;

	// Map the DSP clock detect bits to the generic driver clock detect bits
	ClocksFromDsp = SWAP32(eg->DspCommPage->StatusClocks);

	ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

	if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_SPDIF)
		ClockDetectBits |= ECHO_CLOCK_BIT_SPDIF;

	if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_ADAT)
		ClockDetectBits |= ECHO_CLOCK_BIT_ADAT;

	if (ClocksFromDsp & GML_CLOCK_DETECT_BIT_WORD)
		ClockDetectBits |= ECHO_CLOCK_BIT_WORD;

	return ClockDetectBits;
}



/****************************************************************************

  Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
// WriteControlReg
//
// Most configuration of Gina24, Layla24, or Mona is
// accomplished by writing the control register.  WriteControlReg
// sends the new control register value to the DSP.
//
//===========================================================================

static int WriteControlReg(struct EchoGals *eg, u32 ControlReg)
{
	ECHO_DEBUGPRINTF(("WriteControlReg...\n"));
	if (WaitForHandshake(eg))
		return -EIO;

	// Handle the digital input auto-mute
	if (eg->Flags & ECHOGALS_FLAG_DIGITAL_IN_AUTOMUTE)
		ControlReg |= GML_DIGITAL_IN_AUTO_MUTE;
	else
		ControlReg &= ~GML_DIGITAL_IN_AUTO_MUTE;

	// Write the control register
	ControlReg =  SWAP32(ControlReg);
	if (ControlReg != eg->DspCommPage->ControlReg) {
		eg->DspCommPage->ControlReg = ControlReg;
		ECHO_DEBUGPRINTF(("WriteControlReg: 0x%x\n", ControlReg));
		ClearHandshake(eg);
		return SendVector(eg, DSP_VC_WRITE_CONTROL_REG);
	}
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// Layla24 has an ASIC on the PCI card and another ASIC in the external box;
// both need to be loaded.
//
//===========================================================================

static int LoadASIC(struct EchoGals *eg)
{
	if (eg->ASICLoaded)
		return TRUE;

	ECHO_DEBUGPRINTF(("LoadASIC\n"));

	// Give the DSP a few milliseconds to settle down
	mdelay(10);

	// Load the ASIC for the PCI card
	if (!LoadASICgeneric(eg, DSP_FNC_LOAD_LAYLA24_PCI_CARD_ASIC, Layla24_1ASIC, sizeof(Layla24_1ASIC)))
		return FALSE;

	eg->Asic = Layla24_2S_ASIC;

	// Now give the new ASIC a little time to set up
	mdelay(10);

	// Do the external one
	if (!LoadASICgeneric(eg, DSP_FNC_LOAD_LAYLA24_EXTERNAL_ASIC, Layla24_2S_ASIC, sizeof(Layla24_2S_ASIC)))
		return FALSE;

	// Now give the external ASIC a little time to set up
	mdelay(10);

	// See if it worked
	CheckAsicStatus(eg);

	// Set up the control register if the load succeeded - 48 kHz, internal clock, S/PDIF RCA mode
	if (eg->ASICLoaded)
		WriteControlReg(eg, GML_CONVERTER_ENABLE | GML_48KHZ);

	ECHO_DEBUGPRINTF(("Finished\n"));
	return eg->ASICLoaded;
}



//===========================================================================
//
// DspSetSampleRate
//
// Set the sample rate for Layla24
//
// Layla24 is simple; just send it the sampling rate (assuming that the clock
// mode is correct).
//
//===========================================================================

static int DspSetSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	u32 ControlReg, NewClock, BaseRate;

	// Only set the clock for internal mode.  If the clock is not set to
	// internal, try and re-set the input clock; this more transparently
	// handles switching between single and double-speed mode
	if (eg->InputClock != ECHO_CLOCK_INTERNAL) {
		ECHO_DEBUGPRINTF(("DspSetSampleRate: Cannot set sample rate - clock not set to CLK_CLOCKININTERNAL\n"));
		// Save the rate anyhow
		eg->DspCommPage->SampleRate = SWAP32(SampleRate);
		// Set the input clock to the current value
		SetInputClock(eg, eg->InputClock);
		return GetSampleRate(eg);
	}
	// Get the control register & clear the appropriate bits
	ControlReg = SWAP32(eg->DspCommPage->ControlReg);
	ControlReg &= GML_CLOCK_CLEAR_MASK | GML_SPDIF_RATE_CLEAR_MASK;

	NewClock = 0;

	switch (SampleRate) {
	case 96000:
		NewClock = GML_96KHZ;
		break;
	case 88200:
		NewClock = GML_88KHZ;
		break;
	case 48000:
		NewClock = GML_48KHZ | GML_SPDIF_SAMPLE_RATE1;
		break;
	case 44100:
		NewClock = GML_44KHZ;
		// Professional mode
		if (ControlReg & GML_SPDIF_PRO_MODE)
			NewClock |= GML_SPDIF_SAMPLE_RATE0;
		break;
	case 32000:
		NewClock = GML_32KHZ | GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
		break;
	case 22050:
		NewClock = GML_22KHZ;
		break;
	case 16000:
		NewClock = GML_16KHZ;
		break;
	case 11025:
		NewClock = GML_11KHZ;
		break;
	case 8000:
		NewClock = GML_8KHZ;
		break;
	default:
		/* If this is a non-standard rate, then the driver needs to
		use Layla24's special "continuous frequency" mode */
		NewClock = LAYLA24_CONTINUOUS_CLOCK;
		if (SampleRate > 50000) {
			BaseRate = SampleRate >> 1;
			ControlReg |= GML_DOUBLE_SPEED_MODE;
		} else {
			BaseRate = SampleRate;
		}

		if (BaseRate < 25000)
			BaseRate = 25000;

		if (WaitForHandshake(eg))
			return -EIO;

		eg->DspCommPage->SampleRate = SWAP32(LAYLA24_MAGIC_NUMBER / BaseRate - 2);

		ClearHandshake(eg);
		SendVector(eg, DSP_VC_SET_LAYLA24_FREQUENCY_REG);
	}

	ControlReg |= NewClock;

	/* Tell the DSP about it */
	if (WriteControlReg(eg, ControlReg) == ECHOSTATUS_OK) {
		eg->DspCommPage->SampleRate = SWAP32(SampleRate);
		ECHO_DEBUGPRINTF(("DspSetSampleRate: %d clock %d\n", SampleRate, ControlReg));
	}

	return GetSampleRate(eg);
}



//===========================================================================
//
// SetInputClock
//
//===========================================================================

static int SetInputClock(struct EchoGals *eg, u16 Clock)
{
	char WriteReg;
	u32 ControlReg, SampleRate;

	ECHO_DEBUGPRINTF(("SetInputClock:\n"));

	ControlReg = SWAP32(eg->DspCommPage->ControlReg);

	// Mask off the clock select bits
	ControlReg &= GML_CLOCK_CLEAR_MASK;

	SampleRate = GetSampleRate(eg);
	WriteReg = TRUE;

	// Pick the new clock
	switch (Clock) {
	case ECHO_CLOCK_INTERNAL:
		ECHO_DEBUGPRINTF(("Set Layla24 clock to INTERNAL\n"));
		// If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
		if (GetSampleRate(eg) < 8000 || GetSampleRate(eg) > 100000)
			eg->DspCommPage->SampleRate = SWAP32(48000);
		WriteReg = FALSE;
		break;
	case ECHO_CLOCK_SPDIF:
		if (eg->DigitalMode == DIGITAL_MODE_ADAT)
			return -EAGAIN;
		ControlReg |= GML_SPDIF_CLOCK;
		/* Layla24 doesn't support 96KHz S/PDIF */
		ControlReg &= ~GML_DOUBLE_SPEED_MODE;
		ECHO_DEBUGPRINTF(("Set Layla24 clock to SPDIF\n"));
		break;
	case ECHO_CLOCK_WORD:
		ControlReg |= GML_WORD_CLOCK;
		if (GetInputClockDetect(eg) & GML_CLOCK_DETECT_BIT_WORD96)
			ControlReg |= GML_DOUBLE_SPEED_MODE;
		else
			ControlReg &= ~GML_DOUBLE_SPEED_MODE;
		ECHO_DEBUGPRINTF(("Set Layla24 clock to WORD\n"));
		break;
	case ECHO_CLOCK_ADAT:
		if (eg->DigitalMode != DIGITAL_MODE_ADAT)
			return -EAGAIN;
		ControlReg |= GML_ADAT_CLOCK;
		ControlReg &= ~GML_DOUBLE_SPEED_MODE;
		ECHO_DEBUGPRINTF(("Set Layla24 clock to ADAT\n"));
		break;
	default:
		ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Layla24\n", Clock));
		return -EINVAL;
	}

	eg->InputClock = Clock;

	// Do things according to the flags
	if (WriteReg)
		WriteControlReg(eg, ControlReg);

	// If the user just switched to internal clock, set the sample rate
	if (Clock == ECHO_CLOCK_INTERNAL)
		DspSetSampleRate(eg, GetSampleRate(eg));

	return ECHOSTATUS_OK;
}


/*
// Layla20 only
static int SetOutputClock(struct EchoGals *eg, u16 Clock)
{
	return -EINVAL;
}
*/

/*
// This card has no vmixer.
static int SetVmixerGain(struct EchoGals *eg, u16 Output, u16 Pipe, int Gain)
{
	return -EINVAL;
}
*/

/*
// This card has no vmixer.
static int UpdateVmixerLevel(struct EchoGals *eg)
{
	return -EINVAL;
}
*/

/*
// Set input bus gain (one unit is 0.5dB !)
static int SetInputGain(struct EchoGals *eg, u16 Channel, int Gain)
{
	return -EINVAL;
}
*/


//===========================================================================
//
// Depending on what digital mode you want, Layla24 needs different ASICs
// loaded.  This function checks the ASIC needed for the new mode and sees
// if it matches the one already loaded.
//
//===========================================================================

static int SwitchAsic(struct EchoGals *eg, u8 *AsicNeeded, u32 AsicSize)
{
	s8 Monitors[MONITOR_ARRAY_SIZE];

	//  Check to see if this is already loaded
	if (AsicNeeded != eg->Asic) {
		memcpy(Monitors, eg->DspCommPage->Monitors, MONITOR_ARRAY_SIZE);
		memset(eg->DspCommPage->Monitors, GENERIC_TO_DSP(ECHOGAIN_MUTED), MONITOR_ARRAY_SIZE);

		// Load the desired ASIC
		if (!LoadASICgeneric(eg, DSP_FNC_LOAD_LAYLA24_EXTERNAL_ASIC, AsicNeeded, AsicSize)) {
			memcpy(eg->DspCommPage->Monitors, Monitors, MONITOR_ARRAY_SIZE);
			return FALSE;
		}
		eg->Asic = AsicNeeded;
		memcpy(eg->DspCommPage->Monitors, Monitors, MONITOR_ARRAY_SIZE);
	}

	return TRUE;
}



//===========================================================================
//
// SetDigitalMode
//
//===========================================================================

static int DspSetDigitalMode(struct EchoGals *eg, u8 NewMode)
{
	u32 ControlReg;

	// Clear the current digital mode
	ControlReg = SWAP32(eg->DspCommPage->ControlReg);
	ControlReg &= GML_DIGITAL_MODE_CLEAR_MASK;

	// Tweak the control reg
	switch (NewMode) {
	case DIGITAL_MODE_SPDIF_OPTICAL:
		ControlReg |= GML_SPDIF_OPTICAL_MODE;
		// fall through
	case DIGITAL_MODE_SPDIF_RCA:
		// If the input clock is set to ADAT, set the
		// input clock to internal and the sample rate to 48 KHz
		if (eg->InputClock == ECHO_CLOCK_ADAT) {
			eg->DspCommPage->SampleRate = SWAP32(48000);
			SetInputClock(eg, ECHO_CLOCK_INTERNAL);
		}
		// If the currently loaded ASIC is the S/PDIF ASIC, switch to the ADAT ASIC
		if (!SwitchAsic(eg, Layla24_2S_ASIC, sizeof(Layla24_2S_ASIC)))
			return -EIO;
		break;
	case DIGITAL_MODE_ADAT:
		// If the input clock is set to S/PDIF, set the
		// input clock to internal and the sample rate to 48 KHz
		if (eg->InputClock == ECHO_CLOCK_SPDIF) {
			eg->DspCommPage->SampleRate = SWAP32(48000);
			SetInputClock(eg, ECHO_CLOCK_INTERNAL);
		}
		// If the currently loaded ASIC is the S/PDIF ASIC, switch to the ADAT ASIC
		if (!SwitchAsic(eg, Layla24_2A_ASIC, sizeof(Layla24_2A_ASIC)))
			return -EIO;

		ControlReg |= GML_ADAT_MODE;
		ControlReg &= ~GML_DOUBLE_SPEED_MODE;
		break;
	default:
		return -EINVAL;
	}

	// Write the control reg
	WriteControlReg(eg, ControlReg);

	eg->DigitalMode = NewMode;

	ECHO_DEBUGPRINTF(("SetDigitalMode to %d\n", NewMode));
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// Set the S/PDIF output format (common for Gina24, Layla24, and Mona)
//
//===========================================================================

static int SetProfessionalSpdif(struct EchoGals *eg, char Prof)
{
	u32 ControlReg;

	// Clear the current S/PDIF flags
	ControlReg = SWAP32(eg->DspCommPage->ControlReg);
	ControlReg &= GML_SPDIF_FORMAT_CLEAR_MASK;

	// Set the new S/PDIF flags depending on the mode
	ControlReg |= GML_SPDIF_TWO_CHANNEL | GML_SPDIF_24_BIT | GML_SPDIF_COPY_PERMIT;
	if (Prof) {
		// Professional mode
		ControlReg |= GML_SPDIF_PRO_MODE;

		switch (GetSampleRate(eg)) {
		case 32000:
			ControlReg |= GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
			break;
		case 44100:
			ControlReg |= GML_SPDIF_SAMPLE_RATE0;
			break;
		case 48000:
			ControlReg |= GML_SPDIF_SAMPLE_RATE1;
			break;
		}
	} else {
		// Consumer mode
		switch (GetSampleRate(eg)) {
		case 32000:
			ControlReg |= GML_SPDIF_SAMPLE_RATE0 | GML_SPDIF_SAMPLE_RATE1;
			break;
		case 48000:
			ControlReg |= GML_SPDIF_SAMPLE_RATE1;
			break;
		}
	}

	WriteControlReg(eg, ControlReg);
	eg->ProfessionalSpdif = Prof;
	ECHO_DEBUGPRINTF(("SetProfessionalSpdif to %s\n", (Prof) ? "Professional" : "Consumer"));
	return ECHOSTATUS_OK;
}



/*
//===========================================================================
//
// Detect MIDI output activity
//
//===========================================================================

BOOL CLayla24DspCommObject::IsMidiOutActive()
{
  ULONGLONG  ullCurTime;

  m_pOsSupport->OsGetSystemTime( &ullCurTime );
  return( ( ( ullCurTime - MidiOutTime ) > MIDI_ACTIVITY_TIMEOUT_USEC ) ? FALSE : TRUE );

}  // BOOL CLayla24DspCommObject::IsMidiOutActive()
*/
