// ****************************************************************************
//
//  	CMiaDspCommObject.cpp
//
// Implementation file for EchoGals generic driver Darla 24 DSP
// interface class.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"

#include "DSP/MiaDSP.c"

static int SetInputClock(struct EchoGals *eg, u16 Clock);
static int SetProfessionalSpdif(struct EchoGals *eg, char Prof);
static int SetNominalLevel(struct EchoGals *eg, u16 Index, char Consumer);
static int UpdateFlags(struct EchoGals *eg);


static int InitHw(struct EchoGals *eg, u16 Device, u16 SubDevice)
{
	int err, i;

	ECHO_DEBUGPRINTF(("InitHw() - Mia\n"));

	if ((SubDevice & 0xfff0) != MIA) {
		ECHO_DEBUGPRINTF(("Wrong card type\n"));
		return -ENODEV;
	}

	// This part is common to all the cards
	if ((err = InitDspCommPage(eg))) {
		ECHO_DEBUGPRINTF(("InitHw - could not initialize DSP comm page\n"));
		return err;
	}

	eg->DeviceId = Device;
	eg->SubDeviceId = SubDevice;
	eg->CardType = MIA;
	eg->BadBoard = 1;
	eg->NumPipesOut = 8;
	eg->NumPipesIn = 4;
	eg->NumBussesOut = 4;
	eg->NumBussesIn = 4;
	eg->FirstDigitalBusOut = 2;
	eg->FirstDigitalBusIn = 2;
	eg->HasVmixer = TRUE;
	eg->NumMidiOut = 0;
	eg->NumMidiIn = 0;
	eg->DspCodeToLoad = MiaDSP;
	// Since this card has no ASIC, mark it as loaded so everything works OK
	eg->ASICLoaded = TRUE;
	eg->Flags =	ECHOGALS_FLAG_BADBOARD |
			ECHOCAPS_OUTPUT_GAIN |
			ECHOCAPS_VU_METER |
			ECHOGALS_FLAG_SPDIF_NODITHER |
			ECHOCAPS_INPUT_NOMINAL_LEVEL |
			ECHOCAPS_OUTPUT_NOMINAL_LEVEL |
			ECHOCAPS_SUPER_INTERLEAVE |
			ECHOCAPS_STEREO_BE32;
	eg->InClockTypes = ECHO_CLOCK_BIT_INTERNAL | ECHO_CLOCK_BIT_SPDIF;

	// Load the DSP and the ASIC on the PCI card
	err = LoadFirmware(eg);
	if (err || eg->BadBoard)
		return -EIO;

	eg->Flags &= ~ECHOGALS_FLAG_BADBOARD;

	// Must call this here after DSP is init to init gains and mutes
	if ((err = InitLineLevels(eg)))
		return err;

	// Set professional nominal levels
	for (i = 0; i < eg->FirstDigitalBusOut; i++)
		err = SetNominalLevel(eg, i, FALSE);	// FALSE is +4dBu here

	for (i = 0; i < eg->FirstDigitalBusIn; i++)
		err = SetNominalLevel(eg, eg->NumBussesOut + i, FALSE);

	// Get default sample rate from DSP
	eg->SampleRate = GetSampleRate(eg);

	ECHO_DEBUGPRINTF(("InitHw done\n"));
	return err;
}



//===========================================================================
//
// QueryAudioSampleRate is used to find out if this card can handle a given sample rate.
//
//===========================================================================

static int QueryAudioSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	if (SampleRate != 8000 && SampleRate != 11025 && SampleRate != 12000 &&
	    SampleRate != 16000 && SampleRate != 22050 && SampleRate != 24000 &&
	    SampleRate != 32000 && SampleRate != 44100 && SampleRate != 48000 &&
	    SampleRate != 88200 && SampleRate != 96000) {
		ECHO_DEBUGPRINTF(("QueryAudioSampleRate() Invalid sample rate\n"));
		return -EINVAL;
	}

	ECHO_DEBUGPRINTF(("QueryAudioSampleRate()\n"));
	return ECHOSTATUS_OK;
}



//===========================================================================
//
// GetInputClockDetect returns a bitmask consisting of all the input
// clocks currently connected to the hardware; this changes as the user
// connects and disconnects clock inputs.
//
// You should use this information to determine which clocks the user is
// allowed to select.
//
// Mia supports S/PDIF input clock.
//
//===========================================================================

static u32 GetInputClockDetect(const struct EchoGals *eg)
{
	u32 ClocksFromDsp, ClockDetectBits;

	// Map the DSP clock detect bits to the generic driver clock detect bits
	ClocksFromDsp = SWAP32(eg->DspCommPage->StatusClocks);

	ClockDetectBits = ECHO_CLOCK_BIT_INTERNAL;

	if (ClocksFromDsp & GLDM_CLOCK_DETECT_BIT_SPDIF)
		ClockDetectBits |= ECHO_CLOCK_BIT_SPDIF;

	return ClockDetectBits;
}



// The Mia has no ASIC. Just do nothing
static int LoadASIC(struct EchoGals *eg)
{
	return TRUE;
}



/****************************************************************************

	Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
// DspSetSampleRate
//
// Set the audio sample rate for Mia
//
//===========================================================================

static int DspSetSampleRate(struct EchoGals *eg, u32 SampleRate)
{
	u32 ControlReg;

	ControlReg = MIA_48000;

	switch (SampleRate) {
	case 96000:
		ControlReg = MIA_96000;
		break;
	case 88200:
		ControlReg = MIA_88200;
		break;
	case 44100:
		ControlReg = MIA_44100;
		break;
	case 32000:
		ControlReg = MIA_32000;
		break;
	case 24000:
		ControlReg = MIA_24000;
		break;
	case 22050:
		ControlReg = MIA_22050;
		break;
	case 16000:
		ControlReg = MIA_16000;
		break;
	case 12000:
		ControlReg = MIA_12000;
		break;
	case 11025:
		ControlReg = MIA_11025;
		break;
	case 8000:
		ControlReg = MIA_8000;
		break;
	}

	// Override the clock setting if this Mia is set to S/PDIF clock
	if (eg->InputClock == ECHO_CLOCK_SPDIF) {
		ControlReg &= MIA_SRC_MASK;
		ControlReg |= MIA_SPDIF;
	}
	// Set the control register if it has changed
	if (ControlReg != SWAP32(eg->DspCommPage->ControlReg)) {
		if (WaitForHandshake(eg))
			return -EIO;

		// Set the values in the comm page; the SampleRate field isn't
		// used by the DSP, but is read by the call to GetSampleRate below
		eg->DspCommPage->SampleRate = SWAP32(SampleRate);
		eg->DspCommPage->ControlReg = SWAP32(ControlReg);
		// Poke the DSP
		ClearHandshake(eg);
		SendVector(eg, DSP_VC_UPDATE_CLOCKS);
	}

	return GetSampleRate(eg);
}



//===========================================================================
//
//      Set the input clock
//
//===========================================================================

static int SetInputClock(struct EchoGals *eg, u16 Clock)
{
	u32 SampleRate;

	ECHO_DEBUGPRINTF(("SetInputClock\n"));
	SampleRate = GetSampleRate(eg);

	switch (Clock) {
	case ECHO_CLOCK_INTERNAL:
		ECHO_DEBUGPRINTF(("Set Mia clock to INTERNAL\n"));
		// If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
		if (SampleRate < 8000 || SampleRate > 96000)
			SampleRate = 48000;
		break;
	case ECHO_CLOCK_SPDIF:
		ECHO_DEBUGPRINTF(("Set Mia clock to SPDIF\n"));
		break;
	default:
		ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Mia\n", Clock));
		return -EINVAL;
	}

	eg->InputClock = Clock;
	DspSetSampleRate(eg, SampleRate);
	return ECHOSTATUS_OK;
}


/*
// Layla20 only
static int SetOutputClock(struct EchoGals *eg, u16 Clock)
{
	return -EINVAL;
}
*/

/*
// Gina24, Mona, Layla24 - only
static int DspSetDigitalMode(struct EchoGals *eg, u8 Mode)
{
	return -EINVAL;
}
*/


// This function routes the sound from a virtual channel to a real output
static int SetVmixerGain(struct EchoGals *eg, u16 Output, u16 Pipe, int Gain)
{
	int Index;

	if (Pipe >= eg->NumPipesOut) {
		ECHO_DEBUGPRINTF(("SetVmixertGain: Invalid Pipe %d\n", Pipe));
		return -EINVAL;
	}

	if (Output >= eg->NumBussesOut) {
		ECHO_DEBUGPRINTF(("SetVmixerGain: Invalid out %d\n", (int)Output));
		return -EINVAL;
	}

	if (WaitForHandshake(eg))
		return -EIO;

	eg->VmixerLevel[Output][Pipe] = Gain;
	Index = Output * eg->NumPipesOut + Pipe;
	eg->DspCommPage->VmixerLevel[Index] = GENERIC_TO_DSP(Gain);

	ECHO_DEBUGPRINTF(("SetVmixerGain: pipe %d, out %d = %d\n", (int)Pipe, (int)Output, Gain));
	return ECHOSTATUS_OK;
}



// Tell the DSP to read and update virtual mixer levels in comm page.
static int UpdateVmixerLevel(struct EchoGals *eg)
{
	//ECHO_DEBUGPRINTF(("UpdateVmixerLevel\n"));
	if (WaitForHandshake(eg))
		return -EIO;
	ClearHandshake(eg);
	return SendVector(eg, DSP_VC_SET_VMIXER_GAIN);
}


/*
// Set input bus gain (one unit is 0.5dB !)
static int SetInputGain(struct EchoGals *eg, u16 Channel, int Gain)
{
	return -EINVAL;
}
*/


// Tell the DSP to reread the flags from the comm page
static int UpdateFlags(struct EchoGals *eg)
{
	//  ECHO_DEBUGPRINTF(("UpdateFlags\n"));
	if (WaitForHandshake(eg))
		return -EIO;
	ClearHandshake(eg);
	return SendVector(eg, DSP_VC_UPDATE_FLAGS);
}



static int SetProfessionalSpdif(struct EchoGals *eg, char Prof)
{
	ECHO_DEBUGPRINTF(("SetProfessionalSpdif %d\n", Prof));
	if (Prof)
		eg->DspCommPage->Flags |= SWAP32(DSP_FLAG_PROFESSIONAL_SPDIF);
	else
		eg->DspCommPage->Flags &= ~SWAP32(DSP_FLAG_PROFESSIONAL_SPDIF);
	return UpdateFlags(eg);
}

