// ****************************************************************************
//
// CEchoGals.cpp
//
// Implementation file for the CEchoGals driver class.
// Set editor tabs to 3 for your viewing pleasure.
//
// Copyright Echo Digital Audio Corporation (c) 1998 - 2002
// All rights reserved
// www.echoaudio.com
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal with the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// - Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimers.
//
// - Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimers in the
// documentation and/or other materials provided with the distribution.
//
// - Neither the name of Echo Digital Audio, nor the names of its
// contributors may be used to endorse or promote products derived from
// this Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//

#ifndef ECHO24_FAMILY
#define ECHOGALS_FAMILY
#endif

#include "Echoaudio.h"
#include <linux/pci.h>



static inline int IsAudioOpen(struct EchoGals *eg, u16 PipeIndex)
{
	return (eg->AudioOpenMask & (1 << PipeIndex));
}



// Stops everything and turns off the DSP.
static void RestInPeace(struct EchoGals *eg)
{
	int i;

	ECHO_DEBUGPRINTF(("Del. open=%x\n", eg->AudioOpenMask));

	// Close all open pipes. We start from lower pipes, so CloseAudio() clears
	// the higher bits if the pipes have more than one channel and it never fails.
	for (i = 0; i < ECHO_MAXAUDIOPIPES; i++)
		if (IsAudioOpen(eg, i))
			FreeAudio(eg, i);
	DspPowerDown(eg);
}



/* Fills the comm page with default values */
static int InitDspCommPage(struct EchoGals *eg)
{
	int i;

	// Check if the compiler added extra padding inside the structure
	if (offsetof(DspCommPage_t, VmixerLevel) != 0xba0L) {
		ECHO_DEBUGPRINTF(("InitDspCommPage() - Invalid DspComm structure\n"));
		return -EPERM;
	}
	if (sizeof(dma_addr_t) != 4) {
		ECHO_DEBUGPRINTF(("InitDspCommPage() - dma_addr_t must be 32 bits long\n"));
		return -EPERM;
	}
	if (PAGE_SIZE < 4096)
		ECHO_DEBUGPRINTF(("InitDspCommPage() - PAGE_SIZE too small\n"));

	// Init all the basic stuff
	eg->BadBoard = TRUE;	// Set TRUE until DSP loaded
	eg->DspCode = NULL;	// Current DSP code not loaded
	eg->DigitalMode = DIGITAL_MODE_NONE;
	eg->InputClock = ECHO_CLOCK_INTERNAL;
	eg->OutputClock = ECHO_CLOCK_WORD;

#ifdef MIDI_SUPPORT
	eg->NextMidiWriteTime = 0;
#endif
	eg->ASICLoaded = 0;
	memset(eg->DspCommPage, 0, COMMPAGE_SIZE);

	// Init the comm page
	eg->DspCommPage->CommSize = SWAP32(sizeof(DspCommPage_t));	// Size of DSP comm page
	eg->DspCommPage->Handshake = 0xffffffff;
	eg->DspCommPage->MidiXmitStatus = 0xffffffff;
	eg->DspCommPage->SampleRate = SWAP32((u32)44100);		// Need this in case we start with ESYNC
	for (i = 0; i < DSP_MAXAUDIOINPUTS; i++)
		eg->DspCommPage->InLineLevel[i] = 0x00;

	// Set line levels so we don't blast any inputs on startup
	memset(eg->DspCommPage->Monitors, GENERIC_TO_DSP(ECHOGAIN_MUTED), MONITOR_ARRAY_SIZE);
	memset(eg->DspCommPage->VmixerLevel, GENERIC_TO_DSP(ECHOGAIN_MUTED), VMIXER_ARRAY_SIZE);

	return ECHOSTATUS_OK;
}



/*===========================================================================
//
// This function initializes the several volume controls for busses and pipes.
// It is called from the InitHw() functions once the DSP is up and running.
//
===========================================================================*/

static int InitLineLevels(struct EchoGals *eg)
{
	int st;
	int i, o;

	ECHO_DEBUGPRINTF(("InitLineLevels\n"));

	// Mute output busses
	for (i = 0; i < eg->NumBussesOut; i++)
		if ((st = SetOutputGain(eg, i, ECHOGAIN_MUTED)))
			return st;
	if ((st = UpdateOutputLineLevel(eg)))
		return st;

#ifdef ECHOCARD_HAS_VMIXER
	// Mute the Vmixer
	for (i = eg->NumPipesOut - 1; i >= 0; i--)
		for (o = eg->NumBussesOut - 1; o >= 0; o--)
			if ((st = SetVmixerGain(eg, o, i, ECHOGAIN_MUTED)))
				return st;
	if ((st = UpdateVmixerLevel(eg)))
		return st;
#endif /* ECHOCARD_HAS_VMIXER */

	// Mute the monitor mixer
	for (o = eg->NumBussesOut - 1; o >= 0; o--)
		for (i = eg->NumBussesIn - 1; i >= 0; i--)
			if ((st = SetMonitorGain(eg, o, i, ECHOGAIN_MUTED)))
				return st;
	if ((st = UpdateOutputLineLevel(eg)))
		return st;

#ifdef ECHOCARD_HAS_INPUT_GAIN
	for (i = 0; i < eg->NumBussesIn; i++)
		if ((st = SetInputGain(eg, i, ECHOGAIN_MUTED)))
			return st;
	if ((st = UpdateInputLineLevel(eg)))
		return st;
#endif /* ECHOCARD_HAS_INPUT_GAIN */

	return ECHOSTATUS_OK;
}



/*===========================================================================
//
// Set the digital mode - currently for Gina24, Layla24, and Mona
//
===========================================================================*/

#ifdef ECHOCARD_HAS_DIGITAL_MODE_SWITCH

static int SetDigitalMode(struct EchoGals *eg, u8 Mode)
{
	int err;
	u8 PreviousDigitalMode;
	int i, o;

	if (eg->BadBoard)
		return -EIO;

	if (eg->AudioOpenMask) {
		ECHO_DEBUGPRINTF(("SetDigitalMode()  All audio channels must be closed before changing the digital mode\n"));
		return -EAGAIN;
	}

	if (!(eg->DigitalModes & (1 << Mode)))
		return -EINVAL;

	PreviousDigitalMode = eg->DigitalMode;
	err = DspSetDigitalMode(eg, Mode);

	// If we successfully changed the digital mode from or to ADAT, then make sure
	// all output, input and monitor levels are updated by the DSP comm object.
	if (!err && PreviousDigitalMode != Mode &&
	    (PreviousDigitalMode == DIGITAL_MODE_ADAT || Mode == DIGITAL_MODE_ADAT)) {
		for (o = eg->NumBussesOut - 1; o >= 0; o--)
			for (i = eg->NumBussesIn - 1; i >= 0; i--)
				SetMonitorGain(eg, o, i, eg->MonitorLevel[o][i]);

#ifdef ECHOCARD_HAS_INPUT_GAIN
		for (i = eg->NumBussesIn - 1; i >= 0; i--)
			SetInputGain(eg, i, eg->InLevel[i]);
		UpdateInputLineLevel(eg);
#endif

		for (i = eg->NumBussesOut - 1; i >= 0; i--)
			SetOutputGain(eg, i, eg->OutLevel[i]);
		UpdateOutputLineLevel(eg);
	}

	return err;
}

#endif /* ECHOCARD_HAS_DIGITAL_MODE_SWITCH */



#ifdef ECHOCARD_HAS_DIGITAL_IO	/* This is currently used only for the nodither switch */

static void SetFlags(struct EchoGals *eg, u16 Flags)
{
	// Mask off the read-only flags so they don't change
	Flags &= ECHOGALS_FLAG_WRITABLE_MASK;
	// Set the flags & mark the flags as changed
	eg->Flags |= Flags;
}



static void ClearFlags(struct EchoGals *eg, u16 Flags)
{
	// Mask off the read-only flags so they don't change
	Flags &= ECHOGALS_FLAG_WRITABLE_MASK;
	// Set the flags & mark the flags as changed
	eg->Flags &= ~Flags;
}

#endif /* ECHOCARD_HAS_DIGITAL_IO */



/*===========================================================================
//
// This isn't the interrupt handler itself; rather, the OS-specific layer
// of the driver has an interrupt handler that calls this function.
//
===========================================================================*/

static int ServiceIrq(struct EchoGals *eg, int *fMidiReceived)
{
	// Read the DSP status register and see if this DSP generated this interrupt
	*fMidiReceived = FALSE;

	if (SWAP32(eg->DspRegBase[CHI32_STATUS_REG]) & CHI32_STATUS_IRQ) {
/*
#ifdef MIDI_SUPPORT

    //
    // If this was a MIDI input interrupt, get the MIDI input data
    //
    DWORD dwMidiInCount;

    pDCO->ReadMidi(0, dwMidiInCount);	// The count is at index 0
    if (0 != dwMidiInCount) {
      m_MidiIn.ServiceIrq();
      fMidiReceived = TRUE;
    }
#endif // MIDI_SUPPORT
*/
		// Clear the hardware interrupt
		eg->DspCommPage->MidiInData[0] = 0;
		SendVector(eg, DSP_VC_ACK_INT);
		return ECHOSTATUS_OK;
	}
	return ECHOSTATUS_IRQ_NOT_OURS;
}




/******************************************************************************

 Functions for opening and closing pipes

 ******************************************************************************/

/*===========================================================================
//
// AllocateAudio is used to reserve audio pipes for your exclusive use.  The call
// will fail if someone else has already opened the pipes.  Calling AllocateAudio
// is the first step if you want to play or record. Note that you can eventually
// set an interleave smaller than the number of channel you have allocated.
//
===========================================================================*/

static int AllocateAudio(struct EchoGals *eg, u16 Channel, u16 Interleave, char IsInput)
{
	int i, PipeIndex, max, err;
	u32 ChannelMask;
	char IsCyclic;

	ECHO_DEBUGPRINTF(("OpenAudio: ch=%d int=%d\n", Channel, Interleave));

	if (eg->BadBoard)
		return -EIO;

	IsCyclic = 1;	// This driver uses cyclic buffers only
	if (IsInput) {
		PipeIndex = Channel + eg->NumPipesOut;
		max = eg->NumPipesOut + eg->NumBussesIn;
	} else {
		max = eg->NumBussesOut;
		PipeIndex = Channel;
	}
/*
	if (PipeIndex + Interleave > max) {
		ECHO_DEBUGPRINTF(("OpenAudio: invalid channel\n"));
		return -EINVAL;
	}
	// If the width is more than two, make sure that this card can handle super interleave
	if (Interleave > 2 && !(eg->Flags & ECHOCAPS_SUPER_INTERLEAVE)) {
		ECHO_DEBUGPRINTF(("OpenAudio: no super interleave\n"));
		return -EINVAL;
	}
*/
	// Compute channel mask for this substream
	for (ChannelMask = i = 0; i < Interleave; i++)
		ChannelMask |= 1 << (PipeIndex + i);
	// Check if the specified pipes are already open
	if (eg->AudioOpenMask & ChannelMask) {
		ECHO_DEBUGPRINTF(("OpenAudio: channel already open\n"));
		return -EAGAIN;
	}

	if ((err = AllocateSGList(eg, PipeIndex)))
		return err;

	// Reset dma counter
	eg->DspCommPage->Position[PipeIndex] = 0;

	// Prep stuff
	eg->AudioOpenMask |= ChannelMask;
	if (IsCyclic)
		eg->AudioCyclicMask |= ChannelMask;
	eg->Pipes[PipeIndex].FirstChannel = Channel;
	eg->Pipes[PipeIndex].Interleave = Interleave;
	eg->Pipes[PipeIndex].IsInput = IsInput;
	eg->Pipes[PipeIndex].IsCyclic = IsCyclic;
	StopAudio(eg, PipeIndex);
	ECHO_DEBUGPRINTF(("OpenAudio: ok\n"));
	return PipeIndex;
}



/*===========================================================================
//
// FreeAudio is, naturally, the inverse of AllocateAudio.
//
===========================================================================*/

static int FreeAudio(struct EchoGals *eg, u16 PipeIndex)
{
	u32 ChannelMask;
	int i;

	ECHO_DEBUGPRINTF(("CloseAudio: Pipe %u\n", PipeIndex));

	if (!IsAudioOpen(eg, PipeIndex))
		return -EINVAL;

	// Compute channel mask for this substream
	for (ChannelMask = i = 0; i < eg->Pipes[PipeIndex].Interleave; i++)
		ChannelMask |= 1 << (PipeIndex + i);

	StopAudio(eg, PipeIndex);
	FreeSGList(eg, PipeIndex);

	eg->AudioOpenMask &= ~ChannelMask;
	eg->AudioCyclicMask &= ~ChannelMask;
	eg->Pipes[PipeIndex].Interleave = 0;
	return ECHOSTATUS_OK;
}



/*===========================================================================
//
// This function does exactly what you think it does.
//
// Note that if the card is not set to internal clock (that is, the hardware
// is synced to word clock or some such), this call has no effect.
//
// Note that all of the inputs and outputs on a single card share the same clock.
//
===========================================================================*/

static int SetAudioSampleRate(struct EchoGals *eg, u32 Rate)
{
	ECHO_DEBUGPRINTF(("SetAudioSampleRate: to %d Hz\n", Rate));

	if (QueryAudioSampleRate(eg, Rate))
		return -EINVAL;

	if (Rate == DspSetSampleRate(eg, Rate)) {
		eg->SampleRate = Rate;
		return ECHOSTATUS_OK;
	}
	return -EINVAL;
}



/*===========================================================================
//
// Validate an audio format.
//
// For comments on audio formats, refer to the definition of Audioformat_t
//
===========================================================================*/

static int QueryAudioFormat(struct EchoGals *eg, u16 PipeIndex, const Audioformat_t *AudioFormat)
{
	u16 Interleave, StoredPipeWidth, MaxPipe;

	ECHO_DEBUGPRINTF(("QueryAudioFormat\n"));

	// If this pipe is open, make sure that this audio format does not exceed the stored pipe width
	Interleave = AudioFormat->DataInterleave;
	StoredPipeWidth = eg->Pipes[PipeIndex].Interleave;

	if (StoredPipeWidth) {
		if (Interleave > StoredPipeWidth) {
			ECHO_DEBUGPRINTF(("QueryAudioFormat: pipe was opened with a"
					" width of %d; interleave of %d invalid.\n",
					StoredPipeWidth, AudioFormat->DataInterleave));
			return -EINVAL;
		}
	}
	// Check for super interleave (i.e. interleave > 2)
	if (Interleave > 2) {
		// Make sure the card is capable of super interleave
		if (!(eg->Flags & ECHOCAPS_SUPER_INTERLEAVE))
			return -EINVAL;

		// Allow super interleave for 32 bit little endian data;
		// the interleave factor must be even.
		if (AudioFormat->BitsPerSample != 32 || AudioFormat->DataAreBigEndian || (Interleave & 1))
			return -EINVAL;

		// Make sure that this interleave factor on this pipe
		// does not exceed the number of pipes for the card
		if (PipeIndex >= eg->NumPipesOut) {
			MaxPipe = eg->NumPipesIn;
			PipeIndex -= eg->NumPipesOut;
		} else {
			MaxPipe = eg->NumPipesOut;
		}

		if (PipeIndex + Interleave > MaxPipe)
			return -EINVAL;

		return ECHOSTATUS_OK;
	}
	// Check the interleave, which must be 1 or 2
	if (AudioFormat->DataInterleave < 1) {
		ECHO_DEBUGPRINTF(("QueryAudioFormat: interleave %d not allowed\n", AudioFormat->DataInterleave));
		return -EINVAL;
	}
	// If the big endian flag is set, the data must be mono or stereo interleave,
	// 32 bits wide, left justified data.  Only the upper 24 bits are used.
	if (AudioFormat->DataAreBigEndian) {
		// Must have 32 bits per sample
		if (AudioFormat->BitsPerSample != 32) {
			ECHO_DEBUGPRINTF(("QueryAudioFormat: Only 32 bits per sample supported for big-endian data\n"));
			return -EINVAL;
		}
		// Mono or stereo only
#ifdef STEREO_BIG_ENDIAN32_SUPPORT
		if (AudioFormat->DataInterleave != 1 && AudioFormat->DataInterleave != 2) {
			ECHO_DEBUGPRINTF(("QueryAudioFormat: Interleave of %d not allowed for big-endian data\n", AudioFormat->DataInterleave));
			return -EINVAL;
		}
#else
		if (AudioFormat->DataInterleave != 1) {
			ECHO_DEBUGPRINTF(("QueryAudioFormat: Interleave of %d not allowed for big-endian data\n", AudioFormat->DataInterleave));
			return -EINVAL;
		}
#endif

		return ECHOSTATUS_OK;
	}
	// Check bits per sample
	if (AudioFormat->BitsPerSample != 8 &&
	    AudioFormat->BitsPerSample != 16 &&
	    AudioFormat->BitsPerSample != 32) {
		ECHO_DEBUGPRINTF(("QueryAudioFormat: No valid format specified, bits per sample %d\n", AudioFormat->BitsPerSample));
		return -EINVAL;
	}

	return ECHOSTATUS_OK;
}



/*===========================================================================
//
// SetAudioFormat sets the format of the audio data in host memory for this pipe.
//
===========================================================================*/

static int SetAudioFormat(struct EchoGals *eg, u16 PipeIndex, const Audioformat_t *AudioFormat)
{
	ECHO_DEBUGPRINTF(("SetAudioFormat: for pipe %d\n", PipeIndex));

	// Check the pipe number
	if (PipeIndex >= eg->NumPipesOut + eg->NumPipesIn) {
		ECHO_DEBUGPRINTF(("SetAudioFormat: Invalid pipe %d\n", PipeIndex));
		return -EINVAL;
	}
	// Make sure this pipe is open
	if (!IsAudioOpen(eg, PipeIndex))
		return -EINVAL;
	// Check the format
	if (QueryAudioFormat(eg, PipeIndex, AudioFormat))
		return -EINVAL;

	DspSetFormat(eg, PipeIndex, AudioFormat, eg->Flags & ECHOGALS_FLAG_SPDIF_NODITHER);
	return 0;
}



/******************************************************************************

 Functions for starting and stopping transport

******************************************************************************/

/*===========================================================================
//
//      Start transport for a group of pipes
//
===========================================================================*/

static int StartAudioMask(struct EchoGals *eg, u32 Mask)
{
	u16 PipeIndex;
	u32 StartMask;

	if (eg->BadBoard)
		return -EIO;

	StartMask = 0;

	//--------------------------------------------------------
	// Process each pipe in the mask
	//--------------------------------------------------------
	for (PipeIndex = 0; PipeIndex < eg->NumPipesOut + eg->NumBussesIn; PipeIndex++) {
		// Skip this pipe if it's not in the mask
		if (!(Mask & (1 << PipeIndex)))
			continue;

		// This pipe must have a scatter-gather list
		if (!eg->Pipes[PipeIndex].sglist) {
			ECHO_DEBUGPRINTF(("StartAudioMask: trying to start pipe index %d but there is no s-g list !\n", PipeIndex));
			return -EINVAL;
		}

		// Do different things to this pipe depending on the state
		switch (eg->Pipes[PipeIndex].State) {
		case PIPE_STATE_STOPPED:
			eg->DspCommPage->Position[PipeIndex] = 0;	// Clean up the DMA counter
		case PIPE_STATE_PAUSED:
			eg->Pipes[PipeIndex].State = PIPE_STATE_STARTED;
			StartMask |= 1 << PipeIndex;
			break;
		case PIPE_STATE_PENDING:
		case PIPE_STATE_STARTED:
			break;
		}
	}

	// Send the result of all this to the DSP
	if (!StartMask)
		return ECHOSTATUS_OK;

	return StartTransport(eg, StartMask, eg->AudioCyclicMask);
}



/*===========================================================================
//
//      Start transport for a single pipe
//
===========================================================================*/

static int StartAudio(struct EchoGals *eg, u16 PipeIndex)
{
	return StartAudioMask(eg, 1 << PipeIndex);
}



/*===========================================================================
//
// Pause several pipes simultaneously
//
===========================================================================*/

static int PauseAudioMask(struct EchoGals *eg, u32 Mask)
{
	int i, Status;

	if (eg->BadBoard)
		return -EIO;

	if ((Status = PauseTransport(eg, Mask)))
		return Status;

	for (i = eg->NumPipesOut + eg->NumPipesIn; i >= 0; i--)
		if (Mask & (1 << i))
			eg->Pipes[i].State = PIPE_STATE_PAUSED;

	return ECHOSTATUS_OK;
}



/*===========================================================================
//
// Pause a single pipe
//
===========================================================================*/

static int PauseAudio(struct EchoGals *eg, u16 PipeIndex)
{
	return PauseAudioMask(eg, 1 << PipeIndex);
}



/*===========================================================================
//
// Stop transport for a group of pipes simultaneously
//
===========================================================================*/

static int StopAudioMask(struct EchoGals *eg, u32 Mask)
{
	int i, Status;

	if (eg->BadBoard)
		return -EIO;

	if ((Status = StopTransport(eg, Mask)))
		return Status;

	for (i = eg->NumPipesOut + eg->NumPipesIn; i >= 0; i--)
		if (Mask & (1 << i))
			eg->Pipes[i].State = PIPE_STATE_STOPPED;

	return ECHOSTATUS_OK;
}



/*===========================================================================
//
// Stop transport for a single pipe
//
===========================================================================*/

static int StopAudio(struct EchoGals *eg, u16 PipeIndex)
{
	return StopAudioMask(eg, 1 << PipeIndex);
}



/*===========================================================================
//
// This function returns a pointer to the memory location where the DSP
// writes the 32 bit DMA position for a pipe.  The DSP is constantly
// updating this value as it moves data.
// The DMA counter is in units of bytes, not samples.
//
===========================================================================*/

static volatile u32 *GetAudioPositionPtr(struct EchoGals *eg, u16 PipeIndex)
{
	if (PipeIndex >= eg->NumPipesOut + eg->NumPipesIn)
		return 0;

	return (&eg->DspCommPage->Position[PipeIndex]);
}




/******************************************************************************

 Functions for managing the scatter-gather list

******************************************************************************/

static int AllocateSGList(struct EchoGals *eg, u16 PipeIndex)
{
	AudioPipe_t *Pipe = &eg->Pipes[PipeIndex];
	struct snd_dma_device dmadev;
	int err;

	dmadev.type = SNDRV_DMA_TYPE_DEV;
	dmadev.dev = snd_dma_pci_data(eg->pci);
	/* Allocate a single page for the scatter-gather list */
	if ((err = snd_dma_alloc_pages(&dmadev, PAGE_SIZE, &Pipe->dmabuf)) < 0) {
		ECHO_DEBUGPRINTF(("SGlist: memory allocation failed\n"));
		return err;
	}
	Pipe->sglist = (sg_entry_t *)Pipe->dmabuf.area;
	Pipe->sglist_phys = Pipe->dmabuf.addr;

	memset(Pipe->sglist, 0, PAGE_SIZE);
	Pipe->sglHead = 0;
	eg->DspCommPage->SGListPhys[PipeIndex].PhysAddr = SWAP32(Pipe->sglist_phys);
	ECHO_DEBUGPRINTF(("SGlist: memory allocated phys=%x\n", Pipe->sglist_phys));
	return ECHOSTATUS_OK;
}



static void FreeSGList(struct EchoGals *eg, u16 PipeIndex)
{
	AudioPipe_t *Pipe = &eg->Pipes[PipeIndex];
	struct snd_dma_device dmadev;

	dmadev.type = SNDRV_DMA_TYPE_DEV;
	dmadev.dev = snd_dma_pci_data(eg->pci);
	snd_dma_free_pages(&dmadev, &Pipe->dmabuf);

	ECHO_DEBUGPRINTF(("SGlist: free phys=%x\n", Pipe->sglist_phys));
	Pipe->sglist = 0;
	Pipe->sglist_phys = 0;
}



static int SGListAddMapping(struct EchoGals *eg, u16 PipeIndex, dma_addr_t Address, size_t Length)
{
	int head;

	head = eg->Pipes[PipeIndex].sglHead;
	//ECHO_DEBUGPRINTF(("SGlist: add[%2d]  phys=%x  len=%d\n", head, Address, Length));
	if (head < MAX_SGLIST_ENTRIES - 1) {
		eg->Pipes[PipeIndex].sglist[head].PhysAddr = SWAP32(Address);
		eg->Pipes[PipeIndex].sglist[head].Size = SWAP32(Length);
		head++;
	} else {
		ECHO_DEBUGPRINTF(("SGlist: too many fragments\n"));
		return -ENOMEM;
	}

	eg->Pipes[PipeIndex].sglHead = head;
	return ECHOSTATUS_OK;
}



static int SGListAddIrq(struct EchoGals *eg, u16 PipeIndex)
{
	return SGListAddMapping(eg, PipeIndex, 0, 0);
}



static int SGListWrap(struct EchoGals *eg, u16 PipeIndex)
{
	return SGListAddMapping(eg, PipeIndex, eg->Pipes[PipeIndex].sglist_phys, 0);
}
