// ****************************************************************************
//
//              CEchoGals.H
//
//              Include file for the CEchoGals generic driver class.
//              Set editor tabs to 3 for your viewing pleasure.
//
//      CEchoGals is the big daddy class of the generic code.  It is the upper
//              edge of the generic code - that is, it is the interface between the
//      operating system-specific code and the generic driver.
//
//              There are a number of terms in this file that won't make any sense unless
//      you go read EchoGalsXface.h first.
//
// ----------------------------------------------------------------------------
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************


/*

   Here's a block diagram of how most of the cards work:

                  +-----------+
           record |           |<-------------------- Inputs
          <-------|           |        |
     PCI          | Transport |        |
     bus          |  engine   |       \|/
          ------->|           |    +-------+
            play  |           |--->|monitor|-------> Outputs
                  +-----------+    | mixer |
                                   +-------+

   The lines going to and from the PCI bus represent "pipes".  A pipe performs
   audio transport - moving audio data to and from buffers on the host via
   bus mastering.

   The inputs and outputs on the right represent input and output "busses."
   A bus is a physical, real connection to the outside world.  An example
   of a bus would be the 1/4" analog connectors on the back of Layla or
   an RCA S/PDIF connector.

   For most cards, there is a one-to-one correspondence between outputs
   and busses; that is, each individual pipe is hard-wired to a single bus.

   Cards that work this way are Darla20, Gina20, Layla20, Darla24, Gina24,
   Layla24, Mona, and Indigo.


   Mia has a feature called "virtual outputs."


                  +-----------+
           record |           |<----------------------------- Inputs
          <-------|           |                  |
     PCI          | Transport |                  |
     bus          |  engine   |                 \|/
          ------->|           |   +------+   +-------+
            play  |           |-->|vmixer|-->|monitor|-------> Outputs
                  +-----------+   +------+   | mixer |
                                             +-------+


   Obviously, the difference here is the box labeled "vmixer."  Vmixer is
   short for "virtual output mixer."  For Mia, pipes are *not* hard-wired
   to a single bus; the vmixer lets you mix any pipe to any bus in any
   combination.

   Note, however, that the left-hand side of the diagram is unchanged.
   Transport works exactly the same way - the difference is in the mixer stage.


   Pipes and busses are numbered starting at zero.



   Pipe index
   ==========

   A number of calls in CEchoGals refer to a "pipe index".  A pipe index is
   a unique number for a pipe that unambiguously refers to a playback or record
   pipe.  Pipe indices are numbered starting with analog outputs, followed by
   digital outputs, then analog inputs, then digital inputs.

   Take Gina24 as an example:

   Pipe index

   0-7            Analog outputs
   8-15           Digital outputs
   16-17          Analog inputs
   18-25          Digital inputs


   You get the pipe index by calling CEchoGals::OpenAudio; the other transport
   functions take the pipe index as a parameter.  If you need a pipe index for
   some other reason, use the handy MakePipeIndex method.


   Some calls take a CChannelMask parameter; CChannelMask is a handy way to group
   pipe indices.



   Digital mode switch
   ===================

   Some cards (right now, Gina24, Layla24, and Mona) have a Digital Mode Switch
   or DMS.  Cards with a DMS can be set to one of three mutually exclusive
   digital modes: S/PDIF RCA, S/PDIF optical, or ADAT optical.

   This may create some confusion since ADAT optical is 8 channels wide and
   S/PDIF is only two channels wide.  Gina24, Layla24, and Mona handle this
   by acting as if they always have 8 digital outs and ins.  If you are in
   either S/PDIF mode, the last 6 channels don't do anything - data sent
   out these channels is thrown away and you will always record zeros.

   Note that with Gina24, Layla24, and Mona, sample rates above 50 kHz are
   only available if you have the card configured for S/PDIF optical or S/PDIF RCA.



   Double speed mode
   =================

   Some of the cards support 88.2 kHz and 96 kHz sampling (Darla24, Gina24, Layla24,
   Mona, Mia, and Indigo).  For these cards, the driver sometimes has to worry about
   "double speed mode"; double speed mode applies whenever the sampling rate is above
   50 kHz.

   For instance, Mona and Layla24 support word clock sync.  However, they actually
   support two different word clock modes - single speed (below 50 kHz) and double
   speed (above 50 kHz).  The hardware detects if a single or double speed word clock
   signal is present; the generic code uses that information to determine which mode
   to use.

   The generic code takes care of all this for you.
*/


#if !defined(ECHO24_FAMILY) && !defined(ECHOGALS_FAMILY)
#error No family defined !
#endif


#ifndef _ECHOAUDIO_H_
#define _ECHOAUDIO_H_

#include <linux/string.h>	// Also defines NULL and size_t
#include <sound/driver.h>
#include <sound/memalloc.h>

#if (CONFIG_SND_DEBUG)
#include <sound/core.h>
#define ECHO_DEBUGPRINTF(strings) snd_printk strings
#else
#define ECHO_DEBUGPRINTF(strings)
#endif


#ifdef __BIG_ENDIAN
#define SWAP16(x) le16_to_cpu(x)
#define SWAP32(x) le32_to_cpu(x)
#else
#define SWAP16(x) (x)
#define SWAP32(x) (x)
#endif

#define TRUE 1
#define FALSE 0

/* Memory barriers

 MB prevents loads and stores being reordered across this point.
 RMB prevents loads being reordered across this point.
 WMB prevents stores being reordered across this point.

 On PowerPC the eieio instruction is a barrier providing an ordering (separately) for (a) cacheable
 stores and (b) loads and stores to non-cacheable memory (e.g. I/O devices).

 For now, "wmb()" doesn't actually do anything, as all
 Intel CPU's follow what Intel calls a *Processor Order*,
 in which all writes are seen in the program order even
 outside the CPU.
 I expect future Intel CPU's to have a weaker ordering,
 but I'd also expect them to finally get their act together
 and add some real memory barriers if so.
#define MB wmb()
#define RMB wmb()
#define WMB wmb()
*/
#define MB
#define RMB
#define WMB








#include "Echo_Dsp.h"


// Debug only - forces the Gina24 driver to export the same controls of the Mia
//#define EMUL

//@@ Disable midi support
#undef MIDI_SUPPORT




//***********************************************************************
//
//	PCI configuration space
//
//***********************************************************************

//
// PCI vendor ID and device IDs for the hardware
//
#define VENDOR_ID		0x1057
#define DEVICE_ID_56301		0x1801
#define DEVICE_ID_56361		0x3410
#define SUBVENDOR_ID		0xECC0


//
// Valid Echo PCI subsystem card IDs
//
#define DARLA20			0x0010
#define GINA20			0x0020
#define LAYLA20			0x0030
#define DARLA24			0x0040
#define GINA24			0x0050
#define LAYLA24			0x0060
#define MONA			0x0070
#define MIA			0x0080
#define INDIGO			0x0090


//***********************************************************************
//
//	Array sizes and so forth
//
//***********************************************************************

//
// Sizes
//
#define ECHO_MAXAUDIOINPUTS	32	// Max audio input channels
#define ECHO_MAXAUDIOOUTPUTS	32	// Max audio output channels
#define ECHO_MAXAUDIOPIPES	32	// Max number of input and output pipes
#define ECHO_MAXMIDIJACKS	1	// Max MIDI ports
#define ECHO_MIDI_QUEUE_SZ 	512	// Max MIDI input queue entries
#define ECHO_MTC_QUEUE_SZ	32	// Max MIDI time code input queue entries

//
// MIDI activity indicator timeout
//
#define MIDI_ACTIVITY_TIMEOUT_USEC	200000


//*****************************************************************************
//
// Clocks
//
//*****************************************************************************

//
// Clock numbers
//
#define ECHO_CLOCK_INTERNAL		0
#define ECHO_CLOCK_WORD			1
#define ECHO_CLOCK_SUPER		2
#define ECHO_CLOCK_SPDIF		3
#define ECHO_CLOCK_ADAT			4
#define ECHO_CLOCK_ESYNC		5
#define ECHO_CLOCK_ESYNC96		6
#define ECHO_CLOCK_NUMBER		7
#define ECHO_CLOCKS			0xffff

//
// Clock bit numbers - used to report capabilities and whatever clocks
// are being detected dynamically.
//
#define ECHO_CLOCK_BIT_INTERNAL		(1 << ECHO_CLOCK_INTERNAL)
#define ECHO_CLOCK_BIT_WORD		(1 << ECHO_CLOCK_WORD)
#define ECHO_CLOCK_BIT_SUPER		(1 << ECHO_CLOCK_SUPER)
#define ECHO_CLOCK_BIT_SPDIF		(1 << ECHO_CLOCK_SPDIF)
#define ECHO_CLOCK_BIT_ADAT		(1 << ECHO_CLOCK_ADAT)
#define ECHO_CLOCK_BIT_ESYNC		(1 << ECHO_CLOCK_ESYNC)
#define ECHO_CLOCK_BIT_ESYNC96		(1 << ECHO_CLOCK_ESYNC96)


//*****************************************************************************
//
// Digital modes
//
//*****************************************************************************

//
// Digital modes for Mona, Layla24, and Gina24
//
#define DIGITAL_MODE_NONE			0xFF
#define DIGITAL_MODE_SPDIF_RCA			0
#define DIGITAL_MODE_SPDIF_OPTICAL		1
#define DIGITAL_MODE_ADAT			2
#define DIGITAL_MODE_SPDIF_CDROM		3
#define DIGITAL_MODES				4

//
//	Digital mode capability masks
//
#define ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA	(1 << DIGITAL_MODE_SPDIF_RCA)
#define ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_OPTICAL	(1 << DIGITAL_MODE_SPDIF_OPTICAL)
#define ECHOCAPS_HAS_DIGITAL_MODE_ADAT		(1 << DIGITAL_MODE_ADAT)
#define ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_CDROM	(1 << DIGITAL_MODE_SPDIF_CDROM)


//*****************************************************************************
//
// Driver flags
//
//*****************************************************************************

//
//	Note that some flags are read-only (look for ROFLAG as part of the name).
// The read-only flags are used to tell you what the card can do; you can't
// set or clear them.
//
// Some cards can only handle mono & stereo interleaved data in host audio buffers; that
// is, you may only specify 1 or 2 in the wPipeWidth field when you open a pipe for
// playing or recording.
//
// If the card has the super interleave flag set, that means that you can interleave
// by more than 2.  For super interleave-capable cards, you can interleave by 1, 2,
// 4, 6, 8, 10, 12, 14, or 16.  Note that you cannot interleave beyond the actual
// number of pipes on the card - if the card has 16 pipes and you open pipe 6,
// the maximum interleave would be 10.
//
//
// Gina24, Layla24, and Mona support digital input auto-mute.  If the digital input
// auto-mute is enabled, the DSP will only enable the digital inputs if the card
// is syncing to a valid clock on the ADAT or S/PDIF inputs.
//
// If the auto-mute is disabled, the digital inputs are enabled regardless of
// what the input clock is set or what is connected.
//
//

#define ECHOGALS_FLAG_BADBOARD			0x0001	// Board could not be init or stopped responding (i.e. DSP crashed)
#define ECHOGALS_FLAG_SPDIF_NODITHER		0x0002	// Used to disable digital input dithering
#define ECHOGALS_FLAG_DIGITAL_IN_NODITHER	0x0002
#define ECHOGALS_FLAG_SYNCH_WAVE		0x0004	// Used to enable wave device synchronization. When set, all open wave devices for a single task start when the start for the last channel is received.
#define ECHOGALS_FLAG_DIGITAL_IN_AUTOMUTE	0x0008	// Set if the console needs to handle the digital input auto-mute

#define ECHOCAPS_SUPER_INTERLEAVE		0x0080
#define ECHOCAPS_VMIXER				0x0100
#define ECHOCAPS_INPUT_GAIN			0x0200
#define ECHOCAPS_OUTPUT_GAIN			0x0400
#define ECHOCAPS_DIGITAL_IN_AUTOMUTE		0x0800	// Set if the console needs to handle the digital input auto-mute
#define ECHOCAPS_STEREO_BE32			0x1000
#define ECHOCAPS_VU_METER 			0x2000
#define ECHOCAPS_INPUT_NOMINAL_LEVEL		0x4000
#define ECHOCAPS_OUTPUT_NOMINAL_LEVEL		0x8000

#define ECHOGALS_FLAG_WRITABLE_MASK		0x000f



//
// Return status values
//
//	Changes here require a change in pStatusStrs in CEchoGals.cpp
//
#define ECHOSTATUS_OK				0
#define ECHOSTATUS_FIRST			0
#define ECHOSTATUS_BAD_FORMAT			-0x01
#define ECHOSTATUS_BAD_BUFFER_SIZE		-0x02
#define ECHOSTATUS_CANT_OPEN			-0x03
#define ECHOSTATUS_CANT_CLOSE			-0x04
#define ECHOSTATUS_CHANNEL_NOT_OPEN		-0x05
#define ECHOSTATUS_BUSY				-0x06
#define ECHOSTATUS_BAD_LEVEL			-0x07
#define ECHOSTATUS_NO_MIDI			-0x08
#define ECHOSTATUS_CLOCK_NOT_SUPPORTED		-0x09
#define ECHOSTATUS_CLOCK_NOT_AVAILABLE		-0x0A
#define ECHOSTATUS_BAD_CARDID			-0x0B
#define ECHOSTATUS_NOT_SUPPORTED		-0x0C
#define ECHOSTATUS_BAD_NOTIFY_SIZE		-0x0D
#define ECHOSTATUS_INVALID_PARAM		-0x0E
#define ECHOSTATUS_NO_MEM			-0x0F
#define ECHOSTATUS_NOT_SHAREABLE		-0x10
#define ECHOSTATUS_FIRMWARE_LOADED		-0x11
#define ECHOSTATUS_DSP_DEAD			-0x12
#define ECHOSTATUS_DSP_TIMEOUT			-0x13
#define ECHOSTATUS_INVALID_CHANNEL		-0x14
#define ECHOSTATUS_CHANNEL_ALREADY_OPEN		-0x15
#define ECHOSTATUS_SGLIST_FULL			-0x16
#define ECHOSTATUS_INVALID_INDEX		-0x17
#define ECHOSTATUS_BAD_CARD_NAME 		-0x18
#define ECHOSTATUS_IRQ_NOT_OURS			-0x19
#define ECHOSTATUS_NO_DATA			-0x1E
#define ECHOSTATUS_BUFFER_OVERFLOW		-0x1F
#define ECHOSTATUS_OPERATION_CANCELED		-0x20
#define ECHOSTATUS_EVENT_NOT_OPEN		-0x21
#define ECHOSTATUS_ASIC_NOT_LOADED		-0x22
#define ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED	-0x23
#define ECHOSTATUS_RESERVED			-0x24
#define ECHOSTATUS_BAD_COOKIE			-0x25
#define ECHOSTATUS_MIXER_DISABLED		-0x26
#define ECHOSTATUS_NO_SUPER_INTERLEAVE		-0x27
#define ECHOSTATUS_DUCK_NOT_WRAPPED		-0x28
#define ECHOSTATUS_ARCH_NOT_SUPPORTED		-0x29	// the Dsp comm page structure has wrong size or dma_addr_t is not 32 bits long
#define ECHOSTATUS_RATE_NOT_SUPPORTED		-0x2A
#define ECHOSTATUS_LAST				0x2B


#define ECHOGAIN_MUTED		DSP_TO_GENERIC(-128)	// Minimum possible gain
#define ECHOGAIN_MINOUT		DSP_TO_GENERIC(-128)	// Min output gain in dB
#define ECHOGAIN_MAXOUT		DSP_TO_GENERIC(6)	// Max output gain in dB
#define ECHOGAIN_MININP		DSP_TO_GENERIC(-25)	// Min input gain in dB
#define ECHOGAIN_MAXINP		DSP_TO_GENERIC(25)	// Max input gain in dB

#define PIPE_STATE_STOPPED	0	// Pipe has been reset
#define PIPE_STATE_PAUSED	1	// Pipe has been stopped
#define PIPE_STATE_STARTED	2	// Pipe has been started
#define PIPE_STATE_PENDING	3	// Pipe has pending start


typedef struct {
	u16 DataInterleave;	/* How the data is arranged in memory Mono = 1, stereo = 2, ... */
	u16 BitsPerSample;	/* 8, 16, 32 (24 bits left aligned) */
	u8 MonoToStereo;	/* Only used if DataInterleave is 1 and if this is an output pipe. */
	u8 DataAreBigEndian;	/* 1 = big endian, 0 = littel endian */
} Audioformat_t;


typedef struct {
	u16 FirstChannel;
	u16 Interleave;
	char IsInput;
	char IsCyclic;
	char State;
	sg_entry_t *sglist;
	dma_addr_t sglist_phys;
	int sglHead;
	struct snd_dma_buffer dmabuf;
} AudioPipe_t;


struct EchoGals {
	DspCommPage_t *DspCommPage;		/* Virtual address of the memory seen by DSP */

	u32 AudioOpenMask;			/* Audio channels open mask */
	u32 AudioCyclicMask;			/* Audio use cyclic buffers mask */
	u32 SampleRate;				/* Card sample rate in Hz */
	AudioPipe_t Pipes[ECHO_MAXAUDIOPIPES];

	u16 MidiInOpen;				/* Midi in channels open mask */
	u16 MidiOutOpen;			/* Midi out channels open mask */
	struct pci_dev *pci;
	u8 DigitalMode;				/* Current digital mode (see DIGITAL_MODE_*) */
	u8 SpdifStatus;				/* Gina20, Darla20, Darla24 - only */
	u8 ClockState;				/* Gina20, Darla20, Darla24 - only */
	char ProfessionalSpdif;			/* Gina24, Layla24, Mona - only */
	u8 InputClock;
	u8 OutputClock;
	char MetersOn;

	char NominalLevel[ECHO_MAXAUDIOPIPES];	/* True == -10dBV  False == +4dBu */
	int InLevel[ECHO_MAXAUDIOINPUTS];	/* Input level -25..+25 dB  24.8 fixed point */
	int OutLevel[ECHO_MAXAUDIOOUTPUTS];	/* Output level -128..+6 dB  24.8 fixed point  (-128==muted) */
	int MonitorLevel[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOINPUTS];	/* -25..+25 */
	int VmixerLevel[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOOUTPUTS];	/* -25..+25 */

	u16 NumPipesOut;
	u16 NumPipesIn;
	u16 NumBussesOut;
	u16 NumBussesIn;
	u16 FirstDigitalBusOut;		/* == Number of analog outputs */
	u16 FirstDigitalBusIn;		/* == Number of analog inputs */
	u16 NumMidiOut;			/* # MIDI out channels */
	u16 NumMidiIn;			/* # MIDI in  channels */
	u16 DigitalModes;		/* Bitmask of supported modes (see ECHOCAPS_HAS_DIGITAL_MODE_*) */
	u16 InClockTypes;		/* Suppoted input clock types */
	u16 OutClockTypes;		/* Suppoted output clock types (layla20 only) */

	u16 Flags;			/* See ECHOGALS_FLAG_??? flags defined above */
	u16 DeviceId, SubDeviceId, CardType;
	char HasVmixer;
	char HasASIC;			/* Set TRUE if card has an ASIC */
	char ASICLoaded;		/* Set TRUE when ASIC loaded */
	char BadBoard;			/* Set TRUE if DSP won't load */

	u16 *DspCode;			/* Current DSP code loaded, NULL if nothing loaded */
	u16 *DspCodeToLoad;		/* DSP code to load */
	u32 CommPagePhys;		/* Physical address of the memory seen by DSP */
	volatile u32 *DspRegBase;	/* DSP's register base */
	u32 ActiveMask;			/* Chs. active mask or punks out */
	u8 *Asic;			/* Current ASIC code */

#ifdef MIDI_SUPPORT
	u16 MidiOnCount;	/* Count MIDI enabled cmds */
	u64 MidiInTime;		/* Last time MIDI in occured */
	u64 MidiOutTime;	/* Last time MIDI out occured */
	u64 NextMidiWriteTime;	/* Next time to try MIDI output */
	u16 MtcState;		/* State for MIDI input parsing state machine */
#endif
};



/* Fills the comm page with default values */
static int InitDspCommPage(struct EchoGals *eg);
/* Initialize the hardware, create the CDspCommObject class, return status. */
//int InitHw(struct EchoGals *eg, u16 Device, u16 SubDevice);
/* Stops all running streams and shuts down the DSP. */
//void RestInPeace(struct EchoGals *eg);
/* Init the line levels. This MUST be called from within any derived
   classes as part of InitHw after the DSP is up and running! */
static int InitLineLevels(struct EchoGals *eg);
/* This is called from within an interrupt handler.  It starts interrupt
   handling.  Returns error status if this interrupt is not ours.
   Returns ECHOSTATUS_OK if it is. */
//int ServiceIrq(struct EchoGals *eg, int *MidiReceived);

//int QueryAudioSampleRate(struct EchoGals *eg, u32 SampleRate);
//int SetAudioSampleRate(struct EchoGals *eg, u32 Rate);
//int QueryAudioFormat(struct EchoGals *eg, u16 PipeIndex, const Audioformat_t *AudioFormat);
//int SetAudioFormat(struct EchoGals *eg, u16 PipeIndex, const Audioformat_t *AudioFormat);

//int SetOutputGain(struct EchoGals *eg, u16 Channel, int Gain);
//int SetInputGain(struct EchoGals *eg, u16 Channel, int Gain);
//int SetMonitorGain(struct EchoGals *eg, u16 OutChannel, u16 InChannel, int Gain);
//int SetVmixerGain(struct EchoGals *eg, u16 Output, u16 Pipe, int Gain);
//int SetProfessionalSpdif(struct EchoGals *eg, char Prof);
//static int SetDigitalMode(struct EchoGals *eg, u8 Mode);
//static int SetNominalLevel(struct EchoGals *eg, u16 Index, char Consumer);
//int SetInputClock(struct EchoGals *eg, u16 Clock);
//int SetOutputClock(struct EchoGals *eg, u16 Clock);
//u32 GetInputClockDetect(const struct EchoGals *eg);
//void SetFlags(struct EchoGals *eg, u16 Flags);
//void ClearFlags(struct EchoGals *eg, u16 Flags);
//void SetMetersOn(struct EchoGals *eg, char On);
//void GetAudioMeters(struct EchoGals *eg, long *Meters);

//int AllocateAudio(struct EchoGals *eg, u16 Channel, u16 Interleave, char IsInput);
static int FreeAudio(struct EchoGals *eg, u16 PipeIndex);
//int StartAudio(struct EchoGals *eg, u16 PipeIndex);
//int StartAudioMask(struct EchoGals *eg, u32 Mask);
//int PauseAudio(struct EchoGals *eg, u16 PipeIndex);
//int PauseAudioMask(struct EchoGals *eg, u32 Mask);
static int StopAudio(struct EchoGals *eg, u16 PipeIndex);
//int StopAudioMask(struct EchoGals *eg, u32 Mask);

static int AllocateSGList(struct EchoGals *eg, u16 PipeIndex);
static void FreeSGList(struct EchoGals *eg, u16 PipeIndex);
/*static int SGListAddMapping(struct EchoGals *eg, u16 PipeIndex, dma_addr_t Address, size_t Length);
static int SGListAddIrq(struct EchoGals *eg, u16 PipeIndex);
static int SGListWrap(struct EchoGals *eg, u16 PipeIndex);*/

//volatile u32 *GetAudioPositionPtr(struct EchoGals *eg, u16 PipeIndex);

/* Dsp functions */

//int DspPowerDown(struct EchoGals *eg);
static int LoadFirmware(struct EchoGals *eg);
//int LoadASIC(struct EchoGals *eg);
//static int LoadASICgeneric(struct EchoGals *eg, u32 Cmd, u8 *Code, u32 Size);
//static int CheckAsicStatus(struct EchoGals *eg);

static int WaitForHandshake(struct EchoGals *eg);
static int SendVector(struct EchoGals *eg, u32 Command);
//int WriteControlReg(struct EchoGals *eg, u32 ControlReg);

//int DspSetSampleRate(struct EchoGals *eg, u32 NewSampleRate);
//int DspSetDigitalMode(struct EchoGals *eg, u8 Mode);
//static void DspSetFlags(struct EchoGals *eg, u32 Flags);
//static void DspClearFlags(struct EchoGals *eg, u32 Flags);
//int SetDigitalInputAutoMute(struct EchoGals *eg, int AutoMute);
//int UpdateOutputLineLevel(struct EchoGals *eg);
//int UpdateInputLineLevel(struct EchoGals *eg);
//int UpdateVmixerLevel(struct EchoGals *eg);
//int UpdateFlags(struct EchoGals *eg);

//void DspSetFormat(struct EchoGals *eg, u16 PipeIndex, const Audioformat_t *Format, char NoDither);
//int StartTransport(struct EchoGals *eg, u32 ChannelMask, u32 CyclicMask);
//int PauseTransport(struct EchoGals *eg, u32 ChannelMask);
//int StopTransport(struct EchoGals *eg, u32 ChannelMask);


static inline void ClearHandshake(struct EchoGals *eg)
{
	eg->DspCommPage->Handshake = 0;
}

static inline u32 GetDspRegister(struct EchoGals *eg, u32 Index)
{
	return SWAP32(eg->DspRegBase[Index]);
}

static inline void SetDspRegister(struct EchoGals *eg, u32 Index, u32 Value)
{
	eg->DspRegBase[Index] = SWAP32(Value);
}

static inline int GetSampleRate(const struct EchoGals *eg)
{
	return SWAP32(eg->DspCommPage->SampleRate);
}

#endif // _ECHOAUDIO_H_
