/*
 *  ALSA driver for Echoaudio soundcards.
 *  Copyright (C) 2003 Giuliano Pochini <pochini@shiny.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Enable MIDI support */
/*#define EMIDI*/


/* Debug initialization */
#if (CONFIG_SND_DEBUG)
#define DE_INIT(x) snd_printk x
#else
#define DE_INIT(x)
#endif

/* Debug hw_params callbacks */
#if (CONFIG_SND_DEBUG)
#define DE_HWP(x) snd_printk x
#else
#define DE_HWP(x)
#endif

/* Debug normal activity (open, start, stop...) */
#if (CONFIG_SND_DEBUG)
#define DE_ACT(x) snd_printk x
#else
#define DE_ACT(x)
#endif


/*@@@ Unconditionally disable power management support for now. */
#undef CONFIG_PM

MODULE_AUTHOR("Giuliano Pochini <pochini@shiny.it>");
MODULE_LICENSE("GPL v2");
MODULE_CLASSES("{sound}");
MODULE_DESCRIPTION("Echoaudio soundcards driver");
MODULE_DEVICES("{{Echoaudio series}}");

struct EchoChip {
	spinlock_t lock;
	struct EchoGals eg;		/* EchoGals structure */
	snd_pcm_substream_t *substream[DSP_MAXPIPES];
	int LastPeriod[DSP_MAXPIPES];
	u16 DigitalModes, DigitalModeList[6];
	u16 ClockSources, ClockSourceList[10];
	snd_pcm_t *AnalogPcm, *DigitalPcm;
	snd_card_t *card;
	struct pci_dev *pci;
	unsigned long iophys, iovirt;
	struct resource *iores;
	struct snd_dma_device commpage_dma_dev;
	struct snd_dma_buffer commpage_dma_buf;
	int irq;
	snd_rawmidi_t *rmidi;
	snd_rawmidi_substream_t *MidiIn, *MidiOut;
	u8 Midibuf[ECHO_MIDI_QUEUE_SZ];
	spinlock_t Midilock;
	struct timer_list timer;
	char tinuse;			/* Timer in use */
};

typedef struct EchoChip chip_t;		/* required by many ALSA macros */

struct subsdata {
	volatile u32 *CounterRegister;	/* Hw register that contains the current dma position (lower 32 bits only) */
	u32 LastCounter;		/* The last position, which is used to compute... */
	u32 Position;			/* ...the number of bytes tranferred by the DMA engine, modulo the size of the buffer */
	short PipeIndex;		/* Index of the first channel */
	char Configured;		/* TRUE if this substream is ready to go */
	snd_pcm_hardware_t hw;
	snd_pcm_hw_constraint_list_t constr;
};

static unsigned int ChannelsList[10] = {1, 2, 4, 6, 8, 10, 12, 14, 16, 999999};


/* will go in /sndmagic.h */
#define chip_t_magic 0xEC0A0D10

char *cardname[] = {
	[1] = "Darla20",
	[2] = "Gina20",
	[3] = "Layla20",
	[4] = "Darla24",
	[5] = "Gina24",
	[6] = "Layla24",
	[7] = "Mona",
	[8] = "Mia",
	[9] = "Indigo"
};


MODULE_DEVICE_TABLE(pci, snd_echo_ids);

static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;
static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;
static int enable[SNDRV_CARDS] = SNDRV_DEFAULT_ENABLE_PNP;




/***************************************************************************************************
*** PCM interface
***************************************************************************************************/

static void ssdata_free(snd_pcm_runtime_t *runtime)
{
	kfree(runtime->private_data);
}



static int hw_rule_capture_format_by_channels(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule)
{
	snd_interval_t *c = hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
	snd_mask_t *f = hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
	snd_mask_t fmt;

	snd_mask_any(&fmt);

#ifndef STEREO_BIG_ENDIAN32_SUPPORT
	/* >=2 channels cannot be S32_BE */
	if (c->min == 2) {
		fmt.bits[0] &= ~SNDRV_PCM_FMTBIT_S32_BE;
		return snd_mask_refine(f, &fmt);
	}
#endif
	/* >2 channels must be S32_LE */
	if (c->min > 2) {
		fmt.bits[0] &= SNDRV_PCM_FMTBIT_S32_LE;
		return snd_mask_refine(f, &fmt);
	}
	/* Mono is ok with any format */
	return 0;
}



static int hw_rule_capture_channels_by_format(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule)
{
	snd_interval_t *c = hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
	snd_mask_t *f = hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
	snd_interval_t ch;

	snd_interval_any(&ch);

	/* S32_BE is mono (and stereo) only */
	if (f->bits[0] == SNDRV_PCM_FMTBIT_S32_BE) {
		ch.min = 1;
#ifdef STEREO_BIG_ENDIAN32_SUPPORT
		ch.max = 2;
#else
		ch.max = 1;
#endif
		ch.integer = 1;
		return snd_interval_refine(c, &ch);
	}
	/* U8 amd S16_LE can be only mono or stereo */
	if (f->bits[0] == SNDRV_PCM_FMTBIT_U8 || f->bits[0] == SNDRV_PCM_FMTBIT_S16_LE) {
		ch.min = 1;
		ch.max = 2;
		ch.integer = 1;
		return snd_interval_refine(c, &ch);
	}
	/* S32_LE supports any number of channels. */
	return 0;
}



static int hw_rule_playback_format_by_channels(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule)
{
	snd_interval_t *c = hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
	snd_mask_t *f = hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
	snd_mask_t fmt;

	snd_mask_any(&fmt);

	/* >2 channels must be S32_LE */
	if (c->min > 2) {
		fmt.bits[0] &= SNDRV_PCM_FMTBIT_S32_LE;
		return snd_mask_refine(f, &fmt);
	}
	/* 1 channel must be S32_BE or S32_LE */
	if (c->max < 2) {
		fmt.bits[0] &= SNDRV_PCM_FMTBIT_S32_LE | SNDRV_PCM_FMTBIT_S32_BE;
		return snd_mask_refine(f, &fmt);
	}
#ifndef STEREO_BIG_ENDIAN32_SUPPORT
	/* 2 channels cannot be S32_BE */
	if (c->min == 2 && c->max == 2) {
		fmt.bits[0] &= ~SNDRV_PCM_FMTBIT_S32_BE;
		return snd_mask_refine(f, &fmt);
	}
#endif
	return 0;
}



static int hw_rule_playback_channels_by_format(snd_pcm_hw_params_t *params, snd_pcm_hw_rule_t *rule)
{
	snd_interval_t *c = hw_param_interval(params, SNDRV_PCM_HW_PARAM_CHANNELS);
	snd_mask_t *f = hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);
	snd_interval_t ch;

	snd_interval_any(&ch);

	/* S32_BE is mono (and stereo) only */
	if (f->bits[0] == SNDRV_PCM_FMTBIT_S32_BE) {
		ch.min = 1;
#ifdef STEREO_BIG_ENDIAN32_SUPPORT
		ch.max = 2;
#else
		ch.max = 1;
#endif
		ch.integer = 1;
		return snd_interval_refine(c, &ch);
	}
	/* U8 and S16_LE are stereo only */
	if (f->bits[0] == SNDRV_PCM_FMTBIT_U8 || f->bits[0] == SNDRV_PCM_FMTBIT_S16_LE) {
		ch.min = ch.max = 2;
		ch.integer = 1;
		return snd_interval_refine(c, &ch);
	}
	/* S32_LE supports any number of channels. */
	return 0;
}



static int pcm_open(snd_pcm_substream_t *substream, int MaxChannels, int adat)
{
	struct EchoChip *chip;
	snd_pcm_runtime_t *runtime;
	struct subsdata *ssdata;
	int err, i;

	if (MaxChannels <= 0)
		return -EAGAIN;
/*
	if (!CheckAsicStatus(chip->eg)) {
		DE_HWP(("HW error. Trying to reload the firmware...\n"));
		err = LoadFirmware(chip->eg);
		if (err || chip->eg.BadBoard)
			return -EIO;
	}
*/
	chip = snd_pcm_substream_chip(substream);
	runtime = substream->runtime;

	if (!(ssdata = kmalloc(sizeof(struct subsdata), GFP_KERNEL)))
		return -ENOMEM;
	ssdata->Configured = 0;

	/* Set up hw capabilities and contraints */
	memcpy(&ssdata->hw, &pcm_hardware_skel, sizeof(snd_pcm_hardware_t));
	DE_HWP(("MaxChannels=%d\n", MaxChannels));
	ssdata->constr.list = ChannelsList;
	ssdata->constr.mask = 0;
	for (i = 0; ChannelsList[i] <= MaxChannels; i++);
	ssdata->constr.count = i;
	if (ssdata->hw.channels_max > MaxChannels)
		ssdata->hw.channels_max = MaxChannels;
	if (adat) {
		ssdata->hw.rate_max = 48000;
		ssdata->hw.rates &= SNDRV_PCM_RATE_8000_48000;
	}
/*
if (chip->eg.AudioOpenMask)
ssdata->hw.rate_max=ssdata->hw.rate_min=chip->eg.SampleRate;
*/
	runtime->hw = ssdata->hw;
	runtime->private_data = ssdata;
	runtime->private_free = ssdata_free;

	/* Only mono and any even number of channels are allowed */
	if ((err = snd_pcm_hw_constraint_list(runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS, &ssdata->constr)) < 0)
		return err;

	/* The hw accesses memory in chunks 32 frames long and they should be
	32-bytes-aligned. It's not a requirement, but it seems that IRQs are
	generated with a resolution of 32 frames. Thus we need the following */
	if ((err = snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_SIZE, 32)) < 0)
		return err;
	if ((err = snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_SIZE, 32)) < 0)
		return err;

	return 0;
}



static int pcm_analog_in_open(snd_pcm_substream_t *substream)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);
	int err;

	DE_ACT(("pcm_analog_in_open\n"));
	if ((err = pcm_open(substream, chip->eg.FirstDigitalBusIn - substream->number, 0)) < 0)
		return err;
	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS,
					hw_rule_capture_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1)) < 0)
		return err;
	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT,
					hw_rule_capture_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1)) < 0)
		return err;
	return 0;
}



static int pcm_analog_out_open(snd_pcm_substream_t *substream)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);
	int err;

	DE_ACT(("pcm_analog_out_open\n"));
	if ((err = pcm_open(substream, chip->eg.FirstDigitalBusOut - substream->number, 0)) < 0)
		return err;
	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS,
					hw_rule_playback_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1)) < 0)
		return err;
	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT,
					hw_rule_playback_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1)) < 0)
		return err;
	return 0;
}



#ifdef ECHOCARD_HAS_DIGITAL_IO

static int pcm_digital_in_open(snd_pcm_substream_t *substream)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);
	int err, MaxChannels;

	DE_ACT(("pcm_digital_in_open\n"));
	MaxChannels = chip->eg.NumBussesIn - chip->eg.FirstDigitalBusIn - substream->number;
	if (chip->eg.DigitalMode == DIGITAL_MODE_ADAT) {
		err = pcm_open(substream, MaxChannels, 1);
	} else if (chip->eg.DigitalModes & ECHOCAPS_HAS_DIGITAL_MODE_ADAT) {	/* and mode==S/PDIF */
		/* Subtract the 6 channels that S/PDIF doesn't have */
		err = pcm_open(substream, MaxChannels - 6, 0);
	} else {		/* It hasn't ADAT */
		err = pcm_open(substream, MaxChannels, 0);
	}

	if (err < 0)
		return err;

	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS,
					hw_rule_capture_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1)) < 0)
		return err;
	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT,
					hw_rule_capture_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1)) < 0)
		return err;
	return 0;
}



#ifndef ECHOCARD_HAS_VMIXER	/* See the note in snd_echo_new_pcm() */
static int pcm_digital_out_open(snd_pcm_substream_t *substream)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);
	int err, MaxChannels;

	DE_ACT(("pcm_digital_out_open\n"));
	MaxChannels = chip->eg.NumBussesOut - chip->eg.FirstDigitalBusOut - substream->number;
	if (chip->eg.DigitalMode == DIGITAL_MODE_ADAT) {
		err = pcm_open(substream, MaxChannels, 1);
	} else if (chip->eg.DigitalModes & ECHOCAPS_HAS_DIGITAL_MODE_ADAT) {	/* and mode==S/PDIF */
		/* Subtract the 6 channels that S/PDIF doesn't have */
		err = pcm_open(substream, MaxChannels - 6, 0);
	} else {		/* It hasn't ADAT */
		err = pcm_open(substream, MaxChannels, 0);
	}

	if (err < 0)
		return err;

	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_CHANNELS,
					hw_rule_playback_channels_by_format, 0, SNDRV_PCM_HW_PARAM_FORMAT, -1)) < 0)
		return err;
	if ((err = snd_pcm_hw_rule_add(substream->runtime, 0, SNDRV_PCM_HW_PARAM_FORMAT,
					hw_rule_playback_format_by_channels, 0, SNDRV_PCM_HW_PARAM_CHANNELS, -1)) < 0)
		return err;
	return 0;
}
#endif /* !ECHOCARD_HAS_VMIXER */

#endif /* ECHOCARD_HAS_DIGITAL_IO */



static int pcm_close(snd_pcm_substream_t *substream)
{
	/* Nothing to do here. Audio is already off and ssdata will be freed by its callback */
	DE_ACT(("pcm_close\n"));
	return 0;
}



/* Scatter-gather list alloc and set-up */
static int InitEngine(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params, u16 Channel, u16 Interleave, char IsInput)
{
	struct EchoChip *chip;
	int err, per, rest, page, edge, offs, PipeIndex;
	struct snd_sg_buf *sgbuf;
	struct subsdata *ssdata;
	unsigned long flags;

	chip = snd_pcm_substream_chip(substream);
	ssdata = (struct subsdata *) substream->runtime->private_data;
	PipeIndex = ssdata->PipeIndex;

	/* Sets up che hardware. If it's already initialized, reset and redo, bacause the app changed its mind */
	spin_lock_irqsave(&chip->lock, flags);
	if (ssdata->Configured) {
		DE_HWP(("hwp_closeaudio(%d)\n", PipeIndex));
		if (FreeAudio(&chip->eg, PipeIndex))	/* should never happen */
			DE_HWP(("hwp_FreeAudio err\n"));
		chip->substream[PipeIndex] = 0;
		ssdata->Configured = 0;
	}

	ssdata->PipeIndex = PipeIndex = AllocateAudio(&chip->eg, Channel, Interleave, IsInput);
	if (PipeIndex < 0) {
		spin_unlock_irqrestore(&chip->lock, flags);
		snd_printk(KERN_ERR "OpenAudio() err=0x%x\n", PipeIndex);
		return PipeIndex;
	}
	DE_HWP(("OpenAudio()=%d\n", PipeIndex));

	ssdata->CounterRegister = GetAudioPositionPtr(&chip->eg, PipeIndex);

	DE_HWP(("pcm_hw_params (bufsize=%dB periods=%d persize=%dB)\n",
		params_buffer_bytes(hw_params), params_periods(hw_params), params_period_bytes(hw_params)));
	err = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
	if (err < 0) {
		snd_printk(KERN_ERR "malloc_pages err=%d\n", err);
		goto IEcloseandexit;
	}

	chip->substream[PipeIndex] = substream;
	chip->LastPeriod[PipeIndex] = 0;
	ssdata->LastCounter = 0;
	ssdata->Position = 0;
	ssdata->Configured = 1;
	spin_unlock_irqrestore(&chip->lock, flags);

	sgbuf = snd_pcm_substream_sgbuf(substream);

	DE_HWP(("pcm_hw_params table size=%d pages=%d\n", sgbuf->size, sgbuf->pages));
	edge = PAGE_SIZE;
	for (offs = page = per = 0; offs < params_buffer_bytes(hw_params); per++) {
		rest = params_period_bytes(hw_params);
		if (offs + rest > params_buffer_bytes(hw_params))
			rest = params_buffer_bytes(hw_params) - offs;
		while (rest) {
			if (rest <= edge - offs) {
				/* printk("pcm_hw_params Add%d (off=%d pa=%x phy=%x sz=%d rest=%d)i\n",per,offs,sgbuf->table[page].addr,snd_sgbuf_get_addr(sgbuf, offs),rest,rest); */
				SGListAddMapping(&chip->eg, PipeIndex, snd_sgbuf_get_addr(sgbuf, offs), rest);
				SGListAddIrq(&chip->eg, PipeIndex);
				offs += rest;
				rest = 0;
			} else {
				/* printk("pcm_hw_params Add%d (off=%d pa=%x phy=%x sz=%d rest=%d)\n",per,offs,sgbuf->table[page].addr,snd_sgbuf_get_addr(sgbuf, offs),edge-offs,rest); */
				SGListAddMapping(&chip->eg, PipeIndex, snd_sgbuf_get_addr(sgbuf, offs), edge - offs);
				rest -= edge - offs;
				offs = edge;
			}
			if (offs == edge) {
				edge += PAGE_SIZE;
				page++;
			}
		}
	}
	/*printk("pcm_hw_params End (off=%d rest=%d)\n",offs,rest); */

	/* Close the ring buffer */
	SGListWrap(&chip->eg, PipeIndex);
#if 0
	/* dump sg list */
	{
		sg_entry *list = chip->eg.Pipes[PipeIndex].sg_entry;
		for (per = 0; per < 80; per++) {
			printk("sglist[%d]=%x, %d\n", per, le32_to_cpu(list[per].PhysAddr), le32_to_cpu(chip->eg.Pipes[PipeIndex].gs_entry[per].Size));
		}
	}
#endif
	DE_HWP(("pcm_hw_params ok\n"));
	    return 0;

IEcloseandexit:
	FreeAudio(&chip->eg, PipeIndex);
	spin_unlock_irqrestore(&chip->lock, flags);
	return err;
}



static int pcm_analog_in_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
	return InitEngine(substream, hw_params, substream->number, params_channels(hw_params), 1);
}



static int pcm_analog_out_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
	return InitEngine(substream, hw_params, substream->number, params_channels(hw_params), 0);
}



#ifdef ECHOCARD_HAS_DIGITAL_IO

static int pcm_digital_in_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);

	return InitEngine(substream, hw_params, chip->eg.FirstDigitalBusIn + substream->number,
				params_channels(hw_params), 1);
}


#ifndef ECHOCARD_HAS_VMIXER	/* See the note in snd_echo_new_pcm() */
static int pcm_digital_out_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);

	return InitEngine(substream, hw_params, chip->eg.FirstDigitalBusOut + substream->number,
				params_channels(hw_params), 0);
}
#endif /* !ECHOCARD_HAS_VMIXER */

#endif /* ECHOCARD_HAS_DIGITAL_IO */



static int pcm_hw_free(snd_pcm_substream_t *substream)
{
	struct EchoChip *chip;
	struct subsdata *ssdata;
	int err;
	unsigned long flags;

	chip = snd_pcm_substream_chip(substream);
	ssdata = (struct subsdata *) substream->runtime->private_data;

	spin_lock_irqsave(&chip->lock, flags);
	if (ssdata->Configured) {
		if ((err = FreeAudio(&chip->eg, ssdata->PipeIndex)) < 0) {
			spin_unlock_irqrestore(&chip->lock, flags);
			DE_HWP(("no free err=0x%x\n", (int) err));
			return err;
		}
		DE_HWP(("pcm_hw_free closeaudio(%d)\n", ssdata->PipeIndex));
		chip->substream[ssdata->PipeIndex] = 0;
		ssdata->Configured = 0;
	}
	spin_unlock_irqrestore(&chip->lock, flags);

	DE_HWP(("pcm_hw_freed\n"));
	snd_pcm_lib_free_pages(substream);
	return 0;
}



static int pcm_prepare(snd_pcm_substream_t *substream)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;
	Audioformat_t format;
	int err;

	DE_HWP(("Prepare rate=%d format=%d channels=%d\n", runtime->rate, runtime->format, runtime->channels));
	SetAudioSampleRate(&chip->eg, runtime->rate);
	format.DataInterleave = runtime->channels;
	format.DataAreBigEndian = 0;
	format.MonoToStereo = 0;
	switch (runtime->format) {
	case SNDRV_PCM_FORMAT_U8:
		format.BitsPerSample = 8;
		break;
	case SNDRV_PCM_FORMAT_S16_LE:
		format.BitsPerSample = 16;
		break;
	case SNDRV_PCM_FORMAT_S32_BE:
		format.DataAreBigEndian = 1;
	case SNDRV_PCM_FORMAT_S32_LE:
		format.BitsPerSample = 32;
		break;
	default:
		DE_HWP(("Prepare error: unsupported format %d\n", runtime->format));
		return -EINVAL;
	}
	if ((err = SetAudioFormat(&chip->eg, ((struct subsdata *) runtime->private_data)->PipeIndex, &format)))
		DE_HWP(("Prepare: setformat error 0x%x\n", (int) err));

	return err;
}



static int pcm_trigger(snd_pcm_substream_t *substream, int cmd)
{
	struct EchoChip *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;
	struct subsdata *ssdata = runtime->private_data;
	int err;

	switch (cmd) {
#ifdef CONFIG_PM
	case SNDRV_PCM_TRIGGER_SUSPEND:
		DE_ACT(("pcm_trigger suspend: %d\n", chip->card->power_state));
		if (chip->card->power_state != SNDRV_CTL_POWER_D0) {
			PowerDown(chip->eg);
			snd_power_change_state(chip->card, SNDRV_CTL_POWER_D2);	/* @@ D2 ? */
		}
		break;
	case SNDRV_PCM_TRIGGER_RESUME:
		DE_ACT(("pcm_trigger resume\n"));
		if (chip->card->power_state != SNDRV_CTL_POWER_D0) {
			PowerUp(chip->eg);
			snd_power_change_state(chip->card, SNDRV_CTL_POWER_D0);
		}
		/* no break */
#endif
	case SNDRV_PCM_TRIGGER_START:
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		DE_ACT(("pcm_trigger start\n"));
		err = StartAudio(&chip->eg, ssdata->PipeIndex);
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		DE_ACT(("pcm_trigger stop\n"));
		err = StopAudio(&chip->eg, ssdata->PipeIndex);
		chip->LastPeriod[ssdata->PipeIndex] = 0;
		ssdata->LastCounter = 0;
		ssdata->Position = 0;
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		DE_ACT(("pcm_trigger pause\n"));
		err = PauseAudio(&chip->eg, ssdata->PipeIndex);
		break;
	default:
		return -EINVAL;
	}
	return err;
}



static snd_pcm_uframes_t pcm_pointer(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	struct subsdata *ssdata = runtime->private_data;
	size_t Cnt, bufsize, pos;

	Cnt = le32_to_cpu(*ssdata->CounterRegister);
	ssdata->Position += Cnt - ssdata->LastCounter;
	ssdata->LastCounter = Cnt;
	bufsize = substream->runtime->buffer_size;
	pos = bytes_to_frames(substream->runtime, ssdata->Position);

	while (pos >= bufsize) {
		ssdata->Position -= frames_to_bytes(substream->runtime, bufsize);
		pos -= bufsize;
	}
/*	pos %= bufsize;
	ssdata->Position %= frames_to_bytes(substream->runtime, bufsize);*/
/*printk("pcm_hw_pointer=%d fr=%d sz=%d dmc=%d\n",Cnt,pos,bufsize,ssdata->Position); */
	return pos;
}



/* pcm *_ops structures */
static snd_pcm_ops_t AnPlayback_ops = {
	.open = pcm_analog_out_open,
	.close = pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = pcm_analog_out_hw_params,
	.hw_free = pcm_hw_free,
	.prepare = pcm_prepare,
	.trigger = pcm_trigger,
	.pointer = pcm_pointer,
	.page = snd_pcm_sgbuf_ops_page,
};
static snd_pcm_ops_t AnCapture_ops = {
	.open = pcm_analog_in_open,
	.close = pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = pcm_analog_in_hw_params,
	.hw_free = pcm_hw_free,
	.prepare = pcm_prepare,
	.trigger = pcm_trigger,
	.pointer = pcm_pointer,
	.page = snd_pcm_sgbuf_ops_page,
};
#ifdef ECHOCARD_HAS_DIGITAL_IO
#ifndef ECHOCARD_HAS_VMIXER
static snd_pcm_ops_t DiPlayback_ops = {
	.open = pcm_digital_out_open,
	.close = pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = pcm_digital_out_hw_params,
	.hw_free = pcm_hw_free,
	.prepare = pcm_prepare,
	.trigger = pcm_trigger,
	.pointer = pcm_pointer,
	.page = snd_pcm_sgbuf_ops_page,
};
#endif /* !ECHOCARD_HAS_VMIXER */
static snd_pcm_ops_t DiCapture_ops = {
	.open = pcm_digital_in_open,
	.close = pcm_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = pcm_digital_in_hw_params,
	.hw_free = pcm_hw_free,
	.prepare = pcm_prepare,
	.trigger = pcm_trigger,
	.pointer = pcm_pointer,
	.page = snd_pcm_sgbuf_ops_page,
};
#endif /* ECHOCARD_HAS_DIGITAL_IO */



/*<--snd_echo_probe */
static int __devinit snd_echo_new_pcm(struct EchoChip *chip)
{
	snd_pcm_t *pcm;
	int err;

#ifdef ECHOCARD_HAS_VMIXER
	/* This card has a Vmixer, that is there is no direct mapping from
	   PCM streams to physical outputs. The user is allowed to mix the
	   streams as he wishes via control interface and it's possible to
	   send any stream to any output, so it makes no sense to keep
	   analog and digital outputs separated */

	/* PCM#0 Virtual outputs and analog inputs */
	if ((err = snd_pcm_new(chip->card, "Analog PCM", 0, chip->eg.NumPipesOut,
				chip->eg.FirstDigitalBusIn, &pcm)) < 0)
		return err;
	pcm->private_data = chip;
	chip->AnalogPcm = pcm;
	strcpy(pcm->name, chip->card->shortname);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &AnPlayback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &AnCapture_ops);
	if ((err = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
							snd_dma_pci_data(chip->pci), 64*1024, 256*1024)) < 0)
		return err;
/*	if ((err=snd_pcm_lib_preallocate_pages(pcm->streams[SNDRV_PCM_STREAM_CAPTURE].substream,
						SNDRV_DMA_TYPE_DEV_SG, snd_dma_pci_data(chip->pci),
						128*1024, 512*1024)) < 0)
		return err;*/

	DE_INIT(("Analog PCM ok\n"));

#ifdef ECHOCARD_HAS_DIGITAL_IO
	/* PCM#1 Digital inputs, no outputs */
	if ((err = snd_pcm_new(chip->card, "Digital PCM", 1, 0,
				chip->eg.NumBussesIn - chip->eg.FirstDigitalBusIn, &pcm)) < 0)
		return err;
	pcm->private_data = chip;
	chip->DigitalPcm = pcm;
	strcpy(pcm->name, chip->card->shortname);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &DiCapture_ops);
	if ((err = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
							snd_dma_pci_data(chip->pci), 64*1024, 256*1024)) < 0)
	DE_INIT(("Digital PCM ok\n"));
#endif /* ECHOCARD_HAS_DIGITAL_IO */

#else /* ECHOCARD_HAS_VMIXER */

	/* The card can manage substreams composed by analog and digital channels at the
	   same time, but I prefer to keep analog and digital channels separated, because
	   that mixed thing is confusing and useless. So we register two PCM devices: */

	/* PCM#0 Analog i/o */
	if ((err = snd_pcm_new(chip->card, "Analog PCM", 0,
				 chip->eg.FirstDigitalBusOut, chip->eg.FirstDigitalBusIn, &pcm)) < 0)
		return err;
	pcm->private_data = chip;
	chip->AnalogPcm = pcm;
	strcpy(pcm->name, chip->card->shortname);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &AnPlayback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &AnCapture_ops);
	if ((err = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
							snd_dma_pci_data(chip->pci), 64*1024, 256*1024)) < 0)
		return err;
	DE_INIT(("Analog PCM ok\n"));

#ifdef ECHOCARD_HAS_DIGITAL_IO
	/* PCM#1 Digital i/o */
	if ((err = snd_pcm_new(chip->card, "Digital PCM", 1,
				chip->eg.NumBussesOut - chip->eg.FirstDigitalBusOut,
				chip->eg.NumBussesIn - chip->eg.FirstDigitalBusIn, &pcm)) < 0)
		return err;
	pcm->private_data = chip;
	chip->DigitalPcm = pcm;
	strcpy(pcm->name, chip->card->shortname);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &DiPlayback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &DiCapture_ops);
	if ((err = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
							snd_dma_pci_data(chip->pci), 64*1024, 256*1024)) < 0)
		return err;
	DE_INIT(("Digital PCM ok\n"));
#endif /* ECHOCARD_HAS_DIGITAL_IO */

#endif /* ECHOCARD_HAS_VMIXER */

	return 0;
}




/***************************************************************************************************
*** Control interface
***************************************************************************************************/

/******************* PCM output volume *******************/

static int pcmout_volume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = chip->eg.NumBussesOut;
	uinfo->value.integer.min = ECHOGAIN_MINOUT;
	uinfo->value.integer.max = ECHOGAIN_MAXOUT;
	return 0;
}

static int pcmout_getvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c;

	chip = snd_kcontrol_chip(kcontrol);
	for (c = chip->eg.NumBussesOut - 1; c >= 0; c--)
		ucontrol->value.integer.value[c] = chip->eg.OutLevel[c];
	return 0;
}

static int pcmout_setvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c, changed;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	for (c = chip->eg.NumBussesOut - 1; c >= 0; c--) {
		if (chip->eg.OutLevel[c] != ucontrol->value.integer.value[c]) {
			SetOutputGain(&chip->eg, c, ucontrol->value.integer.value[c]);
			changed = 1;
		}
	}
	if (changed)
		UpdateOutputLineLevel(&chip->eg);
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

#ifdef ECHOCARD_HAS_VMIXER
/* On Vmixer cards this one controls the line-out volume */
static snd_kcontrol_new_t line_volume_control __devinitdata = {
	.name = "Line Playback Volume",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.info = pcmout_volume_info,
	.get = pcmout_getvolume,
	.put = pcmout_setvolume,
};
#else
static snd_kcontrol_new_t pcmout_volume_control __devinitdata = {
	.name = "PCM Playback Volume",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.info = pcmout_volume_info,
	.get = pcmout_getvolume,
	.put = pcmout_setvolume,
};
#endif



/******************* Analog output nominal level (+4dBu / -10dBV) *******************/
#ifdef ECHOCARD_HAS_OUTPUT_NOMINAL_LEVEL
static int pcmout_nominal_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = chip->eg.FirstDigitalBusOut;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int pcmout_getnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c;

	chip = snd_kcontrol_chip(kcontrol);
	for (c = chip->eg.FirstDigitalBusOut - 1; c >= 0; c--)
		ucontrol->value.integer.value[c] = chip->eg.NominalLevel[c];
	return 0;
}

static int pcmout_setnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c, changed;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	for (c = chip->eg.FirstDigitalBusOut - 1; c >= 0; c--) {
		if (chip->eg.NominalLevel[c] != ucontrol->value.integer.value[c]) {
			SetNominalLevel(&chip->eg, c, ucontrol->value.integer.value[c]);
			changed = 1;
		}
	}
	if (changed)
		UpdateOutputLineLevel(&chip->eg);
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

static snd_kcontrol_new_t pcmout_nominal_control __devinitdata = {
	.name = "Line Playback Switch (-10dBV)",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.info = pcmout_nominal_info,
	.get = pcmout_getnominal,
	.put = pcmout_setnominal,
};
#endif /* ECHOCARD_HAS_OUTPUT_NOMINAL_LEVEL */



/******************* Analog input volume *******************/
#ifdef ECHOCARD_HAS_INPUT_GAIN
static int pcmin_volume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = chip->eg.FirstDigitalBusIn;
	uinfo->value.integer.min = ECHOGAIN_MININP;
	uinfo->value.integer.max = ECHOGAIN_MAXINP;
	return 0;
}

static int pcmin_getvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c;

	chip = snd_kcontrol_chip(kcontrol);
	for (c = chip->eg.FirstDigitalBusIn - 1; c >= 0; c--)
		ucontrol->value.integer.value[c] = chip->eg.InLevel[c];
	return 0;
}

static int pcmin_setvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c, changed;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	for (c = chip->eg.FirstDigitalBusIn - 1; c >= 0; c--) {
		if (chip->eg.InLevel[c] != ucontrol->value.integer.value[c]) {
			SetInputGain(&chip->eg, c, ucontrol->value.integer.value[c]);
			changed = 1;
		}
	}
	if (changed)
		UpdateInputLineLevel(&chip->eg);
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

static snd_kcontrol_new_t pcmin_volume_control __devinitdata = {
	.name = "Line Capture Volume",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.info = pcmin_volume_info,
	.get = pcmin_getvolume,
	.put = pcmin_setvolume,
};
#endif /* ECHOCARD_HAS_INPUT_GAIN */



/******************* Analog input nominal level (+4dBu / -10dBV) *******************/
#ifdef ECHOCARD_HAS_INPUT_NOMINAL_LEVEL
static int pcmin_nominal_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = chip->eg.FirstDigitalBusIn;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int pcmin_getnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c;

	chip = snd_kcontrol_chip(kcontrol);
	for (c = chip->eg.FirstDigitalBusIn - 1; c >= 0; c--)
		ucontrol->value.integer.value[c] = chip->eg.NominalLevel[c + chip->eg.NumBussesOut];
	return 0;
}

static int pcmin_setnominal(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int c, changed;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	for (c = chip->eg.FirstDigitalBusIn - 1; c >= 0; c--) {
		if (chip->eg.NominalLevel[c + chip->eg.NumBussesOut] != ucontrol->value.integer.value[c]) {
			SetNominalLevel(&chip->eg, c + chip->eg.NumBussesOut, ucontrol->value.integer.value[c]);
			changed = 1;
		}
	}
	if (changed)
		UpdateOutputLineLevel(&chip->eg);	/* "Output" is not a mistake here. */
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

static snd_kcontrol_new_t pcmin_nominal_control __devinitdata = {
	.name = "Line Capture Switch (-10dBV)",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.info = pcmin_nominal_info,
	.get = pcmin_getnominal,
	.put = pcmin_setnominal,
};
#endif /* ECHOCARD_HAS_INPUT_NOMINAL_LEVEL */



/******************* Monitor mixer *******************/

static int pcm_mixervolume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 1;
	uinfo->value.integer.min = ECHOGAIN_MINOUT;
	uinfo->value.integer.max = ECHOGAIN_MAXOUT;
	uinfo->dimen.d[0] = chip->eg.NumBussesOut;
	uinfo->dimen.d[1] = chip->eg.NumBussesIn;
	return 0;
}

static int pcm_mixergetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	ucontrol->value.integer.value[0] = chip->eg.MonitorLevel
					[ucontrol->id.index / chip->eg.NumBussesIn]
					[ucontrol->id.index % chip->eg.NumBussesIn];
	return 0;
}

static int pcm_mixersetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int changed;
	u16 out, in;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	out = ucontrol->id.index / chip->eg.NumBussesIn;
	in = ucontrol->id.index % chip->eg.NumBussesIn;
	spin_lock_irqsave(&chip->lock, flags);
	if (chip->eg.MonitorLevel[out][in] != ucontrol->value.integer.value[0]) {
		SetMonitorGain(&chip->eg, out, in, ucontrol->value.integer.value[0]);
		changed = 1;
	}
	if (changed)
		UpdateOutputLineLevel(&chip->eg);
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

static snd_kcontrol_new_t pcm_mixervolume_control __devinitdata = {
	.name = "Monitor Mixer Volume",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.count = 1,
	.info = pcm_mixervolume_info,
	.get = pcm_mixergetvolume,
	.put = pcm_mixersetvolume,
};



/******************* Vmixer *******************/
#ifdef ECHOCARD_HAS_VMIXER
static int pcm_vmixervolume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 1;
	uinfo->value.integer.min = ECHOGAIN_MINOUT;
	uinfo->value.integer.max = ECHOGAIN_MAXOUT;
	uinfo->dimen.d[0] = chip->eg.NumBussesOut;
	uinfo->dimen.d[1] = chip->eg.NumPipesOut;
	return 0;
}

static int pcm_vmixergetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	ucontrol->value.integer.value[0] = chip->eg.VmixerLevel
					[ucontrol->id.index / chip->eg.NumPipesOut]
					[ucontrol->id.index % chip->eg.NumPipesOut];
	return 0;
}

static int pcm_vmixersetvolume(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int changed;
	u16 vch, out;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	out = ucontrol->id.index / chip->eg.NumPipesOut;
	vch = ucontrol->id.index % chip->eg.NumPipesOut;
	spin_lock_irqsave(&chip->lock, flags);
	if (chip->eg.VmixerLevel[out][vch] != ucontrol->value.integer.value[0]) {
		SetVmixerGain(&chip->eg, out, vch, ucontrol->value.integer.value[0]);
		changed = 1;
	}
	if (changed)
		UpdateVmixerLevel(&chip->eg);
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

static snd_kcontrol_new_t pcm_vmixervolume_control __devinitdata = {
	.name = "VMixer Volume",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.count = 1,
	.info = pcm_vmixervolume_info,
	.get = pcm_vmixergetvolume,
	.put = pcm_vmixersetvolume,
};
#endif /* ECHOCARD_HAS_VMIXER */



/******************* Digital mode switch *******************/
#ifdef ECHOCARD_HAS_DIGITAL_MODE_SWITCH
static int pcm_digitalmode_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	static char *DigitalModeName[4] = {"S/PDIF Coaxial", "S/PDIF Optical", "ADAT Optical", "S/PDIF Cdrom"};
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->value.enumerated.items = chip->DigitalModes;
	uinfo->count = 1;
	if (uinfo->value.enumerated.item >= chip->DigitalModes)
		uinfo->value.enumerated.item = chip->DigitalModes - 1;
	strcpy(uinfo->value.enumerated.name, DigitalModeName[chip->DigitalModeList[uinfo->value.enumerated.item]]);
	return 0;
}

static int pcm_getdigitalmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int i, mode;

	chip = snd_kcontrol_chip(kcontrol);
	mode = chip->eg.DigitalMode;
	for (i = chip->DigitalModes - 1; i >= 0; i--)
		if (mode == chip->DigitalModeList[i]) {
			ucontrol->value.enumerated.item[0] = i;
			break;
		}
	return 0;
}

static int pcm_setdigitalmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int changed, mode, st;
	unsigned long flags;

	changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	if ((mode = chip->DigitalModeList[ucontrol->value.enumerated.item[0]]) != chip->eg.DigitalMode) {
		changed = 1;
		st = SetDigitalMode(&chip->eg, mode);
		if (st) {
			/* Usually it occurs when the user tries to change
			digital mode while audio is open */
			snd_printk(KERN_ERR "setmode val%d err 0x%x\n", mode, st);
			changed = -EAGAIN;
		}
	}
	spin_unlock_irqrestore(&chip->lock, flags);
	return changed;
}

static snd_kcontrol_new_t pcmin_digitalmode_control __devinitdata = {
	.name = "Digital mode Switch",
	.iface = SNDRV_CTL_ELEM_IFACE_CARD,
	.info = pcm_digitalmode_info,
	.get = pcm_getdigitalmode,
	.put = pcm_setdigitalmode,
};
#endif /* ECHOCARD_HAS_DIGITAL_MODE_SWITCH */



/******************* S/PDIF mode switch *******************/
#ifdef ECHOCARD_HAS_DIGITAL_IO
static int pcm_spdifmode_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	static char *Modes[2] = {"Consumer", "Professional"};

	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->value.enumerated.items = 2;
	uinfo->count = 1;
	if (uinfo->value.enumerated.item)
		uinfo->value.enumerated.item = 1;
	strcpy(uinfo->value.enumerated.name, Modes[uinfo->value.enumerated.item]);
	return 0;
}

static int pcm_getspdifmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	ucontrol->value.enumerated.item[0] = chip->eg.ProfessionalSpdif;
	return 0;
}

static int pcm_setspdifmode(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int mode;
	unsigned long flags;

	chip = snd_kcontrol_chip(kcontrol);
	mode = ucontrol->value.enumerated.item[0];
	if (mode != chip->eg.ProfessionalSpdif) {
		spin_lock_irqsave(&chip->lock, flags);
		SetProfessionalSpdif(&chip->eg, mode);
		spin_unlock_irqrestore(&chip->lock, flags);
		return 1;
	}
	return 0;
}

static snd_kcontrol_new_t pcmin_spdifmode_control __devinitdata = {
	.name = "S/PDIF mode Switch",
	.iface = SNDRV_CTL_ELEM_IFACE_CARD,
	.info = pcm_spdifmode_info,
	.get = pcm_getspdifmode,
	.put = pcm_setspdifmode,
};
#endif /* ECHOCARD_HAS_DIGITAL_IO */



/******************* Select input clock source *******************/

static int pcm_inclock_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	static char *Source[7] = {"Internal", "Word", "Super", "S/PDIF", "ADAT", "ESync", "ESync96"};
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->value.enumerated.items = chip->ClockSources;
	uinfo->count = 1;
	if (uinfo->value.enumerated.item > chip->ClockSources)
		uinfo->value.enumerated.item = chip->ClockSources;
	strcpy(uinfo->value.enumerated.name, Source[chip->ClockSourceList[uinfo->value.enumerated.item]]);
	return 0;
}

static int pcm_getinclock(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int i, clock;

	chip = snd_kcontrol_chip(kcontrol);
	clock = chip->eg.InputClock;

	for (i = 0; i < chip->ClockSources; i++)
		if (clock == chip->ClockSourceList[i])
			ucontrol->value.enumerated.item[0] = i;

	return 0;
}

static int pcm_setinclock(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int changed, clock, st;
	unsigned long flags;

	st = changed = 0;
	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	clock = chip->ClockSourceList[ucontrol->value.enumerated.item[0]];
	if (chip->eg.InputClock != clock) {
		changed = 1;
		st = SetInputClock(&chip->eg, clock);
	}
	spin_unlock_irqrestore(&chip->lock, flags);

	if (st) {
		DE_ACT(("seticlk val%d err 0x%x\n", clock, (int) st));
		switch (st) {
		case ECHOSTATUS_CLOCK_NOT_SUPPORTED:	/* Not supported by the card */
			changed = -EINVAL;
		case ECHOSTATUS_CLOCK_NOT_AVAILABLE:	/* This clock mode is not compatible with the current digital mode */
			changed = -EAGAIN;
		default:
			changed = -1;
		}
	}
	return changed;
}

static snd_kcontrol_new_t pcm_inclock_control __devinitdata = {
	.name = "Sample Clock Source",
	.iface = SNDRV_CTL_ELEM_IFACE_PCM,
	.info = pcm_inclock_info,
	.get = pcm_getinclock,
	.put = pcm_setinclock,
};



/******************* 16 bits digital capture dithering switch *******************/
#ifdef ECHOCARD_HAS_DIGITAL_IO
static int ditherswitch_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int getditherswitch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	ucontrol->value.integer.value[0] = !(chip->eg.Flags & ECHOGALS_FLAG_DIGITAL_IN_NODITHER);
	return 0;
}

/* @@ This is a per-substream setting, not a global one */
static int setditherswitch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);

	if ((chip->eg.Flags & ECHOGALS_FLAG_DIGITAL_IN_NODITHER) == ucontrol->value.integer.value[0]) {
		/* When value==1 we *clear* the bit */
		spin_lock(&chip->lock);
		if (ucontrol->value.integer.value[0])
			ClearFlags(&chip->eg, ECHOGALS_FLAG_DIGITAL_IN_NODITHER);
		else
			SetFlags(&chip->eg, ECHOGALS_FLAG_DIGITAL_IN_NODITHER);
		spin_unlock(&chip->lock);
		return 1;
	}
	return 0;
}

static snd_kcontrol_new_t ditherswitch_control __devinitdata = {
	.name = "Dithering Switch",
	.iface = SNDRV_CTL_ELEM_IFACE_CARD,
	.info = ditherswitch_info,
	.get = getditherswitch,
	.put = setditherswitch,
};
#endif /* ECHOCARD_HAS_DIGITAL_IO */



/******************* VU-meters on/off switch *******************/

static int vumeterswitch_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int setvumeterswitch(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	unsigned long flags;

	chip = snd_kcontrol_chip(kcontrol);
	spin_lock_irqsave(&chip->lock, flags);
	SetMetersOn(&chip->eg, ucontrol->value.integer.value[0]);
	spin_unlock_irqrestore(&chip->lock, flags);
	return 1;
}

static snd_kcontrol_new_t vumeterswitch_control __devinitdata = {
	.name = "VU-meters Switch",
	.iface = SNDRV_CTL_ELEM_IFACE_CARD,
	.access = SNDRV_CTL_ELEM_ACCESS_WRITE,
	.info = vumeterswitch_info,
	.put = setvumeterswitch,
};



/******************* Read VU-meters (input, output, analog and digital together) *******************/

static int vumeter_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 96;
	uinfo->value.integer.min = ECHOGAIN_MINOUT;
	uinfo->value.integer.max = 0;
	uinfo->dimen.d[0] = chip->eg.HasVmixer ? 3 : 2;	/* Out, In, Virt */
	uinfo->dimen.d[1] = 16;				/* 16 channels */
	uinfo->dimen.d[2] = 2;				/* 0=VU, 1=peak */
	return 0;
}

static int getvumeters(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	GetAudioMeters(&chip->eg, ucontrol->value.integer.value);
	return 0;
}

static snd_kcontrol_new_t vumeters_control __devinitdata = {
	.name = "VU-meters",
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READ | SNDRV_CTL_ELEM_ACCESS_VOLATILE,
	.info = vumeter_info,
	.get = getvumeters,
};



/******************* Channels info - it exports informations about the number of channels *******************/

static int channelsinfo_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	struct EchoChip *chip;

	chip = snd_kcontrol_chip(kcontrol);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 6;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1 << ECHO_CLOCK_NUMBER;
	return 0;
}

static int getchannelsinfo(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol)
{
	struct EchoChip *chip;
	int detected, clocks, bit, src;

	chip = snd_kcontrol_chip(kcontrol);
	ucontrol->value.integer.value[0] = chip->eg.NumBussesIn;
	ucontrol->value.integer.value[1] = chip->eg.FirstDigitalBusIn;
	ucontrol->value.integer.value[2] = chip->eg.NumBussesOut;
	ucontrol->value.integer.value[3] = chip->eg.FirstDigitalBusOut;
	ucontrol->value.integer.value[4] = chip->eg.NumPipesOut;

	/* Compute the bitmask of the currently valid input clocks */
	detected = GetInputClockDetect(&chip->eg);
	clocks = 0;
	src = chip->ClockSources - 1;
	for (bit = ECHO_CLOCK_NUMBER - 1; bit >= 0; bit--)
		if (detected & (1 << bit))
			for (; src >= 0; src--)
				if (bit == chip->ClockSourceList[src]) {
					clocks |= 1 << src;
					break;
				}
	ucontrol->value.integer.value[5] = clocks;

	return 0;
}

static snd_kcontrol_new_t channelsinfo_control __devinitdata = {
	.name = "Channels info",
	.iface = SNDRV_CTL_ELEM_IFACE_HWDEP,
	.access = SNDRV_CTL_ELEM_ACCESS_READ | SNDRV_CTL_ELEM_ACCESS_VOLATILE,
	.info = channelsinfo_info,
	.get = getchannelsinfo,
};




/***************************************************************************************************
*** IRQ Handler
***************************************************************************************************/

static irqreturn_t snd_echo_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct EchoChip *chip;
	snd_pcm_substream_t *substream;
	int MidiReceived, per, ss, st;

	chip = snd_magic_cast(chip_t, dev_id, return IRQ_NONE);
	spin_lock(&chip->lock);
	st = ServiceIrq(&chip->eg, &MidiReceived);
	if (st) {
		spin_unlock(&chip->lock);
		return IRQ_NONE;
	}
	st=0;
	/* The hardware doesn't tell us which substream caused the irq,
	thus we have to check all running substreams. */
	for (ss = 0; ss < DSP_MAXPIPES; ss++) {
		if ((substream = chip->substream[ss])) {
			per = pcm_pointer(substream) / substream->runtime->period_size;
			if (per != chip->LastPeriod[ss]) {
				chip->LastPeriod[ss] = per;
				spin_unlock(&chip->lock);
				snd_pcm_period_elapsed(chip->substream[ss]);
				st=1;
				spin_lock(&chip->lock);
			}
		}
	}
	spin_unlock(&chip->lock);
	if (!st)
		DE_ACT(("Period not elapsed ?!\n"));
#ifdef EMIDI
	snd_echo_midi_input_read(chip);
#endif
	return IRQ_HANDLED;
}




/***************************************************************************************************
*** Module construction / destruction
***************************************************************************************************/

static int snd_echo_free(struct EchoChip *chip)
{
	DE_INIT(("Stop DSP...\n"));
	if (chip->eg.DspCommPage) {
		RestInPeace(&chip->eg);
		snd_dma_free_pages(&chip->commpage_dma_dev, &chip->commpage_dma_buf);
	}
	DE_INIT(("Stopped.\n"));

	if (chip->iovirt)
		iounmap((void *)chip->iovirt);
	if (chip->iores) {
		release_resource(chip->iores);
		kfree_nocheck(chip->iores);
	}
	DE_INIT(("MMIO freed.\n"));

	/* release the irq */
	if (chip->irq >= 0)
		free_irq(chip->irq, (void *)chip);
	/* release chip data */
	snd_magic_kfree(chip);
	DE_INIT(("Chip freed.\n"));
	return 0;
}



static int snd_echo_dev_free(snd_device_t *device)
{
	struct EchoChip *chip;

	DE_INIT(("Device free()...\n"));
	chip = snd_magic_cast(chip_t, device->device_data, return -ENXIO);
	return snd_echo_free(chip);
}



/* <--snd_echo_probe */
static __devinit int snd_echo_create(snd_card_t *card, struct pci_dev *pci, struct EchoChip **rchip)
{
	struct EchoChip *chip;
	int err;
	size_t sz;
	static snd_device_ops_t ops = {
		.dev_free = snd_echo_dev_free,
	};

	*rchip = NULL;

	pci_write_config_byte(pci, PCI_LATENCY_TIMER, 0xC0);

	if ((err = pci_enable_device(pci)) < 0)
		return err;

	/* allocate a chip-specific data with magic-alloc */
	chip = snd_magic_kcalloc(chip_t, 0, GFP_KERNEL);
	if (chip == NULL)
		return -ENOMEM;

	spin_lock_init(&chip->lock);
	chip->card = card;
	chip->pci = pci;
	chip->irq = -1;

	/* PCI resource allocation */
	chip->iophys = pci_resource_start(pci, 0);
	sz = pci_resource_len(pci, 0);
	chip->iovirt = (unsigned long)ioremap_nocache(chip->iophys, sz);
	if ((chip->iores = request_mem_region(chip->iophys, sz, "Echoaudio")) == NULL) {
		snd_echo_free(chip);
		snd_printk(KERN_ERR "cannot get memory region\n");
		return -EBUSY;
	}

	if (request_irq(pci->irq, snd_echo_interrupt, SA_INTERRUPT | SA_SHIRQ, "Echoaudio", (void *)chip)) {
		snd_echo_free(chip);
		snd_printk(KERN_ERR "cannot grab irq\n");
		return -EBUSY;
	}
	chip->irq = pci->irq;
	DE_INIT(("pci=%x irq=%d subdev=%04x Init hardware...\n", (int)chip->pci, chip->irq, chip->pci->subsystem_device));

	chip->eg.pci = pci;				/* Used by sglist functions to allocate memory in the DMA space */
	chip->eg.DspRegBase = (u32 *)chip->iovirt;	/* Virtual memory addr assigned to the card's registers */

	/* Create the DSP comm page - this is the area of memory read and written by the DSP via bus mastering */
	chip->commpage_dma_dev.type = SNDRV_DMA_TYPE_DEV;
	chip->commpage_dma_dev.dev = snd_dma_pci_data(chip->pci);
	if (snd_dma_alloc_pages(&chip->commpage_dma_dev, COMMPAGE_SIZE, &chip->commpage_dma_buf) < 0) {
		ECHO_DEBUGPRINTF((KERN_ERR "Cannot allocate CommPage\n"));
		return -ENOMEM;
	}
	chip->eg.CommPagePhys = chip->commpage_dma_buf.addr;
	chip->eg.DspCommPage = (DspCommPage_t *)chip->commpage_dma_buf.area;

	err = InitHw(&chip->eg, chip->pci->device, chip->pci->subsystem_device);
	if (err) {
		DE_INIT(("InitHw err=%x\n", err));
		goto InitError;
	}
	DE_INIT(("Card init OK\n"));

	if ((err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, chip, &ops)) < 0) {
		snd_echo_free(chip);
		return err;
	}
	*rchip = chip;
	/* Init done ! */
	return 0;

	/* An error occurred. Clean up and exit. */
InitError:
	DE_INIT(("Init error\n"));
	snd_echo_free(chip);
	return -EIO;
}



/* constructor */
static int __devinit snd_echo_probe(struct pci_dev *pci, const struct pci_device_id *pci_id)
{
	static int dev;
	snd_card_t *card;
	struct EchoChip *chip;
	char *dsp;
	int i, err;

	if (dev >= SNDRV_CARDS)
		return -ENODEV;
	if (!enable[dev]) {
		dev++;
		return -ENOENT;
	}

	DE_INIT(("Echoaudio soundcards driver.\n"));
	card = snd_card_new(index[dev], id[dev], THIS_MODULE, 0);
	if (card == NULL)
		return -ENOMEM;

	if ((err = snd_echo_create(card, pci, &chip)) < 0) {
		snd_card_free(card);
		return err;
	}

	strcpy(card->driver, "Echoaudio");
	strcpy(card->shortname, cardname[pci_id->subdevice >> 4]);
	switch (pci_id->device) {
	case 0x1801:
		dsp = "56301";
		break;
	case 0x3410:
		dsp = "56361";
		break;
	default:
		dsp = " unknown";
	}
	sprintf(card->longname, "%s rev.%d (DSP%s) at 0x%lx irq %i",
		card->shortname, pci_id->subdevice & 0x000f, dsp, chip->iophys, chip->irq);

	/* Sets up pcm structures according to capabilities of the card */
	pcm_mixervolume_control.count = chip->eg.NumBussesIn * chip->eg.NumBussesOut;

	DE_HWP(("BussesIn=%d  AnalogBussesIn=%d BussesOut=%d  AnalogBussesOut=%d\n", chip->eg.NumBussesIn,
		chip->eg.FirstDigitalBusIn, chip->eg.NumBussesOut, chip->eg.FirstDigitalBusOut));

	if ((err = snd_echo_new_pcm(chip)) < 0) {
		snd_printk(KERN_ERR "Echoaudio - new pcm error %d\n", err);
		snd_card_free(card);
		return err;
	}
/*
#ifdef EMIDI
  if ((pci_id->subdevice&0xfff0)==LAYLA24 || (pci_id->subdevice&0xfff0)==LAYLA)
    if ((err=snd_echo_new_midi(card, chip))<0) {
      printk("Echoaudio - new midi error %d\n", err);
      snd_card_free(card);
      returnerr;
    }
#endif
*/
	/*@@ These ones should check all channels */

#ifdef ECHOCARD_HAS_VMIXER
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&line_volume_control, chip))) < 0)
		return err;
	/* @ it's static, we should do a copy */
	pcm_vmixervolume_control.count = chip->eg.NumPipesOut * chip->eg.NumBussesOut;
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcm_vmixervolume_control, chip))) < 0)
		return err;
#else
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcmout_volume_control, chip))) < 0)
		return err;
#endif

#ifdef ECHOCARD_HAS_INPUT_GAIN
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_volume_control, chip))) < 0)
		return err;
#endif

#ifdef ECHOCARD_HAS_INPUT_NOMINAL_LEVEL
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_nominal_control, chip))) < 0)
		return err;
#endif

#ifdef ECHOCARD_HAS_OUTPUT_NOMINAL_LEVEL
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcmout_nominal_control, chip))) < 0)
		return err;
#endif

	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&vumeterswitch_control, chip))) < 0)
		return err;

	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&vumeters_control, chip))) < 0)
		return err;

	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcm_mixervolume_control, chip))) < 0)
		return err;

	/* Default routing of the Vchannels on the Mia */
#ifdef ECHOCARD_HAS_VMIXER
	SetVmixerGain(&chip->eg, 0, 0, 0);
	SetVmixerGain(&chip->eg, 1, 1, 0);
	SetVmixerGain(&chip->eg, 0, 2, 0);
	SetVmixerGain(&chip->eg, 1, 3, 0);
	SetVmixerGain(&chip->eg, 2, 4, 0);
	SetVmixerGain(&chip->eg, 3, 5, 0);
	SetVmixerGain(&chip->eg, 2, 6, 0);
	SetVmixerGain(&chip->eg, 3, 7, 0);
	UpdateVmixerLevel(&chip->eg);
#endif /* ECHOCARD_HAS_VMIXER */

	/* Enable automute if possible */
#ifdef ECHOCARD_HAS_DIGITAL_IN_AUTOMUTE
	SetDigitalInputAutoMute(&chip->eg, 1);
#endif

	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&channelsinfo_control, chip))) < 0)
		return err;

#ifdef ECHOCARD_HAS_DIGITAL_MODE_SWITCH
	/* Creates a list of available digital modes */
	chip->DigitalModes = 0;
	for (i = 0; i < 6; i++)
		if (chip->eg.DigitalModes & (1 << i))
			chip->DigitalModeList[chip->DigitalModes++] = i;

	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_digitalmode_control, chip))) < 0)
		return err;
#endif /* ECHOCARD_HAS_DIGITAL_MODE_SWITCH */

	/* Creates a list of available clock sources */
	chip->ClockSources = 0;
	for (i = 0; i < 10; i++)
		if (chip->eg.InClockTypes & (1 << i))
			chip->ClockSourceList[chip->ClockSources++] = i;

	if (chip->ClockSources > 1)
		if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcm_inclock_control, chip))) < 0)
			return err;

#ifdef ECHOCARD_HAS_DIGITAL_IO
	if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&ditherswitch_control, chip))) < 0)
		return err;

	if (chip->eg.DigitalModes & (ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_RCA |
					ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_OPTICAL |
					ECHOCAPS_HAS_DIGITAL_MODE_SPDIF_CDROM))
		if ((err = snd_ctl_add(chip->card, snd_ctl_new1(&pcmin_spdifmode_control, chip))) < 0)
			return err;
#endif /* ECHOCARD_HAS_DIGITAL_IO */

/*
	for (i = 0; i < DSP_MAXPIPES; i++) {
		chip->substream[i] = 0;
		chip->LastPeriod[i] = 0;
	}
*/
	if ((err = snd_card_register(card)) < 0) {
		snd_card_free(card);
		return err;
	}
	snd_printk(KERN_INFO "Card registered: %s\n", card->longname);

	pci_set_drvdata(pci, chip);
	dev++;
	return 0;
}



static void __devexit snd_echo_remove(struct pci_dev *pci)
{
	struct EchoChip *chip;

	chip = snd_magic_cast(chip_t, pci_get_drvdata(pci), return);
	if (chip)
		snd_card_free(chip->card);
	pci_set_drvdata(pci, NULL);
}



/***************************************************************************************************
*** Everything starts and ends here
***************************************************************************************************/

/* pci_driver definition */
static struct pci_driver driver = {
	.name = "Echoaudio",
	.id_table = snd_echo_ids,
	.probe = snd_echo_probe,
	.remove = __devexit_p(snd_echo_remove),
};



/* initialization of the module */
static int __init alsa_card_echo_init(void)
{
	int err;

	if ((err = pci_module_init(&driver)) < 0) {
#ifdef MODULE
		snd_printk(KERN_WARNING "Echoaudio soundcard not found or device busy\n");
#endif
		return err;
	}
	return 0;
}



/* clean up the module */
static void __exit alsa_card_echo_exit(void)
{
	pci_unregister_driver(&driver);
}



module_init(alsa_card_echo_init);
module_exit(alsa_card_echo_exit);







/* Debug and to-do notes
16/64
aplay -v -B92880 -F1451  ~/track1.wav
16/32
aplay -v -B92880 -F2902  ~/track1.wav
16/16
aplay -v -B92880 -F5805  ~/track1.wav
16/2
aplay -v -B92880 -F46440  ~/track1.wav
24/2
aplay -v -B139320 -F69660  ~/track1.wav
24/4
aplay -v -B139320 -F34830  ~/track1.wav

64/4
aplay -v -B371519 -F92880  ~/track1.wav
128/4
aplay -v -B743038 -F185759  ~/track1.wav
64/2
aplay -v -B113770

aplay -v -B 35708 -F1234  referenz.wav


IRQ resolution == 32 frames ?

48/95 Argh
aplay -B 2160 -F 1080 -v sin.wav
48/96 ok ma salta facilmente
aplay -B 2180 -F 1080 -v sin.wav
16/32 fiii -EIO (irq sbagliati ?)
aplay -B 720 -F 360 -v sin64.wav
8/32 fiii -EIO (irq sbagliati ?)
aplay -B 720 -F 180 -v sin64.wav
8/64 ok
aplay -B 1440 -F 180 -v sin64.wav
16/64 ok
aplay -B 1440 -F 360 -v sin64.wav
32/64 ok
aplay -B 1440 -F 720 -v sin64.wav


clock out layla20
do a copy of pcm_mixervolume_control for correctness, even if there are no problems
da update*() si pu togliere waithandsh
spostare sglist hw_param --> prepare ?
if eg->BadBoard
hdsp_check_for_firmware
snd_pcm_lib_preallocate_free_for_all()
SetDspRegister MB
automute
open() race check on digital mode and adat   vedi anche 1249
open() race on used substreams
SetMetersOn,SetInputClock,SetProfessionalSpdif,*samplerate,SetInputClock,DspSetDigitalMode,DspSet/ClearFlags,,,,,,,,,, return error
vedi mixart_wait_nice_for_register_value
dependency between controls: eg. setdigitalmode makes inputclock to change

Sistemare eg->Flags ECHOGALS_FLAG_*  e writeblemask

see snd_emu10k1_pcm_mixer_notify

change controls to returs <changed> and <1 for errors ?

magic
chip_t
DigitalModeList

find -name \*.[ch] -exec echo 'sed s/struct\ CEchoGals/struct\ EchoGals/ {} >xx ; mv xx {}' \;
*/
