/******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of version 2 of the GNU General Public License as 
  published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

******************************************************************************/

#include "ipw2200.h"
#include "ieee80211.h"

#define DRV_DESCRIPTION	"Intel(R) PRO/Wireless 2200 Network Driver"
#define DRV_COPYRIGHT	"Copyright(c) 2003-2004 Intel Corporation"
#define DRV_VERSION     IPW2200_VERSION

MODULE_LICENSE("GPL");

int debug = 0;
int channel = 0;
u32 ipw_debug_level;

static void ipw_rx(struct ipw_priv *priv);
static int ipw_queue_tx_reclaim(struct ipw_priv *priv, 
				struct clx2_tx_queue *txq, int qindex);
static int ipw_queue_reset(struct ipw_priv *priv);

static int ipw_queue_tx_hcmd(struct ipw_priv *priv, int hcmd, void *buf,
			     int len, int sync);

static void ipw_tx_queue_free(struct ipw_priv *);

static struct ipw_rx_queue *ipw_rx_queue_alloc(struct ipw_priv *);
static void ipw_rx_queue_free(struct ipw_priv *, struct ipw_rx_queue *);
static void ipw_rx_queue_replenish(void *);

static inline void ipw_enable_interrupts(struct ipw_priv *priv);

static int ipw_up(struct ipw_priv *);
static void ipw_down(struct ipw_priv *);
static int ipw_config(struct ipw_priv *);

static inline char *snprint_line(char *buf, size_t count, 
				 u8 *data, u32 len, u32 ofs)
{
	int out, i, j, l;
	char c;
	
	out = snprintf(buf, count, "%08X", ofs);

	for (l = 0, i = 0; i < 2; i++) {
		out += snprintf(buf + out, count - out, " ");
		for (j = 0; j < 8 && l < len; j++, l++) 
			out += snprintf(buf + out, count - out, "%02X ", 
					data[(i * 8 + j)]);
		for (; j < 8; j++)
			out += snprintf(buf + out, count - out, "   ");
	}
	
	out += snprintf(buf + out, count - out, " ");
	for (l = 0, i = 0; i < 2; i++) {
		out += snprintf(buf + out, count - out, " ");
		for (j = 0; j < 8 && l < len; j++, l++) {
			c = data[(i * 8 + j)];
			if (!isascii(c) || !isprint(c))
				c = '.';
			
			out += snprintf(buf + out, count - out, "%c", c);
		}

		for (; j < 8; j++)
			out += snprintf(buf + out, count - out, " ");
	}
	
	return buf;
}

static inline void printk_buf(int level, u8 *data, u32 len)
{
	char line[81];
	u32 ofs = 0;
	if (!(ipw_debug_level & level))
		return;

	while (len) {
		printk(KERN_INFO "%s\n",
		       snprint_line(line, sizeof(line), &data[ofs], 
				    min(len, 16U), ofs));
		ofs += 16;
		len -= min(len, 16U);
	}
}

u32 _ipw_read_reg32(struct ipw_priv *priv, u32 reg);
#define ipw_read_reg32(a, b) _ipw_read_reg32(a, b)

u8 _ipw_read_reg8(struct ipw_priv *ipw, u32 reg);
#define ipw_read_reg8(a, b) _ipw_read_reg8(a, b)

static void _ipw_write_reg8(struct ipw_priv *priv, u32 reg, u8 value);
static inline void ipw_write_reg8(struct ipw_priv *a, u32 b, u8 c)
{
	IPW_DEBUG_IO("%s %d: write_indirect8(0x%08X, 0x%08X)\n", __FILE__, __LINE__, (u32)(b), (u32)(c)); 
	_ipw_write_reg8(a, b, c);
}

static void _ipw_write_reg16(struct ipw_priv *priv, u32 reg, u16 value);
static inline void ipw_write_reg16(struct ipw_priv *a, u32 b, u16 c)
{
	IPW_DEBUG_IO("%s %d: write_indirect16(0x%08X, 0x%08X)\n", __FILE__, __LINE__, (u32)(b), (u32)(c)); 
	_ipw_write_reg16(a, b, c);
}

static void _ipw_write_reg32(struct ipw_priv *priv, u32 reg, u32 value);
static inline void ipw_write_reg32(struct ipw_priv *a, u32 b, u32 c)
{
	IPW_DEBUG_IO("%s %d: write_indirect32(0x%08X, 0x%08X)\n", __FILE__, __LINE__, (u32)(b), (u32)(c)); 	
	_ipw_write_reg32(a, b, c);
}

#define _ipw_write8(ipw, ofs, val) writeb((val), (ipw)->hw_base + (ofs))
#define ipw_write8(ipw, ofs, val) \
 IPW_DEBUG_IO("%s %d: write_direct8(0x%08X, 0x%08X)\n", __FILE__, __LINE__, (u32)(ofs), (u32)(val)); \
 _ipw_write8(ipw, ofs, val)

#define _ipw_write16(ipw, ofs, val) writew((val), (ipw)->hw_base + (ofs))
#define ipw_write16(ipw, ofs, val) \
 IPW_DEBUG_IO("%s %d: write_direct16(0x%08X, 0x%08X)\n", __FILE__, __LINE__, (u32)(ofs), (u32)(val)); \
 _ipw_write16(ipw, ofs, val)

#define _ipw_write32(ipw, ofs, val) writel((val), (ipw)->hw_base + (ofs))
#define ipw_write32(ipw, ofs, val) \
 IPW_DEBUG_IO("%s %d: write_direct32(0x%08X, 0x%08X)\n", __FILE__, __LINE__, (u32)(ofs), (u32)(val)); \
 _ipw_write32(ipw, ofs, val)

#define _ipw_read8(ipw, ofs) readb((ipw)->hw_base + (ofs))
static inline u8 __ipw_read8(char *f, u32 l, struct ipw_priv *ipw, u32 ofs) {
	IPW_DEBUG_IO("%s %d: read_direct8(0x%08X)\n", f, l, (u32)(ofs));
	return _ipw_read8(ipw, ofs);
}
#define ipw_read8(ipw, ofs) __ipw_read8(__FILE__, __LINE__, ipw, ofs)

#define _ipw_read16(ipw, ofs) readw((ipw)->hw_base + (ofs))
static inline u16 __ipw_read16(char *f, u32 l, struct ipw_priv *ipw, u32 ofs) {
	IPW_DEBUG_IO("%s %d: read_direct16(0x%08X)\n", f, l, (u32)(ofs));
	return _ipw_read16(ipw, ofs);
}
#define ipw_read16(ipw, ofs) __ipw_read16(__FILE__, __LINE__, ipw, ofs)

#define _ipw_read32(ipw, ofs) readl((ipw)->hw_base + (ofs))
static inline u32 __ipw_read32(char *f, u32 l, struct ipw_priv *ipw, u32 ofs) {
	IPW_DEBUG_IO("%s %d: read_direct32(0x%08X)\n", f, l, (u32)(ofs));
	return _ipw_read32(ipw, ofs);
}
#define ipw_read32(ipw, ofs) __ipw_read32(__FILE__, __LINE__, ipw, ofs)

static void _ipw_read_indirect(struct ipw_priv *, u32, u8 *, int);
#define ipw_read_indirect(a, b, c, d) \
	IPW_DEBUG_IO("%s %d: read_inddirect(0x%08X) %d bytes\n", __FILE__, __LINE__, (u32)(b), d); \
	_ipw_read_indirect(a, b, c, d)

static void _ipw_write_indirect(struct ipw_priv *priv, u32 addr, u8 *data, int num);
#define ipw_write_indirect(a, b, c, d) \
	IPW_DEBUG_IO("%s %d: write_indirect(0x%08X) %d bytes\n", __FILE__, __LINE__, (u32)(b), d); \
        _ipw_write_indirect(a, b, c, d)

/* indirect write s */
static void _ipw_write_reg32(struct ipw_priv *priv, u32 reg,
			     u32 value)
{
	IPW_DEBUG_IO(" %p : reg = 0x%8X : value = 0x%8X\n", 
		     priv, reg, value);
	_ipw_write32(priv, CX2_INDIRECT_ADDR, reg);
	_ipw_write32(priv, CX2_INDIRECT_DATA, value);
}


static void _ipw_write_reg8(struct ipw_priv *priv, u32 reg, u8 value)
{
	IPW_DEBUG_IO(" reg = 0x%8X : value = 0x%8X\n", reg, value);
	_ipw_write32(priv, CX2_INDIRECT_ADDR, reg & CX2_INDIRECT_ADDR_MASK);
	_ipw_write8(priv, CX2_INDIRECT_DATA, value);
	IPW_DEBUG_IO(" reg = 0x%8X : value = 0x%8X\n", 
		     (int)priv->hw_base + CX2_INDIRECT_DATA,
		     value);
}

static void _ipw_write_reg16(struct ipw_priv *priv, u32 reg,
			     u16 value)
{
	IPW_DEBUG_IO(" reg = 0x%8X : value = 0x%8X\n", reg, value);
	_ipw_write32(priv, CX2_INDIRECT_ADDR, reg & CX2_INDIRECT_ADDR_MASK);
	_ipw_write16(priv, CX2_INDIRECT_DATA, value);
}

/* indirect read s */

u8 _ipw_read_reg8(struct ipw_priv *priv, u32 reg)
{
	u32 word;
	_ipw_write32(priv, CX2_INDIRECT_ADDR, reg & CX2_INDIRECT_ADDR_MASK);
	IPW_DEBUG_IO(" reg = 0x%8X : \n", reg);
	word = _ipw_read32(priv, CX2_INDIRECT_DATA);
	return (word >> ((reg & 0x3)*8)) & 0xff;
}

u32 _ipw_read_reg32(struct ipw_priv *priv, u32 reg)
{
	u32 value;

	IPW_DEBUG_IO("%p : reg = 0x%08x\n", priv, reg);

	_ipw_write32(priv, CX2_INDIRECT_ADDR, reg);
	value = _ipw_read32(priv, CX2_INDIRECT_DATA);
	IPW_DEBUG_IO(" reg = 0x%4X : value = 0x%4x \n", reg, value);
	return value;
}

/* iterative/auto-increment 32 bit reads and writes */
static void _ipw_read_indirect(struct ipw_priv *priv, u32 addr, u8 * buf,
			       int num)
{
	u32 aligned_addr = addr & CX2_INDIRECT_ADDR_MASK;
	u32 dif_len = addr - aligned_addr;
	u32 aligned_len;
	u32 i;
	
	IPW_DEBUG_IO("addr = %i, buf = %p, num = %i\n", addr, buf, num);

	/* Read the first nibble byte by byte */
	if (unlikely(dif_len)) {
		/* Start reading at aligned_addr + dif_len */
		_ipw_write32(priv, CX2_INDIRECT_ADDR, aligned_addr);
		for (i = dif_len; i < 4; i++, buf++)
			*buf = _ipw_read8(priv, CX2_INDIRECT_DATA + i);
		num -= dif_len;
		aligned_addr += 4;
	}

	/* Read DWs through autoinc register */
	_ipw_write32(priv, CX2_AUTOINC_ADDR, aligned_addr);
	aligned_len = num & CX2_INDIRECT_ADDR_MASK;
	for (i = 0; i < aligned_len; i += 4, buf += 4, aligned_addr += 4)
		*(u32*)buf = ipw_read32(priv, CX2_AUTOINC_DATA);
	
	/* Copy the last nibble */
	dif_len = num - aligned_len;
	_ipw_write32(priv, CX2_INDIRECT_ADDR, aligned_addr);
	for (i = 0; i < dif_len; i++, buf++)
		*buf = ipw_read8(priv, CX2_INDIRECT_DATA + i);
}

static void _ipw_write_indirect(struct ipw_priv *priv, u32 addr, u8 *buf, int num)
{
	u32 aligned_addr = addr & CX2_INDIRECT_ADDR_MASK;
	u32 dif_len = addr - aligned_addr;
	u32 aligned_len;
	u32 i;
	
	IPW_DEBUG_IO("addr = %i, buf = %p, num = %i\n", addr, buf, num);
	
	/* Write the first nibble byte by byte */
	if (unlikely(dif_len)) {
		/* Start writing at aligned_addr + dif_len */
		_ipw_write32(priv, CX2_INDIRECT_ADDR, aligned_addr);
		for (i = dif_len; i < 4; i++, buf++)
			_ipw_write8(priv, CX2_INDIRECT_DATA + i, *buf);
		num -= dif_len;
		aligned_addr += 4;
	}
	
	/* Write DWs through autoinc register */
	_ipw_write32(priv, CX2_AUTOINC_ADDR, aligned_addr);
	aligned_len = num & CX2_INDIRECT_ADDR_MASK;
	for (i = 0; i < aligned_len; i += 4, buf += 4, aligned_addr += 4)
		_ipw_write32(priv, CX2_AUTOINC_DATA, *(u32*)buf);
	
	/* Copy the last nibble */
	dif_len = num - aligned_len;
	_ipw_write32(priv, CX2_INDIRECT_ADDR, aligned_addr);
	for (i = 0; i < dif_len; i++, buf++)
		_ipw_write8(priv, CX2_INDIRECT_DATA + i, *buf);
}

static inline void ipw_set_bit(struct ipw_priv *priv, u32 reg, u32 mask)
{
	ipw_write32(priv, reg, ipw_read32(priv, reg) | mask);
}

static inline void ipw_clear_bit(struct ipw_priv *priv, u32 reg, u32 mask)
{
	ipw_write32(priv, reg, ipw_read32(priv, reg) & ~mask);
}

static inline void ipw_enable_interrupts(struct ipw_priv *priv)
{
	if (priv->status & STATUS_INT_ENABLED)
		return;
	priv->status |= STATUS_INT_ENABLED;
	ipw_write32(priv, CX2_INTA_MASK_R, CX2_INTA_MASK_ALL);
}

static inline void ipw_disable_interrupts(struct ipw_priv *priv)
{
	if (!(priv->status & STATUS_INT_ENABLED))
		return;
	priv->status &= ~STATUS_INT_ENABLED;
	ipw_write32(priv, CX2_INTA_MASK_R, ~CX2_INTA_MASK_ALL);
}

static char *ipw_error_desc(u32 val)
{
	switch (val) {
	case IPW_FW_ERROR_OK: 
		return "ERROR_OK";
	case IPW_FW_ERROR_FAIL: 
		return "ERROR_FAIL";
	case IPW_FW_ERROR_MEMORY_UNDERFLOW: 
		return "MEMORY_UNDERFLOW";
	case IPW_FW_ERROR_MEMORY_OVERFLOW: 
		return "MEMORY_OVERFLOW";
	case IPW_FW_ERROR_BAD_PARAM: 
		return "ERROR_BAD_PARAM";
	case IPW_FW_ERROR_BAD_CHECKSUM: 
		return "ERROR_BAD_CHECKSUM";
	case IPW_FW_ERROR_NMI_INTERRUPT: 
		return "ERROR_NMI_INTERRUPT";
	case IPW_FW_ERROR_BAD_DATABASE: 
		return "ERROR_BAD_DATABASE";
	case IPW_FW_ERROR_ALLOC_FAIL: 
		return "ERROR_ALLOC_FAIL";
	case IPW_FW_ERROR_DMA_UNDERRUN: 
		return "ERROR_DMA_UNDERRUN";
	case IPW_FW_ERROR_DMA_STATUS: 
		return "ERROR_DMA_STATUS";
	case IPW_FW_ERROR_DINOSTATUS_ERROR: 
		return "ERROR_DINOSTATUS_ERROR";
	case IPW_FW_ERROR_EEPROMSTATUS_ERROR: 
		return "ERROR_EEPROMSTATUS_ERROR";
	case IPW_FW_ERROR_SYSASSERT: 
		return "ERROR_SYSASSERT";
	case IPW_FW_ERROR_FATAL_ERROR: 
		return "ERROR_FATALSTATUS_ERROR";
	default: 
		return "UNKNOWNSTATUS_ERROR";
	}
}

static void ipw_dump_nic_error_log(struct ipw_priv *priv)
{
	u32 desc, time, blink1, blink2, ilink1, ilink2, idata, i, count, base;

	base = ipw_read32(priv, IPWSTATUS_ERROR_LOG);
	count = ipw_read_reg32(priv, base);

	IPW_DEBUG_ERROR("Start IPW Error Log Dump:\n");
	for( i=ERROR_START_OFFSET; i<=count*ERROR_ELEM_SIZE; i+=ERROR_ELEM_SIZE) {
		desc   = ipw_read_reg32(priv, base + i);
		time   = ipw_read_reg32(priv, base + i + 1*sizeof(u32));
		blink1 = ipw_read_reg32(priv, base + i + 2*sizeof(u32));
		blink2 = ipw_read_reg32(priv, base + i + 3*sizeof(u32));
		ilink1 = ipw_read_reg32(priv, base + i + 4*sizeof(u32));
		ilink2 = ipw_read_reg32(priv, base + i + 5*sizeof(u32));
		idata =  ipw_read_reg32(priv, base + i + 6*sizeof(u32));

		IPW_DEBUG_ERROR(
			"%s %i 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x\n", 
			ipw_error_desc(desc), time, blink1, blink2, 
			ilink1, ilink2, idata);
	}
}

static inline struct ipw_priv *ipw_netdev(struct net_device *dev) 
{
	return (struct ipw_priv *)netdev_priv(dev);
}

static int ipw_get_ordinal(struct ipw_priv *priv, u32 ord, void *val,
			   u32 *len)
{
	u32 addr, field_info, field_len, field_count, total_len;

	IPW_DEBUG_ORD("ordinal = %i\n", ord);

	if (!priv || !val || !len) {
		IPW_DEBUG_ORD("Invalid argument\n");
		return -EINVAL;
	}
	
	/* verify device ordinal tables have been initialized */
	if (!priv->table0_addr || !priv->table1_addr || !priv->table2_addr) {
		IPW_DEBUG_ORD("Access ordinals before initialization\n");
		return -EINVAL;
	}

	switch ( IPW_ORD_TABLE_ID_MASK & ord ) {
	case IPW_ORD_TABLE_0_MASK:
		/*
		 * TABLE 0: Direct access to a table of 32 bit values
		 *
		 * This is a very simple table with the data directly 
		 * read from the table
		 */

		/* remove the table id from the ordinal */
		ord &= IPW_ORD_TABLE_VALUE_MASK;

		/* boundary check */
		if (ord - 1 > priv->table0_len) {
			IPW_DEBUG_ORD("ordinal value (%i) longer then "
				      "max (%i)\n", ord, priv->table0_len);
			return -EINVAL;
		}

		/* verify we have enough room to store the value */
		if (*len < sizeof(u32)) {
			IPW_DEBUG_ORD("ordinal buffer length too small, "
				      "need %d\n", sizeof(u32));
			return -EINVAL;
		}

		IPW_DEBUG_ORD("Reading TABLE0[%i] from offset 0x%08x\n",
			      ord, priv->table0_addr + ((ord - 1) << 2));

		*len = sizeof(u32);
		*((u32 *)val) = ipw_read_reg32(priv, priv->table0_addr+((ord-1)<<2));
		break;

	case IPW_ORD_TABLE_1_MASK:
		/*
		 * TABLE 1: Indirect access to a table of 32 bit values
		 * 
		 * This is a fairly large table of u32 values each 
		 * representing starting addr for the data (which is
		 * also a u32)
		 */

		/* remove the table id from the ordinal */
		ord &= IPW_ORD_TABLE_VALUE_MASK;
		
		/* boundary check */
		if (ord > priv->table1_len) {
			IPW_DEBUG_ORD("ordinal value too long\n");
			return -EINVAL;
		}

		/* verify we have enough room to store the value */
		if (*len < sizeof(u32)) {
			IPW_DEBUG_ORD("ordinal buffer length too small, "
				      "need %d\n", sizeof(u32));
			return -EINVAL;
		}

		*((u32 *)val) = ipw_read_reg32(priv, (priv->table1_addr + (ord << 2)));
		*len = sizeof(u32);
		break;

	case IPW_ORD_TABLE_2_MASK:
		/*
		 * TABLE 2: Indirect access to a table of variable sized values
		 *
		 * This table consist of six values, each containing
		 *     - dword containing the starting offset of the data
		 *     - dword containing the lengh in the first 16bits
		 *       and the count in the second 16bits
		 */

		/* remove the table id from the ordinal */
		ord &= IPW_ORD_TABLE_VALUE_MASK;

		/* boundary check */
		if (ord > priv->table2_len) {
			IPW_DEBUG_ORD("ordinal value too long\n");
			return -EINVAL;
		}

		/* get the address of statistic */
		addr = ipw_read_reg32(priv, priv->table2_addr + (ord << 3));
		
		/* get the second DW of statistics ; 
		 * two 16-bit words - first is length, second is count */
		field_info = ipw_read_reg32(priv, priv->table2_addr + (ord << 3) + sizeof(u32));
		
		/* get each entry length */
		field_len = *((u16 *)&field_info);
		
		/* get number of entries */
		field_count = *(((u16 *)&field_info) + 1);
		
		/* abort if not enought memory */
		total_len = field_len * field_count;
		if (total_len > *len) {
			*len = total_len;
			return -EINVAL;
		}
		
		*len = total_len;
		if (!total_len)
			return 0;

		IPW_DEBUG_ORD("addr = 0x%08x, total_len = %i, "
			      "field_info = 0x%08x\n", 
			      addr, total_len, field_info);
		ipw_read_indirect(priv, addr, val, total_len);
		break;

	default:
		IPW_DEBUG_ORD("Invalid ordinal!\n");
		return -EINVAL;

	}

	
	return 0;
}

static void ipw_init_ordinals(struct ipw_priv *priv)
{
	priv->table0_addr = ipw_read32(priv, IPW_ORDINALS_TABLE_0);
	priv->table0_len = ipw_read_reg32(priv, priv->table0_addr);

	IPW_DEBUG_ORD("table 0 offset at 0x%08x, len = %i\n",
		      priv->table0_addr, priv->table0_len);

	priv->table1_addr = ipw_read32(priv, IPW_ORDINALS_TABLE_1);
	priv->table1_len = ipw_read_reg32(priv, priv->table1_addr);

	IPW_DEBUG_ORD("table 1 offset at 0x%08x, len = %i\n",
		      priv->table1_addr, priv->table1_len);

	priv->table2_addr = ipw_read32(priv, IPW_ORDINALS_TABLE_2);
	priv->table2_len = ipw_read_reg32(priv, priv->table2_addr);
	priv->table2_len &= 0x0000ffff; /* use first two bytes */

	IPW_DEBUG_ORD("table 2 offset at 0x%08x, len = %i\n",
		      priv->table2_addr, priv->table2_len);

}

/*
 * The following adds a new attribute to the sysfs representation
 * of this device driver (i.e. a new file in /sys/bus/pci/drivers/ipw/)
 * used for controling the debug level.
 * 
 * See the level definitions in ipw for details.
 */
static ssize_t show_debug_level(struct device_driver *d, char *buf)
{
	return sprintf(buf, "0x%08X\n", ipw_debug_level);
}
static ssize_t store_debug_level(struct device_driver *d, const char *buf, 
				 size_t count)
{
	char *p = (char *)buf;
	u32 val;

	if (p[1] == 'x' || p[1] == 'X' || p[0] == 'x' || p[0] == 'X') {
		p++;
		if (p[0] == 'x' || p[0] == 'X')
			p++;
		val = simple_strtoul(p, &p, 16);
	} else
		val = simple_strtoul(p, &p, 10);
	if (p == buf) 
		printk(KERN_INFO DRV_NAME 
		       ": %s is not in hex or decimal form.\n", buf);
	else
		ipw_debug_level = val;

	return strnlen(buf, count);
}
static DRIVER_ATTR(debug_level, S_IWUSR|S_IRUGO, 
		   show_debug_level,store_debug_level);

static ssize_t show_status(struct device *d, char *buf)
{
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;
	return sprintf(buf, "0x%08x\n", (int)p->status);
}
static DEVICE_ATTR(status, S_IRUGO, show_status,NULL);

static ssize_t show_nic_type(struct device *d, char *buf)
{
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;
	u8 type = p->eeprom[EEPROM_NIC_TYPE];

	switch (type) {
	case EEPROM_NIC_TYPE_STANDARD:
		return sprintf(buf, "STANDARD\n");
	case EEPROM_NIC_TYPE_DELL:
		return sprintf(buf, "DELL\n");
	case EEPROM_NIC_TYPE_FUJITSU:
		return sprintf(buf, "FUJITSU\n");
	case EEPROM_NIC_TYPE_IBM:
		return sprintf(buf, "IBM\n");
	case EEPROM_NIC_TYPE_HP:
		return sprintf(buf, "HP\n");
	}
		
	return sprintf(buf, "UNKNOWN\n");
}
static DEVICE_ATTR(nic_type, S_IRUGO, show_nic_type, NULL);

/*
 * The following adds a new attribute to the sysfs representation
 * of the device (i.e. a new file called "eeprom" in 
 * /sys/bus/pci/drivers/ipw/XXXX:XX:XX.X/) that when read will
 * trigger a dump of priv->eeprom 
 *
 * To use this, simply cat the "eeprom" file, and then "dmesg" to
 * see the contents of the eeprom
 */
static ssize_t show_eeprom(struct device *d, char *buf)
{
	int i;
	u8 eeprom[0x100];

	memcpy(eeprom,((struct ipw_priv*)d->driver_data)->eeprom,0x100);
	printk(KERN_INFO "Dumping EEPROM:\n");
	for ( i=0; i<256; i+=8 ) {
		printk(KERN_INFO "0x%2x 0x%2x 0x%2x 0x%2x "
		       "0x%2x 0x%2x 0x%2x 0x%2x\n",
		       eeprom[i+0], 
		       eeprom[i+1],
		       eeprom[i+2],
		       eeprom[i+3],
		       eeprom[i+4],
		       eeprom[i+5],
		       eeprom[i+6],
		       eeprom[i+7]);
	}
	return 0;
}
static DEVICE_ATTR(eeprom, S_IWUSR|S_IRUGO, show_eeprom, NULL);

static ssize_t show_eeprom_sram(struct device *d, char *buf)
{
	int i;
	u8 eeprom[0x100];
	u32 base = ((struct ipw_priv*)d->driver_data)->hw_base + IPW_EEPROM_DATA;
	
	for( i=0; i<CX2_EEPROM_IMAGE_SIZE; i++ )
		eeprom[i] = readb(base+i);

	printk("Dumping EEPROM from SRAM:\n");
	for ( i=0; i<256; i+=8 ) {
		printk(KERN_INFO "0x%2x 0x%2x 0x%2x 0x%2x "
		       "0x%2x 0x%2x 0x%2x 0x%2x\n",
		       eeprom[i+0], 
		       eeprom[i+1],
		       eeprom[i+2],
		       eeprom[i+3],
		       eeprom[i+4],
		       eeprom[i+5],
		       eeprom[i+6],
		       eeprom[i+7]);
	}	

	return 0;
}
static DEVICE_ATTR(eeprom_sram, S_IWUSR|S_IRUGO, show_eeprom_sram, NULL);

static ssize_t show_eeprom_clear(struct device *d, char *buf)
{
	int i;

	for (i = IPW_EEPROM_DATA; 
	     i < IPW_EEPROM_DATA + CX2_EEPROM_IMAGE_SIZE; 
	     i++)
		ipw_write8((struct ipw_priv *)d->driver_data, i, 0);

	return 0;
}
static DEVICE_ATTR(eeprom_clear, S_IWUSR|S_IRUGO, show_eeprom_clear, NULL);

static ssize_t show_error_log(struct device *d, char *buf)
{
	ipw_dump_nic_error_log((struct ipw_priv*)d->driver_data);
	return 0;
}
static DEVICE_ATTR(error_log, S_IWUSR|S_IRUGO, show_error_log, NULL);

static ssize_t show_fw_version(struct device *d, char *buf)
{
	u32 len;
	struct ipw_priv *p = (struct ipw_priv*)d->driver_data;

	if(ipw_get_ordinal(p, IPW_ORD_STAT_FW_VERSION, buf, &len))
		return 0;

	buf[len] = '\n';
	return len + 1;
}
static DEVICE_ATTR(fw_version, S_IWUSR|S_IRUGO, show_fw_version, NULL);

static ssize_t show_fw_date(struct device *d, char *buf)
{
	u32 len;
	struct ipw_priv *p = (struct ipw_priv*)d->driver_data;

	if(ipw_get_ordinal(p, IPW_ORD_STAT_FW_DATE, buf, &len))
		return 0;
	
	buf[len] = '\n';
	return len + 1;
}
static DEVICE_ATTR(fw_date, S_IWUSR|S_IRUGO, show_fw_date, NULL);

static ssize_t show_ucode_version(struct device *d, char *buf)
{
	u32 len = sizeof(u32), tmp = 0;
	struct ipw_priv *p = (struct ipw_priv*)d->driver_data;

	if(ipw_get_ordinal(p, IPW_ORD_STAT_UCODE_VERSION, &tmp, &len))
		return 0;

	return sprintf(buf, "0x%08x\n", tmp);
}
static DEVICE_ATTR(ucode_version, S_IWUSR|S_IRUGO, show_ucode_version, NULL);

static ssize_t show_rtc(struct device *d, char *buf)
{
	u32 len = sizeof(u32), tmp = 0;
	struct ipw_priv *p = (struct ipw_priv*)d->driver_data;

	if(ipw_get_ordinal(p, IPW_ORD_STAT_RTC, &tmp, &len))
		return 0;

	return sprintf(buf, "0x%08x\n", tmp);
}
static DEVICE_ATTR(rtc, S_IWUSR|S_IRUGO, show_rtc, NULL);

/*
 * Add a device attribute to view/control the delay between eeprom
 * operations.
 */
static ssize_t show_eeprom_delay(struct device *d, char *buf)
{
	int n = ((struct ipw_priv*)d->driver_data)->eeprom_delay;
	return sprintf(buf, "%i\n", n);
}
static ssize_t store_eeprom_delay(struct device *d, const char *buf, 
				  size_t count)
{
	struct ipw_priv *p = (struct ipw_priv*)d->driver_data;
	sscanf(buf, "%i", &p->eeprom_delay);
	return strnlen(buf, count);
}
static DEVICE_ATTR(eeprom_delay, S_IWUSR|S_IRUGO, 
		   show_eeprom_delay,store_eeprom_delay);

static ssize_t show_command_event_reg(struct device *d, char *buf)
{
	u32 reg = 0;
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	reg = ipw_read_reg32(p, CX2_INTERNAL_CMD_EVENT);
	return sprintf(buf, "0x%08x\n", reg);
}
static ssize_t store_command_event_reg(struct device *d, 
				       const char *buf, 
				       size_t count)
{
	u32 reg;
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	sscanf(buf, "%x", &reg);
	ipw_write_reg32(p, CX2_INTERNAL_CMD_EVENT, reg);
	return strnlen(buf, count);
}
static DEVICE_ATTR(command_event_reg, S_IWUSR|S_IRUGO, 
		   show_command_event_reg,store_command_event_reg);

static ssize_t show_mem_gpio_reg(struct device *d, char *buf)
{
	u32 reg = 0;
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	reg = ipw_read_reg32(p, 0x301100);
	return sprintf(buf, "0x%08x\n", reg);
}
static ssize_t store_mem_gpio_reg(struct device *d, 
				       const char *buf, 
				       size_t count)
{
	u32 reg;
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	sscanf(buf, "%x", &reg);
	ipw_write_reg32(p, 0x301100, reg);
	return strnlen(buf, count);
}
static DEVICE_ATTR(mem_gpio_reg, S_IWUSR|S_IRUGO,
		   show_mem_gpio_reg,store_mem_gpio_reg);

static ssize_t show_indirect_word(struct device *d, char *buf)
{
	u32 reg = 0;
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	reg = ipw_read_reg32(p, p->indirect_word);
	return sprintf(buf, "0x%08x\n", reg);
}
static ssize_t store_indirect_word(struct device *d, 
				   const char *buf, 
				   size_t count)
{
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	sscanf(buf, "%x", &p->indirect_word);
	return strnlen(buf, count);
}
static DEVICE_ATTR(indirect_word, S_IWUSR|S_IRUGO, 
		   show_indirect_word,store_indirect_word);

static ssize_t show_indirect_byte(struct device *d, char *buf)
{
	u8 reg = 0;
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	reg = ipw_read_reg8(p, p->indirect_byte);
	return sprintf(buf, "0x%02x\n", reg);
}
static ssize_t store_indirect_byte(struct device *d, 
				   const char *buf, 
				   size_t count)
{
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	sscanf(buf, "%x", &p->indirect_byte);
	return strnlen(buf, count);
}
static DEVICE_ATTR(indirect_byte, S_IWUSR|S_IRUGO, 
		   show_indirect_byte,store_indirect_byte);

static ssize_t show_direct_word(struct device *d, char *buf)
{
	u32 reg = 0;
	struct ipw_priv *priv = (struct ipw_priv *)d->driver_data;

	reg = ipw_read32(priv, priv->direct_word);
	return sprintf(buf, "0x%08x\n", reg);
}
static ssize_t store_direct_word(struct device *d, 
				 const char *buf, 
				 size_t count)
{
	struct ipw_priv *p = (struct ipw_priv *)d->driver_data;

	sscanf(buf, "%x", &p->direct_word);
	return strnlen(buf, count);
}
static DEVICE_ATTR(direct_word, S_IWUSR|S_IRUGO, 
		   show_direct_word,store_direct_word);


static inline int rf_kill_active(struct ipw_priv *priv)
{
	return 0 == (ipw_read32(priv, 0x30) & 0x10000);
}

static ssize_t show_rf_kill(struct device *d, char *buf)
{
	/* 0 - RF kill not enabled
	   1 - HW based RF kill active
	   2 - SW based RF kill active (sysfs) */
	struct ipw_priv *priv = (struct ipw_priv *)d->driver_data;
	int val = (priv->status & STATUS_RF_KILL) ? 1 : 0;
	if (val && !rf_kill_active(priv)) 
		val++;
	return sprintf(buf, "%i\n", val);
}

static ssize_t store_rf_kill(struct device *d, const char *buf, size_t count)
{
	struct ipw_priv *priv = (struct ipw_priv *)d->driver_data;
	int state = (buf[0] == '1') ? 1 : 0;
	
	IPW_DEBUG_RF_KILL("RF kill state request: '%c'\n", buf[0]);

	if (state && !(priv->status & STATUS_RF_KILL)) {
		priv->status |= STATUS_RF_KILL;
		IPW_DEBUG_RF_KILL("Recieved RF Kill request via sysfs\n");
		if (priv->workqueue) { 
			cancel_delayed_work(&priv->request_scan);
		}
		wake_up_interruptible(&priv->wait_command_queue);
		queue_work(priv->workqueue, &priv->down);
	} else if (!state && (priv->status & STATUS_RF_KILL)) {
		if (rf_kill_active(priv)) {
			IPW_DEBUG_RF_KILL("Can not turn radio back on - "
					  "disabled by HW switch\n");
		} else {
			IPW_DEBUG_RF_KILL("Turning radio back on\n");
			priv->status &= ~STATUS_RF_KILL;
			queue_work(priv->workqueue, &priv->up);
		}
	}
	return strnlen(buf, count);
}
static DEVICE_ATTR(rf_kill, S_IWUSR|S_IRUGO, show_rf_kill, store_rf_kill);

static void ipw_irq_tasklet(struct ipw_priv *priv)
{
	u32 inta, inta_mask, handled = 0;
	unsigned long flags;
	int rc = 0;

	spin_lock_irqsave(&priv->lock, flags);

	inta = ipw_read32(priv, CX2_INTA_RW);
	inta_mask = ipw_read32(priv, CX2_INTA_MASK_R);
	inta &= (CX2_INTA_MASK_ALL & inta_mask);

	/* Add any cached INTA values that need to be handled */
	inta |= priv->isr_inta;

	/* handle all the justifications for the interrupt */
	if (inta & CX2_INTA_BIT_RX_TRANSFER) {
		ipw_rx(priv);
		handled |= CX2_INTA_BIT_RX_TRANSFER;
	}

	if (inta & CX2_INTA_BIT_TX_CMD_QUEUE) {
		IPW_DEBUG_HC("Command completed.\n");
		rc = ipw_queue_tx_reclaim( priv, &priv->txq_cmd, -1);
		priv->status &= ~STATUS_HCMD_ACTIVE;
		wake_up_interruptible(&priv->wait_command_queue);
		handled |= CX2_INTA_BIT_TX_CMD_QUEUE;
	}

	if (inta & CX2_INTA_BIT_TX_QUEUE_1) {
		IPW_DEBUG_TX("TX_QUEUE_1\n");
		rc = ipw_queue_tx_reclaim( priv, &priv->txq[0], 0);
		handled |= CX2_INTA_BIT_TX_QUEUE_1;
	}

	if (inta & CX2_INTA_BIT_TX_QUEUE_2) {
		IPW_DEBUG_TX("TX_QUEUE_2\n");
		rc = ipw_queue_tx_reclaim( priv, &priv->txq[1], 1);
		handled |= CX2_INTA_BIT_TX_QUEUE_2;
	}

	if (inta & CX2_INTA_BIT_TX_QUEUE_3) {
		IPW_DEBUG_TX("TX_QUEUE_3\n");
		rc = ipw_queue_tx_reclaim( priv, &priv->txq[2], 2);
		handled |= CX2_INTA_BIT_TX_QUEUE_3;
	}

	if (inta & CX2_INTA_BIT_TX_QUEUE_4) {
		IPW_DEBUG_TX("TX_QUEUE_4\n");
		rc = ipw_queue_tx_reclaim( priv, &priv->txq[3], 3);
		handled |= CX2_INTA_BIT_TX_QUEUE_4;
	}

	if (inta & CX2_INTA_BIT_STATUS_CHANGE) {
		IPW_DEBUG_WARNING("STATUS_CHANGE\n");
		handled |= CX2_INTA_BIT_STATUS_CHANGE;
	}

	if (inta & CX2_INTA_BIT_BEACON_PERIOD_EXPIRED) {
		IPW_DEBUG_WARNING("TX_PERIOD_EXPIRED\n");
		handled |= CX2_INTA_BIT_BEACON_PERIOD_EXPIRED;
	}

	if (inta & CX2_INTA_BIT_SLAVE_MODE_HOST_CMD_DONE) {
		IPW_DEBUG_WARNING("HOST_CMD_DONE\n");
		handled |= CX2_INTA_BIT_SLAVE_MODE_HOST_CMD_DONE;
	}

	if (inta & CX2_INTA_BIT_FW_INITIALIZATION_DONE) {
		IPW_DEBUG_WARNING("FW_INITIALIZATION_DONE\n");
		priv->status |= STATUS_FW_READY;
		handled |= CX2_INTA_BIT_FW_INITIALIZATION_DONE;
	}

	if (inta & CX2_INTA_BIT_FW_CARD_DISABLE_PHY_OFF_DONE) {
		IPW_DEBUG_WARNING("PHY_OFF_DONE\n");
		handled |= CX2_INTA_BIT_FW_CARD_DISABLE_PHY_OFF_DONE;
	}

	if (inta & CX2_INTA_BIT_RF_KILL_DONE) {
		IPW_DEBUG_RF_KILL("RF_KILL_DONE\n");
		priv->status |= STATUS_RF_KILL;
		if (priv->workqueue) { 
			cancel_delayed_work(&priv->request_scan);
		}
		wake_up_interruptible(&priv->wait_command_queue);
		netif_carrier_off(priv->net_dev);
		netif_stop_queue(priv->net_dev);
		queue_delayed_work(priv->workqueue, &priv->rf_kill, 2 * HZ);
		handled |= CX2_INTA_BIT_RF_KILL_DONE;
	}
	
	if (inta & CX2_INTA_BIT_FATAL_ERROR) {
		IPW_DEBUG_ERROR("Fatal error\n");
		ipw_dump_nic_error_log(priv);	
		queue_work(priv->workqueue, &priv->adapter_restart);
		handled |= CX2_INTA_BIT_FATAL_ERROR;
	}

	if (inta & CX2_INTA_BIT_PARITY_ERROR) {
		IPW_DEBUG_ERROR("Parity error\n");
		handled |= CX2_INTA_BIT_PARITY_ERROR;
	}

	if (handled != inta) {
		IPW_DEBUG_ERROR("Unhandled INTA bits 0x%08x\n", 
				inta & ~handled);
	}

	/* enable all interrupts */
	ipw_enable_interrupts(priv);

	spin_unlock_irqrestore(&priv->lock, flags);
}
 
#ifdef CONFIG_IPW2200_DEBUG
#define IPW_CMD(x) case IPW_CMD_ ## x : return #x
static char *get_cmd_string(u8 cmd)
{
	switch (cmd) {
		IPW_CMD(HOST_COMPLETE);
		IPW_CMD(POWER_DOWN); 
		IPW_CMD(SYSTEM_CONFIG); 
		IPW_CMD(MULTICAST_ADDRESS); 
		IPW_CMD(SSID); 
		IPW_CMD(ADAPTER_ADDRESS); 
		IPW_CMD(PORT_TYPE); 
		IPW_CMD(RTS_THRESHOLD); 
		IPW_CMD(FRAG_THRESHOLD); 
		IPW_CMD(POWER_MODE); 
		IPW_CMD(WEP_KEY); 
		IPW_CMD(TGI_TX_KEY); 
		IPW_CMD(SCAN_REQUEST); 
		IPW_CMD(ASSOCIATE); 
		IPW_CMD(SUPPORTED_RATES); 
		IPW_CMD(SCAN_ABORT); 
		IPW_CMD(TX_FLUSH); 
		IPW_CMD(QOS_PARAMETERS); 
		IPW_CMD(DINO_CONFIG); 
		IPW_CMD(RSN_CAPABILITIES); 
		IPW_CMD(RX_KEY); 
		IPW_CMD(CARD_DISABLE); 
		IPW_CMD(SEED_NUMBER); 
		IPW_CMD(TX_POWER); 
		IPW_CMD(COUNTRY_INFO); 
		IPW_CMD(AIRONET_INFO); 
		IPW_CMD(AP_TX_POWER); 
		IPW_CMD(CCKM_INFO); 
		IPW_CMD(CCX_VER_INFO); 
		IPW_CMD(SET_CALIBRATION); 
		IPW_CMD(SENSITIVITY_CALIB); 
		IPW_CMD(RETRY_LIMIT); 
		IPW_CMD(IPW_PRE_POWER_DOWN); 
		IPW_CMD(VAP_BEACON_TEMPLATE); 
		IPW_CMD(VAP_DTIM_PERIOD); 
		IPW_CMD(EXT_SUPPORTED_RATES); 
		IPW_CMD(VAP_LOCAL_TX_PWR_CONSTRAINT); 
		IPW_CMD(VAP_QUIET_INTERVALS); 
		IPW_CMD(VAP_CHANNEL_SWITCH); 
		IPW_CMD(VAP_MANDATORY_CHANNELS); 
		IPW_CMD(VAP_CELL_PWR_LIMIT); 
		IPW_CMD(VAP_CF_PARAM_SET); 
		IPW_CMD(VAP_SET_BEACONING_STATE); 
		IPW_CMD(MEASUREMENT); 
		IPW_CMD(POWER_CAPABILITY); 
		IPW_CMD(SUPPORTED_CHANNELS); 
		IPW_CMD(TPC_REPORT); 
		IPW_CMD(WME_INFO); 
		IPW_CMD(PRODUCTION_COMMAND); 
	default: 
		return "UNKNOWN";
	}
}
#endif /* CONFIG_IPW2200_DEBUG */

#define HOST_COMPLETE_TIMEOUT HZ
static int ipw_send_cmd(struct ipw_priv *priv, struct host_cmd *cmd)
{
	int rc = 0;

	if (priv->status & STATUS_HCMD_ACTIVE) {
		IPW_DEBUG_ERROR("Already sending a command\n");
		return -1;
	}

	priv->status |= STATUS_HCMD_ACTIVE;
	
	IPW_DEBUG_HC("Sending %s command (#%d), %d bytes\n", 
		     get_cmd_string(cmd->cmd), cmd->cmd, cmd->len);
	printk_buf(IPW_DL_HOST_COMMAND, (u8*)cmd->param, cmd->len);

	rc = ipw_queue_tx_hcmd(priv, cmd->cmd, &cmd->param, cmd->len, 0);
	if (rc)
		return rc;

	rc = wait_event_interruptible_timeout(
		priv->wait_command_queue, !(priv->status & STATUS_HCMD_ACTIVE),
		HOST_COMPLETE_TIMEOUT);
	if (rc == 0) {
		IPW_DEBUG_INFO("Command completion failed out after %ds.\n",
			       HOST_COMPLETE_TIMEOUT);
		return -EIO;
	}
	if (priv->status & STATUS_RF_KILL) {
		IPW_DEBUG_INFO("Command aborted due to RF Kill Switch\n");
		return -EIO;
	}

	return 0;
}

static int ipw_send_host_complete(struct ipw_priv *priv)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_HOST_COMPLETE,
		.len = 0
	};

	if (!priv) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send HOST_COMPLETE command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_system_config(struct ipw_priv *priv, 
				  struct ipw_sys_config *config)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_SYSTEM_CONFIG,
		.len = sizeof(*config)
	};

	if (!priv || !config) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,config,sizeof(*config));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send SYSTEM_CONFIG command\n");
		return -1;
	}

	return 0;
}

static int ipw_send_ssid(struct ipw_priv *priv, u8 *ssid, int len)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_SSID,
		.len = min(len, IW_ESSID_MAX_SIZE)
	};

	if (!priv || !ssid) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param, ssid, cmd.len);
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send SSID command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_adapter_address(struct ipw_priv *priv, u8 *mac)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_ADAPTER_ADDRESS,
		.len = ETH_ALEN
	};

	if (!priv || !mac) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	printk(KERN_INFO "%s: Setting MAC to " MAC_FMT "\n",
	       priv->net_dev->name, MAC_ARG(mac));

	memcpy(&cmd.param, mac, ETH_ALEN);

	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send ADAPTER_ADDRESS command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_scan_request(struct ipw_priv *priv,
				 struct ipw_scan_request *request)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_SCAN_REQUEST,
		.len = sizeof(*request)
	};

	if (!priv || !request) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,request,sizeof(*request));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send SCAN_REQUEST command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_scan_abort(struct ipw_priv *priv)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_SCAN_ABORT,
		.len = 0
	};

	if (!priv) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send SCAN_ABORT command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_associate(struct ipw_priv *priv,
			      struct ipw_associate *associate)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_ASSOCIATE,
		.len = sizeof(*associate)
	};

	if (!priv || !associate) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,associate,sizeof(*associate));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send ASSOCIATE command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_supported_rates(struct ipw_priv *priv,
				    struct ipw_supported_rates *rates)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_SUPPORTED_RATES,
		.len = sizeof(*rates)
	};

	if (!priv || !rates) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,rates,sizeof(*rates));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send SUPPORTED_RATES command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_card_disable(struct ipw_priv *priv, u32 phy_off)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_CARD_DISABLE,
		.len = sizeof(u32)
	};

	if (!priv) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,&phy_off,sizeof(u32));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send CARD_DISABLE command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_seed_number(struct ipw_priv *priv, u32 seed)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_SEED_NUMBER,
		.len = sizeof(u32)
	};

	if (!priv) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,&seed,sizeof(u32));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send SEED_NUMBER command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_tx_power(struct ipw_priv *priv,
			     struct ipw_tx_power *power)
{
	struct host_cmd cmd = {
		.cmd = IPW_CMD_TX_POWER,
		.len = sizeof(*power)
	};

	if (!priv || !power) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param,power,sizeof(*power));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send TX_POWER command\n");
		return -1;
	}
	
	return 0;
}

static int ipw_send_rts_threshold(struct ipw_priv *priv, u16 rts)
{
	struct ipw_rts_threshold rts_threshold = {
		.rts_threshold = rts,
	};
	struct host_cmd cmd = {
		.cmd = IPW_CMD_RTS_THRESHOLD,
		.len = sizeof(rts_threshold)
	};

	if (!priv) {
		IPW_DEBUG_ERROR("Invalid args\n");
		return -1;
	}

	memcpy(&cmd.param, &rts_threshold, sizeof(rts_threshold));
	if (ipw_send_cmd(priv, &cmd)) {
		IPW_DEBUG_ERROR("failed to send RTS_THRESHOLD command\n");
		return -1;
	}

	return 0;
}

/*
 * The IPW device contains a Microwire compatible EEPROM that stores
 * various data like the MAC address.  Usually the firmware has exclusive
 * access to the eeprom, but during device initialization (before the
 * device driver has sent the HostComplete command to the firmware) the
 * device driver has read access to the EEPROM by way of indirect addressing
 * through a couple of memory mapped registers.
 *
 * The following is a simplified implementation for pulling data out of the
 * the eeprom, along with some helper functions to find information in
 * the per device private data's copy of the eeprom.
 *
 * NOTE: To better understand how these functions work (i.e what is a chip
 *       select and why do have to keep driving the eeprom clock?), read
 *       just about any data sheet for a Microwire compatible EEPROM.
 */

/* write a 32 bit value into the indirect accessor register */
static inline void eeprom_write_reg(struct ipw_priv *p, u32 data)
{
	ipw_write_reg32(p,FW_MEM_REG_EEPROM_ACCESS,data);
	
	/* the eeprom requires some time to complete the operation */
	udelay(p->eeprom_delay);

	return;
}

/* perform a chip select operation */
static inline void eeprom_cs(struct ipw_priv* priv)
{
	eeprom_write_reg(priv,0);
	eeprom_write_reg(priv,EEPROM_BIT_CS);
	eeprom_write_reg(priv,EEPROM_BIT_CS|EEPROM_BIT_SK);
	eeprom_write_reg(priv,EEPROM_BIT_CS);
}

/* perform a chip select operation */
static inline void eeprom_disable_cs(struct ipw_priv* priv)
{
	eeprom_write_reg(priv,EEPROM_BIT_CS);
	eeprom_write_reg(priv,0);
	eeprom_write_reg(priv,EEPROM_BIT_SK);
}

/* push a single bit down to the eeprom */
static inline void eeprom_write_bit(struct ipw_priv *p,u8 bit)
{
	int d = ( bit ? EEPROM_BIT_DI : 0);
	eeprom_write_reg(p,EEPROM_BIT_CS|d);
	eeprom_write_reg(p,EEPROM_BIT_CS|d|EEPROM_BIT_SK);
}

/* push an opcode followed by an address down to the eeprom */
static void eeprom_op(struct ipw_priv* priv, u8 op, u8 addr)
{
	int i;

	eeprom_cs(priv);
	eeprom_write_bit(priv,1);
	eeprom_write_bit(priv,op&2);
	eeprom_write_bit(priv,op&1);
	for ( i=7; i>=0; i-- ) {
		eeprom_write_bit(priv,addr&(1<<i));
	}
}

/* pull 16 bits off the eeprom, one bit at a time */
static u16 eeprom_read_u16(struct ipw_priv* priv, u8 addr)
{
	int i;
	u16 r=0;
	
	/* Send READ Opcode */
	eeprom_op(priv,EEPROM_CMD_READ,addr);

	/* Send dummy bit */
	eeprom_write_reg(priv,EEPROM_BIT_CS);

	/* Read the byte off the eeprom one bit at a time */
	for ( i=0; i<16; i++ ) {
		u32 data = 0;
		eeprom_write_reg(priv,EEPROM_BIT_CS|EEPROM_BIT_SK);
		eeprom_write_reg(priv,EEPROM_BIT_CS);
		data = ipw_read_reg32(priv,FW_MEM_REG_EEPROM_ACCESS);
		r = (r<<1) | ((data & EEPROM_BIT_DO)?1:0);
	}
	
	/* Send another dummy bit */
	eeprom_write_reg(priv,0);
	eeprom_disable_cs(priv);
	
	return r;
}

/* helper function for pulling the mac address out of the private */
/* data's copy of the eeprom data                                 */
static void eeprom_parse_mac(struct ipw_priv* priv, u8* mac)
{
	u8* ee = (u8*)priv->eeprom;
	memcpy(mac, &ee[EEPROM_MAC_ADDRESS], 6);
}

/*
 * Either the device driver (i.e. the host) or the firmware can
 * load eeprom data into the designated region in SRAM.  If neither
 * happens then the FW will shutdown with a fatal error.
 *
 * In order to signal the FW to load the EEPROM, the EEPROM_LOAD_DISABLE
 * bit needs region of shared SRAM needs to be non-zero.
 */
static void ipw_eeprom_init_sram(struct ipw_priv *priv)
{
	int i;
	u16 *eeprom = (u16 *)priv->eeprom;
  
	IPW_DEBUG_TRACE(">>\n");

	/* read entire contents of eeprom into private buffer */
	for ( i=0; i<128; i++ )
		eeprom[i] = eeprom_read_u16(priv,(u8)i);

	/* 
	   If the data looks correct, then copy it to our private 
	   copy.  Otherwise let the firmware know to perform the operation
	   on it's own
	*/
	if ((priv->eeprom + EEPROM_VERSION) != 0) {
		IPW_DEBUG_INFO("Writing EEPROM data into SRAM\n");

		/* write the eeprom data to sram */
		for( i=0; i<CX2_EEPROM_IMAGE_SIZE; i++ )
			ipw_write8(priv, IPW_EEPROM_DATA + i, 
				   priv->eeprom[i]);

		/* Do not load eeprom data on fatal error or suspend */
		ipw_write32(priv, IPW_EEPROM_LOAD_DISABLE, 0);
	} else {
		IPW_DEBUG_INFO("Enabling FW initializationg of SRAM\n");

		/* Load eeprom data on fatal error or suspend */
		ipw_write32(priv, IPW_EEPROM_LOAD_DISABLE, 1);
	}

	IPW_DEBUG_TRACE("<<\n");
}


static inline void ipw_zero_memory(struct ipw_priv *priv, u32 start, u32 count)
{
	count >>= 2;
	if (!count) return;
	_ipw_write32(priv, CX2_AUTOINC_ADDR, start);
	while (count--) 
		_ipw_write32(priv, CX2_AUTOINC_DATA, 0);
}

static inline void ipw_fw_dma_reset_command_blocks(struct ipw_priv *priv)
{
	ipw_zero_memory(priv, CX2_SHARED_SRAM_DMA_CONTROL,
			CB_NUMBER_OF_ELEMENTS_SMALL * 
			sizeof(struct command_block));
}

static int ipw_fw_dma_enable(struct ipw_priv *priv)
{ /* start dma engine but no transfers yet*/

	IPW_DEBUG_FW(">> : \n");
    
	/* Start the dma */
	ipw_fw_dma_reset_command_blocks(priv);
	
	/* Write CB base address */
	ipw_write_reg32(priv, CX2_DMA_I_CB_BASE, CX2_SHARED_SRAM_DMA_CONTROL);

	IPW_DEBUG_FW("<< : \n");
	return 0;
}

static void ipw_fw_dma_abort(struct ipw_priv *priv)
{
	u32 control = 0;

	IPW_DEBUG_FW(">> :\n");
    
	//set the Stop and Abort bit	
	control = DMA_CONTROL_SMALL_CB_CONST_VALUE | DMA_CB_STOP_AND_ABORT;
	ipw_write_reg32(priv,CX2_DMA_I_DMA_CONTROL,control);
	priv->sram_desc.last_cb_index = 0;
	
	IPW_DEBUG_FW("<< \n");
}

static int ipw_fw_dma_write_command_block(struct ipw_priv *priv, int index, struct command_block *cb)
{
	u32 address = CX2_SHARED_SRAM_DMA_CONTROL + (sizeof(struct command_block) * index); 
	IPW_DEBUG_FW(">> :\n");

	ipw_write_indirect(priv, address, (u8*)cb, sizeof(struct command_block));

	IPW_DEBUG_FW("<< :\n");
	return 0;

}

static int ipw_fw_dma_kick(struct ipw_priv *priv)
{
	u32 control = 0;
	u32 index=0;

	IPW_DEBUG_FW(">> :\n");
    
	for (index = 0; index < priv->sram_desc.last_cb_index; index++)
		ipw_fw_dma_write_command_block(priv, index, &priv->sram_desc.cb_list[index]);

	/* Enable the DMA in the CSR register */
	ipw_clear_bit(priv, CX2_RESET_REG,CX2_RESET_REG_MASTER_DISABLED | CX2_RESET_REG_STOP_MASTER);
	
        /* Set the Start bit. */
	control = DMA_CONTROL_SMALL_CB_CONST_VALUE | DMA_CB_START;
	ipw_write_reg32(priv, CX2_DMA_I_DMA_CONTROL, control);

	IPW_DEBUG_FW("<< :\n");
	return 0;
}

static void ipw_fw_dma_dump_command_block(struct ipw_priv *priv)
{
	u32 address;
	u32 register_value=0;
	u32 cb_fields_address=0;

	IPW_DEBUG_FW(">> :\n");
	address = ipw_read_reg32(priv,CX2_DMA_I_CURRENT_CB);
	IPW_DEBUG_FW_INFO("Current CB is 0x%x \n",address);

	/* Read the DMA Controlor register */
	register_value = ipw_read_reg32(priv, CX2_DMA_I_DMA_CONTROL);
	IPW_DEBUG_FW_INFO("CX2_DMA_I_DMA_CONTROL is 0x%x \n",register_value);

	/* Print the CB values*/
	cb_fields_address = address;
	register_value = ipw_read_reg32(priv, cb_fields_address);
	IPW_DEBUG_FW_INFO("Current CB ControlField is 0x%x \n",register_value);

	cb_fields_address += sizeof(u32);
	register_value = ipw_read_reg32(priv, cb_fields_address);
	IPW_DEBUG_FW_INFO("Current CB Source Field is 0x%x \n",register_value);

	cb_fields_address += sizeof(u32);
	register_value = ipw_read_reg32(priv, cb_fields_address);
	IPW_DEBUG_FW_INFO("Current CB Destination Field is 0x%x \n",register_value);

	cb_fields_address += sizeof(u32);
	register_value = ipw_read_reg32(priv, cb_fields_address);
	IPW_DEBUG_FW_INFO("Current CB Status Field is 0x%x \n",register_value);

	IPW_DEBUG_FW(">> :\n");
}

static int ipw_fw_dma_command_block_index(struct ipw_priv *priv)
{
	u32 current_cb_address = 0;
	u32 current_cb_index = 0;

	IPW_DEBUG_FW("<< :\n");
	current_cb_address= ipw_read_reg32(priv, CX2_DMA_I_CURRENT_CB);
	
	current_cb_index = (current_cb_address - CX2_SHARED_SRAM_DMA_CONTROL ) /
		sizeof (struct command_block);
	
	IPW_DEBUG_FW_INFO("Current CB index 0x%x address = 0x%X \n",current_cb_index, current_cb_address );

	IPW_DEBUG_FW(">> :\n");
	return current_cb_index;

}

static int ipw_fw_dma_add_command_block(struct ipw_priv *priv,
					u32 src_address,
					u32 dest_address,
					u32 length,
					int interrupt_enabled,
					int is_last)
{

	u32 control = CB_VALID | CB_SRC_LE | CB_DEST_LE | CB_SRC_AUTOINC | 
		CB_SRC_IO_GATED | CB_DEST_AUTOINC | CB_SRC_SIZE_LONG | 
		CB_DEST_SIZE_LONG;
	struct command_block *cb;
	u32 last_cb_element=0;

	IPW_DEBUG_FW(">>:\n");

	IPW_DEBUG_FW_INFO(">>: src_address=0x%x dest_address=0x%x length=0x%x\n",
			  src_address, dest_address, length);

	if (priv->sram_desc.last_cb_index >= CB_NUMBER_OF_ELEMENTS_SMALL)
		return -1;

	last_cb_element = priv->sram_desc.last_cb_index;
	cb = &priv->sram_desc.cb_list[last_cb_element];
	priv->sram_desc.last_cb_index++;

	/* Calculate the new CB control word */
	if (interrupt_enabled )
		control |= CB_INT_ENABLED;

	if (is_last)
		control |= CB_LAST_VALID;
	
	control |= length;

	/* Calculate the CB Element's checksum value */
	cb->status = control ^src_address ^dest_address;

	/* Copy the Source and Destination addresses */
	cb->dest_addr = (void *)dest_address;
	cb->source_addr = (void *)src_address;

	/* Copy the Control Word last */
	cb->control = control;

	IPW_DEBUG_FW("<< : \n");
	return 0;
}

static int ipw_fw_dma_add_buffer(struct ipw_priv *priv,
				 u32 src_phys,
				 u32 dest_address,
				 u32 length)
{
	u32 bytes_left = length;
	u32 src_offset=0;
	u32 dest_offset=0;
	int status = 0;
	IPW_DEBUG_FW(">> \n");
	IPW_DEBUG_FW_INFO("src_phys=0x%x dest_address=0x%x length=0x%x\n",
			  src_phys, dest_address, length);
	while (bytes_left > CB_MAX_LENGTH) {
		status = ipw_fw_dma_add_command_block( priv,
						       src_phys + src_offset,
						       dest_address + dest_offset,
						       CB_MAX_LENGTH, 0, 0);
		if (status) {
			IPW_DEBUG_FW_INFO(": Failed\n");
			return -1;
		} else 
			IPW_DEBUG_FW_INFO(": Added new cb\n");

		src_offset += CB_MAX_LENGTH;
		dest_offset += CB_MAX_LENGTH;
		bytes_left -= CB_MAX_LENGTH;
	}

	/* add the buffer tail */
	if (bytes_left > 0) {
		status = ipw_fw_dma_add_command_block( priv,	
						       src_phys + src_offset,
						       dest_address + dest_offset,
						       bytes_left, 0, 0);
		if (status) {
			IPW_DEBUG_FW_INFO(": Failed on the buffer tail\n");
			return -1;
		} else 
			IPW_DEBUG_FW_INFO(": Adding new cb - the buffer tail\n");
	}
	
	
	IPW_DEBUG_FW("<< \n");
	return 0;
}

static int ipw_fw_dma_wait(struct ipw_priv *priv)
{
	u32 current_index = 0;
	u32 watchdog = 0;

	IPW_DEBUG_FW(">> : \n");

	current_index = ipw_fw_dma_command_block_index(priv);
	IPW_DEBUG_FW_INFO("sram_desc.last_cb_index:0x%8X\n", 
			  (int) priv->sram_desc.last_cb_index);
	while (current_index < priv->sram_desc.last_cb_index) {
		udelay(50);
		current_index = ipw_fw_dma_command_block_index(priv);

		watchdog++;
		if (watchdog > 400) {
			IPW_DEBUG_FW_INFO("Timeout\n");
			ipw_fw_dma_dump_command_block(priv);
			ipw_fw_dma_abort(priv);
			return -1;
		}
	}

	ipw_fw_dma_abort(priv);
	/*Disable the DMA in the CSR register*/
 	ipw_set_bit(priv, CX2_RESET_REG, CX2_RESET_REG_MASTER_DISABLED | CX2_RESET_REG_STOP_MASTER);

	IPW_DEBUG_FW("<< dmaWaitSync \n");
	return 0;
}

/**
 * Check that card is still alive. 
 * Reads debug register from domain0.
 * If card is present, pre-defined value should
 * be found there.
 * 
 * @param priv
 * @return 1 if card is present, 0 otherwise
 */
static inline int ipw_alive(struct ipw_priv *priv)
{
	return ipw_read32(priv, 0x90) == 0xd55555d5;
}

static inline int ipw_poll_bit(struct ipw_priv *priv, u32 addr, u32 mask,
			       int timeout)
{
	int i = 0;

	do {
		if ((ipw_read32(priv, addr) & mask) == mask) 
			return i;
		mdelay(10);
		i += 10;
	} while (i < timeout);
	
	return -ETIME;
}

/* These functions load the firmware and micro code for the operation of 
 * the ipw hardware.  It assumes the buffer has all the bits for the
 * image and the caller is handling the memory allocation and clean up.
 */


static int ipw_stop_master(struct ipw_priv * priv)
{
	int rc;
	
	IPW_DEBUG_TRACE(">> \n");
	/* stop master. typical delay - 0 */
	ipw_set_bit(priv, CX2_RESET_REG, CX2_RESET_REG_STOP_MASTER);

	rc = ipw_poll_bit(priv, CX2_RESET_REG,
			  CX2_RESET_REG_MASTER_DISABLED, 100);
	if (rc < 0) {
		IPW_DEBUG_ERROR("stop master failed in 10ms\n");
		return -1;
	}

	IPW_DEBUG_INFO("stop master %dms\n", rc);

	return rc;
}

static void ipw_arc_release(struct ipw_priv *priv)
{
	IPW_DEBUG_TRACE(">> \n");
	mdelay(5);

	ipw_clear_bit(priv, CX2_RESET_REG, CBD_RESET_REG_PRINCETON_RESET);

	/* no one know timing, for safety add some delay */
	mdelay(5);
}

struct fw_header {
	u32 version;
	u32 mode;
};

struct fw_chunk {
	u32 address;
	u32 length;
};


/* TODO: UCODE DOWNLOAD SHOULD BE ON WORK QUEUE TO AVOID THE POLING BS 
 * also, the locking and busy waits in this function give me a rash 
 */
static int ipw_load_ucode(struct ipw_priv *priv, u8 * data,
			  size_t len)
{
	int rc = 0;
	int i;
	u8 cr = 0;
	u16 *image;
	unsigned long flags;

	image = (u16 *)( (u32)data + sizeof(struct fw_header));
	
	IPW_DEBUG_TRACE(">> \n");

	rc = ipw_stop_master(priv);

	if (rc < 0)
		return rc;
	/* zero whole domain 1 */
	{
		int addr;
		for (addr = CX2_SHARED_LOWER_BOUND;
		     addr < CX2_REGISTER_DOMAIN1_END; addr += 4) {
			ipw_write32(priv, addr, 0);
		}
	} 

	/* no ucode (yet) */
	memset(&priv->dino_alive, 0, sizeof(priv->dino_alive));
	/* destroy DMA queues */
	/* reset sequence */
	spin_lock_irqsave(&priv->lock, flags);

	ipw_write_reg32(priv, CX2_MEM_HALT_AND_RESET ,CX2_BIT_HALT_RESET_ON);
	ipw_arc_release(priv);
	ipw_write_reg32(priv, CX2_MEM_HALT_AND_RESET, CX2_BIT_HALT_RESET_OFF);

	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	/* reset PHY */
	ipw_write_reg32(priv, CX2_INTERNAL_CMD_EVENT, CX2_BASEBAND_POWER_DOWN);
	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	ipw_write_reg32(priv, CX2_INTERNAL_CMD_EVENT, 0);
	
	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	/* enable ucode store */
	ipw_write_reg8(priv, DINO_CONTROL_REG, 0x0);
	ipw_write_reg8(priv, DINO_CONTROL_REG, DINO_ENABLE_CS);
	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	/* write ucode */
	/**
	 * @bug
	 * Do NOT set indirect address register once and then
	 * store data to indirect data register in the loop.
	 * It seems very reasonable, but in this case DINO do not
	 * accept ucode. It is essential to set address each time.
	 */
	/* load new ipw uCode */
	for (i = 0; i < len / 2; i++)
		ipw_write_reg16(priv, CX2_BASEBAND_CONTROL_STORE,
				image[i]);

	
	/* enable DINO */
	ipw_write_reg8(priv, CX2_BASEBAND_CONTROL_STATUS, 0);
	ipw_write_reg8(priv, CX2_BASEBAND_CONTROL_STATUS,
		       DINO_ENABLE_SYSTEM );

	/* this is where the igx / win driver deveates from the VAP driver.*/

	/* wait for alive response */
	for (i = 0; i < 100; i++) {
		/* poll for incoming data */
		cr = ipw_read_reg8(priv, CX2_BASEBAND_CONTROL_STATUS);
		if (cr & DINO_RXFIFO_DATA)
			break;
		spin_unlock_irqrestore(&priv->lock, flags);
		mdelay(1);
		spin_lock_irqsave(&priv->lock, flags);
	}

	if (cr & DINO_RXFIFO_DATA) {
		/* alive_command_responce size is NOT multiple of 4 */
		u32 responce_buffer[(sizeof(priv->dino_alive) + 3) / 4];

		for (i = 0; i < (sizeof(priv->dino_alive) + 3) / 4; i++) {
			responce_buffer[i] =
				ipw_read_reg32(priv, CX2_BASEBAND_RX_FIFO_READ);
		}
		memcpy(&priv->dino_alive, responce_buffer,
		       sizeof(priv->dino_alive));
		if (priv->dino_alive.alive_command == 1
		    && priv->dino_alive.ucode_valid == 1) {
			rc = 0;
			IPW_DEBUG_INFO
				("Microcode OK, rev. %d (0x%x) dev. %d (0x%x) of %02d/%02d/%02d %02d:%02d\n",
				 priv->dino_alive.software_revision,
				 priv->dino_alive.software_revision,
				 priv->dino_alive.device_identifier,
				 priv->dino_alive.device_identifier,
				 priv->dino_alive.time_stamp[0],
				 priv->dino_alive.time_stamp[1],
				 priv->dino_alive.time_stamp[2],
				 priv->dino_alive.time_stamp[3],
				 priv->dino_alive.time_stamp[4]);
		} else {
			IPW_DEBUG_INFO("Microcode is not alive\n");
			rc = -EINVAL;
		}
	} else {
		IPW_DEBUG_INFO("No alive response from DINO\n");
		rc = -ETIME;
	}
	/* disable DINO, otherwise for some reason
	   firmware have problem getting alive resp. */
	ipw_write_reg8(priv, CX2_BASEBAND_CONTROL_STATUS, 0);
	spin_unlock_irqrestore(&priv->lock, flags);
	IPW_DEBUG_TRACE(">>\n");

	return rc;
}

static int ipw_load_firmware(struct ipw_priv *priv, u8 * data,
			     size_t len)
{
	int rc = -1;
	int offset = 0;
	struct fw_header *header;
	struct fw_chunk *chunk;
	u32 shared_phys;
	u8 *shared_virt;
	
	IPW_DEBUG_TRACE("<< : \n");
	shared_virt = pci_alloc_consistent(priv->pci_dev, len, &shared_phys);

	if (!shared_virt)
		return -ENOMEM;

	memmove(shared_virt, data, len);

	/* Start the Dma */
	rc = ipw_fw_dma_enable(priv);

	header = (struct fw_header *)data;
	offset += sizeof(struct fw_header);
	
	if (priv->sram_desc.last_cb_index > 0) {
		/* the DMA is already ready this would be a bug. */
		BUG();
		goto out;
	}

	do {
		chunk = (struct fw_chunk *)(data + offset);
		offset += sizeof(struct fw_chunk);
		/* build DMA packet and queue up for sending */
		/* dma to chunk->address, the chunk->length bytes  from data + offeset*/
		/* Dma loading */
		rc = ipw_fw_dma_add_buffer(priv, shared_phys + offset,
					   chunk->address, chunk->length);
		if (rc) {
			IPW_DEBUG_INFO("dmaAddBuffer Failed\n");
			goto out;
		}
		
		offset += chunk->length;
	} while (offset < len);

	/* Run the DMA and wait for the answer*/
	rc = ipw_fw_dma_kick(priv);
	if (rc) {
		IPW_DEBUG_ERROR("dmaKick Failed\n");
		goto out;
	}

	rc = ipw_fw_dma_wait(priv);
	if (rc) {
		IPW_DEBUG_ERROR("dmaWaitSync Failed\n");
		goto out;
	}
 out:
	pci_free_consistent( priv->pci_dev, len, shared_virt, shared_phys);
	return rc;
}

/* stop nic */
static int ipw_stop_nic(struct ipw_priv *priv)
{
	int rc = 0;

	/* stop*/
	ipw_write32(priv, CX2_RESET_REG, CX2_RESET_REG_STOP_MASTER);
	
	rc = ipw_poll_bit(priv, CX2_RESET_REG, 
			  CX2_RESET_REG_MASTER_DISABLED, 500); 
	if (rc < 0) {
		IPW_DEBUG_ERROR("wait for reg master disabled failed\n");
		return rc;
	}   

	ipw_set_bit(priv, CX2_RESET_REG, CBD_RESET_REG_PRINCETON_RESET);
	
	return rc;
}

static void ipw_start_nic(struct ipw_priv *priv)
{
	IPW_DEBUG_TRACE(">>\n");

	/* prvHwStartNic  release ARC*/
	ipw_clear_bit(priv, CX2_RESET_REG,
		      CX2_RESET_REG_MASTER_DISABLED | 
		      CX2_RESET_REG_STOP_MASTER | 
		      CBD_RESET_REG_PRINCETON_RESET);
	
	/* enable power management */
	ipw_set_bit(priv, CX2_GP_CNTRL_RW, CX2_GP_CNTRL_BIT_HOST_ALLOWS_STANDBY);

	IPW_DEBUG_TRACE("<<\n");
}
	
static int ipw_init_nic(struct ipw_priv *priv)
{
	int rc;

	IPW_DEBUG_TRACE(">>\n");
	/* reset */	
	/*prvHwInitNic */
	/* set "initialization complete" bit to move adapter to D0 state */
	ipw_set_bit(priv, CX2_GP_CNTRL_RW, CX2_GP_CNTRL_BIT_INIT_DONE);

	/* low-level PLL activation */
	ipw_write32(priv, CX2_READ_INT_REGISTER,  CX2_BIT_INT_HOST_SRAM_READ_INT_REGISTER);

	/* wait for clock stabilization */
	rc = ipw_poll_bit(priv, CX2_GP_CNTRL_RW, 
			  CX2_GP_CNTRL_BIT_CLOCK_READY, 250); 
	if (rc < 0 )
		IPW_DEBUG_INFO("FAILED wait for clock stablization\n");

	/* assert SW reset */
	ipw_set_bit(priv, CX2_RESET_REG, CX2_RESET_REG_SW_RESET);

	udelay(10);

	/* set "initialization complete" bit to move adapter to D0 state */
	ipw_set_bit(priv, CX2_GP_CNTRL_RW, CX2_GP_CNTRL_BIT_INIT_DONE);

	IPW_DEBUG_TRACE(">>\n");
	return 0;
}


/* Call this function from process context, it will sleep in request_firmware. 
 * Probe is an ok place to call this from.
 */
static int ipw_reset_nic(struct ipw_priv *priv)
{
	int rc = 0;

	IPW_DEBUG_TRACE(">>\n");
	
	rc = ipw_init_nic(priv);

	IPW_DEBUG_TRACE("<<\n");
	return rc;
} 

static int ipw_load(struct ipw_priv *priv)
{
	int rc = -1;
	const struct firmware *bootfw = NULL;
	const struct firmware *firmware = NULL;
	const struct firmware *ucode = NULL;
	int retries = 3;

	/* disable interrupts */
	ipw_disable_interrupts(priv);

 retry:
	/* ack pending interrupts */
	ipw_write32(priv, CX2_INTA_RW, CX2_INTA_MASK_ALL);
	
	ipw_stop_nic(priv);

	rc = ipw_reset_nic(priv);
	if (rc) {
		IPW_DEBUG_ERROR("Unable to reset NIC\n");
		goto error;
	}

	ipw_zero_memory(priv, CX2_NIC_SRAM_LOWER_BOUND, 
			CX2_NIC_SRAM_UPPER_BOUND - CX2_NIC_SRAM_LOWER_BOUND);

	/* ask firmware_class module to get the boot firmware off disk */
	rc = request_firmware(&bootfw, "ipw2200_boot.fw", &priv->pci_dev->dev);
	if (rc < 0) {
		IPW_DEBUG_ERROR("ipw2200_boot.fw load failed\n");
		goto error;
	} 
	IPW_DEBUG_INFO("boot fw data %p size %d \n", bootfw->data, bootfw->size);

	/* DMA the initial boot firmware into the device */
	rc = ipw_load_firmware(priv, bootfw->data, bootfw->size);
	if (rc < 0) {
		IPW_DEBUG_ERROR("Unable to load boot firmware\n");
		goto error;
	}

	/* kick start the device */
	ipw_start_nic(priv);

	/* wait for the device to finish it's initial startup sequence */
	rc = ipw_poll_bit(priv, CX2_INTA_RW, 
			  CX2_INTA_BIT_FW_INITIALIZATION_DONE, 500); 
	if (rc < 0) {
		IPW_DEBUG_ERROR("device failed to boot initial fw image\n");
		goto error;
	}
	IPW_DEBUG_INFO("initial device response after %dms\n", rc);

	/* ack fw init done interrupt */	
	ipw_write32(priv, CX2_INTA_RW, CX2_INTA_BIT_FW_INITIALIZATION_DONE);

	/* ask firmware_class module to get the ucode image off disk */
	rc = request_firmware(&ucode, "ipw2200_ucode.fw", &priv->pci_dev->dev);
	if (rc < 0) {
		IPW_DEBUG_ERROR("ipw_ucode.fw load failed\n");
		goto error;
	} 
	IPW_DEBUG_INFO("ucode data %p size %d \n",ucode->data,ucode->size);

	/* DMA the ucode into the device */
	rc = ipw_load_ucode(priv, ucode->data, ucode->size);
	if (rc < 0) {
		IPW_DEBUG_ERROR("Unable to load ucode\n");
		goto error;
	}
	
	/* stop nic */
	ipw_stop_nic(priv);

	/* ask firmware_class module to get the bss firmware image off disk */
	rc = request_firmware(&firmware, "ipw2200_bss.fw",&priv->pci_dev->dev);
	if (rc < 0) {
		IPW_DEBUG_ERROR("ipw_bss.fw load failed\n");
		goto error;
	} 
	IPW_DEBUG_INFO("bss data %p size %d \n",firmware->data,firmware->size);

	/* DMA bss firmware into the device */
	rc = ipw_load_firmware(priv, firmware->data, firmware->size);
	if (rc < 0 ) {
		IPW_DEBUG_ERROR("Unable to load bss firmware\n");
		goto error;
	}

	ipw_write32(priv, IPW_EEPROM_LOAD_DISABLE, 0);

	rc = ipw_queue_reset(priv);
	if (rc) {
		IPW_DEBUG_ERROR("Unable to initialize queues\n");
		goto error;
	}
	
	priv->rxq = ipw_rx_queue_alloc(priv);
	if (!priv->rxq) {
		IPW_DEBUG_ERROR("Unable to initialize Rx queue\n");
		goto error;
	}

	/* kick start the device */
	ipw_start_nic(priv);

	if (ipw_read32(priv, CX2_INTA_RW) & CX2_INTA_BIT_PARITY_ERROR) {
		if (retries > 0) {
			IPW_DEBUG_INFO("Parity error during load -- retrying.\n");
			retries--;
			goto retry;
		}
		IPW_DEBUG_ERROR("TODO: Handle parity error -- schedule restart?\n");
		goto error;
	}

	/* wait for the device */
	rc = ipw_poll_bit(priv, CX2_INTA_RW, 
			  CX2_INTA_BIT_FW_INITIALIZATION_DONE, 500); 
	if (rc < 0) {
		IPW_DEBUG_ERROR("device failed to start after 500ms\n");
		goto error;
	}
	IPW_DEBUG_INFO("device response after %dms\n", rc);

	/* ack fw init done interrupt */
	ipw_write32(priv, CX2_INTA_RW, CX2_INTA_BIT_FW_INITIALIZATION_DONE);

	/* read eeprom data and initialize the eeprom region of sram */
	priv->eeprom_delay = 1;
	ipw_eeprom_init_sram(priv);	

	/* enable interrupts */
	ipw_enable_interrupts(priv);

	ipw_rx_queue_replenish(priv);
	ipw_write32(priv, CX2_RX_READ_INDEX, priv->rxq->read);

	/* ack pending interrupts */
	ipw_write32(priv, CX2_INTA_RW, CX2_INTA_MASK_ALL);

	release_firmware(bootfw);
	release_firmware(ucode);
	release_firmware(firmware);
	return 0;

 error:
	if (priv->rxq)
		ipw_rx_queue_free(priv, priv->rxq);
	ipw_tx_queue_free(priv);
	if (bootfw)
		release_firmware(bootfw);
	if (ucode)
		release_firmware(ucode);
	if (firmware)
		release_firmware(firmware);
	return -1;
}

/** 
 * DMA services
 *
 * Theory of operation
 *
 * A queue is a circular buffers with 'Read' and 'Write' pointers.
 * 2 empty entries always kept in the buffer to protect from overflow.
 *
 * For Tx queue, there are low mark and high mark limits. If, after queuing
 * the packet for Tx, free space become < low mark, Tx queue stopped. When 
 * reclaiming packets (on 'tx done IRQ), if free space become > high mark, 
 * Tx queue resumed.
 *
 * The IPW operates with six queues, one receive queue in the device's
 * sram, one transmit queue for sending commands to the device firmware,
 * and four transmit queues for data.  
 *
 * The four transmit queues allow for performing quality of service (qos) 
 * transmissions as per the 802.11 protocol.  Currently Linux does not
 * provide a mechanism to the user for utilizing prioritized queues, so 
 * we only utilize the first data transmit queue (queue1).
 */

/**
 * Driver allocates buffers of this size for Rx
 */
#define CX2_RX_BUF_SIZE (3000)

static inline int ipw_queue_space(struct clx2_queue *q)
{
	int s = q->last_used - q->first_empty;
	if (s <= 0)
		s += q->n_bd;
	s -= 2;			/* keep some reserve to not confuse empty and full situations */
	if (s < 0)
		s = 0;
	return s;
}

static inline int ipw_queue_inc_wrap(int index, int n_bd)
{
	return (++index == n_bd) ? 0 : index;
}

/**
 * Initialize common DMA queue structure
 * 
 * @param q                queue to init
 * @param count            Number of BD's to allocate. Should be power of 2
 * @param read_register    Address for 'read' register
 *                         (not offset within BAR, full address)
 * @param write_register   Address for 'write' register
 *                         (not offset within BAR, full address)
 * @param base_register    Address for 'base' register
 *                         (not offset within BAR, full address)
 * @param size             Address for 'size' register
 *                         (not offset within BAR, full address)
 */
static void ipw_queue_init(struct ipw_priv *priv, struct clx2_queue *q, 
			   int count, u32 read, u32 write,
			   u32 base, u32 size)
{
	q->n_bd = count;

	q->low_mark = q->n_bd / 4;
	if (q->low_mark < 4)
		q->low_mark = 4;

	q->high_mark = q->n_bd / 8;
	if (q->high_mark < 2)
		q->high_mark = 2;

	q->first_empty = q->last_used = 0;
	q->reg_r = read;
	q->reg_w = write;

	ipw_write32(priv, base, q->dma_addr);
	ipw_write32(priv, size, count);
	ipw_write32(priv, read, 0);
	ipw_write32(priv, write, 0);

	_ipw_read32(priv, 0x90);
}

static int ipw_queue_tx_init(struct ipw_priv *priv, 
			     struct clx2_tx_queue *q,
			     int count, u32 read, u32 write,
			     u32 base, u32 size)
{
	struct pci_dev *dev = priv->pci_dev;

	q->txb = vmalloc(sizeof(q->txb[0]) * count);
	if (!q->txb) {
		IPW_DEBUG_ERROR("vmalloc for auxilary BD structures failed\n");
		return -ENOMEM;
	}

	q->bd = pci_alloc_consistent(dev,sizeof(q->bd[0])*count, &q->q.dma_addr);
	if (!q->bd) {
		IPW_DEBUG_ERROR("pci_alloc_consistent(%d) failed\n",
				sizeof(q->bd[0]) * count);
		vfree(q->txb);
		q->txb = NULL;
		return -ENOMEM;
	}

	ipw_queue_init(priv, &q->q, count, read, write, base, size);
	return 0;
}

/**
 * Free one TFD, those at index [txq->q.last_used].
 * Do NOT advance any indexes
 * 
 * @param dev
 * @param txq
 */
static inline void ipw_queue_tx_free_tfd(struct ipw_priv *priv,
					 struct clx2_tx_queue *txq)
{
	struct tfd_frame *bd = &txq->bd[txq->q.last_used];
	struct pci_dev *dev = priv->pci_dev;
	int i;
	
	/* classify bd */
	if (bd->control_flags.message_type == TX_HOST_COMMAND_TYPE)
		/* nothing to cleanup after for host commands */
		return;

	/* sanity check */
	if (bd->u.data.num_chunks > NUM_TFD_CHUNKS) {
		IPW_DEBUG_ERROR("Too many chunks: %i\n", bd->u.data.num_chunks);
		/** @todo issue fatal error, it is quite serious situation */
		return;
	}

	/* unmap chunks if any */
	for (i = 0; i < bd->u.data.num_chunks; i++) {
		pci_unmap_single(dev, bd->u.data.chunk_ptr[i],
				 bd->u.data.chunk_len[i], PCI_DMA_TODEVICE);
		if (txq->txb[txq->q.last_used]) {
			ieee80211_txb_free(txq->txb[txq->q.last_used]);
			txq->txb[txq->q.last_used] = NULL;
		}
	}
}

/**
 * Deallocate DMA queue.
 * 
 * Empty queue by removing and destroying all BD's.
 * Free all buffers.
 * 
 * @param dev
 * @param q
 */
static void ipw_queue_tx_free(struct ipw_priv *priv,
			    struct clx2_tx_queue *txq)
{
	struct clx2_queue *q = &txq->q;
	struct pci_dev *dev = priv->pci_dev;

	if (q->n_bd == 0) 
		return;	

	/* first, empty all BD's */
	for (; q->first_empty != q->last_used;
	     q->last_used = ipw_queue_inc_wrap(q->last_used, q->n_bd))
		ipw_queue_tx_free_tfd(priv, txq);
	
	/* free buffers belonging to queue itself */
	pci_free_consistent(dev, sizeof(txq->bd[0])*q->n_bd, txq->bd, 
			    q->dma_addr);
	vfree(txq->txb);

	/* 0 fill whole structure */
	memset(txq, 0, sizeof(*txq));
}


/**
 * Destroy all DMA queues and structures
 * 
 * @param priv
 */
static void ipw_tx_queue_free(struct ipw_priv *priv)
{
	/* Tx CMD queue */
	ipw_queue_tx_free(priv, &priv->txq_cmd);

	/* Tx queues */
	ipw_queue_tx_free(priv, &priv->txq[0]);
	ipw_queue_tx_free(priv, &priv->txq[1]);
	ipw_queue_tx_free(priv, &priv->txq[2]);
	ipw_queue_tx_free(priv, &priv->txq[3]);
}

static void inline __maybe_wake_tx(struct ipw_priv *priv)
{
	if (netif_running(priv->net_dev)) {
		switch (priv->port_type) {
		case DCR_TYPE_MU_BSS:
		case DCR_TYPE_MU_IBSS:
			if (!(priv->status & STATUS_ASSOCIATED)) {
				return;
			}
		}
		netif_wake_queue(priv->net_dev);
	}

}

static int ipw_tx_skb(struct ipw_priv *priv, struct sk_buff *skb)
{
	int rc = 0;
	struct tfd_frame *tfd;
	struct clx2_tx_queue *txq = &priv->txq[0];
	struct clx2_queue *q = &txq->q;
	struct ieee80211_txb *txb;
	u16 frame_control;
	u8 dst[ETH_ALEN];
	u8 src[ETH_ALEN];

	memcpy(dst, skb->data, ETH_ALEN);
	memcpy(src, skb->data + ETH_ALEN, ETH_ALEN);

	txb = ieee80211_skb_to_txb(priv->ieee, skb);
	if (txb == NULL) {
		IPW_DEBUG_DROP("Failed to Tx packet\n");
		priv->ieee->stats.tx_errors++;
		goto out_free_skb;
	}

	if (ipw_queue_space(q) < 2) {
		IPW_DEBUG_ERROR("No space for Tx\n");
		rc = -EBUSY;
		goto out_free_txb;
	}

	if (txb->nr_frags > 1) {
		printk(KERN_WARNING "Tx fragmentation not yet supported.\n");
		rc = -EIO;
		goto out_free_txb;
	}

	tfd = &txq->bd[q->first_empty];
	txq->txb[q->first_empty] = txb;

	memset(tfd, 0, sizeof(*tfd));

	tfd->control_flags.message_type = TX_FRAME_TYPE;
	tfd->control_flags.control_bits = TFD_NEED_IRQ_MASK;
	tfd->u.data.cmd_id = DINO_CMD_TX;
	tfd->u.data.len = txb->fragments[0]->len;
	tfd->u.data.tx_flags = DCT_FLAG_NO_WEP | DCT_FLAG_ACK_REQD;
	if (priv->config & CFG_PREAMBLE)
		tfd->u.data.tx_flags |= DCT_FLAG_SHORT_PREMBL;
	frame_control = IEEE802_11_FTYPE_DATA | IEEE802_11_STYPE_DATA;
	if (priv->ieee->iw_mode == IW_MODE_INFRA) {
		frame_control |= IEEE802_11_FCTL_TODS;
		/* To DS: Addr1 = BSSID, Addr2 = SA, Addr3 = DA */
		memcpy(tfd->u.data.tfd.tfd_24.mchdr.addr1, priv->ieee->bssid, ETH_ALEN);
		memcpy(tfd->u.data.tfd.tfd_24.mchdr.addr2, src, ETH_ALEN);
		memcpy(tfd->u.data.tfd.tfd_24.mchdr.addr3, dst, ETH_ALEN);
	} else if (priv->ieee->iw_mode == IW_MODE_ADHOC) {
		/* not From/To DS: Addr1 = DA, Addr2 = SA, Addr3 = BSSID */
		memcpy(tfd->u.data.tfd.tfd_24.mchdr.addr1, dst, ETH_ALEN);
		memcpy(tfd->u.data.tfd.tfd_24.mchdr.addr2, src, ETH_ALEN);
		memcpy(tfd->u.data.tfd.tfd_24.mchdr.addr3, priv->ieee->bssid, ETH_ALEN);
	} else {
		IPW_DEBUG_ERROR("Unknown ieee->iw_mode in Tx.\n");
	}
	if (txb->encrypted) 
		frame_control |= IEEE802_11_FCTL_WEP;
	tfd->u.data.tfd.tfd_24.mchdr.ctrl1 = frame_control & 0xff;
	tfd->u.data.tfd.tfd_24.mchdr.ctrl2 = (frame_control >> 8) & 0xff;

	IPW_DEBUG_TX("Dumping TX packet (%d bytes):\n", txb->fragments[0]->len);
	printk_buf(IPW_DL_TX, txb->fragments[0]->data, txb->fragments[0]->len);
	/* payload */
	tfd->u.data.num_chunks = 1;
	tfd->u.data.chunk_ptr[0] = pci_map_single(
		priv->pci_dev, txb->fragments[0]->data,
		txb->fragments[0]->len, PCI_DMA_TODEVICE);
	tfd->u.data.chunk_len[0] = txb->fragments[0]->len;

	/* kick DMA */
	q->first_empty = ipw_queue_inc_wrap(q->first_empty, q->n_bd);
	ipw_write32(priv, q->reg_w, q->first_empty);

	/* Keep TIMEOUT from happening while we're still developing... */
	priv->net_dev->trans_start = jiffies;

	if (ipw_queue_space(q) < q->high_mark) {
		IPW_DEBUG_INFO("Stopping network queue\n");
		netif_stop_queue(priv->net_dev);
	}

	goto done;

 out_free_txb:
	ieee80211_txb_free(txb);

 out_free_skb:
	dev_kfree_skb(skb);

 done:
	IPW_DEBUG_TRACE("<<\n");
	return rc;
}

static void notify_wx_assoc_event(struct ipw_priv *priv)
{
	union iwreq_data wrqu;
	wrqu.ap_addr.sa_family = ARPHRD_ETHER;
	if (priv->status & STATUS_ASSOCIATED)
		memcpy(wrqu.ap_addr.sa_data, priv->bssid, ETH_ALEN);
	else
		memset(wrqu.ap_addr.sa_data, 0, ETH_ALEN);
	wireless_send_event(priv->net_dev, SIOCGIWAP, &wrqu, NULL);
}

/**
 * Handle host notification packet.
 * Called from interrupt routine
 * 
 * @param notif
 */
static inline void ipw_rx_notification(struct ipw_priv* priv,
				       struct ipw_rx_notification *notif)
{
	IPW_DEBUG_NOTIF("type = %i (%d bytes)\n", 
			notif->subtype, notif->size);
	
	switch (notif->subtype) {
	case HOST_NOTIFICATION_STATUS_ASSOCIATEDIATE: {
		struct notif_association *assoc = &notif->u.assoc;
		
		switch (assoc->state) {
		case CMAS_ASSOCIATED: {
			IPW_DEBUG_NOTIF("assoc: associated: " MAC_FMT " \n", 
					MAC_ARG(priv->bssid));
			
			if (priv->ieee->iw_mode == IW_MODE_INFRA) 
				memcpy(priv->ieee->bssid, priv->bssid, 
				       ETH_ALEN);

			priv->status &= ~STATUS_ASSOCIATING;
			priv->status |= STATUS_ASSOCIATED;

			/* TODO: Set connected state... */

			netif_carrier_on(priv->net_dev);
			if (netif_queue_stopped(priv->net_dev)) {
				IPW_DEBUG_NOTIF("waking queue\n");
				netif_wake_queue(priv->net_dev);
			} else {
				IPW_DEBUG_NOTIF("starting queue\n");
				netif_start_queue(priv->net_dev);
			}

			notify_wx_assoc_event(priv);
			break;
		}
			
		case CMAS_AUTHENTICATED: {
			IPW_DEBUG_NOTIF("assoc: authenticated.\n");
			break;
		}
		
		case CMAS_INIT: {
			IPW_DEBUG_NOTIF("assoc: disassociated.\n");

			priv->status &= ~(STATUS_ASSOCIATING | STATUS_ASSOCIATED);
			
			netif_carrier_off(priv->net_dev);
			netif_stop_queue(priv->net_dev);
			notify_wx_assoc_event(priv);
			queue_work(priv->workqueue, &priv->request_scan);
			break;
		}
			
		default: 
			IPW_DEBUG_ERROR("assoc: unknown (%d)\n",
					assoc->state);
			break;
		}

		break;
	}

	case HOST_NOTIFICATION_STATUS_AUTHENTICATE: {
		struct notif_authenticate *auth = &notif->u.auth;
		switch (auth->state) {
		case CMAS_AUTHENTICATED:
			IPW_DEBUG_NOTIF("auth: authenticated.\n");
			priv->status |= STATUS_AUTH;
			break;

		case CMAS_INIT:
			IPW_DEBUG_NOTIF("auth: disassociated.\n");
			/* disassoc as well */
			if (priv->port_type == DCR_TYPE_MU_BSS) {
				priv->bss_aid=0;
				priv->supp.num=0;
				priv->extended.num=0;
				/* BSS will clear 'active' flag in sweep table by itself */
			}
			priv->status &= ~(STATUS_ASSOCIATING | STATUS_ASSOCIATED | 
					  STATUS_AUTH);

			netif_carrier_off(priv->net_dev);
			netif_stop_queue(priv->net_dev);
			notify_wx_assoc_event(priv);
			queue_work(priv->workqueue, &priv->request_scan);
			break;
		default:
			IPW_DEBUG_NOTIF("auth: failure - %d\n", auth->state);
			break;
		}
		break;
	}

	case HOST_NOTIFICATION_STATUS_SCAN_CHANNEL_RESULT: {
		struct notif_channel_result *x = &notif->u.channel_result;
		
		if (notif->size == sizeof(*x)) {
			IPW_DEBUG_SCAN("Scan result for channel %d\n", 
				       x->channel_num);
			//__print_stats(&x->stats);
		} else {
			IPW_DEBUG_SCAN("Scan result of wrong size %d (should be %d)\n",
				       notif->size,sizeof(*x));
		}
		break;
	}

	case HOST_NOTIFICATION_STATUS_SCAN_COMPLETED: {
		struct notif_scan_complete* x = &notif->u.scan_complete;
		if (notif->size == sizeof(*x)) {
			IPW_DEBUG_SCAN("Scan completed: type %d, %d channels, %d status\n",
				       x->scan_type, 
				       x->num_channels, 
				       x->status);
			IPW_DEBUG_SCAN("age %d\n",(int)(jiffies - priv->ts_scan_abort));
		} else {
			IPW_DEBUG_ERROR("Scan completed of wrong size %d (should be %d)\n",
					notif->size,sizeof(*x));
		}
	
		priv->status &= ~(STATUS_SCANNING | STATUS_SCAN_ABORTING);
		if (priv->status & STATUS_SCAN_PENDING)
			queue_work(priv->workqueue, &priv->request_scan);
		else if (!(priv->status & 
			   (STATUS_ASSOCIATED | STATUS_ASSOCIATING)))
			queue_work(priv->workqueue, &priv->associate);
		priv->scans++;
		break;
	}

	case HOST_NOTIFICATION_STATUS_FRAG_LENGTH: {
		struct notif_frag_length *x = &notif->u.frag_len;

		if (notif->size == sizeof(*x)) {
			IPW_DEBUG_ERROR("Frag length: %d\n", x->frag_length);
		} else {
			IPW_DEBUG_ERROR("Frag length of wrong size %d (should be %d)\n",
					notif->size, sizeof(*x));
		}
		break;
	}

	case HOST_NOTIFICATION_STATUS_LINK_DETERIORATION: {
		struct notif_link_deterioration *x = &notif->u.link_deterioration;
		if (notif->size==sizeof(*x)) {
			memcpy(&priv->last_link_deterioration,x,sizeof(*x));
		} else {
			IPW_DEBUG_ERROR("Link Deterioration of wrong size %d (should be %d)\n",
					notif->size,sizeof(*x));
		}
		break;
	}

	case HOST_NOTIFICATION_DINO_CONFIG_RESPONSE: {
		IPW_DEBUG_ERROR("Dino config\n");
		if (priv->hcmd && priv->hcmd->cmd == HOST_CMD_DINO_CONFIG) {
			/* TODO: Do anything special? */
		} else {
			IPW_DEBUG_ERROR("Unexpected DINO_CONFIG_RESPONSE\n");
		}
		break;
	}

	case HOST_NOTIFICATION_STATUS_BEACON_STATE: {
		struct notif_beacon_state *x = &notif->u.beacon_state;
		if (notif->size == sizeof(*x)) {
			if (priv->port_type != DCR_TYPE_AP) { /* AP send it each beacon, no reason to display */
				IPW_DEBUG_ERROR("Beacon state: 0x%08x, number %d\n",(int)x->state, (int)x->number);
			}
		} else {
			IPW_DEBUG_ERROR("Beacon state of wrong size %d (should be %d)\n",
					notif->size, sizeof(*x));
		}
		break;
	}

	case HOST_NOTIFICATION_STATUS_TGI_TX_KEY: {
		struct notif_tgi_tx_key *x = &notif->u.tgi_tx_key;
		if (notif->size==sizeof(*x)) {
			IPW_DEBUG_ERROR("TGi Tx Key: state 0x%02x sec type 0x%02x station %d\n",
					x->key_state,x->security_type,x->station_index);
		} else {
			IPW_DEBUG_ERROR("TGi Tx Key of wrong size %d (should be %d)\n",
					notif->size,sizeof(*x));
		}
		break;
	}

	case HOST_NOTIFICATION_CALIB_KEEP_RESULTS: {
		struct notif_calibration *x = &notif->u.calibration;

		if (notif->size == sizeof(*x)) {
			priv->calib_jiffies = jiffies;
			memcpy(&priv->calib, x, sizeof(*x));
			IPW_DEBUG_ERROR("Calibration\n");
		} else {
			IPW_DEBUG_ERROR("Calibration of wrong size %d (should be %d)\n",
					notif->size,sizeof(*x));
		}
		break;
	}

	default:
		IPW_DEBUG_ERROR("Unknown notification subtype=%d,flags=0x%2x,size=%d\n",
				notif->subtype,
				notif->flags,
				notif->size);
	}
}

/**
 * Destroys all DMA structures and initialise them again. Fill Rx queue.
 * 
 * @param priv
 * @return error code
 */
static int ipw_queue_reset(struct ipw_priv *priv)
{
	int rc = 0;
	/** @todo customize queue sizes */
	int nTx = 64, nTxCmd = 8;
	ipw_tx_queue_free(priv);
	/* Tx CMD queue */
	rc = ipw_queue_tx_init(priv, &priv->txq_cmd, nTxCmd,
			     CX2_TX_CMD_QUEUE_READ_INDEX,
			     CX2_TX_CMD_QUEUE_WRITE_INDEX,
			     CX2_TX_CMD_QUEUE_BD_BASE,
			     CX2_TX_CMD_QUEUE_BD_SIZE);
	if (rc) {
		IPW_DEBUG_ERROR("Tx Cmd queue init failed\n");
		goto error;
	}
	/* Tx queue(s) */
	rc = ipw_queue_tx_init(priv, &priv->txq[0], nTx,
			       CX2_TX_QUEUE_0_READ_INDEX,
			       CX2_TX_QUEUE_0_WRITE_INDEX,
			       CX2_TX_QUEUE_0_BD_BASE,
			       CX2_TX_QUEUE_0_BD_SIZE);
	if (rc) {
		IPW_DEBUG_ERROR("Tx 0 queue init failed\n");
		goto error;
	}
	rc = ipw_queue_tx_init(priv, &priv->txq[1], nTx,
			       CX2_TX_QUEUE_1_READ_INDEX,
			       CX2_TX_QUEUE_1_WRITE_INDEX,
			       CX2_TX_QUEUE_1_BD_BASE,
			       CX2_TX_QUEUE_1_BD_SIZE);
	if (rc) {
		IPW_DEBUG_ERROR("Tx 1 queue init failed\n");
		goto error;
	}
	rc = ipw_queue_tx_init(priv, &priv->txq[2], nTx,
			       CX2_TX_QUEUE_2_READ_INDEX,
			       CX2_TX_QUEUE_2_WRITE_INDEX,
			       CX2_TX_QUEUE_2_BD_BASE,
			       CX2_TX_QUEUE_2_BD_SIZE);
	if (rc) {
		IPW_DEBUG_ERROR("Tx 2 queue init failed\n");
		goto error;
	}
	rc = ipw_queue_tx_init(priv, &priv->txq[3], nTx,
			       CX2_TX_QUEUE_3_READ_INDEX,
			       CX2_TX_QUEUE_3_WRITE_INDEX,
			       CX2_TX_QUEUE_3_BD_BASE,
			       CX2_TX_QUEUE_3_BD_SIZE);
	if (rc) {
		IPW_DEBUG_ERROR("Tx 3 queue init failed\n");
		goto error;
	}
	/* statistics */
	priv->rx_bufs_min = 0;
	priv->rx_pend_max = 0;
	return rc;

 error:
	ipw_tx_queue_free(priv);
	return rc;
}

/**
 * Reclaim Tx queue entries no more used by NIC.
 * 
 * When FW adwances 'R' index, all entries between old and
 * new 'R' index need to be reclaimed. As result, some free space
 * forms. If there is enough free space (> low mark), wake Tx queue.
 * 
 * @note Need to protect against garbage in 'R' index
 * @param priv
 * @param txq
 * @param qindex
 * @return Number of used entries remains in the queue
 */
static int ipw_queue_tx_reclaim(struct ipw_priv *priv, 
				struct clx2_tx_queue *txq, int qindex)
{
	u32 hw_tail;
	int used;
	struct clx2_queue *q = &txq->q;

	hw_tail = ipw_read32(priv, q->reg_r);
	if (hw_tail >= q->n_bd) {
		IPW_DEBUG_ERROR
			("Read index for DMA queue (%d) is out of range [0-%d)\n",
			 hw_tail, q->n_bd);
		goto done;
	}
	for (; q->last_used != hw_tail;
	     q->last_used = ipw_queue_inc_wrap(q->last_used, q->n_bd)) {
		ipw_queue_tx_free_tfd(priv, txq);
	}
 done:
	if (ipw_queue_space(q) > q->low_mark && qindex >= 0) {
		__maybe_wake_tx(priv);
	}
	used = q->first_empty - q->last_used;
	if (used < 0)
		used += q->n_bd;

	return used;
}

static int ipw_queue_tx_hcmd(struct ipw_priv *priv, int hcmd, void *buf,
			     int len, int sync)
{
	struct clx2_tx_queue *txq = &priv->txq_cmd;
	struct clx2_queue *q = &txq->q;
	struct tfd_frame *tfd;

	if (ipw_queue_space(q) < (sync ? 1 : 2)) {
		IPW_DEBUG_ERROR("No space for Tx\n");
		return -EBUSY;
	}

	tfd = &txq->bd[q->first_empty];
	txq->txb[q->first_empty] = NULL;

	memset(tfd, 0, sizeof(*tfd));
	tfd->control_flags.message_type = TX_HOST_COMMAND_TYPE;
	tfd->control_flags.control_bits = TFD_NEED_IRQ_MASK;
	priv->hcmd_seq++;
	tfd->u.cmd.index = hcmd;
	tfd->u.cmd.length = len;
	memcpy(tfd->u.cmd.payload, buf, len);
	q->first_empty = ipw_queue_inc_wrap(q->first_empty, q->n_bd);
	ipw_write32(priv, q->reg_w, q->first_empty);
	_ipw_read32(priv, 0x90);

	return 0;
}



/* 
 * Rx theory of operation
 *
 * The host allocates 32 DMA target addresses and passes the host address
 * to the firmware at register CX2_RFDS_TABLE_LOWER + N * RFD_SIZE where N is
 * 0 to 31
 *
 * Rx Queue Indexes
 * The host/firmware share two index registers for managing the Rx buffers.
 *
 * The READ index maps to the first position that the firmware may be writing 
 * to -- the driver can read up to (but not including) this position and get 
 * good data.  
 * The READ index is managed by the firmware once the card is enabled.
 *
 * The WRITE index maps to the last position the driver has read from -- the
 * position preceding WRITE is the last slot the firmware can place a packet.
 *
 * The queue is empty (no good data) if WRITE = READ - 1, and is full if
 * WRITE = READ.  
 *
 * During initialization the host sets up the READ queue position to the first 
 * INDEX position, and WRITE to the last (READ - 1 wrapped)
 *
 * When the firmware places a packet in a buffer it will advance the READ index
 * and fire the RX interrupt.  The driver can then query the READ index and
 * process as many packets as possible, moving the WRITE index forward as it
 * resets the Rx queue buffers with new memory.
 * 
 * The management in the driver is as follows:
 * + A list of pre-allocated SKBs is stored in ipw->rxq->rx_free.  When 
 *   ipw->rxq->free_count drops to or below RX_LOW_WATERMARK, work is scheduled
 *   to replensish the ipw->rxq->rx_free.  
 * + In ipw_rx_queue_replenish (scheduled) if 'processed' != 'read' then the
 *   ipw->rxq is replenished and the READ INDEX is updated (updating the
 *   'processed' and 'read' driver indexes as well)
 * + A received packet is processed and handed to the kernel network stack,
 *   detached from the ipw->rxq.  The driver 'processed' index is updated.
 * + The Host/Firmware ipw->rxq is replenished at tasklet time from the rx_free
 *   list. If there are no allocated buffers in ipw->rxq->rx_free, the READ 
 *   INDEX is not incremented and ipw->status(RX_STALLED) is set.  If there 
 *   were enough free buffers and RX_STALLED is set it is cleared.
 *
 *
 * Driver sequence:
 *
 * ipw_rx_queue_alloc()       Allocates rx_free 
 * ipw_rx_queue_replenish()   Replenishes rx_free list from rx_used, and calls
 *                            ipw_rx_queue_restock
 * ipw_rx_queue_restock()     Moves available buffers from rx_free into Rx
 *                            queue, updates firmware pointers, and updates
 *                            the WRITE index.  If insufficient rx_free buffers
 *                            are available, schedules ipw_rx_queue_replenish
 *
 * -- enable interrupts --
 * ISR - ipw_rx()             Detach ipw_rx_mem_buffers from pool up to the
 *                            READ INDEX, detaching the SKB from the pool. 
 *                            Moves the packet buffer from queue to rx_used.
 *                            Calls ipw_rx_queue_restock to refill any empty
 *                            slots.
 * ...
 *
 */

/* 
 * If there are slots in the RX queue that  need to be restocked,
 * and we have free pre-allocated buffers, fill the ranks as much
 * as we can pulling from rx_free.
 *
 * This moves the 'write' index forward to catch up with 'processed', and
 * also updates the memory address in the firmware to reference the new
 * target buffer.
 */
static void ipw_rx_queue_restock(struct ipw_priv *priv)
{
	struct ipw_rx_queue *rxq = priv->rxq;
	struct list_head *element;
	struct ipw_rx_mem_buffer *rxb;

	int write = rxq->write;

	while ((rxq->write != rxq->processed) && (rxq->free_count)) {
		element = rxq->rx_free.next;
		rxb = list_entry(element, struct ipw_rx_mem_buffer, list);
		list_del(element);

		ipw_write32(priv, CX2_RFDS_TABLE_LOWER + rxq->write * RFD_SIZE,
			    rxb->dma_addr);
		rxq->queue[rxq->write] = rxb;
		rxq->write = (rxq->write + 1) % RX_QUEUE_SIZE;
		rxq->free_count--;
	}

	/* If the pre-allocated buffer pool is dropping low, schedule to 
	 * refill it */
	if (rxq->free_count <= RX_LOW_WATERMARK)
		queue_work(priv->workqueue, &priv->rx_replenish);

	/* If we've added more space for the firmware to place data, tell it */
	if (write != rxq->write) 
		ipw_write32(priv, CX2_RX_WRITE_INDEX, rxq->write);
}

/*
 * Move all used packet from rx_used to rx_free, allocating a new SKB for each.
 * Also restock the Rx queue via ipw_rx_queue_restock.  
 * 
 * This is called as a scheduled work item (except for during intialization)
 */
static void ipw_rx_queue_replenish(void *data)
{
	struct ipw_priv *priv = data;
	struct ipw_rx_queue *rxq = priv->rxq;
	struct list_head *element;
	struct ipw_rx_mem_buffer *rxb;

	while (!list_empty(&rxq->rx_used)) {
		element = rxq->rx_used.next;
		rxb = list_entry(element, struct ipw_rx_mem_buffer, list);
		rxb->skb = alloc_skb(CX2_RX_BUF_SIZE, GFP_ATOMIC | GFP_DMA);
		if (!rxb->skb) {
			printk(KERN_CRIT "%s: Can not allocate SKB buffers.\n",
			       priv->net_dev->name);
			/* We don't reschedule replenish work here -- we will
			 * call the restock method and if it still needs
			 * more buffers it will schedule replenish */
			break;
		}
		list_del(element);
		
		rxb->rxb = (struct ipw_rx_buffer *)rxb->skb->data;
		rxb->dma_addr = pci_map_single(
			priv->pci_dev, rxb->skb->data, CX2_RX_BUF_SIZE,
			PCI_DMA_FROMDEVICE);
		
		list_add_tail(&rxb->list, &rxq->rx_free);
		rxq->free_count++;
	}

	ipw_rx_queue_restock(priv);
}

/* Assumes that the skb field of the buffers in 'pool' is kept accurate.
 * If an SKB has been detached, the POOL needs to have it's SKB set to NULL
 * This free routine walks the list of POOL entries and if SKB is set to 
 * non NULL it is unmapped and freed
 */
static void ipw_rx_queue_free(struct ipw_priv *priv, 
			      struct ipw_rx_queue *rxq)
{
	int i;

	if (!rxq)
		return;
	
	for (i = 0; i < RX_QUEUE_SIZE + RX_FREE_BUFFERS; i++) {
		if (rxq->pool[i].skb != NULL) {
			pci_unmap_single(priv->pci_dev, rxq->pool[i].dma_addr,
					 CX2_RX_BUF_SIZE,
					 PCI_DMA_FROMDEVICE);
			dev_kfree_skb(rxq->pool[i].skb);
		}
	}

	kfree(rxq);
}

struct ipw_rx_queue *ipw_rx_queue_alloc(struct ipw_priv *priv)
{
	struct ipw_rx_queue *rxq;
	int i;

	rxq = (struct ipw_rx_queue *)kmalloc(sizeof(*rxq), GFP_KERNEL);
	memset(rxq, 0, sizeof(*rxq));
	INIT_LIST_HEAD(&rxq->rx_free);
	INIT_LIST_HEAD(&rxq->rx_used);

	/* Fill the rx_used queue with _all_ of the Rx buffers */
	for (i = 0; i < RX_FREE_BUFFERS + RX_QUEUE_SIZE; i++) 
		list_add_tail(&rxq->pool[i].list, &rxq->rx_used);

	/* Set us so that we have processed and used all buffers, but have
	 * not restocked the Rx queue with fresh buffers */
	rxq->read = rxq->write = 0;
	rxq->processed = RX_QUEUE_SIZE - 1;
	rxq->free_count = 0;

	return rxq;
}

static int ipw_beacons_allocate(struct ipw_priv *priv)
{
	if (priv->beacons)
		return 0;

	IPW_DEBUG_SCAN("Allocating %d beacon slots.\n", 
		       MAX_BEACON_COUNT);

	priv->beacons = (struct ieee80211_beacon *)kmalloc(
		MAX_BEACON_COUNT * sizeof(struct ieee80211_beacon),
		GFP_KERNEL);
	if (!priv->beacons) {
		printk(KERN_WARNING "%s: Out of memory allocating beacons\n",
		       priv->net_dev->name);
		return -ENOMEM;
	}

	memset(priv->beacons, 0, 
	       MAX_BEACON_COUNT * sizeof(struct ieee80211_beacon));

	return 0;
}

static void ipw_beacons_free(struct ipw_priv *priv)
{
	if (!priv->beacons)
		return;
	kfree(priv->beacons);
	priv->beacons = NULL;
}

static void ipw_beacons_initialize(struct ipw_priv *priv)
{
	int i;

	INIT_LIST_HEAD(&priv->beacon_free_list);
	INIT_LIST_HEAD(&priv->beacon_list);
	for (i = 0; i < MAX_BEACON_COUNT; i++) 
		list_add_tail(&priv->beacons[i].list, &priv->beacon_free_list);
}

static void ipw_disassociate(void *data)
{
	struct ipw_priv *priv = data;
	int err;

	if (!(priv->status & (STATUS_ASSOCIATING | STATUS_ASSOCIATED))) {
		IPW_DEBUG_INFO("Attempt to associate while not associated.\n");
		return;
	}

	IPW_DEBUG_INFO("Disassocation attempt from " MAC_FMT " "
		       "on channel %d.\n",
		       MAC_ARG(priv->assoc_request.bssid), 
		       priv->assoc_request.channel);
	priv->assoc_request.assoc_type = HC_DISASSOCIATE;
	err = ipw_send_associate(priv, &priv->assoc_request);
	if (err) {
		IPW_DEBUG_HC("Attempt to send [dis]associate command failed.\n");
		return;
	}

	priv->status &= ~(STATUS_ASSOCIATING | STATUS_ASSOCIATED);
}

static void ipw_associate(void *data)
{
	struct ipw_priv *priv = data;

	struct ieee80211_beacon *beacon = NULL;
	struct ipw_supported_rates rates;
	int err;

	list_for_each_entry(beacon, &priv->beacon_list, list) {
		if ((priv->config & CFG_STATIC_CHANNEL) && 
		    (beacon->channel != priv->channel)) 
			continue;
		if ((priv->config & CFG_STATIC_ESSID) && 
		    memcmp(beacon->ssid, priv->essid, 
			   min(beacon->ssid_len, priv->essid_len)))
			continue;
		if ((priv->config & CFG_STATIC_BSSID) && 
		    memcmp(beacon->bssid, priv->bssid, ETH_ALEN))
			continue;
		/* Verify privacy compatability */
		if ((priv->capability & CAP_PRIVACY_ON) && 
		    !(beacon->capability & WLAN_CAPABILITY_PRIVACY))
			continue;
		if (!(priv->capability & CAP_PRIVACY_ON) && 
		    (beacon->capability & WLAN_CAPABILITY_PRIVACY))
			continue;
		break;
	}

	/* If we reached the end of the list, then we don't have any valid
	 * matching APs */
	if (&beacon->list == &priv->beacon_list) {
#ifdef CONFIG_IPW2200_DEBUG
		IPW_DEBUG_INFO("Scan completed, no valid APs matched "
			       "[CFG 0x%08lX]\n", priv->config);
		if (priv->config & CFG_STATIC_CHANNEL)
			IPW_DEBUG_INFO("Channel locked to %d\n", priv->channel);
		else
			IPW_DEBUG_INFO("Channel unlocked.\n");
		if (priv->config & CFG_STATIC_ESSID)
			IPW_DEBUG_INFO("ESSID locked to '%s'\n", escape_essid(priv->essid, priv->essid_len));
		else
			IPW_DEBUG_INFO("ESSID unlocked.\n");
		if (priv->config & CFG_STATIC_BSSID)
			IPW_DEBUG_INFO("BSSID locked to %d\n", priv->channel);
		else
			IPW_DEBUG_INFO("BSSID unlocked.\n");
		if (priv->capability & CAP_PRIVACY_ON)
			IPW_DEBUG_INFO("PRIVACY on\n");
		else
			IPW_DEBUG_INFO("PRIVACY off\n");
#endif

		queue_delayed_work(priv->workqueue, &priv->request_scan, 
				   5 * HZ);
		
		return;
	}
	
	IPW_DEBUG_INFO("Assocation attempt to '%s' on channel %d.\n",
		       escape_essid(beacon->ssid, beacon->ssid_len), 
		       beacon->channel);
	memset(&priv->assoc_request, 0, sizeof(priv->assoc_request));
	priv->assoc_request.channel = beacon->channel;
	priv->assoc_request.auth_type = AUTH_OPEN;
	priv->assoc_request.assoc_type = HC_ASSOCIATE;
	priv->assoc_request.band_mode = IPW_BAND_B; /* 802.11b for now */

	priv->assoc_request.beacon_interval = beacon->beacon_interval;
	priv->assoc_request.assoc_tsf_msw = beacon->time_stamp[1];
	priv->assoc_request.assoc_tsf_lsw = beacon->time_stamp[0];
	priv->assoc_request.capability = beacon->capability;
	priv->assoc_request.listen_interval = beacon->listen_interval;
	
	memcpy(&priv->assoc_request.bssid, beacon->bssid, ETH_ALEN);
	memcpy(&priv->assoc_request.dest, beacon->bssid, ETH_ALEN);

	err = ipw_send_ssid(priv, beacon->ssid, beacon->ssid_len);
	if (err) {
		IPW_DEBUG_HC("Attempt to send SSID command failed.\n");
		return;
	}

	/* Hard code B rates for now */
	memset(&rates, 0, sizeof(rates));
        rates.band_mode = IPW_BAND_B;
	rates.num_rates = 4;
	rates.purpose = IPW_RATE_CONNECT;
	rates.supported_rates[0] = 0x82;
	rates.supported_rates[1] = 0x84;
	rates.supported_rates[2] = 0x8b;
	rates.supported_rates[3] = 0x96;
	ipw_send_supported_rates(priv, &rates);
	
	err = ipw_send_associate(priv, &priv->assoc_request);
	if (err) {
		IPW_DEBUG_HC("Attempt to send associate command failed.\n");
		return;
	}

	priv->channel = beacon->channel;
	memcpy(priv->bssid, beacon->bssid, ETH_ALEN);
	priv->essid_len = min(beacon->ssid_len, (u8)IW_ESSID_MAX_SIZE);
	memcpy(priv->essid, beacon->ssid, priv->essid_len);

	priv->status |= STATUS_ASSOCIATING;
}


static inline void isr_beacon_add(
	struct ipw_priv *priv,
	struct ieee80211_probe_response *new_beacon,
	struct ieee80211_rx_stats *stats)
{
	struct list_head *element;
	struct ieee80211_beacon *beacon;
	struct ieee80211_info_element *info_element;

	element = priv->beacon_free_list.next;
	
	beacon = list_entry(element, struct ieee80211_beacon, list);
	
	/* Pull out fixed field data */
	memcpy(beacon->bssid, new_beacon->header.addr3, ETH_ALEN);
	beacon->capability = new_beacon->capability;
	beacon->last_scanned = priv->scans;
	beacon->time_stamp[0] = new_beacon->time_stamp[0];
	beacon->time_stamp[1] = new_beacon->time_stamp[1];
	beacon->beacon_interval = new_beacon->beacon_interval;
	beacon->listen_interval = 0x0A; /* Where to pull this? new_beacon->listen_interval;*/
/*
  if ((priv->long_preamble == 0) &&
  !(beacon->capability & WLAN_CAPABILITY_SHORT_PREAMBLE)) {
  printk(KERN_INFO "%s: Beacon received from AP (" MACSTR ") "
  "that has preable length conflict.\n",
  priv->ndev->name, MAC2STR(beacon->bssid));
  }
*/
	/* First info element is the SSID */
	info_element = &new_beacon->info_element;
	if (info_element->id != MFIE_TYPE_SSID) {
		IPW_DEBUG(IPW_DL_INFO | IPW_DL_SCAN, 
			  "invalid MFIE type for SSID: %d\n",
			  info_element->id);
		return;
	}
	beacon->ssid_len = min(info_element->len, (u8)IW_ESSID_MAX_SIZE);
	memcpy(beacon->ssid, info_element->data, beacon->ssid_len);
	if (beacon->ssid_len < IW_ESSID_MAX_SIZE)
		memset(beacon->ssid + beacon->ssid_len, 0, 
		       IW_ESSID_MAX_SIZE - beacon->ssid_len);

	/* Second info element is the RATES */
	info_element = (struct ieee80211_info_element *)
		&info_element->data[info_element->len];
	if (info_element->id != MFIE_TYPE_RATES) {
		IPW_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
			  "invalid MFIE type for RATES: %d\n",
			  info_element->id);
		return;
	}
	beacon->rates_len = min(info_element->len, MAX_RATES_LENGTH);
	memcpy(beacon->rates, info_element->data, beacon->rates_len);

	/* Third info element is either FH or DS, based on BSS capability */
	info_element = (struct ieee80211_info_element *)
		&info_element->data[info_element->len];
	if (beacon->capability & WLAN_CAPABILITY_BSS) {
		/* BSS - Direct Sequence */
		if (info_element->id != MFIE_TYPE_DS_SET) {
			IPW_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
				  ": invalid MFIE type for DIRECT SEQUENCE: %d\n",
				  info_element->id);
			return;
		}
		
		beacon->channel = info_element->data[0];
	} else if (beacon->capability & WLAN_CAPABILITY_IBSS) {
		/* IBSS - Frequency Hop */
#ifdef NOT_YET
		if (info_element->id != MFIE_TYPE_FH_SET) {
			IPW_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
				  "invalid MFIE type for FREQ HOP: %d\n",
				  info_element->id);
			return;
		}
#endif
	} else {
		/* Invalid ESS/IBSS capability! */
		IPW_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
			  "invalid ESS/IBSS capability (both 0).\n");
		return;
	}

	/* For now, we ignore the optional CF and IBSS parameter sets */

	list_del(element);

	IPW_DEBUG_SCAN("adding '%s' to beacon list.\n", beacon->ssid);
	memcpy(&beacon->stats, stats, sizeof(beacon->stats));

	list_add_tail(&beacon->list, &priv->beacon_list);
}

static void isr_process_probe_response(
	struct ipw_priv *priv,
	struct ieee80211_probe_response *new_beacon,
	struct ieee80211_rx_stats *stats)
{
	struct ieee80211_info_element *ssid_info;
	struct ieee80211_beacon *beacon;
	u8 ssid_len;

	IPW_DEBUG_SCAN(
		"\n"
		"Time Stamp 1    : %08X\n"
		"Time Stamp 2    : %08X\n"
		"Beacon Interval : %04X\n"
		"Capabilities    : %c%c%c%c-%c%c%c%c\n",
		new_beacon->time_stamp[0],
		new_beacon->time_stamp[1],
		new_beacon->beacon_interval,
		(new_beacon->capability & BIT(7)) ? '1' : '0',
		(new_beacon->capability & BIT(6)) ? '1' : '0',
		(new_beacon->capability & BIT(5)) ? '1' : '0',
		(new_beacon->capability & BIT(4)) ? '1' : '0',
		(new_beacon->capability & BIT(3)) ? '1' : '0',
		(new_beacon->capability & BIT(2)) ? '1' : '0',
		(new_beacon->capability & BIT(1)) ? '1' : '0',
		(new_beacon->capability & BIT(0)) ? '1' : '0');
	
	ssid_info = &new_beacon->info_element;
	if (ssid_info->id != MFIE_TYPE_SSID) {
		IPW_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
			  "invalid MFIE type for SSID: %d\n",
			  ssid_info->id);
		return;
	}

	/* Search for this entry in the list and nuke it if it is 
	 * already there */
	ssid_len = min(ssid_info->len, (u8)IW_ESSID_MAX_SIZE);
	list_for_each_entry(beacon, &priv->beacon_list, list) {
		if (!memcmp(beacon->bssid, new_beacon->header.addr3, 
			    ETH_ALEN)) {
			list_del(&beacon->list);
			list_add_tail(&beacon->list, &priv->beacon_free_list);
			break;
		}
	}

	if (list_empty(&priv->beacon_free_list)) {
		IPW_DEBUG_INFO("TODO: Expire oldest beacon to add new.\n");
		printk(KERN_WARNING "%s: Beacon free list is empty.\n",
		       priv->net_dev->name);
	} else 
		isr_beacon_add(priv, new_beacon, stats);

	IPW_DEBUG_SCAN("exit\n");
}
	
static void ipw_handle_mgt_packet(
	struct ipw_priv *priv, 
	struct ieee80211_hdr *header,
	struct ieee80211_rx_stats *stats)
{

	char *msg = "UNKNOWN";
	
	switch (WLAN_FC_GET_STYPE(header->frame_control)) {
	case IEEE802_11_STYPE_ASSOC_RESP:
		msg = "ASSOCIATION RESPONSE";
		break;

	case IEEE802_11_STYPE_REASSOC_RESP:
		msg = "REASSOCIATION RESPONSE";
		break;

	case IEEE802_11_STYPE_PROBE_RESP:
		msg = "PROBE RESPONSE";
		IPW_DEBUG_SCAN("Probe response\n");
		isr_process_probe_response(
			priv, (struct ieee80211_probe_response *)header, stats);
		break;

	case IEEE802_11_STYPE_BEACON:
		msg = "BEACON";
		IPW_DEBUG_SCAN("Beacon\n");
		isr_process_probe_response(
			priv, (struct ieee80211_probe_response *)header, stats);
		break;

	default:
		IPW_DEBUG_WARNING("%s: Unknown management packet: %d\n",
				  priv->net_dev->name, 
				  WLAN_FC_GET_STYPE(header->frame_control));
		break;
	}

	IPW_DEBUG_RX("received %s(%d)\n", msg, 
		     WLAN_FC_GET_STYPE(header->frame_control));
}
	
static inline void ipw_handle_data_packet(struct ipw_priv *priv, 
					      struct ipw_rx_mem_buffer *rxb,
					      struct ieee80211_rx_stats *stats)
{
	struct ipw_rx_packet *pkt = (struct ipw_rx_packet *)rxb->skb->data;

	/* We only process data packets if the 
	 * interface is open */
	if (unlikely((pkt->u.frame.length + IPW_RX_FRAME_SIZE) > 
		     skb_tailroom(rxb->skb))) {
		priv->ieee->stats.rx_errors++;
		priv->wstats.discard.misc++;
		IPW_DEBUG_DROP("Corruption detected! Oh no!\n");
		return;
	} else if (unlikely(!(priv->status & STATUS_DEV_OPEN))) {
		priv->ieee->stats.rx_errors++;
		priv->wstats.discard.misc++;
		IPW_DEBUG_DROP("Dropping packet while interface is not up.\n");
		return;
	}

	/* Advance skb->data to the start of the actual payload */
	skb_reserve(rxb->skb, (u32)&pkt->u.frame.data[0] - (u32)pkt);

	/* Set the size of the skb to the size of the frame */
	skb_put(rxb->skb, pkt->u.frame.length);

	IPW_DEBUG_RX("Rx packet of %d bytes.\n", rxb->skb->len);

	if (!ieee80211_rx(priv->ieee, rxb->skb, stats)) 
		priv->ieee->stats.rx_errors++;
	else /* ieee80211_rx succeeded, so it now owns the SKB */
		rxb->skb = NULL;
}


/*
 * Main entry function for recieving a packet with 80211 headers.  This
 * should be called when ever the FW has notified us that there is a new
 * skb in the recieve queue.
 */
static void ipw_rx(struct ipw_priv *priv)
{
	struct ipw_rx_mem_buffer *rxb;
	struct ipw_rx_packet *pkt;
	struct ieee80211_hdr *header;
	u32 r, w, i;

	r = ipw_read32(priv, CX2_RX_READ_INDEX);
	w = ipw_read32(priv, CX2_RX_WRITE_INDEX);
	i = (priv->rxq->processed + 1) % RX_QUEUE_SIZE;

	while (i != r) {
		rxb = priv->rxq->queue[i];
#ifdef CONFIG_IPW2200_DEBUG
		if (unlikely(rxb == NULL)) {
			printk(KERN_CRIT "Queue not allocated!\n");
			break;
		}
#endif
		priv->rxq->queue[i] = NULL;

		pci_dma_sync_single_for_cpu(priv->pci_dev, rxb->dma_addr,
					    CX2_RX_BUF_SIZE, 
					    PCI_DMA_FROMDEVICE);

		pkt = (struct ipw_rx_packet *)rxb->skb->data;
		IPW_DEBUG_RX("Packet: type=%02X seq=%02X bits=%02X\n",
			     pkt->header.message_type,
			     pkt->header.rx_seq_num,
			     pkt->header.control_bits);

		switch (pkt->header.message_type) {
		case RX_FRAME_TYPE: /* 802.11 frame */ {
			struct ieee80211_rx_stats stats = {
				.rssi = pkt->u.frame.rssi,
				.signal = pkt->u.frame.signal,
				.noise = pkt->u.frame.noise,
				.rate = pkt->u.frame.rate,
				.mac_time = jiffies,
			};
			
			IPW_DEBUG_RX("Frame: len=%u\n", pkt->u.frame.length);
			
			if (pkt->u.frame.length < sizeof(struct ieee80211_hdr)) {
				IPW_DEBUG_DROP("Received packet is too small. "
					       "Dropping.\n");
				priv->ieee->stats.rx_errors++;
				priv->wstats.discard.misc++;
				break;
			}
			
			header = (struct ieee80211_hdr *)(rxb->skb->data + 
							  IPW_RX_FRAME_SIZE);
			switch (WLAN_FC_GET_TYPE(header->frame_control)) {
			case IEEE802_11_FTYPE_MGMT:
				ipw_handle_mgt_packet(priv, header, &stats);
				break;
				
			case IEEE802_11_FTYPE_CTL:
				break;
				
			case IEEE802_11_FTYPE_DATA:
				ipw_handle_data_packet(priv, rxb, &stats);
				break;
			}
			break;
		}

		case RX_HOST_NOTIFICATION_TYPE: {
			IPW_DEBUG_RX("Notification: subtype=%02X flags=%02X size=%d\n",
				     pkt->u.notification.subtype,
				     pkt->u.notification.flags,
				     pkt->u.notification.size);
			ipw_rx_notification(priv, &pkt->u.notification);
			break;
		}

		default:
			IPW_DEBUG_RX("Bad Rx packet of type %d\n",
				     pkt->header.message_type);
			break;
		}
		
		/* For now we just don't re-use anything.  We can tweak this 
		 * later to try and re-use notification packets and SKBs that 
		 * fail to Rx correctly */
		if (rxb->skb != NULL) {
			dev_kfree_skb_any(rxb->skb);
			rxb->skb = NULL;
		}
		
		pci_unmap_single(priv->pci_dev, rxb->dma_addr,
				 CX2_RX_BUF_SIZE, PCI_DMA_FROMDEVICE);
		list_add_tail(&rxb->list, &priv->rxq->rx_used);
		
		i = (i + 1) % RX_QUEUE_SIZE;
	}

	/* Backtrack one entry */
	priv->rxq->processed = (i ? i : RX_QUEUE_SIZE) - 1;

	ipw_rx_queue_restock(priv);
}

inline int ipw_abort_scan(struct ipw_priv *priv)
{
	int err;

	if (priv->status & STATUS_SCAN_ABORTING) {
		IPW_DEBUG_HC("Ignoring concurrent scan abort request.\n");
		return 0;
	}

	err = ipw_send_scan_abort(priv);
	if (err) {
		IPW_DEBUG_HC("Request to abort scan failed.\n");
		return err;
	}
	
	priv->status |= STATUS_SCAN_ABORTING;

	return 0;
}

static int ipw_request_scan(struct ipw_priv *priv)
{
	struct ipw_scan_request scan;
	int i, err;

	if (priv->status & STATUS_EXIT_PENDING) {
		IPW_DEBUG_SCAN("Aborting scan due to device shutdown\n");
		priv->status |= STATUS_SCAN_PENDING;
		return 0;
	}

	if (priv->status & STATUS_SCANNING) {
		IPW_DEBUG_HC("Concurrent scan requested.  Aborting first.\n");
		priv->status |= STATUS_SCAN_PENDING;
		ipw_abort_scan(priv);
		return 0;
	}
	
	if (priv->status & STATUS_SCAN_ABORTING) {
		IPW_DEBUG_HC("Scan request while abort pending.  Queuing.\n");
		priv->status |= STATUS_SCAN_PENDING;
		return 0;
	}

	memset(&scan, 0, sizeof(scan));

	scan.scan_type = 3;
	scan.dwell_time = 20;
	for (i = 0; i < 14; i++)
		scan.channels_list[i] = i + 1;

	err = ipw_send_scan_request(priv, &scan);
	if (err) {
		IPW_DEBUG_HC("Sending scan command failed: %08X\n",
			     err);
		return -EIO;
	}

	priv->status |= STATUS_SCANNING;
	priv->status &= ~STATUS_SCAN_PENDING;

	return 0;
}

/*
 * This file defines the Wireless Extension handlers.  It does not
 * define any methods of hardware manipulation and relies on the
 * functions defined in ipw_main to provide the HW interaction.
 * 
 * The exception to this is the use of the ipw_get_ordinal() 
 * function used to poll the hardware vs. making unecessary calls.
 *
 */
static inline char *ipw_translate_scan(char *start, char *stop, 
				       struct ieee80211_beacon *beacon)
{
	struct iw_event iwe;
	int i;

	/* First entry *MUST* be the AP MAC address */
	iwe.cmd = SIOCGIWAP;
	iwe.u.ap_addr.sa_family = ARPHRD_ETHER; 
	memcpy(iwe.u.ap_addr.sa_data, beacon->bssid, ETH_ALEN);
	start = iwe_stream_add_event(start, stop, &iwe, IW_EV_ADDR_LEN);
	
	/* Remaining entries will be displayed in the order we provide them */

	/* Add the ESSID */
        iwe.u.data.length = beacon->ssid_len;
        if (iwe.u.data.length > 32)
		iwe.u.data.length = 32;
        iwe.cmd = SIOCGIWESSID;
        iwe.u.data.flags = 1;
        start = iwe_stream_add_point(start, stop, &iwe, beacon->ssid);

        /* Add mode */
        iwe.cmd = SIOCGIWMODE;
        if (beacon->capability & 
	    (WLAN_CAPABILITY_BSS | WLAN_CAPABILITY_IBSS)) {
		if (beacon->capability & WLAN_CAPABILITY_BSS)
			iwe.u.mode = IW_MODE_MASTER;
		else
			iwe.u.mode = IW_MODE_ADHOC;

		start = iwe_stream_add_event(start, stop, &iwe, 
					     IW_EV_UINT_LEN);
	}
	
        /* Add frequency */
	iwe.cmd = SIOCGIWFREQ;
	iwe.u.freq.m = beacon->channel;
	iwe.u.freq.e = 0;
	start = iwe_stream_add_event(start, stop, &iwe, IW_EV_FREQ_LEN);

	/* Add rates */
	iwe.cmd = SIOCGIWRATE;
	iwe.u.bitrate.fixed = iwe.u.bitrate.disabled = 0;
	for (i = 0; i < beacon->rates_len; i++) {
		iwe.u.bitrate.value = (beacon->rates[i] & 0x7F) * 500000;
		start = iwe_stream_add_event(start, stop, &iwe, 
					     IW_EV_PARAM_LEN);
	}

	/* Add quality statistics */
	/* TODO: Fix these values... */
	iwe.cmd = IWEVQUAL;
	iwe.u.qual.qual = beacon->stats.rssi;
	iwe.u.qual.level = beacon->stats.signal;
	iwe.u.qual.noise = beacon->stats.noise;
	iwe.u.qual.updated = 7;	

	start = iwe_stream_add_event(start, stop, &iwe, IW_EV_QUAL_LEN);

	/* Add encryption capability */
	iwe.cmd = SIOCGIWENCODE;
	if (beacon->capability & WLAN_CAPABILITY_PRIVACY)
		iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
	else
		iwe.u.data.flags = IW_ENCODE_DISABLED;
	iwe.u.data.length = 0;
	start = iwe_stream_add_point(start, stop, &iwe, beacon->ssid);

	return start;
}

static int ipw_wx_get_name(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{
	strcpy(wrqu->name, "IEEE 802.11bg");
	IPW_DEBUG_WX("Name: %s\n", wrqu->name);
	return 0;
}

static int ipw_set_channel(struct ipw_priv *priv, u8 channel)
{
	if (channel == 0) {
		IPW_DEBUG_INFO("Setting channel to ANY (0)\n");
		priv->config &= ~CFG_STATIC_CHANNEL;
		if (!(priv->status & (STATUS_SCANNING | STATUS_ASSOCIATED |
				      STATUS_ASSOCIATING))) {
			IPW_DEBUG_INFO("Attempting to associate with new "
				       "parameters.\n");
			ipw_associate(priv);
		}

		return 0;
	}

	priv->config |= CFG_STATIC_CHANNEL;

	if (priv->channel == channel) {
		IPW_DEBUG_INFO("Request to set channel to current value (%d)\n",
			       channel);
		return 0;
	}

	IPW_DEBUG_INFO("Setting channel to %i\n", (int)channel);
	priv->channel = channel;

	/* If we are currently associated, or trying to associate
	 * then see if this is a new channel (causing us to disassociate) */
	if (priv->status & (STATUS_ASSOCIATED | STATUS_ASSOCIATING)) {
		IPW_DEBUG_INFO("Disassociating due to channel change.\n");
		ipw_disassociate(priv);
	}


	return 0;
}

static int ipw_wx_set_freq(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra) 
{
	struct ipw_priv *priv = dev->priv;
	struct iw_freq *fwrq = &wrqu->freq;
	
	/* if setting by freq convert to channel */
	if (fwrq->e == 1) {
		if ((fwrq->m >= (int) 2.412e8 &&
		     fwrq->m <= (int) 2.487e8)) {
			int f = fwrq->m / 100000;
			int c = 0;
			
			while ((c < REG_MAX_CHANNEL) &&
			       (f != ipw_frequencies[c]))
				c++;
			
			/* hack to fall through */
			fwrq->e = 0;
			fwrq->m = c + 1;
		}
	}
	
	if (fwrq->e > 0 || fwrq->m > 1000) 
		return -EOPNOTSUPP;

	IPW_DEBUG_WX("SET Freq/Channel -> %d \n", fwrq->m);
	return ipw_set_channel(priv, (u8)fwrq->m);
	
	return 0;
}


static int ipw_wx_get_freq(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{
	struct ipw_priv *priv = ipw_netdev(dev);

	wrqu->freq.e = 0;

	/* If we are associated, trying to associate, or have a statically
	 * configured CHANNEL then return that; otherwise return ANY */
	if (priv->config & CFG_STATIC_CHANNEL ||
	    priv->status & (STATUS_ASSOCIATING | STATUS_ASSOCIATED))
		wrqu->freq.m = priv->channel;
	else 
		wrqu->freq.m = 0;

	IPW_DEBUG_WX("GET Freq/Channel -> %d \n", priv->channel);
	return 0;
}

static int ipw_wx_set_mode(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP;
}

static int ipw_wx_get_mode(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);

	/* We currently only support managed mode */
	wrqu->mode = IW_MODE_INFRA;
	return 0;
}


#define DEFAULT_FRAG_THRESHOLD 2342U
#define MIN_FRAG_THRESHOLD     256U
#define	MAX_FRAG_THRESHOLD     2342U
#define DEFAULT_RTS_THRESHOLD     2304U
#define DEFAULT_BEACON_INTERVAL   100U
#define	DEFAULT_SHORT_RETRY_LIMIT 7U
#define	DEFAULT_LONG_RETRY_LIMIT  4U

static int ipw_wx_get_range(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{
	struct iw_range *range = (struct iw_range *)extra;
	u16 val;
	int i;

	wrqu->data.length = sizeof(*range);
	memset(range, 0, sizeof(*range));

	/* Let's try to keep this struct in the same order as in
	 * linux/include/wireless.h
	 */
	
	/* TODO: See what values we can set, and remove the ones we can't
	 * set, or fill them with some default data.
	 */

	/* 54Mbs == ~27 Mb/s real (802.11g) */
	range->throughput = 27 * 1000 * 1000;     

	range->max_qual.qual = 100;
	/* TODO: Find real max RSSI and stick here */
	range->max_qual.level = 0;
	range->max_qual.noise = 0;
	range->max_qual.updated = 7; /* Updated all three */

	range->avg_qual.qual = 70; /* > 8% missed beacons is 'bad' */
	/* TODO: Find real 'good' to 'bad' threshol value for RSSI */
	range->avg_qual.level = 0; /* FIXME to real average level */
	range->avg_qual.noise = 0;
	range->avg_qual.updated = 7; /* Updated all three */

	range->num_bitrates = RATE_COUNT;

	for (i = 0; i < RATE_COUNT && i < IW_MAX_BITRATES; i++) {
		range->bitrate[i] = ipw_rates_11b[i];
	}
	
	range->max_rts = DEFAULT_RTS_THRESHOLD;
	range->min_frag = MIN_FRAG_THRESHOLD;
	range->max_frag = MAX_FRAG_THRESHOLD;

	range->encoding_size[0] = 5;
	range->encoding_size[1] = 13;           /* Different token sizes */
	range->num_encoding_sizes = 2;		/* Number of entry in the list */
	range->max_encoding_tokens = WEP_KEYS;  /* Max number of tokens */

	/* Set the Wireless Extension versions */
	range->we_version_compiled = WIRELESS_EXT;
	range->we_version_source = 16;

        range->num_channels = FREQ_COUNT;

	val = 0;
	for (i = 0; i < FREQ_COUNT; i++) {
		range->freq[val].i = i + 1;
		range->freq[val].m = ipw_frequencies[i] * 100000;
		range->freq[val].e = 1;
		val++;

		if (val == IW_MAX_FREQUENCIES)
			break;
	}
	range->num_frequency = val;

	IPW_DEBUG_WX("GET Range\n");
	return 0;
}

static int ipw_wx_set_wap(struct net_device *dev, 
			  struct iw_request_info *info, 
			  union iwreq_data *wrqu, char *extra)
{
	struct ipw_priv *priv = ipw_netdev(dev);

	static const unsigned char any[] = {
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	};
	static const unsigned char off[] = {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

	if (wrqu->ap_addr.sa_family != ARPHRD_ETHER) 
		return -EINVAL;

	if (!memcmp(any, wrqu->ap_addr.sa_data, ETH_ALEN) ||
	    !memcmp(off, wrqu->ap_addr.sa_data, ETH_ALEN)) {
		/* we disable mandatory BSSID association */
		IPW_DEBUG_WX("Setting AP BSSID to ANY\n");
		priv->config &= ~CFG_STATIC_BSSID;
		if (!(priv->status & (STATUS_SCANNING | STATUS_ASSOCIATED |
				      STATUS_ASSOCIATING))) {
			IPW_DEBUG_INFO("Attempting to associate with new "
				       "parameters.\n");
			ipw_associate(priv);
		}

		return 0;
	}

	priv->config |= CFG_STATIC_BSSID;
	if (!memcmp(priv->bssid, wrqu->ap_addr.sa_data, ETH_ALEN)) {
		IPW_DEBUG_WX("BSSID set to current BSSID.\n");
		return 0;
	}

	IPW_DEBUG_WX("Setting mandatory BSSID to " MAC_FMT "\n",
		     MAC_ARG(wrqu->ap_addr.sa_data));

	memcpy(priv->bssid, wrqu->ap_addr.sa_data, ETH_ALEN);

	/* If we are currently associated, or trying to associate
	 * then see if this is a new BSSID (causing us to disassociate) */
	if (priv->status & (STATUS_ASSOCIATED | STATUS_ASSOCIATING)) {
		IPW_DEBUG_INFO("Disassociating due to mandatory BSSID change.\n");
		ipw_disassociate(priv);
	}

	return 0;
}

static int ipw_wx_get_wap(struct net_device *dev, 
			  struct iw_request_info *info, 
			  union iwreq_data *wrqu, char *extra)
{
	struct ipw_priv *priv = ipw_netdev(dev);
	/* If we are associated, trying to associate, or have a statically
	 * configured BSSID then return that; otherwise return ANY */
	if (priv->config & CFG_STATIC_BSSID || 
	    priv->status & (STATUS_ASSOCIATED | STATUS_ASSOCIATING)) 
		memcpy(wrqu->ap_addr.sa_data, &priv->bssid, ETH_ALEN);
	else
		memset(wrqu->ap_addr.sa_data, 0, ETH_ALEN);

	IPW_DEBUG_WX("Getting WAP BSSID: " MAC_FMT "\n",
		     MAC_ARG(wrqu->ap_addr.sa_data));

	return 0;
}

static int ipw_wx_set_essid(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{
	struct ipw_priv *priv = dev->priv;
	char *essid = ""; /* ANY */
	int length = 0;
  
	if (wrqu->essid.flags && wrqu->essid.length) {
		length = wrqu->essid.length - 1;
		essid = extra;
	}

	if (length == 0) {
		IPW_DEBUG_WX("Setting ESSID to ANY\n");
		priv->config &= ~CFG_STATIC_ESSID;
		if (!(priv->status & (STATUS_SCANNING | STATUS_ASSOCIATED |
				      STATUS_ASSOCIATING))) {
			IPW_DEBUG_INFO("Attempting to associate with new "
				       "parameters.\n");
			ipw_associate(priv);
		}

		return 0;
	}

	length = min(length, IW_ESSID_MAX_SIZE);

	priv->config |= CFG_STATIC_ESSID;

	if (priv->essid_len == length && !memcmp(priv->essid, extra, length)) {
		IPW_DEBUG_WX("ESSID set to current ESSID.\n");
		return 0;
	}

	IPW_DEBUG_WX("Setting ESSID: '%s' (%d)\n", escape_essid(essid, length),
		     length);

	priv->essid_len = length;
	memcpy(priv->essid, essid, priv->essid_len);
	
	/* If we are currently associated, or trying to associate
	 * then see if this is a new ESSID (causing us to disassociate) */
	if (priv->status & (STATUS_ASSOCIATED | STATUS_ASSOCIATING)) {
		IPW_DEBUG_INFO("Disassociating due to ESSID change.\n");
		ipw_disassociate(priv);
	}

	return 0;
}

static int ipw_wx_get_essid(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{
	struct ipw_priv *priv = (struct ipw_priv *)dev->priv;

	/* If we are associated, trying to associate, or have a statically
	 * configured ESSID then return that; otherwise return ANY */
	if (priv->config & CFG_STATIC_ESSID ||
	    priv->status & (STATUS_ASSOCIATED | STATUS_ASSOCIATING)) { 
		IPW_DEBUG_WX("Getting essid: '%s'\n", 
			     escape_essid(priv->essid, priv->essid_len));
		memcpy(extra, priv->essid, priv->essid_len); 
		wrqu->essid.length = priv->essid_len;
		wrqu->essid.flags = 1; /* active */
	} else {
		IPW_DEBUG_WX("Getting essid: ANY\n");
		wrqu->essid.length = 0;
		wrqu->essid.flags = 0; /* active */
	}

	return 0;
}

static int ipw_wx_set_nick(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{ 
	struct ipw_priv *priv = dev->priv;

	IPW_DEBUG_WX("Setting nick to '%s'\n", extra);
	if (wrqu->data.length > IW_ESSID_MAX_SIZE)
		return -E2BIG;

	wrqu->data.length = min((size_t)wrqu->data.length, sizeof(priv->nick));
	memset(priv->nick, 0, sizeof(priv->nick));
	memcpy(priv->nick, extra,  wrqu->data.length);
	IPW_DEBUG_TRACE("<<\n");
	return 0;

}


static int ipw_wx_get_nick(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{ 
	struct ipw_priv *priv = dev->priv;
	IPW_DEBUG_WX("Getting nick\n");
	wrqu->data.length = strlen(priv->nick) + 1;
	memcpy(extra, priv->nick, wrqu->data.length);
	wrqu->data.flags = 1; /* active */
	return 0;
}


static int ipw_wx_set_rate(struct net_device *dev,
			   struct iw_request_info *info,
			   union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}



static int ipw_wx_get_rate(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP;  
}


static int ipw_wx_set_rts(struct net_device *dev, 
			  struct iw_request_info *info, 
			  union iwreq_data *wrqu, char *extra)
{ 
	struct ipw_priv *priv = dev->priv;
	int err;

	if (wrqu->rts.value < 1 || 
	    wrqu->rts.value > 2304)
		return -EINVAL;

	if (wrqu->rts.disabled) {
		IPW_DEBUG_WX("SET RTS Threshold -> disabled\n");
		priv->config &= CFG_RTS;
		return 0;
	}

	priv->config |= CFG_RTS;
	priv->rts_threshold = wrqu->rts.value;
	err = ipw_send_rts_threshold(priv, wrqu->rts.value);
	if (err)
		return err;

	IPW_DEBUG_WX("SET RTS Threshold -> %d \n", wrqu->rts.value);
	return 0;
}


static int ipw_wx_get_rts(struct net_device *dev, 
			  struct iw_request_info *info, 
			  union iwreq_data *wrqu, char *extra)
{ 
	struct ipw_priv *priv = dev->priv;

	if (!(priv->config & CFG_RTS)) {
		IPW_DEBUG_WX("GET RTS Threshold -> disabled\n");
		wrqu->rts.disabled = 1;
		return 0;
	}

	/* We don't query the HW since we cache all configuration flags
	 * and use those to init the HW */
	wrqu->rts.value = priv->rts_threshold;
	wrqu->rts.fixed = 1; /* no auto select */
	wrqu->rts.disabled = 0; /* can't be disabled */

	IPW_DEBUG_WX("GET RTS Threshold -> %d \n", wrqu->rts.value);

	return 0;
}


static int ipw_wx_set_txpow(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return 0; 
}


static int ipw_wx_get_txpow(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}


static int ipw_wx_set_frag(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}


static int ipw_wx_get_frag(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}


static int ipw_wx_set_retry(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}


static int ipw_wx_get_retry(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}


static int ipw_wx_set_scan(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{
	struct ipw_priv *priv = dev->priv;
	IPW_DEBUG_WX("Start scan\n");
	if (ipw_request_scan(priv))
		return -EIO;
	return 0;
}

static int ipw_wx_get_scan(struct net_device *dev, 
			   struct iw_request_info *info, 
			   union iwreq_data *wrqu, char *extra)
{ 
	struct ipw_priv *priv = dev->priv;
	struct ieee80211_beacon *beacon;
	unsigned long flags;

	char *ev = extra;
	char *stop = ev + IW_SCAN_MAX_DATA;
	int i = 0;

	IPW_DEBUG_WX("Getting scan\n");
	
	spin_lock_irqsave(&priv->lock, flags);

	list_for_each_entry(beacon, &priv->beacon_list, list) {
		i++;
		if (priv->scan_age == 0 ||
		    beacon->last_scanned + priv->scan_age >= priv->scans) 
			ev = ipw_translate_scan(ev, stop, beacon);
	}

	spin_unlock_irqrestore(&priv->lock, flags);

	wrqu->data.length = ev -  extra;
	wrqu->data.flags = 0;
	
	IPW_DEBUG_WX("exit: %d beacons returned.\n", i);

	return 0;
}

static int ipw_wx_set_encode(struct net_device *dev, 
				 struct iw_request_info *info, 
				 union iwreq_data *wrqu, char *key)
{
	struct ipw_priv *priv = ipw_netdev(dev);
	return ieee80211_wx_set_encode(priv->ieee, info, wrqu, key);
}

static int ipw_wx_get_encode(struct net_device *dev, 
				 struct iw_request_info *info, 
				 union iwreq_data *wrqu, char *key)
{
	struct ipw_priv *priv = ipw_netdev(dev);
	return ieee80211_wx_get_encode(priv->ieee, info, wrqu, key);
}

static int ipw_wx_set_power(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}

static int ipw_wx_get_power(struct net_device *dev, 
			    struct iw_request_info *info, 
			    union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}

#ifdef CONFIG_IPW_PROMISC
static int ipw_wx_set_promisc(struct net_device *dev, 
			      struct iw_request_info *info, 
			      union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}


static int ipw_wx_reset(struct net_device *dev, 
			struct iw_request_info *info, 
			union iwreq_data *wrqu, char *extra)
{ 
	IPW_DEBUG_WX("0x%p, 0x%p, 0x%p\n", dev, info, wrqu);
	return -EOPNOTSUPP; 
}
#endif // CONFIG_IPW_PROMISC

static iw_handler ipw_wx_handlers[] =
{
        NULL,                     /* SIOCSIWCOMMIT */
        ipw_wx_get_name,      /* SIOCGIWNAME */
        NULL,                     /* SIOCSIWNWID */
        NULL,                     /* SIOCGIWNWID */
        ipw_wx_set_freq,      /* SIOCSIWFREQ */
        ipw_wx_get_freq,      /* SIOCGIWFREQ */
        ipw_wx_set_mode,      /* SIOCSIWMODE */
        ipw_wx_get_mode,      /* SIOCGIWMODE */
        NULL,                     /* SIOCSIWSENS */
        NULL,                     /* SIOCGIWSENS */
        NULL,                     /* SIOCSIWRANGE */
        ipw_wx_get_range,     /* SIOCGIWRANGE */
        NULL,                     /* SIOCSIWPRIV */
        NULL,                     /* SIOCGIWPRIV */
        NULL,                     /* SIOCSIWSTATS */
        NULL,                     /* SIOCGIWSTATS */
        NULL,                     /* SIOCSIWSPY */
        NULL,                     /* SIOCGIWSPY */
        NULL,                     /* SIOCGIWTHRSPY */
        NULL,                     /* SIOCWIWTHRSPY */
        ipw_wx_set_wap,       /* SIOCSIWAP */
        ipw_wx_get_wap,       /* SIOCGIWAP */
        NULL,                     /* -- hole -- */
        NULL,                     /* SIOCGIWAPLIST -- depricated */
        ipw_wx_set_scan,      /* SIOCSIWSCAN */
        ipw_wx_get_scan,      /* SIOCGIWSCAN */
        ipw_wx_set_essid,     /* SIOCSIWESSID */
        ipw_wx_get_essid,     /* SIOCGIWESSID */
        ipw_wx_set_nick,      /* SIOCSIWNICKN */
        ipw_wx_get_nick,      /* SIOCGIWNICKN */
        NULL,                     /* -- hole -- */
        NULL,                     /* -- hole -- */
        ipw_wx_set_rate,      /* SIOCSIWRATE */
        ipw_wx_get_rate,      /* SIOCGIWRATE */
        ipw_wx_set_rts,       /* SIOCSIWRTS */
        ipw_wx_get_rts,       /* SIOCGIWRTS */
        ipw_wx_set_frag,      /* SIOCSIWFRAG */
        ipw_wx_get_frag,      /* SIOCGIWFRAG */
        ipw_wx_set_txpow,     /* SIOCSIWTXPOW */
        ipw_wx_get_txpow,     /* SIOCGIWTXPOW */
        ipw_wx_set_retry,     /* SIOCSIWRETRY */
        ipw_wx_get_retry,     /* SIOCGIWRETRY */
        ipw_wx_set_encode,    /* SIOCSIWENCODE */
        ipw_wx_get_encode,    /* SIOCGIWENCODE */
        ipw_wx_set_power,     /* SIOCSIWPOWER */
        ipw_wx_get_power,     /* SIOCGIWPOWER */
};

#ifdef CONFIG_IPW_PROMISC

static const struct iw_priv_args ipw_priv_args[] = { 
	{
		SIOCIWFIRSTPRIV + 0x0, 
		IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 2, 0, "monitor" 
	}, 
	{
		SIOCIWFIRSTPRIV + 0x1, 
		IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 0, 0, "reset" 
	},
};

static iw_handler ipw_priv_handler[] = {
	ipw_wx_set_promisc,  /* SIOCIWFIRSTPRIV */
	ipw_wx_reset,        /* SIOCIWFIRSTPRIV + 1*/ 
};

#endif /* CONFIG_IPW_PROMISC */

struct iw_handler_def ipw_wx_handler_def = 
{
	.standard = ipw_wx_handlers,
	.num_standard = sizeof(ipw_wx_handlers) / sizeof(iw_handler),
#ifdef CONFIG_IPW_PROMISC
	.num_private = sizeof(ipw_priv_handler) / sizeof(iw_handler),
 	.num_private_args = sizeof(ipw_priv_args) /
	sizeof(struct iw_priv_args),
	.private = (iw_handler *)ipw_priv_handler,
	.private_args = (struct iw_priv_args *)ipw_priv_args,	
#else
	.num_private = 0,
	.num_private_args = 0,
	.private = NULL,
 	.private_args = NULL,
#endif /* CONFIG_IPW_PROMISC */
};

/*
 * Get wireless statistics.
 * Called by /proc/net/wireless
 * Also called by SIOCGIWSTATS
 */
struct iw_statistics *ipw_get_wireless_stats(struct net_device * dev)
{
	enum {
		POOR = 30,
		FAIR = 60,
		GOOD = 80,
		VERY_GOOD = 90,
		EXCELLENT = 95,
		PERFECT = 100
	};
	int rssi_qual;
	int tx_qual;
	int beacon_qual;

	struct ipw_priv *priv = ipw_netdev(dev);
	struct iw_statistics *wstats;
	u32 rssi, quality, tx_retries, missed_beacons, tx_failures;
	u32 ord_len = sizeof(u32);

	if (!priv)
		return (struct iw_statistics *) NULL;

	wstats = &priv->wstats;

	if (ipw_get_ordinal(priv, IPW_ORD_STAT_MISSED_BEACONS,
			    &missed_beacons, &ord_len))
		goto fail_get_ordinal;

        /* If we don't have a connection the quality and level is 0*/
	if (!(priv->status & STATUS_ASSOCIATED)) {
		wstats->qual.qual = 0; 
		wstats->qual.level = 0;
	} else {
		if (ipw_get_ordinal(priv, IPW_ORD_STAT_CURR_RSSI_RAW,
				    &rssi, &ord_len))
			goto fail_get_ordinal;
		wstats->qual.level = rssi;
		if (rssi < 10) 
			rssi_qual = rssi * POOR / 10;
		else if (rssi < 15)
			rssi_qual = (rssi - 10) * (FAIR - POOR) / 5 + POOR;
		else if (rssi < 20)
			rssi_qual = (rssi - 15) * (GOOD - FAIR) / 5 + FAIR;
		else if (rssi < 30)
			rssi_qual = (rssi - 20) * (VERY_GOOD - GOOD) / 
				10 + GOOD;
		else
			rssi_qual = (rssi - 30) * (PERFECT - VERY_GOOD) / 
				10 + VERY_GOOD;
		
		if (ipw_get_ordinal(priv, IPW_ORD_STAT_TX_RETRY,
				    &tx_retries, &ord_len))
			goto fail_get_ordinal;
		
		if (tx_retries > 75) 
			tx_qual = (90 - tx_retries) * POOR / 15;
		else if (tx_retries > 70)
			tx_qual = (75 - tx_retries) * (FAIR - POOR) / 5 + POOR;
		else if (tx_retries > 65)
			tx_qual = (70 - tx_retries) * (GOOD - FAIR) / 5 + FAIR;
		else if (tx_retries > 50)
			tx_qual = (65 - tx_retries) * (VERY_GOOD - GOOD) / 
				15 + GOOD;
		else
			tx_qual = (50 - tx_retries) * 
				(PERFECT - VERY_GOOD) / 50 + VERY_GOOD;
		
		if (missed_beacons > 50) 
			beacon_qual = (60 - missed_beacons) * POOR / 10;
		else if (missed_beacons > 40)
			beacon_qual = (50 - missed_beacons) * (FAIR - POOR) / 
				10 + POOR;
		else if (missed_beacons > 32)
			beacon_qual = (40 - missed_beacons) * (GOOD - FAIR) / 
				18 + FAIR;
		else if (missed_beacons > 20)
			beacon_qual = (32 - missed_beacons) * 
				(VERY_GOOD - GOOD) / 20 + GOOD;
		else
			beacon_qual = (20 - missed_beacons) * 
				(PERFECT - VERY_GOOD) / 20 + VERY_GOOD;
		
		quality = min(beacon_qual, min(tx_qual, rssi_qual));

#ifdef CONFIG_IPW2200_DEBUG
		if (beacon_qual == quality)
			IPW_DEBUG_WX("Quality clamped by Missed Beacons\n");
		else if (tx_qual == quality)
			IPW_DEBUG_WX("Quality clamped by Tx Retries\n");
		else if (quality != 100)
			IPW_DEBUG_WX("Quality clamped by Signal Strength\n");
		else
			IPW_DEBUG_WX("Quality not clamped.\n");
#endif

		wstats->qual.qual = quality;
		wstats->qual.level = rssi;
	}
	
	wstats->qual.noise = IPW_RSSI_TO_DBM;
	wstats->qual.updated = 7;	

        /* FIXME: this is percent and not a # */
	wstats->miss.beacon = missed_beacons; 
	
	if (ipw_get_ordinal(priv, IPW_ORD_STAT_TX_FAILURE, 
			    &tx_failures, &ord_len))
		goto fail_get_ordinal;
	wstats->discard.retries = tx_failures;
	
	return wstats;

 fail_get_ordinal:
	IPW_DEBUG_WX("failed querying ordinals.\n");

	return NULL;
}


/* net device stuff */

static inline void init_sys_config(struct ipw_sys_config *sys_config)
{
        memset(sys_config, 0, sizeof(struct ipw_sys_config));
	sys_config->bt_coexistence = 1; /* We may need to look into prvStaBtConfig */
	sys_config->answer_broadcast_ssid_probe = 0;
	sys_config->accept_all_data_frames = 0;
	sys_config->accept_non_directed_frames = 1;
	sys_config->exclude_unicast_unencrypted = 0;
	sys_config->disable_unicast_decryption = 1;
	sys_config->exclude_multicast_unencrypted = 0;
	sys_config->disable_multicast_decryption = 1;
	sys_config->antenna_diversity = CFG_SYS_ANTENNA_BOTH;
	sys_config->pass_crc_to_host = 0; /* TODO: See if 1 gives us FCS */
	sys_config->dot11g_auto_detection = 0;
	sys_config->enable_cts_to_self = 0; 
	sys_config->bt_coexist_collision_thr = 0;
}

static void add_cck_scan_rates(struct ipw_supported_rates *rates,
			       u8 modulation, u32 rate_mask)
{
	u8 basic_ratemask = (IPW_OFDM_MODULATION == modulation) ? IPW_BASIC_RATE_MASK : 0;
  
	if (rate_mask & IPW_CCK_RATE_1MB_MASK)
		rates->supported_rates[rates->num_rates++] = (IPW_BASIC_RATE_MASK | IPW_CCK_RATE_1MB);
	if (rate_mask & IPW_CCK_RATE_2MB_MASK)
		rates->supported_rates[rates->num_rates++] = (IPW_BASIC_RATE_MASK | IPW_CCK_RATE_2MB);
	if (rate_mask & IPW_CCK_RATE_5MB_MASK)
		rates->supported_rates[rates->num_rates++] = (basic_ratemask | IPW_CCK_RATE_5MB);
	if (rate_mask & IPW_CCK_RATE_11MB_MASK)
		rates->supported_rates[rates->num_rates++] = (basic_ratemask | IPW_CCK_RATE_11MB);
}

static void add_ofdm_scan_rates(struct ipw_supported_rates *rates,
				u8 modulation, u32 rate_mask)
{
	u8 basic_ratemask = (IPW_OFDM_MODULATION == modulation) ? IPW_BASIC_RATE_MASK : 0;

	if (rate_mask & IPW_OFDM_RATE_6MB_MASK)
		rates->supported_rates[rates->num_rates++] = (basic_ratemask | IPW_OFDM_RATE_6MB);
	if (rate_mask & IPW_OFDM_RATE_9MB_MASK)
		rates->supported_rates[rates->num_rates++] = IPW_OFDM_RATE_9MB;
	if (rate_mask & IPW_OFDM_RATE_12MB_MASK)
		rates->supported_rates[rates->num_rates++] = (basic_ratemask | IPW_OFDM_RATE_12MB);
	if (rate_mask & IPW_OFDM_RATE_18MB_MASK)
		rates->supported_rates[rates->num_rates++] = IPW_OFDM_RATE_18MB;
	if (rate_mask & IPW_OFDM_RATE_24MB_MASK)
		rates->supported_rates[rates->num_rates++] = (basic_ratemask | IPW_OFDM_RATE_24MB);
	if (rate_mask & IPW_OFDM_RATE_36MB_MASK)
		rates->supported_rates[rates->num_rates++] = (IPW_OFDM_RATE_36MB);
	if (rate_mask & IPW_OFDM_RATE_48MB_MASK)
		rates->supported_rates[rates->num_rates++] = IPW_OFDM_RATE_48MB;
	if (rate_mask & IPW_OFDM_RATE_54MB_MASK)
		rates->supported_rates[rates->num_rates++] = IPW_OFDM_RATE_54MB;
}

static int ipw_net_open(struct net_device *dev)
{
	struct ipw_priv *priv = dev->priv;
	IPW_DEBUG_INFO("dev->open\n");
	priv->status |= STATUS_DEV_OPEN;
	/* we should be verifying the device is ready to be opened */
	netif_start_queue(dev);
	return 0;
}

static int ipw_net_stop(struct net_device *dev)
{
	struct ipw_priv *priv = dev->priv;
	IPW_DEBUG_INFO("dev->close\n");
	priv->status &= ~STATUS_DEV_OPEN;
	netif_stop_queue(dev);
	return 0;
}

static int ipw_net_hard_start_xmit(struct sk_buff *skb,
				   struct net_device *dev)
{
	struct ipw_priv *priv = dev->priv;
	unsigned long flags;

	IPW_DEBUG_TX("dev->xmit(%d bytes)\n", skb->len);

	spin_lock_irqsave(&priv->lock, flags);

	if (!(priv->status & STATUS_ASSOCIATED)) {
		IPW_DEBUG_INFO("Tx attempt while not associated.\n");
		priv->ieee->stats.tx_carrier_errors++;
		goto fail_unlock;
	}

	ipw_tx_skb(priv, skb);

	spin_unlock_irqrestore(&priv->lock, flags);
	return 0;

 fail_unlock:
	netif_stop_queue(dev);
	spin_unlock_irqrestore(&priv->lock, flags);
	return 1;
}

static struct net_device_stats *ipw_net_get_stats(struct net_device *dev)
{
	struct ipw_priv *priv = dev->priv;
	
	return &priv->ieee->stats;
}

static void ipw_net_set_multicast_list(struct net_device *dev)
{

}

static int ipw_net_set_mac_address(struct net_device *dev, void *p)
{
	struct ipw_priv *priv = dev->priv;
	struct sockaddr *addr = p;
	if (!is_valid_ether_addr(addr->sa_data))
		return -EADDRNOTAVAIL;
	priv->config |= CFG_CUSTOM_MAC;
	memcpy(priv->net_dev->dev_addr, addr->sa_data, ETH_ALEN);
	return ipw_send_adapter_address(priv, priv->net_dev->dev_addr);
}

static int ipw_net_change_mtu(struct net_device *dev, int new_mtu)
{
	if (new_mtu < ETH_ZLEN || new_mtu > ETH_DATA_LEN)
		return -EINVAL;
	dev->mtu = new_mtu;
	return 0;
}

static int ipw_net_do_ioctl(struct net_device *dev, struct ifreq *ifr,
			    int cmd)
{
	return -EOPNOTSUPP;
}

static u32 ipw_ethtool_get_link(struct net_device *dev)
{
	struct ipw_priv *priv = dev->priv;
	return priv->status & STATUS_ASSOCIATED;
}

static struct ethtool_ops ipw_ethtool_ops = {
	.get_link        = ipw_ethtool_get_link,
};

static irqreturn_t ipw_isr(int irq, void *data, struct pt_regs *regs)
{
	struct ipw_priv *priv = (struct ipw_priv *)data;
	u32 inta, inta_mask;
	unsigned long flags;
	
	if (!priv)
		return IRQ_NONE;

	spin_lock_irqsave(&priv->lock, flags);

	if (!(priv->status & STATUS_INT_ENABLED)) {
		IPW_DEBUG_WARNING("IRQ ignored!\n");
		spin_unlock_irqrestore(&priv->lock, flags);
		return IRQ_NONE;
	}

	inta = ipw_read32(priv, CX2_INTA_RW);
	inta_mask = ipw_read32(priv, CX2_INTA_MASK_R);
	
	if (inta == 0xFFFFFFFF) {
		/* Hardware disappeared */
		IPW_DEBUG_WARNING("IRQ INTA == 0xFFFFFFFF\n");
		spin_unlock_irqrestore(&priv->lock, flags);
		return IRQ_NONE;
	}

	if (!(inta & (CX2_INTA_MASK_ALL & inta_mask))) {
		/* Shared interrupt */
		spin_unlock_irqrestore(&priv->lock, flags);
		return IRQ_NONE;
	}

	/* tell the device to stop sending interrupts */
	ipw_disable_interrupts(priv);
	    
	/* ack current interrupts */
	inta &= (CX2_INTA_MASK_ALL & inta_mask);
	ipw_write32(priv, CX2_INTA_RW, inta);
	    
	/* Cache INTA value for our tasklet */
	priv->isr_inta = inta;

	tasklet_schedule(&priv->irq_tasklet);

 	spin_unlock_irqrestore(&priv->lock, flags);

	return IRQ_HANDLED;
}

static void ipw_adapter_restart(void *adapter)
{
	struct ipw_priv *priv = adapter;

	ipw_down(priv);
	if (ipw_up(priv)) {
		IPW_DEBUG_ERROR("Failed to up device\n");
		return;
	}
}

static void ipw_rf_kill(void *adapter)
{
	struct ipw_priv *priv = adapter;
	unsigned long flags;
	
	if (rf_kill_active(priv)) {
		IPW_DEBUG_RF_KILL("RF Kill active, rescheduling GPIO check\n");
		spin_lock_irqsave(&priv->lock, flags);
		if (priv->workqueue)
			queue_delayed_work(priv->workqueue,
					   &priv->rf_kill, 2 * HZ);
		spin_unlock_irqrestore(&priv->lock, flags);
		return;
	}

	/* 
	 * RF Kill is now disabled, so bring the device back up 
	 */
	IPW_DEBUG_RF_KILL("RF Kill no longer active, restarting device\n");

	priv->status &= ~STATUS_RF_KILL;

	ipw_adapter_restart(priv);
}

static int ipw_setup_deferred_work(struct ipw_priv *priv)
{
	int ret = 0;

#ifdef _LINUX_SWSUSP_H	
	priv->workqueue = create_workqueue(DRV_NAME, 0);
#else
	priv->workqueue = create_workqueue(DRV_NAME);
#endif	
	init_waitqueue_head(&priv->wait_command_queue);

	INIT_WORK(&priv->associate, ipw_associate, priv);
	INIT_WORK(&priv->rx_replenish, ipw_rx_queue_replenish, priv);
	INIT_WORK(&priv->adapter_restart, ipw_adapter_restart, priv);
	INIT_WORK(&priv->rf_kill, ipw_rf_kill, priv);
	INIT_WORK(&priv->up, (void (*)(void *))ipw_up, priv);
	INIT_WORK(&priv->down, (void (*)(void *))ipw_down, priv);
	INIT_WORK(&priv->request_scan, (void (*)(void *))ipw_request_scan, priv);

	tasklet_init(&priv->irq_tasklet, (void (*)(unsigned long))
		     ipw_irq_tasklet, (unsigned long)priv);

	return ret;
}


static void shim__set_security(struct ieee80211_device *ieee, 
			       struct ieee80211_security *sec)
{
	struct ipw_priv *priv = ieee->priv;
	if (sec->flags & SEC_ENABLED) {
		if (sec->enabled)
			priv->capability |= CAP_PRIVACY_ON;
		else
			priv->capability &= ~CAP_PRIVACY_ON;

		/* If the privacy state has switched from what we are 
		 * associated/ing with, then disassociate */
		if ((priv->status & (STATUS_ASSOCIATED | STATUS_ASSOCIATING)) &&
		    (((priv->assoc_request.capability & 
		       WLAN_CAPABILITY_PRIVACY) && !sec->enabled) ||
		     (!((priv->assoc_request.capability & 
			 WLAN_CAPABILITY_PRIVACY) && sec->enabled)))) {
			IPW_DEBUG_INFO("Disassociating due to capability change.\n");
			ipw_disassociate(priv);
		}
	}
}

static struct ieee80211_helper_functions ipw_ieee_callbacks = {
	.set_security = shim__set_security,
	.card_present = NULL,
	.cor_sreset = NULL,
	.dev_open = NULL,
	.dev_close = NULL,
	.genesis_reset = NULL,
	.set_unencrypted_filter = NULL,
	.hw_enable = NULL,
	.hw_config = NULL,
	.hw_reset = NULL,
	.hw_shutdown = NULL,
	.reset_port = NULL,
	.tx = NULL,
	.schedule_reset = NULL,
	.tx_80211 = NULL
};

static int ipw_config(struct ipw_priv *priv) 
{
	int i;
	struct ipw_supported_rates rates;
	struct ipw_sys_config sys_config;
	struct ipw_tx_power tx_power;

	memset(&rates, 0, sizeof(rates));
	memset(&sys_config, 0, sizeof(sys_config));
	memset(&tx_power, 0, sizeof(tx_power));

	priv->scan_age = DEFAULT_MAX_SCAN_AGE;

	/* move device from INIT to HOST_COMPLETE state */
	if (ipw_send_host_complete(priv))
		goto error;
	
	/* move device from HOST_COMPLETE to CARD_DISABLE state */
	if (ipw_send_card_disable(priv, 0))
		goto error;

	/* configure device for 'G' band */
	tx_power.band_mode = IPW_BAND_G;
	tx_power.num_channels = 11;
	for (i = 0; i < 11; i++) {
		tx_power.channels_tx_power[i].channel_number = i + 1;
		tx_power.channels_tx_power[i].tx_power = 0x14;
	}
	if (ipw_send_tx_power(priv, &tx_power))
		goto error;

	/* configure device to also handle 'B' band */
	tx_power.band_mode = IPW_BAND_B;
	if (ipw_send_tx_power(priv, &tx_power))
		goto error;

	/* initialize adapter address */
	if (ipw_send_adapter_address(priv, priv->net_dev->dev_addr))
		goto error;

	/* set basic system config settings */
	init_sys_config(&sys_config);
	if (ipw_send_system_config(priv, &sys_config))
		goto error;

	/* configure supported rates */
	rates.band_mode = IPW_BAND_G;
	rates.purpose = IPW_RATE_CAPABILITIES;
	add_cck_scan_rates(&rates, IPW_CCK_MODULATION, IPW_CCK_DEFAULT_RATES_MASK);
	add_ofdm_scan_rates(&rates, IPW_CCK_MODULATION, IPW_OFDM_DEFAULT_RATES_MASK);
	if (ipw_send_supported_rates(priv, &rates))
		goto error;

	/* Set request-to-send threshold */
	if (priv->rts_threshold) {
		if (ipw_send_rts_threshold(priv, priv->rts_threshold))
			goto error;
	}

        /* TODO: Make this random*/
	if (ipw_send_seed_number(priv, 0x2c171baa))
		goto error;
	
	/* final state transition to the RUN state */
	if (ipw_send_host_complete(priv))
		goto error;

	/* kick off an active scan */
	if (ipw_request_scan(priv))
		goto error;

	return 0;
	
 error:
	return -EIO;
}

#define MAX_HW_RESTARTS 20
static int ipw_up(struct ipw_priv *priv)
{
	int rc, i;

	if (priv->status & STATUS_EXIT_PENDING)
		return -EIO;

	for (i = 0; i < MAX_HW_RESTARTS; i++ ) {
		/* Load the microcode, firmware, and eeprom.  
		 * Also start the clocks. */
		rc = ipw_load(priv);
		if (rc) {
			IPW_DEBUG_ERROR("Unable to load firmware: 0x%08X\n",
					rc);
			return rc;
		}

		ipw_init_ordinals(priv);
		eeprom_parse_mac(priv, priv->net_dev->dev_addr);

		rc = ipw_config(priv);
		if (!rc) {
			IPW_DEBUG_INFO("Configured device on count %i\n", i);

			return 0;
		} else {
			IPW_DEBUG_INFO("Device configuration failed: 0x%08X\n",
				       rc);
		}
		
		IPW_DEBUG_INFO("Failed to config device on retry %d of %d\n",
			       i, MAX_HW_RESTARTS);

		/* We had an error bringing up the hardware, so take it
		 * all the way back down so we can try again */
		ipw_down(priv);
	}

	/* tried to restart and config the device for as long as our 
	 * patience could withstand */
	IPW_DEBUG_ERROR("Unable to initialize device after %d attempts.\n",
			i);
	return -EIO;
}

static void ipw_down(struct ipw_priv *priv)
{
	if (priv->status & STATUS_RF_KILL)
		/* we are already down */
		return;

	/* tell the device to stop sending interrupts */
	ipw_disable_interrupts(priv);

	/* Clear all bits but the RF Kill and device->open bit */
	priv->status &= (STATUS_RF_KILL | STATUS_DEV_OPEN);

	netif_carrier_off(priv->net_dev);
	netif_stop_queue(priv->net_dev);

	ipw_stop_nic(priv);
}


/* Called by register_netdev() */
static int ipw_net_init(struct net_device *dev)
{
	struct ipw_priv *priv = dev->priv;

	if (rf_kill_active(priv)) {
		IPW_DEBUG_WARNING("Radio Frequency Kill Switch is On:\n"
				  "Kill switch must be turned off for "
				  "wireless networking to work.");
		queue_delayed_work(priv->workqueue, &priv->rf_kill, 2 * HZ);
		return 0;
	}

	if (ipw_up(priv))
		return -EIO;

	return 0;
}

/* PCI driver stuff */
static struct pci_device_id card_ids[] = {
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2701, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2702, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2711, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2712, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2721, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2722, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2731, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2732, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2741, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x103c, 0x2741, 0, 0, 0},/* very old pre-releas hw */
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2742, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2751, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2752, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2753, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2754, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2761, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x1043, 0x8086, 0x2762, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x104f, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{PCI_VENDOR_ID_INTEL, 0x4220, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	
	/* required last entry */
	{0,}
};

MODULE_DEVICE_TABLE(pci, card_ids);

static int ipw_pci_probe(struct pci_dev *pdev,
			 const struct pci_device_id *ent)
{
	int err = 0;
	struct net_device *net_dev;
	void *base;
	u32 length, val;
	struct ipw_priv *priv;

	net_dev = alloc_etherdev(sizeof(struct ipw_priv));
	if (net_dev == NULL) {
		err = -ENOMEM;
		goto out;
	}

	priv = net_dev->priv;
	memset(priv, 0, sizeof(*priv));
	priv->net_dev = net_dev;
	priv->pci_dev = pdev;
#ifdef CONFIG_IPW2200_DEBUG
	ipw_debug_level = debug;
#endif
	spin_lock_init(&priv->lock);

	/* Initialize module parameter values here */
	if (channel != 0) {
		priv->config |= CFG_STATIC_CHANNEL;
		priv->channel = channel;
		IPW_DEBUG_INFO("Bind to static channel %d\n", channel);
		/* TODO: Validate that provided channel is in range */
	}

	if (pci_enable_device(pdev)) {
		err = -ENODEV;
		goto out_free_netdev;
	}

	pci_set_master(pdev);
	pci_set_drvdata(pdev, priv);

	err = pci_request_regions(pdev, DRV_NAME);
	if (err) 
		goto out_pci_disable_device;

	/* We disable the RETRY_TIMEOUT register (0x41) to keep 
	 * PCI Tx retries from interfering with C3 CPU state */
	pci_read_config_dword(pdev, 0x40, &val); 
	if ((val & 0x0000ff00) != 0) 
		pci_write_config_dword(pdev, 0x40, val & 0xffff00ff);
	
	length = pci_resource_len(pdev, 0);
	priv->hw_len = length;
	
	base = ioremap_nocache(pci_resource_start(pdev, 0), length);
	if (!base) {
		err = -ENODEV;
		goto out_pci_release_regions;
	}

	priv->hw_base = (unsigned long)base;
	IPW_DEBUG_INFO("pci_resource_len = 0x%08x\n", length);
	IPW_DEBUG_INFO("pci_resource_base = %p\n", base);

	/* These seem to cause spurios INTs during load, which don't get 
	 * handled, so the IRQ gets disabled
	 pci_set_power_state(pdev, 0);
	 pci_restore_state(pdev, NULL); */

	err = ipw_setup_deferred_work(priv);
	if (err) {
		IPW_DEBUG_ERROR("Unable to setup deferred work\n");
		goto out_iounmap;
	}


	/* Initialize WEP */
	priv->ieee = ieee80211_alloc(priv->net_dev, priv);
	if (!priv->ieee) {
		IPW_DEBUG_ERROR("Unable to allocate ieee80211 data\n");
		err = -ENOMEM;
		goto out_destroy_workqueue;
	}

	priv->ieee->func = &ipw_ieee_callbacks;
	priv->ieee->tx_payload_only = 1;
	priv->ieee->iw_mode = IW_MODE_INFRA;

	err = ipw_beacons_allocate(priv);
	if (err) {
		IPW_DEBUG_ERROR("Unable to allocate beacon storage\n");
		goto out_free_ieee80211;
	}
	ipw_beacons_initialize(priv);

	err = request_irq(pdev->irq, ipw_isr, SA_SHIRQ, DRV_NAME, 
			  priv);
	if (err) {
		IPW_DEBUG_ERROR("Error allocating IRQ %d\n", pdev->irq);
		goto out_release_beacons;
	}

	SET_MODULE_OWNER(net_dev);

	strcpy(net_dev->name, "eth%d");
	net_dev->open = ipw_net_open;
	net_dev->stop = ipw_net_stop;
	net_dev->init = ipw_net_init;
	net_dev->hard_start_xmit = ipw_net_hard_start_xmit;
	net_dev->get_stats = ipw_net_get_stats;
	net_dev->set_multicast_list = ipw_net_set_multicast_list;
	net_dev->set_mac_address = ipw_net_set_mac_address;
	net_dev->change_mtu = ipw_net_change_mtu;
	net_dev->do_ioctl = ipw_net_do_ioctl;
	net_dev->get_wireless_stats = ipw_get_wireless_stats;
	net_dev->wireless_handlers = &ipw_wx_handler_def;
	net_dev->ethtool_ops = &ipw_ethtool_ops;
	net_dev->irq = pdev->irq;
	net_dev->base_addr = priv->hw_base;
	net_dev->mem_start = pci_resource_start(pdev, 0);
	net_dev->mem_end = net_dev->mem_start + pci_resource_len(pdev, 0) - 1;

	priv->hw_id = *ent;

	device_create_file(&pdev->dev, &dev_attr_rf_kill);
	device_create_file(&pdev->dev, &dev_attr_direct_word);
	device_create_file(&pdev->dev, &dev_attr_indirect_byte);
	device_create_file(&pdev->dev, &dev_attr_indirect_word);
	device_create_file(&pdev->dev, &dev_attr_mem_gpio_reg);
	device_create_file(&pdev->dev, &dev_attr_command_event_reg);
	device_create_file(&pdev->dev, &dev_attr_nic_type);
	device_create_file(&pdev->dev, &dev_attr_status);
	device_create_file(&pdev->dev, &dev_attr_error_log);
	device_create_file(&pdev->dev, &dev_attr_eeprom_clear);
	device_create_file(&pdev->dev, &dev_attr_eeprom_sram);
	device_create_file(&pdev->dev, &dev_attr_eeprom);
	device_create_file(&pdev->dev, &dev_attr_eeprom_delay);
	device_create_file(&pdev->dev, &dev_attr_fw_version);
	device_create_file(&pdev->dev, &dev_attr_fw_date);
	device_create_file(&pdev->dev, &dev_attr_ucode_version);
	device_create_file(&pdev->dev, &dev_attr_rtc);

	if (register_netdev(net_dev)) {
		IPW_DEBUG_ERROR("failed to register network device\n");
		err = -ENODEV;
		goto out_release_irq;
	}

	return 0;

 out_release_irq:
	free_irq(pdev->irq, priv);
 out_release_beacons:
	ipw_beacons_free(priv);
 out_free_ieee80211:
	ieee80211_free(priv->ieee);
	priv->ieee = NULL;
 out_destroy_workqueue:
	destroy_workqueue(priv->workqueue);
	priv->workqueue = NULL;
 out_iounmap:
	iounmap((void *)priv->hw_base);
 out_pci_release_regions:
	pci_release_regions(pdev);
 out_pci_disable_device:
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
 out_free_netdev:
	free_netdev(priv->net_dev);
 out:
	return err;
}

static void ipw_pci_remove(struct pci_dev *pdev)
{
	struct ipw_priv *priv = pci_get_drvdata(pdev);
	if (!priv)
		return;

	priv->status |= STATUS_EXIT_PENDING;

	device_remove_file(&pdev->dev, &dev_attr_rtc);
	device_remove_file(&pdev->dev, &dev_attr_ucode_version);
	device_remove_file(&pdev->dev, &dev_attr_fw_date);		
	device_remove_file(&pdev->dev, &dev_attr_fw_version);		
	device_remove_file(&pdev->dev, &dev_attr_eeprom_delay);	
	device_remove_file(&pdev->dev, &dev_attr_eeprom);
	device_remove_file(&pdev->dev, &dev_attr_eeprom_sram);
	device_remove_file(&pdev->dev, &dev_attr_eeprom_clear);
	device_remove_file(&pdev->dev, &dev_attr_error_log);
	device_remove_file(&pdev->dev, &dev_attr_command_event_reg);
	device_remove_file(&pdev->dev, &dev_attr_mem_gpio_reg);
	device_remove_file(&pdev->dev, &dev_attr_indirect_word);
	device_remove_file(&pdev->dev, &dev_attr_indirect_byte);
	device_remove_file(&pdev->dev, &dev_attr_direct_word);
	device_remove_file(&pdev->dev, &dev_attr_rf_kill);
	device_remove_file(&pdev->dev, &dev_attr_status);
	device_remove_file(&pdev->dev, &dev_attr_nic_type);

	unregister_netdev(priv->net_dev);

	ipw_down(priv);

	ipw_rx_queue_free(priv, priv->rxq);
	ipw_tx_queue_free(priv);
	ipw_beacons_free(priv);
	ieee80211_free(priv->ieee);

	/* ipw_down will ensure that there is no more pending work
	 * in the workqueue's, so we can safely remove them now. */
	if (priv->workqueue) { 
		cancel_delayed_work(&priv->request_scan);
		cancel_delayed_work(&priv->rf_kill);
		destroy_workqueue(priv->workqueue);
		priv->workqueue = NULL;
	}

	free_irq(pdev->irq, priv);
	iounmap((void *)priv->hw_base);
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	free_netdev(priv->net_dev);
}

static int ipw_pci_suspend(struct pci_dev *pdev, u32 state)
{
	int err = 0;

	return err;
}

static int ipw_pci_resume(struct pci_dev *pdev)
{
	int err = 0;

	return err;
}

/* driver initialization stuff */
static struct pci_driver ipw_driver = {
	.name = DRV_NAME,
	.id_table = card_ids,
	.probe = ipw_pci_probe,
	.remove = __devexit_p(ipw_pci_remove),
	.suspend = ipw_pci_suspend,
	.resume = ipw_pci_resume,
};

static int __init ipw_init(void)
{
	int ret;

	printk(KERN_INFO DRV_NAME ": " DRV_DESCRIPTION ", " DRV_VERSION "\n");
	printk(KERN_INFO DRV_NAME ": " DRV_COPYRIGHT "\n");

	ret = pci_module_init(&ipw_driver);
	if (ret) {
		IPW_DEBUG_ERROR("Unable to initialize PCI module\n");
		return ret;
	}

	ret = driver_create_file(&ipw_driver.driver, 
				 &driver_attr_debug_level);
	if (ret) {
		IPW_DEBUG_ERROR("Unable to create driver sysfs file\n");
		pci_unregister_driver(&ipw_driver);
		return ret;
	}

	return ret;
}

static void __exit ipw_exit(void)
{
	driver_remove_file(&ipw_driver.driver, &driver_attr_debug_level);
	pci_unregister_driver(&ipw_driver);
}

module_param(debug, int, 644);
MODULE_PARM_DESC(debug, "debug output mask");

module_param(channel, int, 644);
MODULE_PARM_DESC(channel, "channel to limit associate to (default 0 [ANY])"); 

module_exit(ipw_exit);
module_init(ipw_init);
